/*
* MPIA - GEIRS project
*
* "$Id: FitsImg.cxx 795M 2020-10-28 12:54:23Z (local) $
*
*/

#undef PACKAGE_BUGREPORT
#undef PACKAGE_NAME
#undef PACKAGE_STRING
#undef PACKAGE_TARNAME
#undef PACKAGE_URL
#undef PACKAGE_VERSION


#include "config.h"

#include <string>
#include <cstring>
#include <cmath>
#include <ctime>
#include <cstdlib>
#include <fstream>
#include <sstream>
#include <limits>
#include <valarray>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_TIFF_H
extern "C" {
#include <tiffio.h>
}
#endif

#include "FitsImg.h"
#include "ColorLUTSet.h"
#include "Subwin.h"

/** If this is defined, the baseline palette-color class is used.
* If this is no defined, the triple of RGBcolors itself is inserted in the scan lines.
* This should result basically in a 3-fold reduction of the size,
* because not 3 RGB values (each 8 bytes) but only one index into the LUT is stored.
* The truth is that after the use of a LZW compression that advantage is essentially
* void: often the TIFF files with palette-color are larger than those without....
* A further disadvantage of the palette-color class seems to be that tiff2pdf
* does not handle thes.
* @since 2020-08-25
#define FITS2TIFF_USE_PALETTE
*/

using namespace std ;

/** Constructor.
* @param[in] fitsname The name of the FITS file to be converted.
* @since 2012-11-25
* @author Richard J. Mathar
*/
FitsImg::FitsImg(char *fitsIname) : iname()
{
	iname.push_back(fitsIname) ;
} /* Fits2Img::Fits2Img */

/** Constructor.
* @param[in] argc The number of names in the fitsname argument.
* @param[in] fitsname The names of the FITS files to be converted.
* @since 2013-01-29
* @author Richard J. Mathar
*/
FitsImg::FitsImg(int argc,char *fitsIname[]) : iname()
{
	for(int f=0 ; f < argc ; f++)
		iname.push_back(fitsIname[f]) ;
} /* Fits2Img::Fits2Img */

/** Dump the entire image as a serialized stream of floating point numbers to a file, or in TIFF format.
* The output are 32-bit floating point values, starting with the top row left to
*  right, then the 2nd row left to right etc. So the serialized order of the values is
*  as defined by most graphic applications (i.e., flipped up-down compared to the FITS convention).
*  The representation is what the endiness of the current processor is.
* @param[in] ofile The name of the binary output file to be created/overwritten.
* @param[in] rf The rotation and flip transformation on the original image.
* @param[in] cuts The lower and higher ADU cuts to be mapped to 0 and 1 on output.
*  If both are the same value, the program uses a histogram and discards 5 percent of the data
*  to define the cuts levels.
* @param[in] raw If true output is a raw stream of 32-bit floating point numbers, else a TIFF file.
* @param[in] lutidx An index in the range 0 to one less than the number of static color tables.
* @param[in] gammExp Exponent of a gamma value
* @param[in] bin Binning factor. Length of the tiles that are merged into single values in the output.
* @param[in] compr Compression integer, 1=NONE, 5=LZW, 32773=PACKBITS
* @since 2019-11-05
* @since 2020-03-06 rephrased to work with Rotflip specs and cfitsio (no CCfits required)
* @since 2020-03-27 Implemented gamma correction and binning
* @author Richard J. Mathar
*/
void FitsImg::dumpRaw(const string & ofile, const Rotflip &rf, float cuts[2], bool raw, int lutidx, 
		double gammExp, const int bin, bool verb, int compr)
{
	/* status of cfitsio command library
	*/
	int status =0;
	fitsfile *ifits ;

	/* try to access (read only)  the FITS file of the command line
	*/
	fits_open_file(&ifits, iname[0].c_str(), READONLY, & status) ;
	if ( status)
	{
		/* couldn't open the FITS file for reading
		*/
		fits_report_error(stderr, status) ;
		return ;
	}
		
	/* number of extensions (e.g. 1 if there is only a PHDU)
	*/
	int nex ;
	fits_get_num_hdus(ifits, & nex, &status) ;
	if ( status)
	{
		/* couldn't obtain the number of HDU's
		*/
		cerr << "Could not read number of HDU's in " << iname[0] << endl ;
		fits_report_error(stderr, status) ;
		return ;
	}


	/* cfitsio enumerates HDU's starting at 1 for the PHDU. Loop thru all of them.
	*/
	for(int exIdx=1 ; exIdx <= nex ; exIdx++)
	{
		status =0 ;

		int hdutype ;
		/* move local fits pointer to the extension number exIdx
		*/
		fits_movabs_hdu(ifits, exIdx, &hdutype, &status) ;

		if (status)
		{
			/* could not reach that HDU; just assume some general problem with the FITS input file
			*/
			cerr << "Could not open HDU's  number " << exIdx << " of " << iname[0] << endl ;
			fits_report_error(stderr, status) ;
			break ;
		}
	
		/* number of axes, 2 for simple images, 3 for cubes
		*/
		int naxes ;
		fits_get_img_dim(ifits, & naxes, &status) ;

		/* This program cannot handle 1D and >=4D cubes
		* naxes=0 probably means that MEF are in use here.
		*/
		if ( naxes < 2 || naxes > 3 || status)
			continue ;

		dumpRaw(ifits,rf,cuts, ofile,exIdx,raw,lutidx,gammExp,bin,verb,compr) ;
	} /* end of loop over HDUs */

	/* close FITS input file
	*/
	fits_close_file(ifits,&status) ;
} /* FitsImg::dumpRaw */

/** Dump the image as a raw sequence of floating point values or a TIFF image into a file.
* These are 32-bit floating point values, starting with the top row left to
*  right, then the 2nd row left to right etc. So the order of the values is
*  as defined by most graphic applications (flipped along x relative to the FITS convention)
*  and what the endiness of the current processor is.
* @param[in] iex Pointer to the actual FITS handle header-data-unit of the data.
* @param[in] rf The roation and/or flip of the original image.
* @param[in] cuts Lower and upper cut level to be mapped to dark and white on output.
* @param[in] ofile Name of the file to be written.
*  One way of looking at the numbers is "od -t f4 --endian=big"
*  Further conversion like "raw2tiff -w 4096 -l 4096 -d float tst_0.bin tst_0.tif"
* @param[in] exIdx HDU number, starting at 0=primary, 1=first extension etc.
* @param[in] raw If true output is a raw stream of 32-bit floating point numbers, else a TIFF file.
* @param[in] lutidx An index in the range 0 to one less than the number of static color tables.
* @param[in] gammExp Exponent of a gamma value.
*       Equals 1 if the brighness in the TIFF file is linear to the brightness of the FITS file.
* @param[in] bin Integer binning factor. Small positive integer.
* @param[in] compr A compression specification, 1 =NONE, 32773=PACKBITS, 5=LZW
* @since 2019-11-05
* @since 2020-03-06 rephrased to work with Rotflip specs and cfitsio (no CCfits required)
* @since 2020-03-27 Added gamma correction and binning.
* @since 2020-10-06 Added compression option
* @since 2020-10-28 Corrected rendering of TIFF timestamp
* @see https://libtiff.gitlab.io/libtiff/
* @author Richard J. Mathar
*/
void FitsImg::dumpRaw(fitsfile * iex, const Rotflip &rf, float cuts[2], const string & ofile, int exIdx,
	bool raw,int lutidx, double gammExp, const int bin, bool verb, int compr)
{
	/* the size along x, along y (both pixels) and z (default 1, larger if cubes)
	* The first 2 parameters will be filled in by cfitsio later on.
	*/
	long axs[3] = {0L,0L,1L} ;

	int naxes ;
	int bitpix ;
	int status =0 ;

	/* read BITPIX and the number and lengts of the image axes.
	*/
	fits_get_img_param(iex, 3, & bitpix, & naxes, axs, &status) ;
	if ( status )
	{
		cerr << "Could not read major keywords (bitpix or lengths of axes) in HDU number " << exIdx << endl ;
		fits_report_error(stderr, status) ;
		return ;
	}


	/* number of pixels in the data cube = product of axes lengths
	*/
	long firstp[3] = {1L,1L,1L} ;
	long long nelements = axs[0]*axs[1]*axs[2] ;

	if ( nelements == 0)
		/* no output, exit from function, if no data...
		*/
		return ;

	float * farr = new float[nelements] ;
	int anynul ;
	/* convert all data to float values, independent of BITPIX encoding within the input 
	*/
	fits_read_pix(iex, TFLOAT, firstp, nelements, 0, farr,&anynul, &status  ) ;

	if ( status )
	{
		delete [] farr ;
		/* no output if no data...
		*/
		fits_report_error(stderr, status) ;
		return ;
	}

	/* FITS keyword CREATOR
	*/
	char creator[1+FLEN_VALUE] ;

	/* FITS keyword DATE
	*/
	char date[1+FLEN_VALUE] ;

	/* FITS keyword OBJECT
	*/
	char object[1+FLEN_VALUE] ;

	/* FITS keyword INSTRUMENT
	*/
	char instrum[1+FLEN_VALUE] ;

	/* FITS keyword ORIGFILE
	*/
	char  origfile[1+FLEN_VALUE] ;

	/* FITS keyword EXTNAME
	*/
	char extname[1+FLEN_VALUE] ;

	char comment[1+FLEN_COMMENT] ;
	/* the following is probably only available in the PHDU. Ignore status errors.
	* Collect some FITS standard header keywords that may be equivalent to
	* standard TIFF properties.
	*/
	fits_read_keyword(iex, "CREATOR", creator, comment, &status) ;
	status =0 ;
	fits_read_keyword(iex, "DATE", date, comment, &status) ;
	status =0 ;
	fits_read_keyword(iex, "OBJECT", object, comment, &status) ;
	status =0 ;
	fits_read_keyword(iex, "INSTRUME", instrum, comment, &status) ;
	status =0 ;
	fits_read_keyword(iex, "FILENAME", origfile, comment, &status) ;
	status =0 ;


	/* flip array farr along horiz and/or vertical axis. We want the output
	* to be always in X11 order, so the number of up-down-flips is opposite
	* to what the parameter of the input demands. So bit number 1 of the rf.flipxy
	* is toggled here, i.e., exclusive-ored with the numerical value 2.
	*/
	Rotflip rfx11(rf.rot90, rf.flipxy ^2);

	/* each slice is a axs[0]xaxs[1] array. On output, rotations may have swapped these dimensions.
	*/
	Subwin subin(0,0,axs[0],axs[1]) ;
	Subwin subout = subin.rotflip(rfx11,subin) ;
	flip(farr,axs,rfx11,subout) ;

	/* redefine the cut levels if upper and lower value are the same
	* (if the user did not use the -c option of fits2tiff)
	*/
	if ( isnan(cuts[0]) )
	{
		/* sort on a local copy of the values ... consider all in a cube at the same time...
		*/
		float * arr4sort = new float[nelements] ;
		for( int i=0 ; i < nelements ; i++)
			arr4sort[i] = farr[i] ;

		qsort(arr4sort, nelements, sizeof(float),qsortCmpFlt ) ;

		const int perc100 = nelements-1 ;
		/* print a numerical histogram 
		*/
		if ( verb)
		{
			cout << "percentiles of " << nelements << " pixels:" << endl ;
			for(float perc = 0.05 ; perc <= 0.99 ; perc += 0.05)
				/* 0 maps to arr4sort[0] and 1 maps to arr4sort[n-1]
				*/
				cout << perc << " " << arr4sort[(int)(perc*perc100)] << endl ;
		}
		cuts[0] = arr4sort[(int)(0.025*perc100)] ;
		cuts[1] = arr4sort[(int)(0.975*perc100)] ;
		if ( verb)
			cout << "cuts set to " << cuts[0] << ":" << cuts[1] << endl ;
		delete [] arr4sort ;
	}

	/* loop over the slices of any FITS cube
	* (single instance if not a cube....)
	*/
	for ( int slic =0 ; slic < axs[2] ; slic++)
	{
		if ( raw )
		{
			/** attach current HDU and slice index to file name.
			*/
			ostringstream ofile2 ;
			ofile2 << ofile << "_" << (exIdx-1) << "_" << slic << ".bin" ;
		
			ofstream of(ofile2.str().c_str(), ios_base::trunc | ios_base::binary) ;
	
			/* pack the values in the range into an array, mapped to the value interval (0,1),
			* independent of any BLANK value that might have been defined in the FITS header.
			*/
			valarray<float> farrslic = serial(farr,subout,cuts,axs,slic,gammExp,bin) ;
	
			/* put all values to the cout.
			*/
			for(int i= 0 ; i < (int)farrslic.size() ; i++)
				of.write( (char *)& farrslic[i],sizeof(float)) ;
			of.close() ;
		}
		else
		{
#ifdef HAVE_TIFF_H

			ColorLUTSet luts ;

			/** attach current HDU and slice index to file name
			*/
			ostringstream ofile2 ;
			ofile2 << ofile << "_" << (exIdx-1) << "_" << slic << ".tif" ;
		
			int err ;
			TIFF * of = TIFFOpen(ofile2.str().c_str(),"w") ;
			err = TIFFSetField(of, TIFFTAG_IMAGEWIDTH,subout.xsize/bin) ;
			if ( err <= 0 )
				cerr << "SetField error IMAGWIDTH " << subout.xsize/bin << endl ;
			err = TIFFSetField(of, TIFFTAG_IMAGELENGTH,subout.ysize/bin) ;
			if ( err <= 0 )
				cerr << "SetField error IMAGLENGTH " << subout.ysize/bin << endl ;
			err = TIFFSetField(of, TIFFTAG_ROWSPERSTRIP,1) ;
			if ( err <= 0 )
				cerr << "SetField error ROWSPERSTRIP " << 1 << endl ;

			if ( strlen(date) > 0 )
			{
				/* remove the leading and trailing tic-marks and eliminate the T in the ISO string
				*/
				memmove(&date[0],&date[1],19*sizeof(char)) ;
				date[10] = date[19] = ' ' ;
				TIFFSetField(of, TIFFTAG_DATETIME,date) ;
			}

			if  ( strlen(creator) > 0 )
				TIFFSetField(of, TIFFTAG_SOFTWARE,creator) ;

			if  ( strlen(object) > 0 )
				TIFFSetField(of, TIFFTAG_IMAGEDESCRIPTION,object) ;

			if  ( strlen(instrum) > 0 )
				TIFFSetField(of, TIFFTAG_ARTIST,instrum) ;

			if  ( strlen(origfile) > 0 )
				TIFFSetField(of, TIFFTAG_DOCUMENTNAME,origfile) ;

			if  ( strlen(extname) > 0 )
				TIFFSetField(of, TIFFTAG_PAGENAME,extname) ;

			TIFFSetField(of, TIFFTAG_PAGENUMBER,slic) ;
		
			/* pack the values in the range into an array, mapped to the value interval (0,1),
			* independent of any BLANK value that might have been defined in the FITS header.
			*/
			valarray<float> farrslic = serial(farr,subout,cuts,axs,slic,gammExp,bin) ;

			if ( lutidx <0 || lutidx >= GEIRS_COLORLUTSET_CNT )
			{
				/* gray scale */
				err = TIFFSetField(of, TIFFTAG_BITSPERSAMPLE,8*TIFFDataWidth(TIFF_FLOAT)) ;
				if ( err <= 0 )
					cerr << "SetField error BITSPERSAMPEL " << endl ;
	
				err = TIFFSetField(of, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_IEEEFP);
				if ( err <= 0 )
					cerr << "SetField error SAMPLEFORMAT " << endl ;
	
				/* preset default. not needed
				err = TIFFSetField(of, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
				if ( err <= 0 )
					cerr << "SetField error ORIENTATION " << endl ;
				*/
	
				err = TIFFSetField(of, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG) ;
				if ( err <= 0 )
					cerr << "SetField error PLNARCONFIG " << endl ;
	
				err = TIFFSetField(of, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK) ;
				if ( err <= 0 )
					cerr << "SetField error PHOTOMETRIC " << endl ;
	
				// err = TIFFSetField(of, TIFFTAG_COMPRESSION, COMPRESSION_PACKBITS) ;
				// err = TIFFSetField(of, TIFFTAG_COMPRESSION, COMPRESSION_LZW) ;
				err = TIFFSetField(of, TIFFTAG_COMPRESSION, compr) ;
				if ( err <= 0 )
					cerr << "SetField error COMPRESSION " << endl ;
	
				/* any default will do
				err = TIFFSetField(of, TIFFTAG_FILLORDER,  FILLORDER_LSB2MSB);
				if ( err <= 0 )
					cerr << "SetField error FILLORDER " << endl ;
				*/
		
				float * row = (float *) _TIFFmalloc( (subout.xsize/bin) *sizeof(float) ) ;
		
				/* put all values to the cout, one per row.
				*/
				for(int y= 0 ; y < subout.ysize/bin ; y++)
				{
					#pragma omp parallel for
					for(int x=0 ; x < subout.xsize/bin ; x++)
						row[x] = farrslic[x+y*subout.xsize/bin] ;
					int err =TIFFWriteScanline(of, row, y, 0);
					if ( err < 0 )
					{
						cerr << "TIFFWritescanline error row " << y << endl ;
						break ;
					}
				}
				_TIFFfree(row) ;
			}
			else 
			{
#ifdef FITS2TIFF_USE_PALETTE
				/* RGB with color map in a palette of 256 colors
				*/
				err = TIFFSetField(of, TIFFTAG_SAMPLESPERPIXEL,1) ;
#else
				err = TIFFSetField(of, TIFFTAG_SAMPLESPERPIXEL,3) ;
#endif
				if ( err <= 0 )
					cerr << "SetField error SAMPLESPERPIXEL " << endl ;

				err = TIFFSetField(of, TIFFTAG_BITSPERSAMPLE,8) ;
				if ( err <= 0 )
					cerr << "SetField error BITSPERSAMPEL " << endl ;
	
				err = TIFFSetField(of, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
				if ( err <= 0 )
					cerr << "SetField error SAMPLEFORMAT " << endl ;
	
				err = TIFFSetField(of, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG) ;
				if ( err <= 0 )
					cerr << "SetField error PLANARCONFIG " << endl ;
	
#ifdef FITS2TIFF_USE_PALETTE
				err = TIFFSetField(of, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_PALETTE) ;
#else
				err = TIFFSetField(of, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB) ;
#endif
				if ( err <= 0 )
					cerr << "SetField error PHOTOMETRIC " << endl ;
	
				// err = TIFFSetField(of, TIFFTAG_COMPRESSION, COMPRESSION_PACKBITS) ;
				err = TIFFSetField(of, TIFFTAG_COMPRESSION, COMPRESSION_LZW) ;
				if ( err <= 0 )
					cerr << "SetField error COMPRESSION " << endl ;

#ifdef FITS2TIFF_USE_PALETTE
				/* add the color map table
				* type of each datum is SHORT (=ushort 16bit 0..65535). total number of shorts is
				* 3*(2^bitspersample)=3*256. all red values first, then green values, then blue values.
				*/
				unsigned short cmp[3*256] ;
				luts.tiff(lutidx,cmp) ;
				err = TIFFSetField(of, TIFFTAG_COLORMAP, cmp, & cmp[256], & cmp[512]) ;
				if ( err <= 0 )
					cerr << "SetField error COLORMAP " << endl ;
	
				char * row = (char *) _TIFFmalloc( (subout.xsize/bin) *sizeof(char) ) ;
#else
				char * row = (char *) _TIFFmalloc( (subout.xsize/bin) *sizeof(char[3]) ) ;
#endif
				/* put all values to the cout, one per row.
				*/
				for(int y= 0 ; y < subout.ysize/bin ; y++)
				{
					for(int x=0 ; x < subout.xsize/bin ; x++)
					{
						float gray = farrslic[x+y*subout.xsize/bin] ;
#ifdef FITS2TIFF_USE_PALETTE
						const int rgbidx = max(0,min(255,(int)(gray*256.0f))) ;
						row[x] = rgbidx ;
#else
						luts.map(lutidx,gray,&row[3*x]) ;
#endif
					}
					int err =TIFFWriteScanline(of, row, y, 0);
					if ( err < 0 )
					{
						cerr << "TIFFWritescanline error row " << y << endl ;
						break ;
					}
				}
				_TIFFfree(row) ;
			}
			TIFFClose(of) ;
#else
			cerr << "Error: not implemented: TIFF package devel header files not found at compile time" << endl ;
#endif
		} /* end if-else raw or tiff */
	} /* end of loop over slices of cubes */

	delete [] farr ;
} /* FitsImg::dumpRaw */

/** Extract a slice of a data cube, scaled to the [0..1] range.
* These are 32-bit floating point values, starting with the top row left to
*  right, then the 2nd row left to right etc. So the order of the values is
*  as defined by most graphic applications (flipped along x relative to the FITS convention)
*  and what the endiness of the current processor is.
* @param[in] farr The array of the input cube of data.
* @param[in] subout The origin and dimension of the region of the array.
*  This is implicitly a X11 convention where (0,0) = top left.
* @param[in] slic Index of the slice (3d index into the cube of farr[][][]).
* @param[in] gammExp gamma factor (exponent)
* @param[in] bin Binning factor. binxbin input values are merged into a single output value.
* @return the portion of the data cube constrained to range[0.1] in the first coordinate,
*  to range[2.3] in the second coordinate, and to index slic in the 3rd coordinate.
*  All values are scale so the input value of cuts[0] becomes 0 and cuts[1] becomes 1.0 on output.
*  If there is more than 1 slice in the farr input array, the number of values
*  returned is smaller (only the portion of the enumerted slice).
* @since 2019-11-06
* @since 2020-03-06 rephrased to work with farr pointer instead of farr valarray
* @since 2020-03-27 implemented gammExp power and binning.
* @author Richard J. Mathar
*/
valarray<float> FitsImg::serial(const float * farr, const Subwin & subout,
	const float cuts[2], const long axs[3], const int slic, const double gammExp, const int bin)
{
	valarray<float> out((subout.xsize/bin) * (subout.ysize/bin)) ;

	/* not strictly needed, but for performance reasons we slit the unbinned
	* case from the rest of the code...
	*/
	if ( bin == 1 )
	{
		/* put all values to the out, applying the cut levels. 
		* If there needs to be a flip of any kind,
		* that should have been done before calling this..
		*/
		for(int y= 0 ; y < subout.ysize ; y++)
		{
			#pragma omp parallel for
			for(int x= 0 ; x < subout.xsize ; x++)
			{
				/* output mapped to (0,1) is (value-cut[0])/(cuts[1]-cuts[0]) if cuts[1]>cuts[0]
				* This is correct for cuts[0]<cuts[1] and also cuts[0]>cuts[1] (black on white background)
				* Serial index of (x,y,z) is x+axs[0]*(y+z*axs[1]).
				*/
				float scal = (farr[x+subout.xsize *(y+slic*subout.ysize )]-cuts[0]) / (cuts[1]-cuts[0]) ;
				/* clamp to the interval [0...1] */
				scal = max(0.0f,min(scal,1.0f)) ;
				if ( cuts[0] > cuts[1])
					scal = pow(scal,gammExp) ;
				else
					scal = 1.f-pow(1.0-scal,gammExp) ;
				out[ x+ subout.xsize *y] = scal ;
			}
		}
	}
	else
	{
		const int tilarea = bin*bin ;
		float * til = new float[tilarea] ;
		/* put all values to the out, applying the cut levels. 
		* If there needs to be a flip of any kind,
		* that should have been done before calling this..
		*/
		for(int y= 0 ; y < subout.ysize/bin ; y++)
		{
			for(int x= 0 ; x < subout.xsize/bin ; x++)
			{
				/* determine a median value of the bin x bin tile
				*/
				#pragma omp parallel for
				for(int ysub =0 ; ysub < bin ; ysub++)
				for(int xsub =0 ; xsub < bin ; xsub++)
					til[xsub + bin*ysub] = farr[x*bin+xsub +subout.xsize *((y*bin+ysub)+slic*subout.ysize )] ;

				qsort(til, tilarea, sizeof(float),qsortCmpFlt ) ;
				float val = ( tilarea % 2) ? til[tilarea/2] :  ( til[tilarea/2] + til[1+tilarea/2])/2. ;

				/* output mapped to (0,1) is (value-cut[0])/(cuts[1]-cuts[0]) if cuts[1]>cuts[0]
				* This is correct for cuts[0]<cuts[1] and also cuts[0]>cuts[1] (black on white background)
				* Serial index of (x,y,z) is x+axs[0]*(y+z*axs[1]).
				*/
				float scal = (val-cuts[0]) / (cuts[1]-cuts[0]) ;
				/* clamp to the interval [0...1] */
				scal = max(0.0f,min(scal,1.0f)) ;
				if ( cuts[0] > cuts[1])
					scal = pow(scal,gammExp) ;
				else
					scal = 1.f-pow(1.0-scal,gammExp) ;
				out[ x+ subout.xsize/bin *y] = scal ;
			}
		}
		delete [] til ;
	}
	return out ;
} /* FitsImg::serial */


/** Flip-rotate an array (cube of data).
* @param[in,out] farr The array to be flipped.
*   On return the swapped/flipped pixels have overwritten the old contents.
* @param[in] axs axs[0] width of pixels, axs[1] height of pixels, axs[2] the number of slices.
*   These lengths are for the input data, prior to rotation/flip.
*   The output data may refer to a swapped variant of axs[0] and axs[1] if a
*   a 90 or 270 degree rotation is implied by rf.
* @param[in] rf The rotation by mulitples of 90 degrees and/or right-left or up-down flip.
* @param[in] subout Dimensions of the region after application of the rotation/flip.
* @since 2019-11-05
* @since 2020-03-06 augmented to work with general Rotflip arguments and float pointer farr
* @author Richard J. Mathar
*/
void FitsImg::flip(float * farr, const long axs[3], const Rotflip &rf, const Subwin & subout)
{ 
	/* If there is no rotation nor flip: nothing to be done; farr remains unchanged.
	*/
	if ( rf.rot90 ==0 && rf.flipxy ==0 )
		return ;

	const Subwin subin(0,0,axs[0],axs[1]) ;
	/* Array of transformed image pixel values.
	* The same number of pixels as the original.
	* We cannot use in-place 2-element swaps if there are rotations with order 4 (90 or 270 degrees).
	*/
	float * fout = new float[axs[0]*axs[1]*axs[2]] ;

	for(int slic=0 ; slic < axs[2] ; slic++)
	{
		/* Apply rf to this slice
		*/
		for(int x=0 ; x < axs[0] ; x++)
		{
			/* todo: could there be a pragma omp parallel for here?
			*/
			for(int y=0 ; y < axs[1] ; y++)
			{
				/* serialized pixel index in the source data
				*/
				const int src = x+axs[0]*(y+slic*axs[1]) ;
				/* compute new pixel 2D coordinates in rotated array
				*/
				int xy[2] = {x, y} ;
				subin.rotflipPix(xy, rf) ;

				/* serialized pixel index in the rotated array
				*/
				const int dest = xy[0]+subout.xsize*(xy[1]+slic*subout.ysize) ;
				fout[dest] = farr[src] ;
			}
		}
	}
	/* copy intermediate array back to original. Deallocate temporary storage.
	*/
	memcpy(farr, fout, axs[0]*axs[1]*axs[2]*sizeof(float)) ;
	delete [] fout ;
} /* FitsImg::flip */

/*!*************************************
* Compare two integers in the support variant for qsort()
* @param[in] i The first numerical value.
* @param[in] j The second numerical value.
* @return +1, 0 or -1 depending on whether the contents of i (as a float)
*   is larger than, equal to or smaller than the contents of j.
* @since 2019-11-08
* @author R. J. Mathar
* @todo replace this a call to Sutil::qsortCmpFlt
*/
int FitsImg::qsortCmpFlt(const void *i, const void *j)
{
	if ( *((const float*) i) > *((const float*) j) )
		return 1 ;
	else if ( *((const float*) i) == *((const float*) j) )
		return 0 ;
	else
		return -1 ;
} /* qsortCmpFlt */
