/** @file
* @brief A parametrization of the group operations of multiple 90 degree rotations and flips.
* This is a group of order 8 where the generators are the rotation by 90 degrees
* and a flip along the horizontal or vertical axis. These operations can be performed
* on the pixel arrays of (mosaics of) detectors of rectangular shape by introducing an intervening
* 1-to-1 index map.
*/

/*
* MPIA - GEIRS project
*
* $Header: https://svn.mpia.de/gulli/geirs/src/trunk/Rotflip.cxx 792 2020-03-16 10:58:44Z mathar $
*
*/

#include "config.h"
#include <cstring>

#include <cstdio>

#include "Rotflip.h"

using namespace std ;

/*!**************************************** 
* @brief Default constructor.
* This defines a transformation which is unknown.
* @since 2015-01-24
* @author Richard J. Mathar
*/
Rotflip::Rotflip() 
	: rot90(-1), flipxy(-1)
{
	/* no normalize() here because we keep the negative values to indicate non-initialized
	* status
	*/
} /* Rotflip::Rotflip */

/*!**************************************** 
* @brief Constructor with explicit number of rotations and flips.
* @param rot a number from 0 to 3 giving the number of right-angle rotations
* @param flips A number from 0 to 2. 0=no flip. 1=right-left flip. 2=up-down flip.
* @since 2015-01-24
* @author Richard J. Mathar
*/
Rotflip::Rotflip(int rot, int flips) 
	: rot90(rot), flipxy(flips)
{
	normalize() ;
} /* Rotflip::Rotflip */

#ifdef HAVE_CCFITS_FITS_H
/*!**************************************** 
* @brief Constructor reading a primary FITS HDU.
*
* @param ffil The FITS file to be scanned
* @since 2015-05-26
* @author Richard J. Mathar
*/
Rotflip::Rotflip(FITS & ffil) : rot90(0), flipxy(0)
{
	PHDU & phdu = ffil.pHDU() ;

	/* try to catch DETXYFLI in the primary header.
	*/
	phdu.readAllKeys() ;
	map<String, Keyword*> & allkw = phdu.keyWord() ;
	map<String, Keyword*>::iterator kwitr = allkw.begin() ;
	for( ; kwitr != allkw.end() ; kwitr++)
	{
		const string & thisk = kwitr->first ;
		// cerr <<__FILE__ << " " << __LINE__ << " " << thisk << endl ;
		// for some reason, these strings do not contain the "HIERARCH" leading part (?)
		const Keyword * thisv = kwitr->second ;

		int v ;
		if ( thisk.find("DETROT90") != string::npos )
		{
			/* CAM_DETROT90 value */
			rot90 = thisv->value(v) ;
		}

		if ( thisk.find("DETXYFLI") != string::npos )
		{
			/* CAM_DETXYFLIP specification */
			flipxy = thisv->value(v) ;
		}
	}

	normalize() ;

} /* Rotflip::Rotflip */
#endif /* HAVE_CCFITS_FITS_H */

/*!****************************************
* @brief Map all possible combinations of 4 rotation and 3 flip parameters to common set.
* This means if the flipxy parameter is nonzero, the rotation component is reduced to 0 or 1
* by an equivalence transformation in the dihedral group.
* @since 2015-11-06
* @author Richard J. Mathar
*/
void Rotflip::normalize()
{
	/* standardize the rotations to a value from 0 to 3
	* If they are in the range -3 to -1.
	*/
	if ( rot90 < 0)
		rot90 += 4 ;

	rot90 %= 4 ;
	flipxy &= 3 ;

	if ( flipxy == 3)
	{
		/* replace combined flips by a single rotation */
		flipxy=0 ;
		rot90 = (rot90 +2) %4 ;
	}


	/* normalize further to one of the 8 elements of the dihedral group
	* rot90=2 flip = 1 becomes rot90=0 flip =2
	* rot90=3 flip = 1 becomes rot90=1 flip =2
	* rot90=2 flip = 2 becomes rot90=0 flip =1
	* rot90=3 flip = 2 becomes rot90=1 flip =1
	* Default for instruments defined in GENERIC startup script (values of 0 if CAM_DETROT90 or CAM_DETXYFLIP not set)
	*/
	if ( rot90 > 1 && flipxy > 0)
	{
		flipxy = 3-flipxy ;
		rot90 = (rot90 +2) %4 ;
	}
} /* Rotflip::normalize */

/*!**************************************** 
* @brief Construct the inverse of the operation.
*
* This implies (because the rotation and flip don't commute) that
* the inverse flip is done first, then the inverse rotation.
* @since 2015-01-24
* @author Richard J. Mathar
*/
Rotflip Rotflip::inverse() const
{
	/* most of the 8 relevant operations of the group are self-inverse.
	* The exceptions are the pure rotations by 90 or 270 degrees.
	*/
	if ( flipxy == 0 && (rot90 %2 ) != 0 )
	{
		return Rotflip(4-rot90,0) ;
	}
	else
		return *this ;

} /* Rotflip::inverse */

/*!**************************************** 
* @brief Construct the operation as if occuring with the other handedness.
*
* This means that the operation is transformed from the X11 (left-handed)
* to the FITS (right) coordinate system or vice versa.
* @since 2015-01-24
* @author Richard J. Mathar
*/
Rotflip Rotflip::toFits() const
{
	/* The only thing needed is to complement the angle,
	* which means reversing is sense which is the same as taking the
	* 4-complement.
	* The flips are "neutral" in this transformation.
	*/
	return Rotflip(4-rot90,flipxy) ;
} /* Rotflip::toFits */

/*!************************************************
* @brief Comparison of two image rotation-flip parameter sets.
*
* @return True (considered equal) if the number of rotations and the flips agree.
*   Since the constructor normalizes the parameters to a unique set of
*   eight possible pairs of parameters, the image transformations can
*   be considered equal if the two integer parameters are equal.
*/
bool operator== (const Rotflip & first , const Rotflip & secnd)
{
        return ( first.flipxy == secnd.flipxy && first.rot90 == secnd.rot90 ) ;
} /* Rotflip::operator== */

/*!************************************************
* @brief Comparison of two image rotation-flip parameter sets.
*
* @return True (considered unequal) of the number of rotations or the flips disagree.
*/
bool operator!= (const Rotflip & first , const Rotflip & secnd)
{
        return ( first.flipxy != secnd.flipxy || first.rot90 != secnd.rot90 ) ;
} /* Rotflip::operator!= */
