/*!*****************************************************
* @file
* @brief The class defines a rectangular area in pixel units.
*
* The operations offered are of geometric type: intersection, deciding
* if a point is in- or outside, supporting the roto-flip-transformations,
* deciding whether a point is in the 4-pixel reference pixel frame of a Hawaii2 RG.
*/

/*
* MPIA - GEIRS project
*
* $Header: https://svn.mpia.de/gulli/geirs/src/trunk/Subwin.cxx 792 2020-03-16 10:58:44Z mathar $
*
*/

#include <cstring>
#include <cstdio>
#include <sstream>
#include <iostream>

#include "Subwin.h"

using namespace std ;

/*!**************************************** 
* @brief Default constructor.

* This defines a window of zero area (height and width zero).
* @since 2013-10-30
* @author Richard J. Mathar
*/
Subwin::Subwin() : x1(0), y1(0), xsize(0), ysize(0)
{
} /* Subwin::Subwin */

/*!**************************************** 
* @brief Constructor with explicit layout of dimensions.

* @param xll 0-based lower left edge x-coordinate.
* @param yll 0-based lower left edge y-coordinate.
* @param xwidth Pixel width along x.
* @param yheight Pixel width along y.
* @since 2013-10-30
* @author Richard J. Mathar
*/
Subwin::Subwin(int xll, int yll, int xwidth, int yheight) : x1(xll), y1(yll), xsize(xwidth), ysize(yheight)
{
} /* Subwin::Subwin */


#ifdef HAVE_CCFITS_FITS_H
/*!**************************************** 
* @brief Constructor taken the dimensions from the DETSEC of a FITS header.

* @param iex The image extension with a DETSEC string.
* @since 2015-05-02
* @author Richard J. Mathar
*/
Subwin::Subwin(ExtHDU & iex) : x1(0), y1(0), xsize(0), ysize(0)
{
	/* try to catch DETSEC in one of the HIERARCH modifications.
	* If there is no such thing, the member initializsation list takes over
	* and the result is an empty rectangle.
	*/
	iex.readAllKeys() ;
	map<String, Keyword*> & allkw = iex.keyWord() ;
	map<String, Keyword*>::iterator kwitr = allkw.begin() ;
	for( ; kwitr != allkw.end() ; kwitr++)
	{
		const string & thisk = kwitr->first ;
		const Keyword * thisv = kwitr->second ;
		if ( thisk.find("DETSEC") == 0 )
		{
			/* window specification, '[x1:x2,y1:y2]' */
			string recta ;
			thisv->value(recta) ;
			/* run Subwin ctor with a string of that type
			*/
			init(recta) ;
		}
	}
} /* Subwin::Subwin */
#endif /* HAVE_CCFITS_FITS_H */

/*!**************************************** 
* @brief Constructor with explicit layout of dimensions.
*
* @param fts A string in the standard FITS [x1:x2,y1:y2] format
* @since 2014-04-16
* @since 2015-02-18 downsize windows with improper x2<x or y2<y1 coordinates to 0 area.
* @author Richard J. Mathar
*/
Subwin::Subwin(const string & fts)
{
	init(fts) ;
} /* Subwin::Subwin */

/*!**************************************** 
* @brief Constructor with explicit layout of dimensions.

* @param fts A string in the standard FITS [x1:x2,y1:y2] format
* @since 2014-04-16
* @since 2015-02-18 downsize windows with improper x2<x or y2<y1 coordinates to 0 area.
* @author Richard J. Mathar
*/
void Subwin::init(const string & fts)
{
	int coo[4] ;
	/* eliminate everything but digits and the minus sign
	*/
	char * cfts = new char[fts.length()+1] ;
	cfts[fts.length()] = '\0' ;
	for(int i=0 ; i < (int) fts.length() ; i++)
	{
		if ( isdigit(fts[i]) || fts[i] == '-')
			cfts[i]=fts[i] ;
		else
			cfts[i]=' ' ;
	}
	/* at this point a specification with more than two colons or anything
	* similar would also pass: [2:300,5:100:2] will be rendered as [2:300,5:100] for example
	*/
	istringstream parsrecta(cfts) ;
	parsrecta >> coo[0] >> coo[1] >> coo[2] >> coo[3] ;
	delete [] cfts ;

	/* note that coo[2] and coo[3] will be zero if the user forgot the
	* quotation marks in the FITS template line... TODO: catch this
	*/

	/* convert to 0-based format
	*/
	x1 = coo[0]-1 ;
	y1 = coo[2]-1 ;
	/* if dimensions appear to be negative, actually move them up to 0 so
	* the window will be flagged to have 0 pixels inside most of the other functionality
	*/
	xsize = max(0,coo[1]-coo[0]+1) ;
	ysize = max(0,coo[3]-coo[2]+1) ;

} /* Subwin::Subwin */

/*!**************************************** 
* @brief Number of pixels in the window.

* @return The product of pixels along x and along y.
* @since 2013-11-03
* @author Richard J. Mathar
*/
int Subwin::area() const
{
	return xsize*ysize ;
} /* Subwin::area */

/*!**************************************** 
* @brief Flags whether a pixel is inside this window.

* @return True if the pixel (x,y) is inside the window.
* @since 2014-06-23
* @author Richard J. Mathar
*/
bool Subwin::contains(int x, int y) const
{
	return (x >= x1 && x < x1+xsize && y >= y1 && y < y1+ysize) ;
} /* Subwin::contains */

/*!**************************************
* @brief Compute the common overlap with another window.

* @param oth The other window to be intersected with this.
* @return The window with the common area.
*  If there is no overlap, the area of the new window will be zero.
* @since 2013-11-03
* @author Richard J. Mathar
*/
Subwin Subwin::intersection(const Subwin &oth) const
{
	/* bbox[0] = lower left coordinates of intersection
	* bbox[1] = upper right coordinates of intersection.
	* Upper right coordinates are exclusive (the maximum allowed index is 1 less).
	*/
	int bbox[2][2] ;

	/* x coordinate lower left */
	bbox[0][0] = max(x1,oth.x1) ;
	/* y coordinate lower left */
	bbox[0][1] = max(y1,oth.y1) ;
	/* x coordinate upper right */
	bbox[1][0] = min(x1+xsize,oth.x1+oth.xsize) ;
	/* y coordinate upper right */
	bbox[1][1] = min(y1+ysize,oth.y1+oth.ysize) ;

	/* there is some overlap if the x upper right coordinate is larger than the
	* x lower left coordinate and the y upper right coordinate is larger than the
	* y lower left coordinate.
	*/
	if ( bbox[1][0] > bbox[0][0] && bbox[1][1] > bbox[0][1])
		return Subwin(bbox[0][0], bbox[0][1], bbox[1][0]-bbox[0][0], bbox[1][1]-bbox[0][1]) ;
	else
		return Subwin() ;
} /* Subwin::intersection */

/*!*********************************************
* @brief Decide whether the window is entirely inside the reference px frame.
*
* The coordinates are reduced modulo 2048 or 4096 (the Hawaii2/4-RG pixel size)
* and intersected with the inner 2040x2040 or 4092x402 pixels. If the intersection is
* empty, return true.
* This is mainly of interest to filter the filler subwindows in srre mode.
* @param[in] isH4rg If true this is a Hawaii-4RG, otherwise a Hawaii-2RG.
* @return true If there is no intersection with the inner pixels.
*/
bool Subwin::inHaw2rgRef(const bool isH4rg) const
{
	const int chipsz = isH4rg ? h4rgsize : h2rgsize ;

	/* a window shifted by multiples of the Hawaii2RG size
	*/
	const Subwin shfted(x1 % chipsz, y1 % chipsz, xsize, ysize) ;
	/* the inner frame of the "science" pixels of the Hawaii2RG
	*/
	const Subwin h2scipix(4,4,chipsz-4,chipsz-4) ;
	/* area of the intersection in units of pixels.
	*/
	const int overlpx = shfted.intersection(h2scipix).area() ;
	/* false (=not in the reference pixels) if the overlap is positive
	*/
	return ( overlpx > 0) ? false : true ;
} /* inHaw2rgRef */

/*!*********************************************
* @brief Remove pixels in the reference regions.
*
* The coordinates are reduced modulo 2048 or 4096 (the Hawaii2/4-RG pixel size)
* and intersected with the inner 2040x2040 or 4092x4092 pixels. Only the intersection with
* that inner pixel region is returned.
* @param[in] isH4rg If true this is a Hawaii-4RG, otherwise a Hawaii-2RG.
* @return The intersection of this window (supposed not to cross boundaries) with the inner DETSEC.
* @since 2015-10-12
*/
Subwin Subwin::clipRef(const bool isH4rg) const
{
	const int chipsz = isH4rg ? h4rgsize : h2rgsize ;
	/* The location of this window in a square grid with 2048x2048 granularity
	* This would be detx=0,1 and dety=0 for the CARMENES standard, and detx=dety=0 for Luci's.
	*/
	const int detx = x1/chipsz ;
	const int dety = y1/chipsz ;

	/* the inner frame of the "science" pixels of the Hawaii2RG
	* which we want to keep. Is smaller by 8 pixels in width and height, 2040x2040 or 4092x4092
	*/
	const Subwin h2scipix(4+detx*chipsz,4+dety*chipsz,chipsz-8,chipsz-8) ;

	/* pixels in the intersection.
	*/
	return  intersection(h2scipix) ;
} /* clipRef */

/*!*********************************************
* @brief Decide whether a pixel is inside the reference px frame.
*
* The coordinates are reduced modulo 2048 (the Hawaii2-RG pixel size)
* and intersected with the inner 2040x2040 pixels. If the intersection is
* empty, return true.
* @param x The 0-based x coordinate of the pixel
* @param y The 0-based y coordinate of the pixel
* @param isH4rg If true this is a Hawaii-4RG, otherwise a Hawaii-2RG.
* @return true If the pixel is outside the 2040x2040 core pixels of the H2RG chips.
*/
bool Subwin::inHaw2rgRef(int x, int y, const bool isH4rg)
{
	const int chipsz = isH4rg ? h4rgsize : h2rgsize ;
	x = (x+ chipsz) % chipsz ;
	y = (y+ chipsz) % chipsz ;
	if ( x < 4 || x >= chipsz-4 || y < 4 || y >= chipsz-4)
		/* 0, 1, 2, 3 and 2044, 2045, 2046, 2047 are in
		*/
		return true;
	else
		return false;
} /* inHaw2rgRef */

/*!**************************************
* @brief Compute the window which is the common bounding box of this and another.

* @param oth The other window to be combined with this.
* @return The window which unites both rectangular windows.
* @since 2014-03-27
* @author Richard J. Mathar
*/
Subwin Subwin::boundBox(const Subwin &oth) const
{
	/* bbox[0] = lower left coordinates of bbox
	* bbox[1] = upper right coordinates of bbox.
	* Upper right coordinates are exclusive (the maximum allowed index is 1 less).
	*/
	int bbox[2][2] ;

	/* x coordinate lower left */
	bbox[0][0] = min(x1,oth.x1) ;
	/* y coordinate lower left */
	bbox[0][1] = min(y1,oth.y1) ;
	/* x coordinate upper right */
	bbox[1][0] = max(x1+xsize,oth.x1+oth.xsize) ;
	/* y coordinate upper right */
	bbox[1][1] = max(y1+ysize,oth.y1+oth.ysize) ;

	/* there is some overlap if the x upper right coordinate is larger than the
	* x lower left coordinate and the y upper right coordinate is larger than the
	* y lower left coordinate.
	*/
	if ( bbox[1][0] > bbox[0][0] && bbox[1][1] > bbox[0][1])
		return Subwin(bbox[0][0], bbox[0][1], bbox[1][0]-bbox[0][0], bbox[1][1]-bbox[0][1]) ;
	else
		return Subwin() ;
} /* Subwin::boundBox */

/*!**************************************
* @brief Compute the window rotated and/or flipped.
*
* Modify the corner coordinates of this window as if they were actively rotated
* counter-clock-wise a multiple of 90 degrees around the center of a host frame.
* For non-square windows, the algebra assumes that (for 90 or 270 degree rotations)
* the host frame is co-rotated, such that the new coordinates refer to the lower left
* corner of the rotated host-frame.
* The name of the function refers to the alternative interpretation to rotate the
* window clock-wise (to the right) if all window coordinates are measured in a
* left-handed coordinate system where x points to the right and y down.
*
* If flipxy is nonzero:
*  The four corners of the window (represented by lower left x and y coordinate
*  and edge lengths) are manipulated to represent a window that is flipped
*  around one of the center lines of the host window (represented by its x and y
*  dimensions).
*
* @param detrot90 The number of 90 degree left rotations of this window.
*   Basically this should be a number from 0 up to 3 (inclusive).
* @param flipxy If the LSB is set, flip additionally right-left.
*   If the second bit is set, flip additionally up-down. This value should
*   be a number from 0 up to 2 (inclusive). A value of 3 is reinterpreted
*   as an additional 180 degree rotation.
* @param parent The parent window in which the coordinates of the current
*   window are measured.
* @return The window's coordinates if first rotated than flipped around
*   the center of the parent window. For parent windows that are not of
*   square shape, the new window coordinates are obtained by rotating the
*   parent window (with this window hung fixed to it), and then measuring
*   this window with respect to the lower left (rotated...) coordinate of
*   the parent window.
* @since 2014-02-10
* @author Richard J. Mathar
*/
Subwin Subwin::rotflip(int detrot90, int flipxy, const Subwin & parent) const
{
	/* xyll[0] x coordinate of the new lower left  window corner, xyll[1]
	* the y coordinate of the new lower left window corner.
	*/
	int xyll[2] ;

	/* 12 different combinations with 8 different results:
	* Rotations always first.
	AB  detrot90=0 flipxy=0 AB
	CD                      CD
	AB  detrot90=1 flipxy=0 BD
	CD                      AC
	AB  detrot90=2 flipxy=0 DC
	CD                      BA
	AB  detrot90=3 flipxy=0 CA
	CD                      DB
	right-left-flips plus 4 cases above:
	AB  detrot90=0 flipxy=1 BA
	CD                      DC
	AB  detrot90=1 flipxy=1 DB
	CD                      CA
	AB  detrot90=2 flipxy=1 CD
	CD                      AB
	AB  detrot90=3 flipxy=1 AC
	CD                      BD
	up-down-flips plus 4 cases above:
	AB  detrot90=0 flipxy=2 CD
	CD                      AB
	AB  detrot90=1 flipxy=2 AC
	CD                      BD
	AB  detrot90=2 flipxy=2 BA
	CD                      DC
	AB  detrot90=3 flipxy=2 DB
	CD                      CA
	Actually each of the 4 in the last group is the same as a further
	* rotation by 180 degrees and a flipxy=1.
	*/

	/* Reduce flipxy to two bits. Replace flipxy =3 by another two turns.
	*/
	flipxy &= 3  ;
	if ( flipxy == 3)
	{
		flipxy =0 ;
		detrot90 += 2 ;
	}

	/* The effect of the rotation studied first. Reduce the 90 degree turns modulo 4.
	*/
	switch (detrot90 % 4)
	{
	case 0 :
		/* new lower left corner is old lower left corner */
		xyll[0] = x1 ;
		xyll[1] = y1 ;
		break ;
	case 1 :
		/* new lower left corner is old upper left corner
		* measured relative to old upper left parent corner */
     /* 90 degrees... The old upper left corner becomes the new lower left corner.
     * After 90 deg left:

-----.----
|  DA.    |
|    .    |
|    .    |
|    .    |
|  CD.    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
..........|
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
----------
     */
		xyll[0] = parent.ysize - (y1+ysize) ;
		xyll[1] = x1 ;
		break ;
	case 2 :
		/* new lower left corner is old upper right corner
		* measured relative to old upper right parent corner */
     /* 180 degrees...
     * The old upper right corner becomes the new lower left corner
     * after 180

----------------------.--------------------
|                     .                   |
|                     .                   |
|                     .                   |
|                     .                   |
..........................................
|    A    B           .                   |
|    D    C           .                   |
|                     .                   |
|                     .                   |
----------------------.--------------------
     * again the interpretation in a coordinate system with upper left corner
     * of the host window at (0,0) with x=right and y=down is a 180 degree
     * rotation to the right.
     */
		xyll[0] = parent.xsize - (x1+xsize) ;
		xyll[1] = parent.ysize - (y1+ysize) ;
		break ;
	default :
		/* case 3: new lower left corner is old lower right corner
		* measured relative to old lower right parent corner */
     /* 270 degrees or -90 degrees.
     * The rotation around the center of coordinates (!) is
     * new x = old y and new y = negative old x.
     * The former lower right corner A of the rectangle at x=ox+swnd.xsize becomes
     * the new lower left corner A of the rotated rectangle. Furthermore we assume for
     * host frames where ysz!=xsz that the host frame is also rotated, such that
     * the old vertical distance to the host edges become the new horizontal distances
     * to the host edges and vice versa. The coordinates are measured anyway with
     * the origin relative to the host frames lower left coordinate:
     * before
^y
----------------------.--------------------
|                     .                   |
|                     .                   |
|                     .           C    D  |
|                     .           B    A  |
..........................................
|                     .                   |
|                     .                   |
|                     .                   |
|                     .                   |
----------------------.-------------------->x

     * after 270 ccw

^y
-----.----
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
..........|
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .BC  |
|    .    |
|    .    |
|    .AD  |
|    .    |
---------- >x
     * There is also an entirely different interpretation with the same
     * arithmetics (implemented exactly in the same formulas): We may assume
     * that the original coordinates are with x=right and y=down (as in X11
     * notation or as with the standard display of a H2RG with fanout up,
     * fast direction =x, and slow direction =y: then this is a rotation by
     * 270 clockwise (to the right!), where C is the "lower left" (minimum ox,oy)
     * corner of the original window and D is the "lower left" corner of the new 
     * window:

----------------------.--------------------> x
|                     .                   |
|                     .                   |
|                     .           C    D  |
|                     .           B    A  |
..........................................
|                     .                   |
|                     .                   |
|                     .                   |
|                     .                   |
----------------------.--------------------
\/ y


-----.---- > x
|  DA.    |
|    .    |
|    .    |
|    .    |
|  CB.    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
..........|
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
|    .    |
---------- 
\/ y
*/
		xyll[0] = y1 ;
		xyll[1] = parent.xsize-(x1+xsize) ;
	}

	/* this flag indicates whether width and height of this
	* window are swapped in the image. If the number of 90 degree
	* rots is odd, the operation creates a swapped ratio. Further
	* flips don't change that.
	*/
	const bool wh_swap = (detrot90 % 2) ? true : false ;

	/* Then add the effect of the left-right or up-down flip within the
	* potentially rotated parent window. Case flipxy=0 does not
	* have an effect, and case flipxy=3 has already been transformed
	* above to a rotation. Two cases of single rotations remain.
	*/
	switch (flipxy & 3 )
	{
	case 1 :
		/* right-left flip: new lower left corner is old lower left
		* corner measured relative to new right parent edge. Without
		* a previous rotation, xsize of parent and this have remained.
		* With a previous rotation, xsize and ysize of parent and
		* this have already swapped meaning.
		* xyll[1] is not changed. */
		xyll[0] = wh_swap ? ( parent.ysize - (xyll[0]+ysize) )
				: ( parent.xsize - (xyll[0]+xsize) ) ;
		break ;
	case 2 :
		/* up-down flip: new lower left corner is old upper left corner
		* measured relative to new upper parent edge.
		* xyll[0] is not changed. As mentioned above, this
		* could be eliminated by adding 2 to detrot90 and using
		* case 1 instead.... */
		xyll[1] = wh_swap ? ( parent.xsize - (xyll[1]+xsize) )
				: (parent.ysize - (xyll[1]+ysize) ) ;
	}

	if ( wh_swap)
		return Subwin(xyll[0], xyll[1], ysize, xsize) ;
	else
		return Subwin(xyll[0], xyll[1], xsize, ysize) ;
} /* Subwin::rotflip */

/*!**************************************
* @brief Compute the window rotated and/or flipped.
*
* @param Rotflip The definition of angle and flip of the rotation.
* @param parent The parent window in which the coordinates of the current
*   window are measured.
* @return The window's coordinates if first rotated than flipped around
*   the center of the parent window.
* @since 2015-01-24
* @author Richard J. Mathar
*/
Subwin Subwin::rotflip(const Rotflip & rf, const Subwin & parent) const
{
	return rotflip(rf.rot90, rf.flipxy, parent) ;
} /* Subwin::rotflip */

/*!**************************************
* @brief Compute the window rotated and/or flipped.
*
* The operation is the inverse operation of rotflip(). This means the inverse
* flip is done first, followed by the inverse rotation, given that the 
* operations of rotation and flip are not commutative.
* The result of using rotflipInv() followed by rotflip() or vice versa
* using the same three (!) arguments gives the original window.
* A short table of the rot-parameter, the flip parameter followed in brackets
* by the associated rot and flip parameter that (if the followup action)
* is the inverse:
*            0, 0, [0, 0]
*            0, 1, [0, 1]  or [2,2]
*            0, 2, [0, 2]  or [2,1]
*            1, 0, [3, 0]
*            1, 1, [1, 1]  or [3,2]
*            1, 2, [1, 2]  or [3,1]
*            2, 0, [2, 0] 
*            2, 1, [0, 2]  or [2,1]
*            2, 2, [0, 1]  or [2,2]
*            3, 0, [1, 0]
*            3, 1, [1, 2]  or [3,1]
*            3, 2, [1, 1]  or [3,2]
* So the simplest rule seems to be:
* If the detrot90 is even, the inverse is the same combined operation.
* If the detrot90 is odd, the inverse is the operation with the complementary angle and complimentary flip.
* (And these left inverses work also as right inverses)
* Alternatively:
* If the flip is 1 or 2, the operation is self-inverse,
* If the flip is 0 , the operation needs the complement of the angle.
*
* The combined operation (yflip)o(op)o(yflip) does not cancel the two
* y-flips at the start and end, but is equivalent to the operation where
* the angular complement of op is combined with its original flips: [r,f]->[(4-r)mod 4,f].
* This is the reason why the cw/ccw equivalence in the X11 and FITS
* rotations is observed.
*
* @param detrot90 The number of 90 degree left rotations of this window
*   in the forward operation. This becomes a number of right rotations
*   in this inverse.
*   Basically this should be a number from 0 up to 3 (inclusive).
* @param flipxy If the LSB is set, flip additionally right-left.
*   If the second bit is set, flip additionally up-down. This value should
*   be a number from 0 up to 2 (inclusive). A value of 3 is reinterpreted
*   as an additional 180 degree rotation.
* @param parent The parent window in which the coordinates of the inversed
*  window are measured. Important for non-square parent windows:
*  This means the window is the same window as used for the forward transform;
*  it is the window after undoing the rotation.
* @return The window's coordinates if first flipped around
*   the center of the parent window, then rotated backwards (to the right)
*   combined with the parent window which then will become the 'parent'.
* @since 2014-02-11
* @author Richard J. Mathar
*/
Subwin Subwin::rotflipInv(int detrot90, int flipxy, const Subwin & parent) const
{
	/* norotxyll[0] x coordinate of the back-rotated lower left  window corner, norotxyll[1]
	* the y coordinate of the back-rotated lower left window corner.
	*/
	int norotxyll[2] ;

	/* Reduce flipxy to two bits. Replace flipxy =3 by another two turns.
	* So the case flipxy & 3 == 3 does not need to be implemented below.
	*/
	flipxy &= 3  ;
	if ( flipxy == 3)
	{
		flipxy =0 ;
		detrot90 += 2 ;
	}

	/* This flag indicates whether width and height of this
	* window are swapped in the image. If the number of 90 degree
	* rotations is odd, the operation creates a swapped ratio. Further
	* flips don't change that.
	*/
	const bool wh_swap = (detrot90 % 2) ? true : false ;

	/* Initialize for case flipxy =0 (including flipxy=3 already reduced above).
	*/
	norotxyll[0] = x1 ;
	norotxyll[1] = y1 ;

	/* First consider the effect of the left-right or up-down flip within the
	* potentially rotated parent window. Case flipxy=0 does not
	* have an effect, and case flipxy=3 has already been transformed
	* above to a rotation. Two cases of single rotations remain.
	*/
	switch (flipxy & 3 )
	{
	case 1 :
		/* right-left flip: original lower left corner is lower left
		* corner measured relative to rotated right parent edge. Without
		* a previous rotation, wh_swap=false, xsize of parent and of this have remained.
		* With a previous rotation, xsize and ysize of parent and
		* this have swapped values.  norotxyll[1] is not changed.
		*  ----------------------------
		*  |                          |
		*  |                          |
		*  |                          |
		*  |                ---       |  dots=norotxyll after flip
		*  |                | |       |
		*  |                .--       |
		*  |.................         |
		*  ----------------------------
		*
		*  ----------------------------
		*  |                          |
		*  |                          |
		*  |                          |
		*  |    ---                   |
		*  |    | |                   |
		*  |    --....................|
		*  |                         .|
		*  ----------------------------
		*/
		norotxyll[0] = (wh_swap ? parent.ysize: parent.xsize) - (x1+xsize) ;
		break ;
	case 2 :
		/* up-down flip: original lower left corner is upper left corner
		* measured relative to rotated upper parent edge.
		* norotxyll[0] is not changed. As mentioned above, this
		* could be eliminated by adding 2 to detrot90 and using
		* case 1 instead....
		*  ----------------------------
		*  |                          |
		*  |                          |
		*  |                          |
		*  |                ---       | dots=norotxyll after flip
		*  |                | |       |
		*  |                .--       |
		*  |.................         |
		*  ----------------------------
		*
		*  ----------------------------
		*  |.................         |
		*  |                .--       |
		*  |                | |       |
		*  |                ---       |
		*  |                          |
		*  |                          |
		*  |                          |
		*  ----------------------------
		*/
		norotxyll[1] = (wh_swap ? parent.xsize : parent.ysize) - (y1+ysize) ;
	}

	int xyll[2] ;

	/* norotxyll now contains the non-flipped but rotated coordinates
	* in the rotated parent window. The effect of the rotation is undone next.
	*/
	switch (detrot90 % 4)
	{
	case 0 :
		/* original lower left corner is lower left corner
		*/
		xyll[0] = norotxyll[0] ;
		xyll[1] = norotxyll[1] ;
		break ;
	case 1 :
		/* Lower right corner measured relative to current parent lower right corner
		* is original lower left corner measured relative to original lower left corner.
		*/
		xyll[0] = norotxyll[1] ;
		xyll[1] = (wh_swap ? parent.ysize : parent.xsize) - (norotxyll[0]+xsize) ;
		break ;
	case 2 :
		/* Upper right corner measured relative to current parent upper right corner
		* is original lower left corner measured relative to parent lower left corner.
		*/
		xyll[0] = (wh_swap ? parent.ysize : parent.xsize) - (norotxyll[0]+xsize) ; 
		xyll[1] = (wh_swap ? parent.xsize : parent.ysize) - (norotxyll[1]+ysize) ;
		break ;
	default :
		/* case 3, 270 deg: Upper left corner relative to current parent upper left corner
		* is original lower left corner measured relative to parent lower left corner.
		*/
		xyll[0] = (wh_swap ? parent.xsize : parent.ysize) - (norotxyll[1]+ysize) ;
		xyll[1] = norotxyll[0] ;
	}

	if ( wh_swap)
		return Subwin(xyll[0], xyll[1], ysize, xsize) ;
	else
		return Subwin(xyll[0], xyll[1], xsize, ysize) ;
} /* Subwin::rotflipInv */

/*!**************************************
* @brief Compute the window rotated and/or flipped.
*
* @param[in] rf The combination of rotation and flip (of the original transform).
* @param parent The parent window in which the coordinates of the inversed
*  window are measured. Important for non-square parent windows:
*  This means the window is the same window as used for the forward transform;
*  it is the window after undoing the rotation.
* @return The window's coordinates if first flipped around
*   the center of the parent window, then rotated backwards (to the right)
*   combined with the parent window which then will become the 'parent'.
* @since 2015-01-24
* @author Richard J. Mathar
*/
Subwin Subwin::rotflipInv(const Rotflip &rf, const Subwin & parent) const
{
	return rotflipInv(rf.rot90, rf.flipxy, parent) ;
}

/*!**************************************
* @brief Compute the new location of a pixel after rotation and/or flipping within this window.
*
* @param[in,out] xy The two integer coordinates of the pixel.
*  On return replaced by the rotated/flipped coordinates.
* @param[in] detrot90 The number of 90 degree left rotations in this window.
*   Basically this should be a number from 0 up to 3 (inclusive).
* @param[in] flipxy If the LSB is set, flip additionally right-left.
*   If the second bit is set, flip additionally up-down. This value should
*   be a number from 0 up to 2 (inclusive). A value of 3 is reinterpreted
*   as an additional 180 degree rotation.
* @return The window's coordinates if first rotated then flipped around
*   the center of this window. For parent windows that are not of
*   square shape, the new window coordinates are obtained by rotating the
*   parent window (with this window hung fixed to it), and then measuring
*   this window with respect to the lower left (rotated...) coordinate of
*   the parent window.
* @since 2014-02-11
* @author Richard J. Mathar
*/
void Subwin::rotflipPix(int xy[2], int detrot90, int flipxy) const
{
	/* The implementation is basically a copy of rotflip() above assuming
	* that a window of xsize X ysize = 1X1 size is rotated/flipped inside a parent window,
	* and that this window here *is* the parent window.
	*/
	int xyll[2] ;

	/* Reduce flipxy to two bits. Replace flipxy =3 by another two turns.
	*/
	flipxy &= 3  ;
	if ( flipxy == 3)
	{
		flipxy =0 ;
		detrot90 += 2 ;
	}

	/* The effect of the rotation studied first. Reduce the 90 degree turns modulo 4.
	*/
	switch (detrot90 % 4)
	{
	case 0 :
		/* new lower left corner is old lower left corner */
		xyll[0] = xy[0] ;
		xyll[1] = xy[1] ;
		break ;
	case 1 :
		/* new lower left corner is old upper left corner
		* measured relative to old upper left parent corner */
		xyll[0] = ysize - (xy[1]+1) ;
		xyll[1] = xy[0] ;
		break ;
	case 2 :
		/* new lower left corner is old upper right corner
		* measured relative to old upper right parent corner */
		xyll[0] = xsize - (xy[0]+1) ;
		xyll[1] = ysize - (xy[1]+1) ;
		break ;
	default :
		/* case 3: new lower left corner is old lower right corner
		* measured relative to old lower right parent corner */
		xyll[0] = xy[1] ;
		xyll[1] = xsize-(xy[0]+1) ;
	}

	/* this flag indicates whether width and height of this
	* window are swapped in the image. If the number of 90 degree
	* rots is odd, the operation creates a swapped ratio. Further
	* flips don't change that.
	*/
	const bool wh_swap = (detrot90 % 2) ? true : false ;

	/* Then add the effect of the left-right or up-down flip within the
	* potentially rotated parent window. Case flipxy=0 does not
	* have an effect, and case flipxy=3 has already been transformed
	* above to a rotation. Two cases of single rotations remain.
	*/
	switch (flipxy & 3 )
	{
	case 1 :
		/* right-left flip: new lower left corner is old lower left
		* corner measured relative to new right parent edge. Without
		* a previous rotation, xsize of parent and this have remained.
		* With a previous rotation, xsize and ysize of parent and
		* this have already swapped meaning.
		* xyll[1] is not changed. */
		xy[0] = wh_swap ? ( ysize - (xyll[0]+1) )
				: ( xsize - (xyll[0]+1) ) ;
		xy[1] = xyll[1] ;
		break ;
	case 2 :
		/* up-down flip: new lower left corner is old upper left corner
		* measured relative to new upper parent edge.
		* xyll[0] is not changed. As mentioned above, this
		* could be eliminated by adding 2 to detrot90 and using
		* case 1 instead.... */
		xy[0] = xyll[0] ;
		xy[1] = wh_swap ? ( xsize - (xyll[1]+1) )
				: (ysize - (xyll[1]+1) ) ;
		break ;
	default:
		xy[0] = xyll[0] ;
		xy[1] = xyll[1] ;
	}

} /* Subwin::rotflipPix */

/*!**************************************
* @brief Compute the new location of a pixel after rotation and/or flipping within this window.
*
* @param[in,out] xy The two integer coordinates of the pixel.
*  On return replaced by the rotated/flipped coordinates.
* @param[in] rf The specification fo the rotation and flip.
* @return The window's coordinates if first rotated than flipped around
*   the center of this window. For parent windows that are not of
*   square shape, the new window coordinates are obtained by rotating the
*   parent window (with this window hung fixed to it), and then measuring
*   this window with respect to the lower left (rotated...) coordinate of
*   the parent window.
* @since 2015-01-24
* @author Richard J. Mathar
*/
void Subwin::rotflipPix(int xy[2], const Rotflip &rf ) const
{
	rotflipPix(xy, rf.rot90, rf.flipxy) ;
}

#ifdef UNUSED
/*!**************************************
* @brief Compute inversed rotation/flip of a pixel coordinate pair.
*
* The operation is the inverse operation of rotflipPix(). This means the inverse
* flip is done first, followed by the inverse rotation, given that the 
* operations of rotation and flip are not commutative.
* The result of using rotflipInv() followed by rotflip() or vice versa
* using the same three (!) arguments gives the original window.
* @param[in,out] xy The cartesian coordinates of the image pixel. On
*  return the cartesian coordinates of the original pixel before the roto-flip.
* @param detrot90 The number of 90 degree left rotations of this window
*   in the forward operation. This becomes a number of right rotations
*   in this inverse.
*   Basically this should be a number from 0 up to 3 (inclusive).
* @param flipxy If the LSB is set, flip additionally right-left.
*   If the second bit is set, flip additionally up-down. This value should
*   be a number from 0 up to 2 (inclusive). A value of 3 is reinterpreted
*   as an additional 180 degree rotation.
* @since 2014-02-11
* @author Richard J. Mathar
*/
void Subwin::rotflipPixInv(int xy[2], int detrot90, int flipxy) const
{
	/* The implementation is basically a copy of rotflipInv() but
	* assuming that the window has xsize X ysize = 1 X 1 size and that the
	* host (parent) window is this window here.
	*/
	/* norotxyll[0] x coordinate of the back-rotated lower left  window corner, norotxyll[1]
	* the y coordinate of the back-rotated lower left window corner.
	*/
	int norotxyll[2] ;

	/* Reduce flipxy to two bits. Replace flipxy =3 by another two turns.
	*/
	flipxy &= 3  ;
	if ( flipxy == 3)
	{
		flipxy =0 ;
		detrot90 += 2 ;
	}

	/* This flag indicates whether width and height of this
	* window are swapped in the image. If the number of 90 degree
	* rotations is odd, the operation creates a swapped ratio. Further
	* flips don't change that.
	*/
	const bool wh_swap = (detrot90 % 2) ? true : false ;

	/* Initialize for case flipxy =0 (including flipxy=3 already reduced above).
	*/
	norotxyll[0] = xy[0] ;
	norotxyll[1] = xy[1] ;

	/* First consider the effect of the left-right or up-down flip within the
	* potentially rotated parent window. Case flipxy=0 does not
	* have an effect, and case flipxy=3 has already been transformed
	* above to a rotation. Two cases of single rotations remain.
	*/
	switch (flipxy & 3 )
	{
	case 1 :
		norotxyll[0] = (wh_swap ? ysize: xsize) - (xy[0]+1) ;
		break ;
	case 2 :
		norotxyll[1] = (wh_swap ? xsize : ysize) - (xy[1]+1) ;
	}

	/* norotxyll now contains the non-flipped but rotated coordinates
	* in the rotated parent window. The effect of the rotation is undone next.
	*/
	switch (detrot90 % 4)
	{
	case 0 :
		/* original lower left corner is lower left corner
		*/
		xy[0] = norotxyll[0] ;
		xy[1] = norotxyll[1] ;
		break ;
	case 1 :
		/* Lower right corner measured relative to current parent lower right corner
		* is original lower left corner measured relative to original lower left corner.
		*/
		xy[0] = norotxyll[1] ;
		xy[1] = (wh_swap ? ysize : xsize) - (norotxyll[0]+1) ;
		break ;
	case 2 :
		/* Upper right corner measured relative to current parent upper right corner
		* is original lower left corner measured relative to parent lower left corner.
		*/
		xy[0] = (wh_swap ? ysize : xsize) - (norotxyll[0]+1) ; 
		xy[1] = (wh_swap ? xsize : ysize) - (norotxyll[1]+1) ;
		break ;
	default :
		/* case 3, 270 deg: Upper left corner relative to current parent upper left corner
		* is original lower left corner measured relative to parent lower left corner.
		*/
		xy[0] = (wh_swap ? xsize : ysize) - (norotxyll[1]+1) ;
		xy[1] = norotxyll[0] ;
	}

} /* Subwin::rotflipPixInv */
#endif /* UNUSED */

/*!**************************************
* @brief Count split windows if this window overlaps a set of others.
*
* @param oth The set of windows to be tested for overlap with this.
* @return The number of (non-empty) subwindows created by intersection of this with each of oth.
* @since 2013-11-03
* @author Richard J. Mathar
*/
int Subwin::countSplit(const vector<Subwin> & oth) const
{
	/* total count of subwindows
	*/
	int s =0 ;
	for (int w=0; w < (int) oth.size() ; w++)
	{
		/* area of the overlap of this with window number w of oth.
		*/
		const int ar = intersection(oth[w]).area() ;
		if ( ar > 0 )
			s++ ;
	}
	return s ;
} /* countSplit */

/*!**************************************
* @brief Detect whether the window crosses chip boundaries.
*
* @param chipsz The x and y size of each chip (pixels).
* @return True if any of the 4 corners of the window lie in different chips.
* @since 2014-03-06
* @author Richard J. Mathar
*/
bool Subwin::crossChip(const int chipsz) const
{
	/* detect whether the lower left and upper right edge
	* are the same modulo the chip size
	*/
	if ( x1/chipsz != (x1+xsize-1)/chipsz)
		return true;
	else if ( y1/chipsz != (y1+ysize-1)/chipsz)
		return true;
	else
		return false;
} /* crossChip */

/*!**************************************
* @brief Construct the windows that each cover one chip.
*
* @param chipsz The x and y size of each chip (pixels).
* @param ndetx The number of chips along x in the mosaic.
* @param ndety The number of chips along y in the mosaic.
* @return The ndetx*ndety windows, one per chip.
* @since 2013-11-03
* @author Richard J. Mathar
*/
vector<Subwin> Subwin::fullMosaWin(int chipsz, int ndetx, int ndety)
{
	/* enumeration here in output: x coordinate fast and y coordinate slow
	*   2  3
	*   0  1
	*/
	/* enumeration DET_HW_NAME
	*   [0]  [3]
	*   [1]  [2]
	*/
	vector<Subwin> w ;
	for (int y=0; y < ndety ; y++)
		for (int x=0; x < ndetx ; x++)
			/* construct a new subwindow at x*ndetx and y*ndety lower left
			*/
			w.push_back( Subwin(x*chipsz, y*chipsz, chipsz, chipsz) ) ;
	return w ;
} /* fullMosaWin */

/*!***********************************************
* @brief Construct a string [x1:x2,y1:y2] of the DATASEC.
*
*  The string is the same as the DETSEC but reduced by the 4 edge
*  pixels if the detector is of the Hawaii2-RG type.
*  The values are also reduced modulo the chip size.
* @param ishaw2rg Flag to indicate whether the chip is of Hawaii2-RG or another (Hawaii-2, Hawaii-4RG).
* @param chipsz The pixel size of each chip of the detector. 2048 or 4096 or Hawaii-2RG or Hawaii-4RG
* @return The standard 1-based 4-pixel range string in brackets.
* @since 2013-11-18
* @since 2014-02-26 Added chipsz parameter.
* @since 2019-11-06 full size for the Hawaii-4RG
* @author Richard J. Mathar
*/
string Subwin::datasec(const bool ishaw2rg, const int chipsz) const
{
	/* if this spreads across more than one chip, dont' produce
	* anything. Actually, one still could remove the 4 pixels of H2RG's..
	*/
	if ( crossChip(chipsz) )
		return string() ;

	/* xymimax[0][0] = minimum x, xymimax[0][1] = maximum x,
	* xymimax[1][0] = minimum y, xymimax[1][1] = maximym y.
	*/
	int xymimax[2][2] ;
	/* first create the 0-based values. So x1 and y1 are actually reached
	* as lower limits, and x1+xsize and y1+ysize are the lowest values not reached.
	*/
	xymimax[0][0] = x1 ;
	xymimax[0][1] = x1+xsize ;
	xymimax[1][0] = y1 ;
	xymimax[1][1] = y1+ysize ;

	if ( ishaw2rg)
	{
		/* for Hawaii-2RG or Hawaii-4RG Push values of 0,1,2,3 up to 4. Push values of 2045,2046,2047,2048 down to 2044.
		*/
		for(int i=0 ; i < 2 ; i++)
		{
			/* reduce any values (perhaps in chips as for PANIC) modulo 2048.
			*/
			int red = xymimax[i][0] % chipsz ;
			if ( red < 4 )
				/* convert lower limits of 0,1,2,3 to 4 */
				xymimax[i][0] += 4-red ;
			else if ( red >= chipsz-4)
				/* convert lower limits of 2044, 2045, 2046, 2047 to 2052 */
				xymimax[i][0] += chipsz+4 -red ;
			
			/* reduce any values (perhaps in chips as for PANIC) modulo 2048.
			*/
			red = xymimax[i][1] % chipsz ;
			if ( red >= chipsz-3 )
				/* convert upper limits of 2045, 2046, 2047 to 2044 */
				xymimax[i][1] -= red-(chipsz-4) ;
			else if ( red <= 4 )
				/* convert upper limits of 2048 (now red=0), 2049 (now 1), .. 2052 (now 4) to 2044 */
				xymimax[i][1] -= red+4 ;
		}
	}

	/* reduction modulo the chip size for compatibility
	* with the DATASEC convertion of IRAF. All coordinates still 0-based.
	*/
	for(int dir =0; dir <2 ; dir++)
	{
		xymimax[dir][0] %= chipsz ;
		/* need to wrap the values that are reached, not the value one larger....
		*/
		xymimax[dir][1] = 1 + (xymimax[dir][1]-1) % chipsz ;
	}

	/* 1-based FITS types; inclusive range */
	ostringstream ostr ;
	ostr << "[" << (xymimax[0][0]+1) << ":" << xymimax[0][1] << ","
		<< (xymimax[1][0]+1) << ":" << xymimax[1][1] << "]" ;
	return ostr.str() ;
} /* Subwin::datasec */

/*!***********************************************
* @brief Construct a string [x1:x2,y1:y2] of the DETSEC.
*
* @return The standard 1-based 4-pixel range string in brackets.
* @since 2013-11-25
* @author Richard J. Mathar
*/
string Subwin::detsec() const
{
	/* 1-based FITS types; inclusive range */
	ostringstream ostr ;
	ostr << "[" << (x1+1) << ":" << (x1+xsize) << ","
		<< (y1+1) << ":" << (y1+ysize) << "]" ;
	return ostr.str() ;
} /* Subwin::detsec */

/*!******************************************
* @brief Enumerate window on a xy grid of a mosaic of detector chips.
* @param[in] chipsz Number of pixels in both directions (per chips).
* @param[out] chipNo The two values chipNo[0] and chipNo[1] contain a 0-based x and y chip number.
*  If the window is crossing chip boundaries, both chipNo components are set to -1.
* @since 2014-03-25
* @author R. J. Mathar
*/
void Subwin::chipNoxy(const int chipsz, int chipNo[2]) const
{
	/* lower left pixel chip number */
	const int llx = x1 / chipsz ;
	const int lly = y1 / chipsz ;

	/* upper right pixel chip number */
	const int urx = (x1 + xsize-1) / chipsz ;
	const int ury = (y1 + ysize-1) / chipsz ;


	/* both corners on the same chip...
	*/
	if ( llx == urx && lly == ury)
	{
		chipNo[0] = llx ;
		chipNo[1] = lly ;
	}
	else
	{
		chipNo[0] = chipNo[1] = -1 ;
	}
} /* chipNoxy */


/*!**********************************************
* @brief String representation x1,y1,xsize,ysize to an output stream
*
* @since 2014-04-14
* @author Richard J. Mathar
*/
ostream & operator<< (ostream & os, const Subwin & win)
{
        os << "(" << win.x1 << "," << win.y1 << "," << win.xsize << "," << win.ysize << ") " ;
        return os ;
}

/*!************************************************
* @brief Comparison of two windows.
*
* Two windows are equal if all 4 corner coordinates match.
* @return True (considered equal) if the windows match in size and order.
* @author Richard J. Mathar
* @since 2015-10-15
*/
bool operator== (const Subwin & first , const Subwin & secnd)
{
        if ( first.x1 != secnd.x1 || first.y1 != secnd.y1
        	|| first.xsize != secnd.xsize || first.ysize != secnd.ysize )
                return false;

        return true ;
} /* Subwin::operator== */

/*!************************************************
* @brief Comparison of two windows.
*
* @return True (considered equal) if  'first' has smaller area or smaller leading y-coordinate
* or smaller leading x-coordinate.
* @author Richard J. Mathar
* @since 2015-10-15
*/
bool operator< (const Subwin & first , const Subwin & secnd)
{
	/* first criterion: area
	*/
        if ( first.area() < secnd.area() )
                return true;
        else if ( first.area() > secnd.area() )
                return false ;
	else
	{
		/* equal area. consider the (x1,y1) coordinates and
		* define the window to be smaller if that coordinate would appear
		* earlier in a streamlined x1+y1*xsize scanner
		*/
		if ( first.y1 < secnd.y1)
			return true;
		else if ( first.y1 > secnd.y1)
			return false;
		else
		{
			if ( first.x1 < secnd.x1)
				return true;
			else
				/* this includes the equal case */
				return false;
		}
	}
} /* Subwin::operator< */
