/** @file
* The program fits2tiff converts images of FITS files to the TIFF format.
*
*/

/*
* MPIA - GEIRS project
*
* "@(#) $Id: fits2tiff.cxx 795M 2020-10-16 15:30:58Z (local) $"
*
*/
#include "config.h"

#include <limits>
#include <unistd.h>
#include <cstring>
#undef PACKAGE_BUGREPORT
#undef PACKAGE_NAME
#undef PACKAGE_STRING
#undef PACKAGE_TARNAME
#undef PACKAGE_URL
#undef PACKAGE_VERSION

#ifdef HAVE_TIFF_H
extern "C" {
#include <tiffio.h>
}
#endif

#include "FitsImg.h"
#include "Rotflip.h"
#include "ColorLUTSet.h"

/** Print a usage syntax message detailing the command line
* @param[in] argv0 The name of the main program.
* @param[in] verb If true, print the names and numbers of the availabel LUT's
* @since 2019-11-05
*/
void usage(char *argv0, bool verb)
{
	cout << "Usage:" << endl ;
	cout << argv0 << " -r [-v] [-F {1|2|3}] [-R {1|2|3}] [-c cutslow:cutshigh] [-g gamma] [-b bin] [-o outStem] fitsIn.fits" << endl ;
#ifdef HAVE_TIFF_H
	// cout << argv0 << " [-l lutname | -L lutFile] [-v] [-F {1|2|3}] [-R {1|2|3}] [-c cutslow:cutshigh] [-o outStem] fitsIn.fits" << endl ;
	cout << argv0 << " [-l {0|1|..|" << (GEIRS_COLORLUTSET_CNT-1) 
		<< "}] [-v] [-F {1|2|3}] [-R {1|2|3}] [-c cutslow:cutshigh] [-g gamma] [-b bin] [-o outStem] [-C compr] fitsIn.fits" << endl ;
	if ( verb)
	{
		ColorLUTSet lut ;
		for(int c=0 ; c < GEIRS_COLORLUTSET_CNT ;c++)
			cout << " " << c << " : " << lut.names[c] ;
		cout << endl ;
	}
#endif
} /* usage */

/** The program fits2tiff converts the image in the primary HDU of a FITS file to an ASCII format.
*
* @param[in] argc The number of command line switches and arguments.
*   All arguments are optional. If the input file is missing, the program
*   prints a terse usage hint.
* @param[in] argv The vector of all command line arguments.
*
* The option -r means that the output files are raw data files with floating
*  point numbers (in the endianess of the computer) written line by line. They could for example
*  be investigated further with "od -t f4 out_0.bin" or "raw2tiff -w 4096 -l 4096 -d float out_0.bit out_0.tiff".
*  The latter assumes full-frame Hawaii-4 RG images, for example, so width and height are 4096 pixels.
*  So using the option -r contradicts to some degree the "fits2tiff" name here.
*
* The option -l specifies a color lookup tables by a 0-based index into some static set.
*   To see what is available, use the option -v without any other command line argument
*
* The option -L specifies a name of a file with a color lookup table (not yet implemented).
*
* If neither -l nor -L are used, the output will be a grayscale TIFF file.
*
* The option -F The images are flipped. If the parameter of the option is 1, the images are flipped
*   left-right, if the parameter is 2, the images are flipped up-down, and if the parameter is 3
*   the images are flipped twice (which effectively is a rotation by 180 degrees).
*
* The option -R means the images are rotated counter-clock wise
*   by as many 90 degree angles are indicated by the argument.
*   So -R 1 means 90 degrees, -R 2 means 180 degrees, -R 3 means 270 degrees.
* 
* The option -c specifies a lower and upper cut value in the original (physical) data of the FITS file
*  which are mapped to 0.0 and 1.0 on output.  if the higher value is arithmetically smaller than the lower,
*  the image is inverted (black on white).
*  If the lower value is negative, one needs to pack the option and its arguments
*  into quotes like this: -c '-1000:300'
*  If the option is missing, the program compiles a histogram of all values in
*  the image, and uses cut levels that discard the lowest 2.5 percent and hightest 2.5 percent.
*
*  To get an idea of what the natural spread of ADU values in the FITS file is,
*  use the -v option and a FITS file. This will print some cumulative density
*  function of the values from 0.05 (5 percent) to 0.95 (95 percent).
*
* The option -o defines a stem of the output files. The output files are of the form
*   outStem_[hdu]_[slic].bin, where the hdu index is counted upwards from 0 starting
*   at the primary HDU (and followed by extensions if present), and the slice index
*   starts at 0 (and followed by higher numbers if there are FITS cubes.)
*   If the option is missing, the program derives the output file name by
*   chopping off the suffix of the FITS file name.
*
* The option -b followed by a small positive integer specifies a binning factor.
*   Square tiles of the corresponding size are replaced by single pixels
*   in the output file. That pixel value is defined by the median of the
*   physical values in the FITS tile.
*   If the option is missing, the program uses the factor 1,
*   so each
*   pixel in the TIFF file corresponds to a pixel in the FITS file.
*
* The option -g defines a gamma brightness correction/mapping exponent.
*   The option argument should be a posive value in approximately the range
*   0.0001 to 1000.0. The effect of the option is that after mapping FITS values
*   to a range between 0 and 1 the brigthness in the TIFF image is set to 
*   that linear value raised to the power gamma. Values of gamma less than 1
*   increase the contrast in the bright FITS regions; values of gamma larger than 1
*   increase the contrast in the dark FITS regions. The roles of "bright" and "dark"
*   are inverted in the TIFF files if the first cut level is larger than the second.
*   If the option is missing, the program uses 1.0, a linear scale.
*
* The option -C followed by an integer defines a compression algorithm in the TIFF
*   file. The supported integers are 1 (=non compression), 2 (=CCITT modified Huffman RLE),
*   3 (= CCITT Group 3 fax encoding), 3(=CCITT Group 3 fax encoding), 4 (=CCITT Group 4 fax), 
*   5 (= Lempel-Ziv  & Welch),
*   6 (=6.0 JPEG), 7 (= JPEG DCT), 32773 =Macintosh RLE), 32809 =ThunderScan RLE).
* 
* The final command line argument must be the FITS file name with the image(s).
* The in-line slicing operations of cfitsio  are available: if the name is of the
* form fitsIn.fits[llx:lrx,ury:uly] with four positive integers specifiying
* the lower left and upper right x and y-indices (in the 1-based FITS convention), the tiff-output will only contain
* that rectangle of the original file. Note that on Linuxes this type of adding a
* bracket to the argument (the file name) needs to be quoted, like for example
* "myfile.fits[100:200,400:800]"
*
* Almost the same functionality can (interactively) be obtained by using the file->export
* menu in ds9.
* @since 2019-11-05
* @todo implement an option -L to point at a RGB color lookup table file of the user.
*/
int main(int argc, char *argv[])
{
	/* option character
	*/
	char oc ;

#ifdef HAVE_TIFF_H
	bool raw=false ;
#else
	bool raw=true ;
#endif
	bool verb=false ;
	string outf ;
	// string lutname ;
	int lutidx = -1;
	string lutfile ;

	/* no flips or rotations by default
	*/
	int flipxy =0;
	int rot90 =0;

	/* bining factor 1 by default
	*/
	int bin =1;

	float binRange[2] ;

	char *colon =0 ;

	/* gamm value. Output = Input ^ gamma.
	*/
	double gammExp=1.0 ;

#ifdef HAVE_TIFF_H
	/* TIFF 6.0 compression scheme
	*/
	int compr = COMPRESSION_LZW ;
#else
	int compr = 0 ;
#endif
	
	binRange[0] = binRange[1] = numeric_limits<double>::quiet_NaN() ;

#ifdef HAVE_TIFF_H
	while (  (oc=getopt(argc,argv,"o:c:l:L:F:R:rvg:b:C:")) != -1 )
#else
	while (  (oc=getopt(argc,argv,"o:c:l:L:F:R:rvg:b:")) != -1 )
#endif
	{
		switch(oc)
		{
		case 'o' :
			outf = string(optarg) ;
			break ;
		case 'l' :
			// lutname = string(optarg) ;
			lutidx = atoi(optarg) ;
			if ( lutidx < 0)
			{
				/* patch LUT index to the default (gray scale)
				*/
				cerr << "Invalid (negative) LUT index " << lutidx << endl ;
				lutidx = -1 ;
			}
			break ;
#ifdef HAVE_TIFF_H
		case 'C' :
			// lutname = string(optarg) ;
			compr = atoi(optarg) ;
			if ( (compr >=1 && compr <=10) || compr == 32766 || compr==32771 || compr==32773 || compr==32809)
				;
			else
				compr = COMPRESSION_PACKBITS ;
			break ;
#endif
		case 'g' :
			gammExp = atof(optarg) ;
			if ( gammExp <=0.)
			{
				/* patch gamma value to the linear map
				*/
				cerr << "Invalid (negative) gamma-value " << gammExp << endl ;
				gammExp = 1.0 ;
			}
			break ;
		case 'L' :
			lutfile = string(optarg) ;
			break ;
		case 'c' :
			colon = strchr(optarg,':') ;
			if ( colon == 0 )
			{
				cerr << "Error: colon missing in cuts options" << endl ;
				return 1;
			}
			binRange[0] = atof(optarg) ;
			binRange[1] = atof(1+colon) ;
			if ( binRange[0] == binRange[1] )
			{
				cerr << "Error: equal lower and upper cut level " << binRange[0] << endl ;
				return 1;
			}
			// cout << "cut levels " << binRange[0] << " " << binRange[1] << endl ;
			break ;
		case 'F' :
			flipxy = atoi(optarg) ;
			break ;
		case 'R' :
			rot90 = atoi(optarg) ;
			break ;
#ifdef HAVE_TIFF_H
		case 'r' :
			raw = true ;
			break ;
#endif
		case 'b' :
			bin = atoi(optarg) ;
			if ( bin <= 0 )
			{
				cerr << "Error: non-positive binning factor options" << bin << endl ;
				bin =1 ;
			}
			break ;
		case 'v' :
			verb = true ;
			break ;
		case '?' :
			cerr << "Invalid command line option " << optopt << endl ;
			break ;
		}
	}

	/* ensure that this is called with at least one file name
	*/
	if ( optind >= argc )
	{
		usage(argv[0],verb) ;
		return 1 ;
	}

	FitsImg fin(argv[optind]) ;

	/* if the stem is empty, get a default name by deleting the .fits suffix
	* from the input file
	*/
	if ( outf.size() <= 0)
	{
		/* obtain a copy of the FITS file name
		*/
		char * deriv = strdup(argv[optind]) ;
		/* search and discard the suffix
		*/
		char * dot = strrchr(deriv,'.') ;
		if ( dot )
			*dot = '\0' ;
		outf=deriv ;
		free(deriv) ;
	}

	/* combine the rotation and flip into an instance of the Rotflip class
	*/
	Rotflip rf(rot90, flipxy) ;

	fin.dumpRaw(outf,rf,binRange,raw,lutidx,gammExp,bin,verb,compr) ;
	return 0 ;
} /* main */
