/*
* MPIA - GEIRS project
*
* "@(#) $Id: ColorLUTSet.cxx 795 2020-09-02 11:29:29Z mathar $
*
*/

#undef PACKAGE_BUGREPORT
#undef PACKAGE_NAME
#undef PACKAGE_STRING
#undef PACKAGE_TARNAME
#undef PACKAGE_URL
#undef PACKAGE_VERSION


#include "config.h"

#include "ColorLUTSet.h"

using namespace std ;

/** Constructor.
* @since 2019-11-07
* @author Richard J. Mathar
*/
ColorLUTSet::ColorLUTSet()
{
} /* ctor */

/** Create a RGB triple by looking up a number in the range 0.0..1.0.
* @param[in] lutidx a 0-based index into the available set of LUT's.
*   Negative indices or those outside the available set are patched
*   to 0 (which is a gray-scale).
* @param[in] gray a floating poitn number in the range 0 (dark) to 1 (bright).
* @param[out] rgb On return rgb[0..2] contain the red, green and blue components,
*   each in the range 0..255.
* @since 2019-11-07
* @author Richard J. Mathar
*/
void ColorLUTSet::map(int lutidx, float gray, char rgb[]) const
{
	/* gray 0.0 -> index 0 , gray 1-1/256 = 0.99609 -> index 255.
	* Clamp the input to the available range of values.
	*/
	const int rgbidx = max(0,min(255,(int)(gray*256.0f))) ;

	/* ensure that the lutidx is within th erange of known tables
	*/
	lutidx = min(max(0,lutidx),GEIRS_COLORLUTSET_CNT-1) ;
	/* look up the 24-bit value of this particular table and gray value
	*/
	const int rgbval = allLUTs[lutidx][rgbidx] ;
	/* extract the individual red , green and blue components.
	*/
	rgb[0] = rgbval & 0xff ;
	rgb[1] = (rgbval >> 8) & 0xff ;
	rgb[2] = (rgbval >> 16) & 0xff ;
} /* map */

/** Convert the 256 RGB triples to a representation understood by libtiff palettes.
* @param[in] lutidx a 0-based index into the available set of LUT's.
*   Negative indices or those outside the available set are patched
*   to 0 (which is a gray-scale).
* @param[out] palRGB If not null, replaced on output by a C-type array of the values.
*   Warning: needs to be allocated of size 3*256 values at least to avoid memory faults.
* @return A vector of size 3*256 containg the red bytes in [0..255], the blue bytes in [256..512]...
*   If lutidx is outside the range of available colors, the vector is empty.
* @since 2020-08-26
* @author Richard J. Mathar
*/
vector<unsigned short> ColorLUTSet::tiff(int lutidx, unsigned short * palRGB ) const
{
	vector<unsigned short> palet(3*256) ;
	if ( lutidx >=0 && lutidx < GEIRS_COLORLUTSET_CNT )
	{
		for(int rgb= 0 ; rgb < 3 ;  rgb++)
			for(int rgbidx=0 ; rgbidx < 256 ; rgbidx++)
			{
				int rgbval = allLUTs[lutidx][rgbidx] ;
				rgbval = (rgbval >> (8*rgb) ) & 0xff ;
				palet[rgbidx + rgb*256] = rgbval ;
				if ( palRGB)
					palRGB[rgbidx + rgb*256] = rgbval ;
			}
	}
	return palet ;
} /* tiff */

/* the following constructed with
* cd ~/GEIRS/trunk-r6788M-21 ; java de.mpg.mpia.geirs.ColorLUTSet | fold -w 132
*/

int ColorLUTSet::allLUTs[GEIRS_COLORLUTSET_CNT][256] = {
/* 0 Grayscale */

{0x0 , 0x10101 , 0x20202 , 0x30303 , 0x40404 , 0x50505 , 0x60606 , 0x70707 , 0x80808 , 0x90909 , 0xa0a0a , 0xb0b0b , 0xc0c0c , 0xd0d0d , 0xe0e0e , 0xf0f0f , 0x101010 , 0x111111 , 0x121212 , 0x131313 , 0x141414 , 0x151515 , 0x161616 , 0x171717 , 0x181818 , 0x191919 , 0x1a1a1a , 0x1b1b1b , 0x1c1c1c , 0x1d1d1d , 0x1e1e1e , 0x1f1f1f , 0x202020 , 0x212121 , 0x222222 , 0x232323 , 0x242424 , 0x252525 , 0x262626 , 0x272727 , 0x282828 , 0x292929 , 0x2a2a2a , 0x2b2b2b , 0x2c2c2c , 0x2d2d2d , 0x2e2e2e , 0x2f2f2f , 0x303030 , 0x313131 , 0x323232 , 0x333333 , 0x343434 , 0x353535 , 0x363636 , 0x373737 , 0x383838 , 0x393939 , 0x3a3a3a , 0x3b3b3b , 0x3c3c3c , 0x3d3d3d , 0x3e3e3e , 0x3f3f3f , 0x404040 , 0x414141 , 0x424242 , 0x434343 , 0x444444 , 0x454545 , 0x464646 , 0x474747 , 0x484848 , 0x494949 , 0x4a4a4a , 0x4b4b4b , 0x4c4c4c , 0x4d4d4d , 0x4e4e4e , 0x4f4f4f , 0x505050 , 0x515151 , 0x525252 , 0x535353 , 0x545454 , 0x555555 , 0x565656 , 0x575757 , 0x585858 , 0x595959 , 0x5a5a5a , 0x5b5b5b , 0x5c5c5c , 0x5d5d5d , 0x5e5e5e , 0x5f5f5f , 0x606060 , 0x616161 , 0x626262 , 0x636363 , 0x646464 , 0x656565 , 0x666666 , 0x676767 , 0x686868 , 0x696969 , 0x6a6a6a , 0x6b6b6b , 0x6c6c6c , 0x6d6d6d , 0x6e6e6e , 0x6f6f6f , 0x707070 , 0x717171 , 0x727272 , 0x737373 , 0x747474 , 0x757575 , 0x767676 , 0x777777 , 0x787878 , 0x797979 , 0x7a7a7a , 0x7b7b7b , 0x7c7c7c , 0x7d7d7d , 0x7e7e7e , 0x7f7f7f , 0x808080 , 0x818181 , 0x828282 , 0x838383 , 0x848484 , 0x858585 , 0x868686 , 0x878787 , 0x888888 , 0x898989 , 0x8a8a8a , 0x8b8b8b , 0x8c8c8c , 0x8d8d8d , 0x8e8e8e , 0x8f8f8f , 0x909090 , 0x919191 , 0x929292 , 0x939393 , 0x949494 , 0x959595 , 0x969696 , 0x979797 , 0x989898 , 0x999999 , 0x9a9a9a , 0x9b9b9b , 0x9c9c9c , 0x9d9d9d , 0x9e9e9e , 0x9f9f9f , 0xa0a0a0 , 0xa1a1a1 , 0xa2a2a2 , 0xa3a3a3 , 0xa4a4a4 , 0xa5a5a5 , 0xa6a6a6 , 0xa7a7a7 , 0xa8a8a8 , 0xa9a9a9 , 0xaaaaaa , 0xababab , 0xacacac , 0xadadad , 0xaeaeae , 0xafafaf , 0xb0b0b0 , 0xb1b1b1 , 0xb2b2b2 , 0xb3b3b3 , 0xb4b4b4 , 0xb5b5b5 , 0xb6b6b6 , 0xb7b7b7 , 0xb8b8b8 , 0xb9b9b9 , 0xbababa , 0xbbbbbb , 0xbcbcbc , 0xbdbdbd , 0xbebebe , 0xbfbfbf , 0xc0c0c0 , 0xc1c1c1 , 0xc2c2c2 , 0xc3c3c3 , 0xc4c4c4 , 0xc5c5c5 , 0xc6c6c6 , 0xc7c7c7 , 0xc8c8c8 , 0xc9c9c9 , 0xcacaca , 0xcbcbcb , 0xcccccc , 0xcdcdcd , 0xcecece , 0xcfcfcf , 0xd0d0d0 , 0xd1d1d1 , 0xd2d2d2 , 0xd3d3d3 , 0xd4d4d4 , 0xd5d5d5 , 0xd6d6d6 , 0xd7d7d7 , 0xd8d8d8 , 0xd9d9d9 , 0xdadada , 0xdbdbdb , 0xdcdcdc , 0xdddddd , 0xdedede , 0xdfdfdf , 0xe0e0e0 , 0xe1e1e1 , 0xe2e2e2 , 0xe3e3e3 , 0xe4e4e4 , 0xe5e5e5 , 0xe6e6e6 , 0xe7e7e7 , 0xe8e8e8 , 0xe9e9e9 , 0xeaeaea , 0xebebeb , 0xececec , 0xededed , 0xeeeeee , 0xefefef , 0xf0f0f0 , 0xf1f1f1 , 0xf2f2f2 , 0xf3f3f3 , 0xf4f4f4 , 0xf5f5f5 , 0xf6f6f6 , 0xf7f7f7 , 0xf8f8f8 , 0xf9f9f9 , 0xfafafa , 0xfbfbfb , 0xfcfcfc , 0xfdfdfd , 0xfefefe , 0xffffff}

,

/* 1 Hot_Iron_DICOM */

{0x0 , 0x2 , 0x4 , 0x6 , 0x8 , 0xa , 0xc , 0xe , 0x10 , 0x12 , 0x14 , 0x16 , 0x18 , 0x1a , 0x1c , 0x1e , 0x20 , 0x22 , 0x24 , 0x26 , 0x28 , 0x2a , 0x2c , 0x2e , 0x30 , 0x32 , 0x34 , 0x36 , 0x38 , 0x3a , 0x3c , 0x3e , 0x40 , 0x42 , 0x44 , 0x46 , 0x48 , 0x4a , 0x4c , 0x4e , 0x50 , 0x52 , 0x54 , 0x56 , 0x58 , 0x5a , 0x5c , 0x5e , 0x60 , 0x62 , 0x64 , 0x66 , 0x68 , 0x6a , 0x6c , 0x6e , 0x70 , 0x72 , 0x74 , 0x76 , 0x78 , 0x7a , 0x7c , 0x7e , 0x80 , 0x82 , 0x84 , 0x86 , 0x88 , 0x8a , 0x8c , 0x8e , 0x90 , 0x92 , 0x94 , 0x96 , 0x98 , 0x9a , 0x9c , 0x9e , 0xa0 , 0xa2 , 0xa4 , 0xa6 , 0xa8 , 0xaa , 0xac , 0xae , 0xb0 , 0xb2 , 0xb4 , 0xb6 , 0xb8 , 0xba , 0xbc , 0xbe , 0xc0 , 0xc2 , 0xc4 , 0xc6 , 0xc8 , 0xca , 0xcc , 0xce , 0xd0 , 0xd2 , 0xd4 , 0xd6 , 0xd8 , 0xda , 0xdc , 0xde , 0xe0 , 0xe2 , 0xe4 , 0xe6 , 0xe8 , 0xea , 0xec , 0xee , 0xf0 , 0xf2 , 0xf4 , 0xf6 , 0xf8 , 0xfa , 0xfc , 0xfe , 0xff , 0x2ff , 0x4ff , 0x6ff , 0x8ff , 0xaff , 0xcff , 0xeff , 0x10ff , 0x12ff , 0x14ff , 0x16ff , 0x18ff , 0x1aff , 0x1cff , 0x1eff , 0x20ff , 0x22ff , 0x24ff , 0x26ff , 0x28ff , 0x2aff , 0x2cff , 0x2eff , 0x30ff , 0x32ff , 0x34ff , 0x36ff , 0x38ff , 0x3aff , 0x3cff , 0x3eff , 0x40ff , 0x42ff , 0x44ff , 0x46ff , 0x48ff , 0x4aff , 0x4cff , 0x4eff , 0x50ff , 0x52ff , 0x54ff , 0x56ff , 0x58ff , 0x5aff , 0x5cff , 0x5eff , 0x60ff , 0x62ff , 0x64ff , 0x66ff , 0x68ff , 0x6aff , 0x6cff , 0x6eff , 0x70ff , 0x72ff , 0x74ff , 0x76ff , 0x78ff , 0x7aff , 0x7cff , 0x7eff , 0x80ff , 0x482ff , 0x884ff , 0xc86ff , 0x1088ff , 0x148aff , 0x188cff , 0x1c8eff , 0x2090ff , 0x2492ff , 0x2894ff , 0x2c96ff , 0x3098ff , 0x349aff , 0x389cff , 0x3c9eff , 0x40a0ff , 0x44a2ff , 0x48a4ff , 0x4ca6ff , 0x50a8ff , 0x54aaff , 0x58acff , 0x5caeff , 0x60b0ff , 0x64b2ff , 0x68b4ff , 0x6cb6ff , 0x70b8ff , 0x74baff , 0x78bcff , 0x7cbeff , 0x80c0ff , 0x84c2ff , 0x88c4ff , 0x8cc6ff , 0x90c8ff , 0x94caff , 0x98ccff , 0x9cceff , 0xa0d0ff , 0xa4d2ff , 0xa8d4ff , 0xacd6ff , 0xb0d8ff , 0xb4daff , 0xb8dcff , 0xbcdeff , 0xc0e0ff , 0xc4e2ff , 0xc8e4ff , 0xcce6ff , 0xd0e8ff , 0xd4eaff , 0xd8ecff , 0xdceeff , 0xe0f0ff , 0xe4f2ff , 0xe8f4ff , 0xecf6ff , 0xf0f8ff , 0xf4faff , 0xf8fcff , 0xfcfeff}

,

/* 2 PET_DICOM */

{0x0 , 0x20200 , 0x40400 , 0x60600 , 0x80800 , 0xa0a00 , 0xc0c00 , 0xe0e00 , 0x101000 , 0x121200 , 0x141400 , 0x161600 , 0x181800 , 0x1a1a00 , 0x1c1c00 , 0x1e1e00 , 0x202000 , 0x222200 , 0x242400 , 0x262600 , 0x282800 , 0x2a2a00 , 0x2c2c00 , 0x2e2e00 , 0x303000 , 0x323200 , 0x343400 , 0x363600 , 0x383800 , 0x3a3a00 , 0x3c3c00 , 0x3e3e00 , 0x404000 , 0x424200 , 0x444400 , 0x464600 , 0x484800 , 0x4a4a00 , 0x4c4c00 , 0x4e4e00 , 0x505000 , 0x525200 , 0x545400 , 0x565600 , 0x585800 , 0x5a5a00 , 0x5c5c00 , 0x5e5e00 , 0x606000 , 0x626200 , 0x646400 , 0x666600 , 0x686800 , 0x6a6a00 , 0x6c6c00 , 0x6e6e00 , 0x707000 , 0x727200 , 0x747400 , 0x767600 , 0x787800 , 0x7a7a00 , 0x7c7c00 , 0x7e7e00 , 0x808000 , 0x827e02 , 0x847c04 , 0x867a06 , 0x887808 , 0x8a760a , 0x8c740c , 0x8e720e , 0x907010 , 0x926e12 , 0x946c14 , 0x966a16 , 0x986818 , 0x9a661a , 0x9c641c , 0x9e621e , 0xa06020 , 0xa25e22 , 0xa45c24 , 0xa65a26 , 0xa85828 , 0xaa562a , 0xac542c , 0xae522e , 0xb05030 , 0xb24e32 , 0xb44c34 , 0xb64a36 , 0xb84838 , 0xba463a , 0xbc443c , 0xbe423e , 0xc04040 , 0xc23e42 , 0xc43c44 , 0xc63a46 , 0xc83848 , 0xca364a , 0xcc344c , 0xce324e , 0xd03050 , 0xd22e52 , 0xd42c54 , 0xd62a56 , 0xd82858 , 0xda265a , 0xdc245c , 0xde225e , 0xe02060 , 0xe21e62 , 0xe41c64 , 0xe61a66 , 0xe81868 , 0xea166a , 0xec146c , 0xee126e , 0xf01070 , 0xf20e72 , 0xf40c74 , 0xf60a76 , 0xf80878 , 0xfa067a , 0xfc047c , 0xfe027e , 0x80 , 0xfc0282 , 0xf80484 , 0xf40686 , 0xf00888 , 0xec0a8a , 0xe80c8c , 0xe40e8e , 0xe01090 , 0xdc1292 , 0xd81494 , 0xd41696 , 0xd01898 , 0xcc1a9a , 0xc81c9c , 0xc41e9e , 0xc020a0 , 0xbc22a2 , 0xb824a4 , 0xb426a6 , 0xb028a8 , 0xac2aaa , 0xa82cac , 0xa42eae , 0xa030b0 , 0x9c32b2 , 0x9834b4 , 0x9436b6 , 0x9038b8 , 0x8c3aba , 0x883cbc , 0x843ebe , 0x8040c0 , 0x7c42c2 , 0x7844c4 , 0x7446c6 , 0x7048c8 , 0x6c4aca , 0x684ccc , 0x644ece , 0x6050d0 , 0x5c52d2 , 0x5854d4 , 0x5456d6 , 0x5058d8 , 0x4c5ada , 0x485cdc , 0x445ede , 0x4060e0 , 0x3c62e2 , 0x3864e4 , 0x3466e6 , 0x3068e8 , 0x2c6aea , 0x286cec , 0x246eee , 0x2070f0 , 0x1c72f2 , 0x1874f4 , 0x1476f6 , 0x1078f8 , 0xc7afa , 0x87cfc , 0x47efe , 0x80ff , 0x482ff , 0x884ff , 0xc86ff , 0x1088ff , 0x148aff , 0x188cff , 0x1c8eff , 0x2090ff , 0x2492ff , 0x2894ff , 0x2c96ff , 0x3098ff , 0x349aff , 0x389cff , 0x3c9eff , 0x40a0ff , 0x44a2ff , 0x48a4ff , 0x4ca6ff , 0x50a8ff , 0x54aaff , 0x58acff , 0x5caeff , 0x60b0ff , 0x64b2ff , 0x68b4ff , 0x6cb6ff , 0x70b8ff , 0x74baff , 0x78bcff , 0x7cbeff , 0x80c0ff , 0x84c2ff , 0x88c4ff , 0x8cc6ff , 0x90c8ff , 0x94caff , 0x98ccff , 0x9cceff , 0xa0d0ff , 0xa4d2ff , 0xa8d4ff , 0xacd6ff , 0xb0d8ff , 0xb4daff , 0xb8dcff , 0xbcdeff , 0xc0e0ff , 0xc4e2ff , 0xc8e4ff , 0xcce6ff , 0xd0e8ff , 0xd4eaff , 0xd8ecff , 0xdceeff , 0xe0f0ff , 0xe4f2ff , 0xe8f4ff , 0xecf6ff , 0xf0f8ff , 0xf4faff , 0xf8fcff , 0xfcfeff}

,

/* 3 Hot_Metal_Blue_DICOM */

{0x0 , 0x10000 , 0x30000 , 0x50000 , 0x70000 , 0x90000 , 0xb0000 , 0xd0000 , 0xf0000 , 0x110000 , 0x130000 , 0x150000 , 0x170000 , 0x190000 , 0x1b0000 , 0x1d0000 , 0x1f0000 , 0x210000 , 0x230000 , 0x250000 , 0x270000 , 0x290000 , 0x2b0000 , 0x2d0000 , 0x2f0000 , 0x310000 , 0x320000 , 0x340000 , 0x360000 , 0x380000 , 0x3a0000 , 0x3c0000 , 0x3e0000 , 0x400000 , 0x420000 , 0x440000 , 0x460000 , 0x480000 , 0x4a0000 , 0x4c0000 , 0x4e0000 , 0x500000 , 0x520000 , 0x540000 , 0x560000 , 0x580000 , 0x5a0000 , 0x5c0000 , 0x5e0000 , 0x600000 , 0x620000 , 0x640000 , 0x650000 , 0x670000 , 0x690000 , 0x6b0000 , 0x6d0000 , 0x6f0000 , 0x710000 , 0x730000 , 0x750000 , 0x770000 , 0x790000 , 0x7b0000 , 0x7d0000 , 0x7f0000 , 0x810000 , 0x830000 , 0x850000 , 0x870000 , 0x890000 , 0x8b0000 , 0x8d0000 , 0x8f0000 , 0x910000 , 0x930000 , 0x950000 , 0x960000 , 0x980000 , 0x9a0000 , 0x9c0000 , 0x9e0000 , 0xa00000 , 0xa20000 , 0xa40000 , 0xa60000 , 0xa80003 , 0xaa0006 , 0xac0009 , 0xae000c , 0xb0000f , 0xb20012 , 0xb40015 , 0xb60018 , 0xb8001b , 0xba001e , 0xbc0021 , 0xbe0024 , 0xc00027 , 0xc2002a , 0xc4002d , 0xc60030 , 0xc80033 , 0xc90036 , 0xc60039 , 0xc3003c , 0xc0003f , 0xbd0042 , 0xba0045 , 0xb70048 , 0xb4004b , 0xb1004e , 0xae0051 , 0xab0054 , 0xa80057 , 0xa5005a , 0xa2005d , 0x9f0060 , 0x9c0063 , 0x990066 , 0x930069 , 0x8d016c , 0x87036f , 0x810572 , 0x7b0775 , 0x750978 , 0x6f0b7b , 0x690d7e , 0x630f81 , 0x5d1184 , 0x571287 , 0x51148a , 0x4b168d , 0x451890 , 0x3f1a93 , 0x391c96 , 0x331e99 , 0x30209c , 0x2d229f , 0x2a23a2 , 0x2725a5 , 0x2427a8 , 0x2129ab , 0x1e2bae , 0x1b2db1 , 0x182fb4 , 0x1531b7 , 0x1233ba , 0xf34bd , 0xc36c0 , 0x938c3 , 0x63ac6 , 0x33cc9 , 0x3ecc , 0x40cf , 0x42d2 , 0x44d5 , 0x45d8 , 0x47db , 0x49de , 0x4be1 , 0x4de4 , 0x4fe7 , 0x51ea , 0x53ed , 0x55f0 , 0x56f3 , 0x58f6 , 0x5af9 , 0x5cfc , 0x5eff , 0x360ff , 0x662ff , 0x964ff , 0xc66ff , 0xf67ff , 0x1269ff , 0x156bff , 0x186dff , 0x1b6fff , 0x1e71ff , 0x2173ff , 0x2475ff , 0x2777ff , 0x2a78ff , 0x2d7aff , 0x307cff , 0x337eff , 0x3680ff , 0x3982ff , 0x3c84ff , 0x3f86ff , 0x4288ff , 0x4589ff , 0x488bff , 0x4b8dff , 0x4e8fff , 0x5191ff , 0x5493ff , 0x5795ff , 0x5a97ff , 0x5d99ff , 0x609aff , 0x639cff , 0x669eff , 0x69a0ff , 0x6ca2ff , 0x6fa4ff , 0x72a6ff , 0x75a8ff , 0x78aaff , 0x7babff , 0x7eadff , 0x81afff , 0x84b1ff , 0x87b3ff , 0x8ab5ff , 0x8db7ff , 0x90b9ff , 0x93bbff , 0x96bcff , 0x99beff , 0x9cc0ff , 0x9fc2ff , 0xa2c4ff , 0xa5c6ff , 0xa8c8ff , 0xabcaff , 0xaeccff , 0xb1cdff , 0xb4cfff , 0xb7d1ff , 0xbad3ff , 0xbdd5ff , 0xc0d7ff , 0xc3d9ff , 0xc6dbff , 0xc9ddff , 0xccdeff , 0xcfe0ff , 0xd2e2ff , 0xd5e4ff , 0xd8e6ff , 0xdbe8ff , 0xdeeaff , 0xe1ecff , 0xe4eeff , 0xe7efff , 0xeaf1ff , 0xedf3ff , 0xf0f5ff , 0xf3f7ff , 0xf6f9ff , 0xf9fbff , 0xfcfdff , 0xffffff}

,

/* 4 mixed */

{0x0 , 0x20202 , 0x40404 , 0x60606 , 0x80808 , 0xa0a0a , 0xc0c0c , 0xe0e0e , 0x101010 , 0x121212 , 0x141414 , 0x161616 , 0x181818 , 0x1a1a1a , 0x1c1c1c , 0x1e1e1e , 0x202020 , 0x222222 , 0x242424 , 0x262626 , 0x282828 , 0x2a2a2a , 0x2c2c2c , 0x2e2e2e , 0x303030 , 0x323232 , 0x343434 , 0x363636 , 0x383838 , 0x3a3a3a , 0x3c3c3c , 0x3e3e3e , 0x404040 , 0x424242 , 0x444444 , 0x464646 , 0x484848 , 0x4a4a4a , 0x4c4c4c , 0x4e4e4e , 0x505050 , 0x525252 , 0x545454 , 0x565656 , 0x585858 , 0x5a5a5a , 0x5c5c5c , 0x5e5e5e , 0x606060 , 0x626262 , 0x646464 , 0x666666 , 0x686868 , 0x6a6a6a , 0x6c6c6c , 0x6e6e6e , 0x707070 , 0x727272 , 0x747474 , 0x767676 , 0x787878 , 0x7a7a7a , 0x7c7c7c , 0x7e7e7e , 0x808080 , 0x828282 , 0x848484 , 0x868686 , 0x888888 , 0x8a8a8a , 0x8c8c8c , 0x8e8e8e , 0x909090 , 0x929292 , 0x949494 , 0x969696 , 0x989898 , 0x9a9a9a , 0x9c9c9c , 0x9e9e9e , 0xa0a0a0 , 0xa2a2a2 , 0xa4a4a4 , 0xa6a6a6 , 0xa8a8a8 , 0xaaaaaa , 0xacacac , 0xaeaeae , 0xb0b0b0 , 0xb2b2b2 , 0xb4b4b4 , 0xb6b6b6 , 0xb8b8b8 , 0xbababa , 0xbcbcbc , 0xbebebe , 0xc0c0c0 , 0xc2c2c2 , 0xc4c4c4 , 0xc6c6c6 , 0xc8c8c8 , 0xcacaca , 0xcccccc , 0xcecece , 0xd0d0d0 , 0xd2d2d2 , 0xd4d4d4 , 0xd6d6d6 , 0xd8d8d8 , 0xdadada , 0xdcdcdc , 0xdedede , 0xe0e0e0 , 0xe2e2e2 , 0xe4e4e4 , 0xe6e6e6 , 0xe8e8e8 , 0xeaeaea , 0xececec , 0xeeeeee , 0xf0f0f0 , 0xf2f2f2 , 0xf4f4f4 , 0xf6f6f6 , 0xf8f8f8 , 0xfafafa , 0xfcfcfc , 0xfefefe , 0x38 , 0x40 , 0x48 , 0x54 , 0x5c , 0x68 , 0x70 , 0x7c , 0x84 , 0x90 , 0x98 , 0xa0 , 0xac , 0xb4 , 0xc0 , 0xc8 , 0xd4 , 0xdc , 0xe8 , 0xf0 , 0xfc , 0xcfc , 0x18fc , 0x24fc , 0x34fc , 0x40fc , 0x4cfc , 0x5cfc , 0x68fc , 0x74fc , 0x84fc , 0x90fc , 0x9cfc , 0xacfc , 0xb8fc , 0xc4fc , 0xd4fc , 0xe0fc , 0xecfc , 0xfcfc , 0xfcf0 , 0xf8e0 , 0xf4d0 , 0xf0c0 , 0xecb0 , 0xe8a0 , 0xe490 , 0x4e080 , 0x4e070 , 0x4dc60 , 0x4d850 , 0x4d440 , 0x4d030 , 0x4cc20 , 0x4c810 , 0x8c400 , 0x2cc800 , 0x50cc00 , 0x74d000 , 0x98d400 , 0xbcd800 , 0xbcd000 , 0xbcc800 , 0xc0bc00 , 0xc0b000 , 0xc4a804 , 0xc49c04 , 0xc89004 , 0xcc8404 , 0xcc7c08 , 0xd07008 , 0xd06408 , 0xd45808 , 0xd8500c , 0xd8440c , 0xdc380c , 0xe02c10 , 0xe42410 , 0xe41810 , 0xe81018 , 0xe01020 , 0xdc1028 , 0xd41030 , 0xd01038 , 0xc81040 , 0xc81044 , 0xc41448 , 0xc4144c , 0xc01854 , 0xc01858 , 0xbc1c5c , 0xb81c60 , 0xb82068 , 0xb4206c , 0xb42470 , 0xb02474 , 0xac287c , 0xac2880 , 0xa82c84 , 0xa4308c , 0xa4348c , 0xa03c8c , 0x9c448c , 0x984888 , 0x985088 , 0x945888 , 0x905c88 , 0x8c6484 , 0x8c6c84 , 0x887084 , 0x847884 , 0x808080 , 0x888888 , 0x909090 , 0x989898 , 0xa0a0a0 , 0xa8a8a8 , 0xb0b0b0 , 0xb8b8b8 , 0xc0c0c0 , 0xc8c8c8 , 0xd0d0d0 , 0xd8d8d8 , 0xe0e0e0 , 0xe8e8e8 , 0xf0f0f0 , 0xf8f8f8 , 0xfcfcfc}

,

/* 5 brown */

{0x20209 , 0x20209 , 0x1030d , 0x10410 , 0x10513 , 0x10616 , 0x1071a , 0x1081d , 0x10920 , 0x10a23 , 0x10b26 , 0x10c2a , 0x10c2d , 0x10d30 , 0x10e33 , 0x10f37 , 0x1103a , 0x1113d , 0x11240 , 0x11342 , 0x11446 , 0x11549 , 0x1164b , 0x1174d , 0x1184f , 0x11952 , 0x11a54 , 0x11b56 , 0x11c58 , 0x11d5a , 0x11e5d , 0x11f5f , 0x12061 , 0x12163 , 0x12166 , 0x12267 , 0x1236a , 0x1246b , 0x1256d , 0x1266e , 0x1276f , 0x12870 , 0x12972 , 0x12a73 , 0x12b74 , 0x12c75 , 0x12d77 , 0x12e78 , 0x12f79 , 0x1307b , 0x1317c , 0x1327d , 0x1337e , 0x13480 , 0x13581 , 0x13682 , 0x13684 , 0x13785 , 0x13886 , 0x13887 , 0x13989 , 0x13a8a , 0x13b8b , 0x13b8c , 0x13c8e , 0x13d8f , 0x13d90 , 0x13e92 , 0x13f92 , 0x13f93 , 0x14094 , 0x14094 , 0x14095 , 0x14196 , 0x14297 , 0x14297 , 0x14398 , 0x14499 , 0x1459a , 0x1459a , 0x1469b , 0x1479c , 0x1479d , 0x1489d , 0x1499e , 0x1499f , 0x14aa0 , 0x14ba0 , 0x14ba1 , 0x14ca2 , 0x14da3 , 0x14ea3 , 0x14ea4 , 0x14fa5 , 0x150a6 , 0x150a7 , 0x151a7 , 0x152a8 , 0x153a9 , 0x154aa , 0x154aa , 0x255ab , 0x256ac , 0x356ad , 0x357ad , 0x357ae , 0x458af , 0x459b0 , 0x559b0 , 0x55ab1 , 0x65bb2 , 0x65bb3 , 0x65cb3 , 0x75cb4 , 0x75db5 , 0x85eb6 , 0x85eb6 , 0x85fb7 , 0x960b8 , 0x960b8 , 0xa61b8 , 0xa61b9 , 0xa62b9 , 0xb62ba , 0xb63ba , 0xc63bb , 0xc64bb , 0xc64bc , 0xd65bc , 0xd65bd , 0xe66bd , 0xe66be , 0xf67be , 0xf67bf , 0xf68bf , 0x1068bf , 0x1069c0 , 0x1169c0 , 0x116ac1 , 0x116ac1 , 0x126bc2 , 0x126bc2 , 0x136cc3 , 0x136cc3 , 0x136dc4 , 0x146dc4 , 0x146ec5 , 0x156ec5 , 0x156fc6 , 0x166fc6 , 0x1670c7 , 0x1670c7 , 0x1771c8 , 0x1771c9 , 0x1872c9 , 0x1872ca , 0x1873ca , 0x1973cb , 0x1974cb , 0x1a74cc , 0x1a75cc , 0x1a75cd , 0x1b76cd , 0x1b76ce , 0x1c77ce , 0x1c77cf , 0x1c78cf , 0x1d78d0 , 0x1d78d0 , 0x1e79d1 , 0x1e79d1 , 0x1f7ad2 , 0x1f7ad2 , 0x1f7bd3 , 0x207bd4 , 0x207cd4 , 0x217cd5 , 0x217dd5 , 0x217dd6 , 0x227ed6 , 0x227ed7 , 0x237fd7 , 0x237fd8 , 0x2380d8 , 0x2480d9 , 0x2481d9 , 0x2581da , 0x2582da , 0x2682db , 0x2683db , 0x2683dc , 0x2784dc , 0x2884dd , 0x2885dd , 0x2985de , 0x2a86df , 0x2a86df , 0x2b87e0 , 0x2b87e0 , 0x2c88e1 , 0x2d88e1 , 0x2d89e2 , 0x2e89e2 , 0x2f8ae3 , 0x2f8ae3 , 0x308be4 , 0x318be4 , 0x318ce5 , 0x328ce5 , 0x328de6 , 0x338de6 , 0x348ee7 , 0x348ee7 , 0x358fe8 , 0x368fe8 , 0x3790e9 , 0x3890ea , 0x3991ea , 0x3b92eb , 0x3d93eb , 0x3f94ec , 0x4095ec , 0x4196ed , 0x4397ed , 0x4498ee , 0x4698ee , 0x4799ef , 0x499aef , 0x4b9bf0 , 0x4c9cf0 , 0x4e9df1 , 0x529ef1 , 0x559ff2 , 0x57a0f2 , 0x5aa2f3 , 0x5da3f4 , 0x5fa4f4 , 0x62a6f5 , 0x65a8f5 , 0x68abf6 , 0x70aef6 , 0x76b1f7 , 0x7bb4f7 , 0x81b6f8 , 0x86b9f8 , 0x8cbcf9 , 0x91bff9 , 0x97c1fa , 0x9cc9fa , 0xa2d0fb , 0xa7d8fb , 0xbee0fc , 0xcde7fc , 0xddeffd , 0xeef6fd , 0xeef6fd}

,

/* 6 isocontour */

{0x10101 , 0x10101 , 0x20202 , 0x30303 , 0x40404 , 0x50505 , 0x60606 , 0x70707 , 0x80808 , 0x90909 , 0xa0a0a , 0xb0b0b , 0xc0c0c , 0xe0e0e , 0xf0f0f , 0x101010 , 0x111111 , 0x121212 , 0x131313 , 0x141414 , 0x151515 , 0x161616 , 0x171717 , 0x181818 , 0x1a1a1a , 0x1b1b1b , 0x1c1c1c , 0x1d1d1d , 0x1e1e1e , 0x1f1f1f , 0x202020 , 0x212121 , 0x222222 , 0x232323 , 0x242424 , 0x262626 , 0x272727 , 0x282828 , 0x292929 , 0x2a2a2a , 0x2b2b2b , 0x2c2c2c , 0x2d2d2d , 0x2f2e2e , 0x632f2f , 0x983030 , 0xcb3232 , 0xfe3333 , 0xcb3434 , 0x993535 , 0x683636 , 0x373737 , 0x383838 , 0x393939 , 0x3a3a3a , 0x3b3b3b , 0x3c3c3c , 0x3e3e3e , 0x3f3f3f , 0x404040 , 0x404040 , 0x414141 , 0x424242 , 0x434343 , 0x444444 , 0x454545 , 0x464646 , 0x474747 , 0x484848 , 0x4a4a4a , 0x4b4b4b , 0x4c4c4c , 0x4d4d4d , 0x4e4e4e , 0x4f4f4f , 0x505050 , 0x515151 , 0x525252 , 0x535353 , 0x545454 , 0x565656 , 0x575757 , 0x585858 , 0x595959 , 0x5a5a5a , 0x5b5b5b , 0x5c5c5c , 0x5d5d5d , 0x5e5e5e , 0x5f5f5f , 0x606060 , 0x628962 , 0x63b163 , 0x64d864 , 0x65fe65 , 0x66d866 , 0x67b367 , 0x688e68 , 0x696969 , 0x6a6a6a , 0x6b6b6b , 0x6c6c6c , 0x6e6e6e , 0x6f6f6f , 0x707070 , 0x717171 , 0x727272 , 0x737373 , 0x747474 , 0x757575 , 0x767676 , 0x777777 , 0x787878 , 0x7a7a7a , 0x7b7b7b , 0x7c7c7c , 0x7d7d7d , 0x7e7e7e , 0x7f7f7f , 0x808080 , 0x818181 , 0x828282 , 0x838383 , 0x848484 , 0x858585 , 0x878787 , 0x888888 , 0x898989 , 0x8a8a8a , 0x8b8b8b , 0x8c8c8c , 0x8d8d8d , 0x8e8e8e , 0x8f8f8f , 0x909090 , 0x919191 , 0x939393 , 0x949494 , 0x95afaf , 0x96caca , 0x97e4e4 , 0x98fefe , 0x99e5e5 , 0x9acccc , 0x9bb4b4 , 0x9c9c9c , 0x9d9d9d , 0x9f9f9f , 0xa0a0a0 , 0xa1a1a1 , 0xa2a2a2 , 0xa3a3a3 , 0xa4a4a4 , 0xa5a5a5 , 0xa6a6a6 , 0xa7a7a7 , 0xa8a8a8 , 0xa9a9a9 , 0xababab , 0xacacac , 0xadadad , 0xaeaeae , 0xafafaf , 0xb0b0b0 , 0xb1b1b1 , 0xb2b2b2 , 0xb3b3b3 , 0xb4b4b4 , 0xb5b5b5 , 0xb7b7b7 , 0xb8b8b8 , 0xb9b9b9 , 0xbababa , 0xbbbbbb , 0xbcbcbc , 0xbdbdbd , 0xbebebe , 0xbfbfbf , 0xbfbfbf , 0xc0c0c0 , 0xc1c1c1 , 0xc3c3c3 , 0xc4c4c4 , 0xc5c5c5 , 0xc6c6c6 , 0xc7c7d5 , 0xc8c8e3 , 0xc9c9f1 , 0xcacafe , 0xcbcbf2 , 0xcccce5 , 0xcdcdda , 0xcfcfcf , 0xd0d0d0 , 0xd1d1d1 , 0xd2d2d2 , 0xd3d3d3 , 0xd4d4d4 , 0xd5d5d5 , 0xd6d6d6 , 0xd7d7d7 , 0xd8d8d8 , 0xd9d9d9 , 0xdbdbdb , 0xdcdcdc , 0xdddddd , 0xdedede , 0xdfdfdf , 0xe0e0e0 , 0xe1e1e1 , 0xe2e2e2 , 0xe3e3e3 , 0xe4e4e4 , 0xe5e5e5 , 0xe7e7e7 , 0xe8e8e8 , 0xe9e9e9 , 0xeaeaea , 0xebebeb , 0xececec , 0xededed , 0xeeeeee , 0xefefef , 0xf0f0f0 , 0xf1f1f1 , 0xf3f3f3 , 0xf4f4f4 , 0xf5f5f5 , 0xf6f6f6 , 0xf7f7f7 , 0xf8f8f8 , 0xf9f9f9 , 0xfafafa , 0xfbfbfb , 0xfcfcfc , 0xfdfdfd , 0xfefefe , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd}

,

/* 7 thermal */

{0x730046 , 0x730046 , 0x740046 , 0x760046 , 0x780046 , 0x7a0046 , 0x7c0046 , 0x7e0046 , 0x800046 , 0x830046 , 0x850046 , 0x880046 , 0x8b0046 , 0x8d0046 , 0x900046 , 0x930046 , 0x970046 , 0x9a0046 , 0x9d0046 , 0xa00046 , 0xa40046 , 0xa70046 , 0xaa0046 , 0xae0046 , 0xb10046 , 0xb50046 , 0xb80046 , 0xbc0046 , 0xc20046 , 0xc80046 , 0xce0046 , 0xd30046 , 0xd90046 , 0xde0046 , 0xe30046 , 0xe80046 , 0xec0046 , 0xf00046 , 0xf40046 , 0xf80045 , 0xfb0045 , 0xfd0044 , 0xff0243 , 0xff0542 , 0xff0940 , 0xff0d3f , 0xff113d , 0xff163b , 0xff1b39 , 0xff2037 , 0xff2635 , 0xff2c33 , 0xff3230 , 0xff392d , 0xff3f2b , 0xff4628 , 0xff4d25 , 0xff5422 , 0xff5b1f , 0xff621c , 0xff6a1a , 0xfe7117 , 0xfd7914 , 0xfc8011 , 0xfb880e , 0xfa900c , 0xf89809 , 0xf7a006 , 0xf6a804 , 0xf5b002 , 0xf3b700 , 0xf2bf00 , 0xf1c600 , 0xf0cd00 , 0xefd400 , 0xeeda00 , 0xede000 , 0xece600 , 0xebeb00 , 0xebf000 , 0xebf500 , 0xeaf900 , 0xeafd00 , 0xeaff01 , 0xeaff04 , 0xeaff07 , 0xebff0b , 0xecff10 , 0xedff15 , 0xeeff1b , 0xefff20 , 0xf0ff27 , 0xf1ff2d , 0xf3ff34 , 0xf4ff3c , 0xf6ff44 , 0xf7ff4c , 0xf9ff54 , 0xfaff5c , 0xfcff65 , 0xfdff6d , 0xfeff75 , 0xfeff7e , 0xfeff86 , 0xfeff8f , 0xfeff98 , 0xfeffa0 , 0xfeffa8 , 0xfeffb0 , 0xfeffb8 , 0xfeffc0 , 0xfeffc7 , 0xfeffce , 0xfeffd5 , 0xfeffdb , 0xfeffe1 , 0xfeffe7 , 0xfeffec , 0xfefff0 , 0xfefff4 , 0xfefff7 , 0xfefffa , 0xfefefc , 0xfefefd , 0xfefefe , 0xfcfefe , 0xf9fefd , 0xf6fefc , 0xf3fefb , 0xeffef9 , 0xecfef6 , 0xe7fef3 , 0xe3fef0 , 0xdffeed , 0xdafee9 , 0xd5fee4 , 0xd0ffdf , 0xcbffdb , 0xc6ffd6 , 0xc0ffd0 , 0xbbffcb , 0xb5ffc4 , 0xafffbe , 0xa9ffb8 , 0xa3ffb2 , 0x9dffab , 0x97ffa5 , 0x91ff9e , 0x8aff97 , 0x84ff90 , 0x7eff8a , 0x76ff81 , 0x6eff78 , 0x66ff70 , 0x5eff67 , 0x57ff5f , 0x4fff57 , 0x48ff4f , 0x41ff47 , 0x3aff40 , 0x33ff39 , 0x2dff33 , 0x26ff2d , 0x20ff27 , 0x1bff23 , 0x16ff1e , 0x11ff1b , 0xdff18 , 0x8ff15 , 0x5ff14 , 0x2ff13 , 0x1ff13 , 0x1ff15 , 0x1ff17 , 0x1ff1b , 0x1ff20 , 0x1ff25 , 0x1ff2c , 0x1ff33 , 0x1ff3b , 0x1ff44 , 0x1ff4d , 0x1ff56 , 0x3ff61 , 0x5ff6b , 0x8ff76 , 0xaff7d , 0xcff83 , 0xeff89 , 0x10ff90 , 0x11ff96 , 0x13ff9c , 0x15ffa2 , 0x17ffa8 , 0x19ffae , 0x1affb4 , 0x1cffb9 , 0x1effbf , 0x1ffec5 , 0x21feca , 0x22fccf , 0x24fcd4 , 0x25fad9 , 0x26f9de , 0x27f8e3 , 0x28f6e7 , 0x29f5eb , 0x2af3ee , 0x2bf1f2 , 0x2beff5 , 0x2bedf8 , 0x2cebfb , 0x2ce9fd , 0x2ce5ff , 0x2ce2ff , 0x2bdeff , 0x2bdaff , 0x2ad5ff , 0x2ad0ff , 0x29cbff , 0x28c6ff , 0x28c0ff , 0x27bbff , 0x26b5ff , 0x25afff , 0x24a9ff , 0x22a2ff , 0x219cff , 0x2095ff , 0x1f8fff , 0x1e88ff , 0x1c81ff , 0x1b7aff , 0x1974ff , 0x186dff , 0x1766ff , 0x155fff , 0x1459ff , 0x1352ff , 0x114cff , 0x1046ff , 0xf3fff , 0xd39ff , 0xc33ff , 0xb2dff , 0xa28ff , 0x923ff , 0x71dff , 0x619ff , 0x614ff , 0x510ff , 0x40cff , 0x308ff , 0x305ff , 0x202ff , 0x202ff}

,

/* 8 rgb-green */

{0x0 , 0x100 , 0x200 , 0x300 , 0x400 , 0x500 , 0x600 , 0x700 , 0x800 , 0x900 , 0xa00 , 0xb00 , 0xc00 , 0xd00 , 0xe00 , 0xf00 , 0x1000 , 0x1100 , 0x1200 , 0x1300 , 0x1400 , 0x1500 , 0x1600 , 0x1700 , 0x1800 , 0x1900 , 0x1a00 , 0x1b00 , 0x1c00 , 0x1d00 , 0x1e00 , 0x1f00 , 0x2000 , 0x2100 , 0x2200 , 0x2300 , 0x2400 , 0x2500 , 0x2600 , 0x2700 , 0x2800 , 0x2900 , 0x2a00 , 0x2b00 , 0x2c00 , 0x2d00 , 0x2e00 , 0x2f00 , 0x3000 , 0x3100 , 0x3200 , 0x3300 , 0x3400 , 0x3500 , 0x3600 , 0x3700 , 0x3800 , 0x3900 , 0x3a00 , 0x3b00 , 0x3c00 , 0x3d00 , 0x3e00 , 0x3f00 , 0x4000 , 0x4100 , 0x4200 , 0x4300 , 0x4400 , 0x4500 , 0x4600 , 0x4700 , 0x4800 , 0x4900 , 0x4a00 , 0x4b00 , 0x4c00 , 0x4d00 , 0x4e00 , 0x4f00 , 0x5000 , 0x5100 , 0x5200 , 0x5300 , 0x5400 , 0x5500 , 0x5600 , 0x5700 , 0x5800 , 0x5900 , 0x5a00 , 0x5b00 , 0x5c00 , 0x5d00 , 0x5e00 , 0x5f00 , 0x6000 , 0x6100 , 0x6200 , 0x6300 , 0x6400 , 0x6500 , 0x6600 , 0x6700 , 0x6800 , 0x6900 , 0x6a00 , 0x6b00 , 0x6c00 , 0x6d00 , 0x6e00 , 0x6f00 , 0x7000 , 0x7100 , 0x7200 , 0x7300 , 0x7400 , 0x7500 , 0x7600 , 0x7700 , 0x7800 , 0x7900 , 0x7a00 , 0x7b00 , 0x7c00 , 0x7d00 , 0x7e00 , 0x7f00 , 0x8000 , 0x8100 , 0x8200 , 0x8300 , 0x8400 , 0x8500 , 0x8600 , 0x8700 , 0x8800 , 0x8900 , 0x8a00 , 0x8b00 , 0x8c00 , 0x8d00 , 0x8e00 , 0x8f00 , 0x9000 , 0x9100 , 0x9200 , 0x9300 , 0x9400 , 0x9500 , 0x9600 , 0x9700 , 0x9800 , 0x9900 , 0x9a00 , 0x9b00 , 0x9c00 , 0x9d00 , 0x9e00 , 0x9f00 , 0xa000 , 0xa100 , 0xa200 , 0xa300 , 0xa400 , 0xa500 , 0xa600 , 0xa700 , 0xa800 , 0xa900 , 0xaa00 , 0xab00 , 0xac00 , 0xad00 , 0xae00 , 0xaf00 , 0xb000 , 0xb100 , 0xb200 , 0xb300 , 0xb400 , 0xb500 , 0xb600 , 0xb700 , 0xb800 , 0xb900 , 0xba00 , 0xbb00 , 0xbc00 , 0xbd00 , 0xbe00 , 0xbf00 , 0xc000 , 0xc100 , 0xc200 , 0xc300 , 0xc400 , 0xc500 , 0xc600 , 0xc700 , 0xc800 , 0xc900 , 0xca00 , 0xcb00 , 0xcc00 , 0xcd00 , 0xce00 , 0xcf00 , 0xd000 , 0xd100 , 0xd200 , 0xd300 , 0xd400 , 0xd500 , 0xd600 , 0xd700 , 0xd800 , 0xd900 , 0xda00 , 0xdb00 , 0xdc00 , 0xdd00 , 0xde00 , 0xdf00 , 0xe000 , 0xe100 , 0xe200 , 0xe300 , 0xe400 , 0xe500 , 0xe600 , 0xe700 , 0xe800 , 0xe900 , 0xea00 , 0xeb00 , 0xec00 , 0xed00 , 0xee00 , 0xef00 , 0xf000 , 0xf100 , 0xf200 , 0xf300 , 0xf400 , 0xf500 , 0xf600 , 0xf700 , 0xf800 , 0xf900 , 0xfa00 , 0xfb00 , 0xfc00 , 0xfd00 , 0xfe00 , 0xff00}

,

/* 9 hot */

{0x10103 , 0x10103 , 0x10106 , 0x10108 , 0x1010b , 0x1010e , 0x10110 , 0x10113 , 0x10116 , 0x10118 , 0x1011b , 0x1011e , 0x10120 , 0x10123 , 0x10126 , 0x10128 , 0x1012b , 0x1012e , 0x10130 , 0x10133 , 0x10136 , 0x10138 , 0x1013b , 0x1013e , 0x10140 , 0x10142 , 0x10145 , 0x10147 , 0x1014a , 0x1014d , 0x1014f , 0x10152 , 0x10155 , 0x10157 , 0x1015a , 0x1015c , 0x1015f , 0x10162 , 0x10164 , 0x10167 , 0x1016a , 0x1016c , 0x1016f , 0x10172 , 0x10174 , 0x10177 , 0x1017a , 0x1017c , 0x1017f , 0x10182 , 0x10184 , 0x10187 , 0x1018a , 0x1018c , 0x1018f , 0x10192 , 0x10194 , 0x10197 , 0x1019a , 0x1019c , 0x1019f , 0x101a2 , 0x101a4 , 0x101a7 , 0x101aa , 0x101ac , 0x101af , 0x101b1 , 0x101b4 , 0x101b7 , 0x101b9 , 0x101bc , 0x101bf , 0x101c0 , 0x101c3 , 0x101c6 , 0x101c8 , 0x101cb , 0x101ce , 0x101d0 , 0x101d3 , 0x101d6 , 0x101d8 , 0x101db , 0x101de , 0x101e0 , 0x101e3 , 0x101e6 , 0x101e8 , 0x101eb , 0x101ee , 0x101f0 , 0x101f3 , 0x101f6 , 0x101f8 , 0x101fb , 0x101fe , 0x103fe , 0x106fe , 0x108fe , 0x10bfe , 0x10efe , 0x110fe , 0x113fe , 0x116fe , 0x118fe , 0x11bfe , 0x11efe , 0x120fe , 0x123fe , 0x126fe , 0x128fe , 0x12bfe , 0x12efe , 0x130fe , 0x133fe , 0x136fe , 0x138fe , 0x13bfe , 0x13efe , 0x140fe , 0x142fe , 0x145fe , 0x147fe , 0x14afe , 0x14dfe , 0x14ffe , 0x152fe , 0x155fe , 0x157fe , 0x15afe , 0x15cfe , 0x15ffe , 0x162fe , 0x164fe , 0x167fe , 0x16afe , 0x16cfe , 0x16ffe , 0x172fe , 0x174fe , 0x177fe , 0x17afe , 0x17cfe , 0x17ffe , 0x182fe , 0x184fe , 0x187fe , 0x18afe , 0x18cfe , 0x18ffe , 0x192fe , 0x194fe , 0x197fe , 0x19afe , 0x19cfe , 0x19ffe , 0x1a2fe , 0x1a4fe , 0x1a7fe , 0x1aafe , 0x1acfe , 0x1affe , 0x1b1fe , 0x1b4fe , 0x1b7fe , 0x1b9fe , 0x1bcfe , 0x1bffe , 0x1c0fe , 0x1c3fe , 0x1c6fe , 0x1c8fe , 0x1cbfe , 0x1cefe , 0x1d0fe , 0x1d3fe , 0x1d6fe , 0x1d8fe , 0x1dbfe , 0x1defe , 0x1e0fe , 0x1e3fe , 0x1e6fe , 0x1e8fe , 0x1ebfe , 0x1eefe , 0x1f0fe , 0x1f3fe , 0x1f6fe , 0x1f8fe , 0x1fbfe , 0x1fefe , 0x4fefe , 0x8fefe , 0xcfefe , 0x10fefe , 0x14fefe , 0x18fefe , 0x1cfefe , 0x20fefe , 0x24fefe , 0x28fefe , 0x2cfefe , 0x30fefe , 0x34fefe , 0x38fefe , 0x3cfefe , 0x40fefe , 0x43fefe , 0x47fefe , 0x4bfefe , 0x4ffefe , 0x53fefe , 0x57fefe , 0x5bfefe , 0x5ffefe , 0x63fefe , 0x67fefe , 0x6bfefe , 0x6ffefe , 0x73fefe , 0x77fefe , 0x7bfefe , 0x7ffefe , 0x83fefe , 0x87fefe , 0x8bfefe , 0x8ffefe , 0x93fefe , 0x97fefe , 0x9bfefe , 0x9ffefe , 0xa3fefe , 0xa7fefe , 0xabfefe , 0xaffefe , 0xb3fefe , 0xb7fefe , 0xbbfefe , 0xbffefe , 0xc2fefe , 0xc6fefe , 0xcafefe , 0xcefefe , 0xd2fefe , 0xd6fefe , 0xdafefe , 0xdefefe , 0xe2fefe , 0xe6fefe , 0xeafefe , 0xeefefe , 0xf2fefe , 0xf6fefe , 0xf6fefe}

,

/* 10 siemens */

{0x200102 , 0x200102 , 0x300102 , 0x400103 , 0x4f0103 , 0x5f0104 , 0x5f0104 , 0x5f0104 , 0x5f0104 , 0x5f0104 , 0x5f0104 , 0x5f0104 , 0x5f0104 , 0x5f0104 , 0x5f0104 , 0x5f0104 , 0x630104 , 0x680104 , 0x6c0104 , 0x710104 , 0x760104 , 0x7a0104 , 0x7f0104 , 0x830105 , 0x880105 , 0x8d0105 , 0x910105 , 0x960105 , 0x9a0105 , 0x9f0105 , 0xa40105 , 0xa80106 , 0xad0106 , 0xb20106 , 0xb60106 , 0xbb0106 , 0xbf0106 , 0xc30106 , 0xc80106 , 0xcc0107 , 0xd10107 , 0xd50107 , 0xda0107 , 0xdf0107 , 0xe30107 , 0xe80107 , 0xed0108 , 0xec0108 , 0xe8010d , 0xe30112 , 0xde0117 , 0xd9011d , 0xd40122 , 0xcf0127 , 0xca012c , 0xc50132 , 0xc00137 , 0xbc013c , 0xb70140 , 0xb20146 , 0xad014b , 0xa80150 , 0xa30155 , 0x9e015b , 0x990160 , 0x940165 , 0x8f016a , 0x8a0170 , 0x850175 , 0x80017a , 0x7b017f , 0x770185 , 0x72018a , 0x6d018f , 0x680194 , 0x63019a , 0x5e019f , 0x5901a4 , 0x5401a9 , 0x4f01af , 0x4a01b4 , 0x4501b9 , 0x4001be , 0x3c01c3 , 0x3701c8 , 0x3201cd , 0x2d01d2 , 0x2801d8 , 0x2301dd , 0x1e01e2 , 0x1901e7 , 0x1401ed , 0xf01f2 , 0xa01f7 , 0x601fd , 0xf01f3 , 0xe02f3 , 0xe04f3 , 0xe06f3 , 0xe08f3 , 0xe09f3 , 0xe0bf3 , 0xe0df3 , 0xe0ff3 , 0xe10f3 , 0xe12f3 , 0xd14f3 , 0xd16f3 , 0xd18f4 , 0xd19f4 , 0xd1bf4 , 0xd1df4 , 0xd1ff4 , 0xd20f4 , 0xd22f4 , 0xd24f4 , 0xc26f4 , 0xc28f4 , 0xc29f4 , 0xc2bf4 , 0xc2df5 , 0xc2ff5 , 0xc30f5 , 0xc32f5 , 0xc34f5 , 0xc36f5 , 0xb38f5 , 0xb39f5 , 0xb3bf5 , 0xb3df5 , 0xb3ff5 , 0xb40f5 , 0xb41f6 , 0xb43f6 , 0xb45f6 , 0xb46f6 , 0xa48f6 , 0xa4af6 , 0xa4cf6 , 0xa4ef6 , 0xa4ff6 , 0xa51f6 , 0xa53f6 , 0xa55f6 , 0xa56f6 , 0xa58f7 , 0x95af7 , 0x95cf7 , 0x95ef7 , 0x95ff7 , 0x961f7 , 0x963f7 , 0x965f7 , 0x966f7 , 0x968f7 , 0x96af7 , 0x86cf7 , 0x86ef8 , 0x86ff8 , 0x871f8 , 0x873f8 , 0x875f8 , 0x876f8 , 0x878f8 , 0x87af8 , 0x87cf8 , 0x77ef8 , 0x77ff8 , 0x781f8 , 0x783f9 , 0x785f9 , 0x786f9 , 0x788f9 , 0x78af9 , 0x78cf9 , 0x78df9 , 0x68ff9 , 0x691f9 , 0x693f9 , 0x695f9 , 0x696f9 , 0x698f9 , 0x69afa , 0x69cfa , 0x69dfa , 0x69ffa , 0x5a1fa , 0x5a3fa , 0x5a5fa , 0x5a6fa , 0x5a8fa , 0x5aafa , 0x5acfa , 0x5adfa , 0x5affb , 0x5b1fb , 0x4b3fb , 0x4b5fb , 0x4b6fb , 0x4b8fb , 0x4bafb , 0x4bcfb , 0x4bdfb , 0x4bffb , 0x4c0fb , 0x4c2fb , 0x4c4fc , 0x4c4fc , 0x3c8fb , 0x3ccfb , 0x3d0fb , 0x3d4fa , 0x3d7fa , 0x2dbfa , 0x2dffa , 0x2e3f9 , 0x2e7f9 , 0x2eaf9 , 0x1eef9 , 0x1f2f8 , 0x1f6f8 , 0x1faf8 , 0x1fef8 , 0x1fef8 , 0x9fef8 , 0x11fef8 , 0x19fef8 , 0x21fef8 , 0x2afef8 , 0x32fef9 , 0x3afef9 , 0x41fef9 , 0x4afef9 , 0x52fef9 , 0x5afefa , 0x62fefa , 0x6afefa , 0x73fefa , 0x7bfefa , 0x83fefb , 0x8bfefb , 0x94fefb , 0x9cfefb , 0xa4fefb , 0xacfefc , 0xb4fefc , 0xbdfefc , 0xc4fefc , 0xccfefc , 0xd4fefd , 0xddfefd , 0xe5fefd , 0xedfefd , 0xf5fefd , 0x10101 , 0x10101}

,

/* 11 cold */

{0x141414 , 0x141414 , 0x272726 , 0x393936 , 0x4a4a44 , 0x5a5a51 , 0x6a6a5d , 0x797968 , 0x878772 , 0x94947a , 0xa1a181 , 0xadad87 , 0xb8b88c , 0xc1c190 , 0xcaca93 , 0xd3d396 , 0xdbdb97 , 0xe2e297 , 0xe8e897 , 0xeeee96 , 0xf3f394 , 0xf7f791 , 0xfafa8e , 0xfcfc8b , 0xfefe87 , 0xfefe82 , 0xfefe7d , 0xfefe78 , 0xfefe73 , 0xfefe6e , 0xfefe69 , 0xfefe64 , 0xfefe5f , 0xfefe5a , 0xfefe55 , 0xfefe50 , 0xfefe4b , 0xfefe46 , 0xfefe41 , 0xfefe3d , 0xfefe38 , 0xfefe33 , 0xfefe2e , 0xfefe29 , 0xfefe24 , 0xfefe1f , 0xfefe1a , 0xfefe15 , 0xfefe10 , 0xfefe0b , 0xfefe06 , 0xfefe01 , 0xfefc03 , 0xfef906 , 0xfef708 , 0xfef40b , 0xfef20d , 0xfeef10 , 0xfeed12 , 0xfeea15 , 0xfee817 , 0xfee51a , 0xfee31c , 0xfee01f , 0xfede21 , 0xfedb24 , 0xfed926 , 0xfed629 , 0xfed42b , 0xfed12e , 0xfecf30 , 0xfecc33 , 0xfeca35 , 0xfec738 , 0xfec53a , 0xfec23d , 0xfec03f , 0xfebe41 , 0xfebc43 , 0xfeb946 , 0xfeb748 , 0xfeb44b , 0xfeb24d , 0xfeaf50 , 0xfead52 , 0xfeaa55 , 0xfea857 , 0xfea55a , 0xfea35c , 0xfea05f , 0xfe9e61 , 0xfe9b64 , 0xfe9966 , 0xfe9669 , 0xfe946b , 0xfe916e , 0xfe8f70 , 0xfe8c73 , 0xfe8a75 , 0xfe8778 , 0xfe857a , 0xfe827d , 0xfe8080 , 0xfe7d82 , 0xfe7a85 , 0xfe7887 , 0xfe758a , 0xfe738c , 0xfe708f , 0xfe6e91 , 0xfe6b94 , 0xfe6996 , 0xfe6699 , 0xfe649b , 0xfe619e , 0xfe5fa0 , 0xfe5ca3 , 0xfe5aa5 , 0xfe57a8 , 0xfe55aa , 0xfe52ad , 0xfe50af , 0xfe4db2 , 0xfe4bb4 , 0xfe48b7 , 0xfe46b9 , 0xfe43bc , 0xfe41be , 0xfe3fc0 , 0xfe3dc2 , 0xfe3ac5 , 0xfe38c7 , 0xfe35ca , 0xfe33cc , 0xfe30cf , 0xfe2ed1 , 0xfe2bd4 , 0xfe29d6 , 0xfe26d9 , 0xfe24db , 0xfe21de , 0xfe1fe0 , 0xfe1ce3 , 0xfe1ae5 , 0xfe17e8 , 0xfe15ea , 0xfe12ed , 0xfe10ef , 0xfe0df2 , 0xfe0bf4 , 0xfe08f7 , 0xfe06f9 , 0xfe03fc , 0xfe01fe , 0xfc03fe , 0xf906fe , 0xf708fe , 0xf40bfe , 0xf20dfe , 0xef10fe , 0xed12fe , 0xea15fe , 0xe817fe , 0xe51afe , 0xe31cfe , 0xe01ffe , 0xde21fe , 0xdb24fe , 0xd926fe , 0xd629fe , 0xd42bfe , 0xd12efe , 0xcf30fe , 0xcc33fe , 0xca35fe , 0xc738fe , 0xc53afe , 0xc23dfe , 0xc03ffe , 0xbe41fe , 0xbc43fe , 0xb946fe , 0xb748fe , 0xb44bfe , 0xb24dfe , 0xaf50fe , 0xad52fe , 0xaa55fe , 0xa857fe , 0xa55afe , 0xa35cfe , 0xa05ffe , 0x9e61fe , 0x9b64fe , 0x9966fe , 0x9669fe , 0x946bfe , 0x916efe , 0x8f70fe , 0x8c73fe , 0x8a75fe , 0x8778fe , 0x857afe , 0x827dfe , 0x7f80fe , 0x7d82fe , 0x7a85fe , 0x7887fe , 0x758afe , 0x738cfe , 0x708ffe , 0x6e91fe , 0x6b94fe , 0x6996fe , 0x6699fe , 0x649bfe , 0x619efe , 0x5fa0fe , 0x5ca3fe , 0x5aa5fe , 0x57a8fe , 0x55aafe , 0x52adfe , 0x50affe , 0x4db2fe , 0x4bb4fe , 0x48b7fe , 0x46b9fe , 0x43bcfe , 0x41befe , 0x3fc0fe , 0x3dc2fe , 0x3cc5fe , 0x3bc8fe , 0x3ccbfe , 0x3ccffe , 0x3ed2fe , 0x40d5fe , 0x42d9fe , 0x46dcfe , 0x4adffe , 0x50e3fe , 0x56e6fe , 0x5de9fe , 0x65ecfe , 0x6eeffe , 0x78f1fe , 0x83f4fe , 0x8ff6fe , 0x9bf8fe , 0xa9fafe , 0xb8fcfe , 0xc7fdfe , 0xd9fefe , 0xebfefe , 0xebfefe}

,

/* 12 003-ice */

{0x8c9c00 , 0x8c9d00 , 0x8d9e00 , 0x8e9f00 , 0x8fa000 , 0x90a100 , 0x91a200 , 0x92a300 , 0x93a500 , 0x94a600 , 0x95a700 , 0x97a900 , 0x98aa00 , 0x99ab00 , 0x9bad00 , 0x9cae00 , 0x9eb000 , 0x9fb100 , 0xa0b200 , 0xa1b300 , 0xa2b400 , 0xa3b500 , 0xa4b600 , 0xa5b700 , 0xa6b800 , 0xa6b800 , 0xa7b900 , 0xa7ba00 , 0xa8bb00 , 0xa8bb00 , 0xa9bc00 , 0xa9bd00 , 0xaabe00 , 0xaabe00 , 0xabbf00 , 0xacc000 , 0xadc100 , 0xadc100 , 0xaec200 , 0xafc300 , 0xb0c400 , 0xb4c302 , 0xb8c304 , 0xbcc207 , 0xc0c209 , 0xc4c20b , 0xc8c10e , 0xccc110 , 0xd1c113 , 0xd2bf14 , 0xd3be15 , 0xd5bd16 , 0xd6bc18 , 0xd7bb19 , 0xd9ba1a , 0xdab91b , 0xdcb81d , 0xddb61f , 0xdfb422 , 0xe1b324 , 0xe3b127 , 0xe4af2a , 0xe6ae2c , 0xe8ac2f , 0xeaab32 , 0xe8a931 , 0xe7a831 , 0xe6a731 , 0xe5a631 , 0xe4a530 , 0xe3a430 , 0xe2a330 , 0xe1a230 , 0xe2a033 , 0xe39e37 , 0xe59c3b , 0xe69a3f , 0xe79843 , 0xe99647 , 0xea944b , 0xec924f , 0xed8f53 , 0xee8c57 , 0xef8a5b , 0xf1875f , 0xf28463 , 0xf38267 , 0xf47f6b , 0xf67d70 , 0xf67a72 , 0xf77875 , 0xf77678 , 0xf8747b , 0xf8717d , 0xf96f80 , 0xf96d83 , 0xfa6b86 , 0xfa6989 , 0xfa678c , 0xfa658f , 0xfa6492 , 0xfa6295 , 0xfa6098 , 0xfa5e9b , 0xfb5d9e , 0xfa5ba1 , 0xfa5aa5 , 0xfa58a8 , 0xfa57ac , 0xfa55af , 0xfa54b3 , 0xfa52b6 , 0xfa51ba , 0xfa51bb , 0xfa52bd , 0xfa53bf , 0xfa54c1 , 0xfa54c3 , 0xfa55c5 , 0xfa56c7 , 0xfa57c9 , 0xf957cb , 0xf858cd , 0xf858cf , 0xf759d1 , 0xf65ad3 , 0xf65ad5 , 0xf55bd7 , 0xf55cd9 , 0xf35cda , 0xf15ddc , 0xef5ddd , 0xed5edf , 0xeb5fe0 , 0xe95fe2 , 0xe760e3 , 0xe661e5 , 0xe660e6 , 0xe660e8 , 0xe660e9 , 0xe660eb , 0xe65fed , 0xe65fee , 0xe65ff0 , 0xe65ff2 , 0xe55ef3 , 0xe45ef4 , 0xe35ef5 , 0xe25ef6 , 0xe15df7 , 0xe05df8 , 0xdf5df9 , 0xde5dfa , 0xdb5dfa , 0xd95dfa , 0xd65dfa , 0xd45dfa , 0xd15dfa , 0xcf5dfa , 0xcc5dfa , 0xca5dfa , 0xc75cfa , 0xc45cfa , 0xc15bfa , 0xbf5bfa , 0xbc5bfa , 0xb95afa , 0xb65afa , 0xb45afa , 0xb159fa , 0xaf58fa , 0xad58fa , 0xab57fa , 0xa956fa , 0xa756fa , 0xa555fa , 0xa355fa , 0xa053fa , 0x9d51fa , 0x9b4ffa , 0x984dfa , 0x954bfa , 0x9349fa , 0x9047fa , 0x8e45fb , 0x8b44fa , 0x8943fa , 0x8643fa , 0x8442fa , 0x8241fa , 0x7f41fa , 0x7d40fa , 0x7b40fa , 0x793efa , 0x783dfa , 0x773cfa , 0x763bfa , 0x7539fa , 0x7438fa , 0x7337fa , 0x7236fa , 0x7135fa , 0x7034fa , 0x6f33fa , 0x6e32fa , 0x6d31fa , 0x6c30fa , 0x6b2ffa , 0x6a2ffa , 0x682dfa , 0x672cfa , 0x652afa , 0x6429fa , 0x6227fa , 0x6126fa , 0x5f24fa , 0x5e23fa , 0x5c21fa , 0x5b1ffa , 0x5a1dfa , 0x591bfa , 0x5719fa , 0x5617fa , 0x5515fa , 0x5413fb , 0x5110fb , 0x4f0efb , 0x4c0bfb , 0x4a09fb , 0x4707fb , 0x4504fb , 0x4202fb , 0x4000fb , 0x3b00fa , 0x3601f9 , 0x3101f8 , 0x2d02f7 , 0x2802f6 , 0x2303f5 , 0x1e03f4 , 0x1a04f3 , 0x1a03f1 , 0x1a03ef , 0x1a02ee , 0x1a02ec , 0x1a01ea , 0x1a01e9 , 0x1a00e7 , 0x1b00e6 , 0x1b00e6 , 0x1b00e6 , 0x1b00e6 , 0x1b00e6 , 0x1b00e6 , 0x1b00e6 , 0x1b00e6}

,

/* 13 16_equal */

{0x0 , 0x7c , 0x7c , 0x7c , 0x7c , 0x7c , 0x7c , 0x7c , 0x7c , 0x7c , 0x7c , 0x7c , 0x7c , 0x7c , 0x7c , 0x7c , 0xc8 , 0xc8 , 0xc8 , 0xc8 , 0xc8 , 0xc8 , 0xc8 , 0xc8 , 0xc8 , 0xc8 , 0xc8 , 0xc8 , 0xc8 , 0xc8 , 0xc8 , 0xc8 , 0x24fc , 0x24fc , 0x24fc , 0x24fc , 0x24fc , 0x24fc , 0x24fc , 0x24fc , 0x24fc , 0x24fc , 0x24fc , 0x24fc , 0x24fc , 0x24fc , 0x24fc , 0x24fc , 0x90fc , 0x90fc , 0x90fc , 0x90fc , 0x90fc , 0x90fc , 0x90fc , 0x90fc , 0x90fc , 0x90fc , 0x90fc , 0x90fc , 0x90fc , 0x90fc , 0x90fc , 0x90fc , 0xfcfc , 0xfcfc , 0xfcfc , 0xfcfc , 0xfcfc , 0xfcfc , 0xfcfc , 0xfcfc , 0xfcfc , 0xfcfc , 0xfcfc , 0xfcfc , 0xfcfc , 0xfcfc , 0xfcfc , 0xfcfc , 0x4e080 , 0x4e080 , 0x4e080 , 0x4e080 , 0x4e080 , 0x4e080 , 0x4e080 , 0x4e080 , 0x4e080 , 0x4e080 , 0x4e080 , 0x4e080 , 0x4e080 , 0x4e080 , 0xe488 , 0x4e080 , 0x8c400 , 0x8c400 , 0x8c400 , 0x8c400 , 0x8c400 , 0x8c400 , 0x8c400 , 0x8c400 , 0x8c400 , 0x8c400 , 0x8c400 , 0x8c400 , 0x8c400 , 0x8c400 , 0x8c400 , 0x8c400 , 0xc0bc00 , 0xc0bc00 , 0xc0bc00 , 0xc0bc00 , 0xc0bc00 , 0xc0bc00 , 0xc0bc00 , 0xc0bc00 , 0xc0bc00 , 0xc0bc00 , 0xc0bc00 , 0xc0bc00 , 0xc0bc00 , 0xc0bc00 , 0xc0bc00 , 0xc0bc00 , 0xd06408 , 0xd06408 , 0xd06408 , 0xd06408 , 0xd06408 , 0xd06408 , 0xd06408 , 0xd06408 , 0xd06408 , 0xd06408 , 0xd06408 , 0xd06408 , 0xd06408 , 0xd06408 , 0xd06408 , 0xd06408 , 0xe81018 , 0xe81018 , 0xe81018 , 0xe81018 , 0xe81018 , 0xe81018 , 0xe81018 , 0xe81018 , 0xe81018 , 0xe81018 , 0xe81018 , 0xe81018 , 0xe81018 , 0xe81018 , 0xe81018 , 0xe81018 , 0xc4144c , 0xc4144c , 0xc4144c , 0xc4144c , 0xc4144c , 0xc4144c , 0xc4144c , 0xc4144c , 0xc4144c , 0xc4144c , 0xc4144c , 0xc4144c , 0xc4144c , 0xc4144c , 0xc4144c , 0xc4144c , 0xb02474 , 0xb02474 , 0xb02474 , 0xb02474 , 0xb02474 , 0xb02474 , 0xb02474 , 0xb02474 , 0xb02474 , 0xb02474 , 0xb02474 , 0xb02474 , 0xb02474 , 0xb02474 , 0xb02474 , 0xb02474 , 0x984888 , 0x984888 , 0x984888 , 0x984888 , 0x984888 , 0x984888 , 0x984888 , 0x984888 , 0x984888 , 0x984888 , 0x984888 , 0x984888 , 0x984888 , 0x984888 , 0x984888 , 0x984888 , 0x808080 , 0x808080 , 0x808080 , 0x808080 , 0x808080 , 0x808080 , 0x808080 , 0x808080 , 0x808080 , 0x808080 , 0x808080 , 0x808080 , 0x808080 , 0x808080 , 0x808080 , 0x808080 , 0xc0c0c0 , 0xc0c0c0 , 0xc0c0c0 , 0xc0c0c0 , 0xc0c0c0 , 0xc0c0c0 , 0xc0c0c0 , 0xc0c0c0 , 0xc0c0c0 , 0xc0c0c0 , 0xc0c0c0 , 0xc0c0c0 , 0xc0c0c0 , 0xc0c0c0 , 0xc0c0c0 , 0xc0c0c0 , 0xfcfcfc , 0xfcfcfc , 0xfcfcfc , 0xfcfcfc , 0xfcfcfc , 0xfcfcfc , 0xfcfcfc , 0xfcfcfc , 0xfcfcfc , 0xfcfcfc , 0xfcfcfc , 0xfcfcfc , 0xfcfcfc , 0xfcfcfc , 0xfcfcfc , 0xfcfcfc}

,

/* 14 Red */

{0x30103 , 0x30103 , 0x30104 , 0x30105 , 0x30106 , 0x30108 , 0x30109 , 0x3010a , 0x3010c , 0x3010d , 0x3020e , 0x3020f , 0x30211 , 0x30312 , 0x30313 , 0x30315 , 0x30316 , 0x30317 , 0x30319 , 0x3031a , 0x3031b , 0x3031c , 0x3031e , 0x3031f , 0x30320 , 0x30322 , 0x30323 , 0x30324 , 0x30325 , 0x30327 , 0x30328 , 0x30329 , 0x3032b , 0x3032c , 0x3032d , 0x3032f , 0x30330 , 0x30331 , 0x30332 , 0x30334 , 0x30335 , 0x30336 , 0x30338 , 0x30339 , 0x3033a , 0x3033c , 0x4043d , 0x4043e , 0x4043f , 0x40440 , 0x40441 , 0x50542 , 0x50544 , 0x50545 , 0x50546 , 0x50547 , 0x60649 , 0x6064a , 0x6064b , 0x6064d , 0x6064e , 0x7074f , 0x70751 , 0x70752 , 0x70753 , 0x70754 , 0x80856 , 0x80857 , 0x80858 , 0x8085a , 0x8085b , 0x9095c , 0x9095e , 0x9095f , 0x90960 , 0x90961 , 0xa0a63 , 0xa0a64 , 0xa0a65 , 0xa0a67 , 0xa0a68 , 0xb0b69 , 0xb0b6a , 0xb0b6c , 0xb0b6d , 0xb0b6e , 0xb0c70 , 0xb0c71 , 0xc0c72 , 0xc0c74 , 0xc0c75 , 0xc0c76 , 0xc0d77 , 0xc0d79 , 0xc0d7a , 0xc0d7b , 0xd0d7d , 0xd0e7e , 0xd0e7f , 0xd0e80 , 0xd0e82 , 0xd0e83 , 0xd0f84 , 0xd0f86 , 0xe0f87 , 0xe0f88 , 0xe0f8a , 0xe108b , 0xe108c , 0xe108d , 0xe108f , 0xf1090 , 0xf1091 , 0xf1193 , 0xf1194 , 0xf1195 , 0xf1197 , 0xf1198 , 0xf1299 , 0x10129a , 0x10129c , 0x10129d , 0x10129e , 0x1013a0 , 0x1013a1 , 0x1013a2 , 0x1014a3 , 0x1014a5 , 0x1014a6 , 0x1015a7 , 0x1015a9 , 0x1015aa , 0x1116ab , 0x1116ad , 0x1116ae , 0x1117af , 0x1117b0 , 0x1117b2 , 0x1118b3 , 0x1118b4 , 0x1118b6 , 0x1118b7 , 0x1119b8 , 0x1119ba , 0x1119bb , 0x111abc , 0x111abd , 0x111abf , 0x121bbf , 0x121bc0 , 0x121bc2 , 0x121cc3 , 0x121cc4 , 0x121cc5 , 0x121dc7 , 0x121dc8 , 0x121dc9 , 0x121ecb , 0x121ecc , 0x121ecd , 0x121fcf , 0x121fd0 , 0x121fd1 , 0x121fd2 , 0x1220d4 , 0x1220d5 , 0x1220d6 , 0x1120d8 , 0x1120d9 , 0x1120da , 0x1120dc , 0x1021dd , 0x1021de , 0x1021df , 0x1021e1 , 0x1021e2 , 0xf21e3 , 0xf21e5 , 0xf22e6 , 0xf22e7 , 0xe22e8 , 0xe22ea , 0xe22eb , 0xe22ec , 0xe22ee , 0xd23ef , 0xd23f0 , 0xd23f2 , 0xd23f3 , 0xc23f4 , 0xc23f5 , 0xc23f7 , 0xc24f8 , 0xc24f9 , 0xc24fb , 0xc24fc , 0xd25fd , 0xf27fd , 0x132bfe , 0x1830fe , 0x1e35fe , 0x243afd , 0x2a40fd , 0x3044fd , 0x374afd , 0x3d50fc , 0x4255fc , 0x485bfc , 0x4f60fc , 0x5566fb , 0x5b6bfb , 0x6171fb , 0x6876fa , 0x6e7cfa , 0x7481fa , 0x7b87fa , 0x818cf9 , 0x8792f9 , 0x8d97f9 , 0x949df9 , 0x9aa2f8 , 0xa0a8f8 , 0xa6adf8 , 0xadb3f7 , 0xb3b8f7 , 0xb9bef7 , 0xbfc3f7 , 0xc5c8f6 , 0xcbcef6 , 0xd1d3f6 , 0xd7d9f6 , 0xdedef5 , 0xe4e4f5 , 0xeae9f5 , 0xf0eff6 , 0xf6f4f8 , 0xfbf9fa , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe}

,

/* 15 Thallium */

{0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10201 , 0x10201 , 0x10301 , 0x10401 , 0x10401 , 0x10501 , 0x10601 , 0x10601 , 0x10701 , 0x10701 , 0x10801 , 0x10901 , 0x10901 , 0x10a01 , 0x10b01 , 0x10c01 , 0x10d01 , 0x10d01 , 0x10e01 , 0x10e01 , 0x10f01 , 0x11001 , 0x11001 , 0x11101 , 0x11201 , 0x11301 , 0x11301 , 0x11301 , 0x11301 , 0x11301 , 0x11401 , 0x11401 , 0x11401 , 0x11401 , 0x11501 , 0x11501 , 0x11501 , 0x11601 , 0x11601 , 0x11601 , 0x11601 , 0x11701 , 0x11701 , 0x11701 , 0x11701 , 0x11801 , 0x11801 , 0x11801 , 0x11801 , 0x11901 , 0x11901 , 0x11901 , 0x11901 , 0x11a01 , 0x11a01 , 0x11b01 , 0x11b01 , 0x11c01 , 0x11c01 , 0x11d01 , 0x11d01 , 0x11e01 , 0x11e01 , 0x11f01 , 0x11f01 , 0x12001 , 0x12001 , 0x12101 , 0x12101 , 0x12201 , 0x12201 , 0x12301 , 0x12301 , 0x12401 , 0x12501 , 0x12501 , 0x12601 , 0x12601 , 0x12701 , 0x12801 , 0x12801 , 0x12901 , 0x22a02 , 0x22a02 , 0x32b03 , 0x32c03 , 0x42d04 , 0x42d04 , 0x52e05 , 0x52f05 , 0x63006 , 0x63006 , 0x63106 , 0x73307 , 0x83308 , 0x83408 , 0x93509 , 0x93609 , 0xa360a , 0xa370a , 0xb380b , 0xb380b , 0xb390b , 0xc3a0c , 0xc3b0c , 0xd3b0d , 0xd3c0d , 0xe3d0e , 0xe3d0e , 0xe3e0e , 0xf3f0f , 0xf400f , 0x104010 , 0x114211 , 0x114311 , 0x124412 , 0x124512 , 0x134613 , 0x134613 , 0x144714 , 0x144814 , 0x154915 , 0x154a15 , 0x164b16 , 0x164c16 , 0x174d17 , 0x174e17 , 0x184f18 , 0x185018 , 0x195119 , 0x195119 , 0x1a521a , 0x1a531a , 0x1c561c , 0x1d571d , 0x1e581e , 0x1f5a1f , 0x215b21 , 0x225d22 , 0x235e23 , 0x246024 , 0x266126 , 0x276327 , 0x286428 , 0x296629 , 0x2b672b , 0x2c692c , 0x2d6a2d , 0x2e6c2e , 0x306d30 , 0x316f31 , 0x327032 , 0x357335 , 0x367536 , 0x377637 , 0x397739 , 0x3a793a , 0x3b7a3b , 0x3c7c3c , 0x3e7d3e , 0x3f7f3f , 0x408040 , 0x418241 , 0x438443 , 0x468646 , 0x488848 , 0x4b8b4b , 0x4d8d4d , 0x508f50 , 0x529152 , 0x559355 , 0x579657 , 0x5c9a5c , 0x5f9c5f , 0x619f61 , 0x64a164 , 0x66a366 , 0x69a569 , 0x6ba76b , 0x6eaa6e , 0x70ac70 , 0x73ae73 , 0x75b075 , 0x78b278 , 0x7ab57a , 0x7db77d , 0x7fb97f , 0x82bb82 , 0x86bf86 , 0x8ac08a , 0x8ec28e , 0x93c493 , 0x9cc99c , 0xa0cba0 , 0xa4cea4 , 0xa9d0a9 , 0xadd2ad , 0xb2d5b2 , 0xb6d7b6 , 0xbad9ba , 0xbfdcbf , 0xc2dec2 , 0xc7e0c7 , 0xcbe3cb , 0xcfe5cf , 0xd4e7d4 , 0xd8ead8 , 0xddecdd , 0xe1eee1 , 0xe5f1e5 , 0xeaf3ea , 0xf4f2fb , 0xfbf7fc , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd}

,

/* 16 nih-image-fire1 */

{0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xfeffff , 0xfeffff , 0xfdffff , 0xfcffff , 0xf9ffff , 0xf6ffff , 0xf3ffff , 0xefffff , 0xecffff , 0xe9ffff , 0xe5ffff , 0xe2ffff , 0xdfffff , 0xdcffff , 0xd9ffff , 0xd6ffff , 0xd3ffff , 0xcfffff , 0xccffff , 0xc9ffff , 0xc6ffff , 0xc2ffff , 0xbfffff , 0xbcffff , 0xb9ffff , 0xb6ffff , 0xb2ffff , 0xafffff , 0xacffff , 0xa9ffff , 0xa6ffff , 0xa2ffff , 0x9fffff , 0x9cffff , 0x99ffff , 0x96ffff , 0x93ffff , 0x90ffff , 0x8cffff , 0x89ffff , 0x85ffff , 0x82ffff , 0x7fffff , 0x7cffff , 0x79ffff , 0x76ffff , 0x73ffff , 0x70ffff , 0x6dffff , 0x6affff , 0x66ffff , 0x63ffff , 0x60ffff , 0x5dffff , 0x5affff , 0x56ffff , 0x53ffff , 0x50ffff , 0x4dffff , 0x49ffff , 0x46ffff , 0x43ffff , 0x40ffff , 0x3dffff , 0x39ffff , 0x36ffff , 0x33ffff , 0x30ffff , 0x2dffff , 0x29ffff , 0x26ffff , 0x23ffff , 0x20ffff , 0x1cffff , 0x19ffff , 0x16ffff , 0x13ffff , 0x10ffff , 0xcffff , 0x9ffff , 0x6ffff , 0x5feff , 0x3fcff , 0x2fbff , 0xf9ff , 0xf6ff , 0xf3ff , 0xf0ff , 0xecff , 0xe9ff , 0xe6ff , 0xe3ff , 0xdfff , 0xdcff , 0xd9ff , 0xd5ff , 0xd2ff , 0xcfff , 0xccff , 0xc9ff , 0xc6ff , 0xc3ff , 0xbfff , 0xbcff , 0xb9ff , 0xb6ff , 0xb2ff , 0xafff , 0xacff , 0xa9ff , 0xa6ff , 0xa2ff , 0x9fff , 0x9cff , 0x99ff , 0x96ff , 0x92ff , 0x8fff , 0x8cff , 0x89ff , 0x86ff , 0x83ff , 0x80ff , 0x7cff , 0x79ff , 0x75ff , 0x72ff , 0x6fff , 0x6cff , 0x69ff , 0x66ff , 0x63ff , 0x60ff , 0x5dff , 0x5aff , 0x56ff , 0x53ff , 0x50ff , 0x4dff , 0x4aff , 0x46ff , 0x43ff , 0x40ff , 0x3dff , 0x39ff , 0x36ff , 0x33ff , 0x30ff , 0x2dff , 0x29ff , 0x26ff , 0x23ff , 0x20ff , 0x1dff , 0x19ff , 0x16ff , 0x13ff , 0x10ff , 0xcff , 0x9ff , 0x6ff , 0x3ff , 0x2fc , 0x1fa , 0x1f7 , 0xf5 , 0xf2 , 0xef , 0xec , 0xe9 , 0xe6 , 0xe3 , 0xe0 , 0xdc , 0xd9 , 0xd6 , 0xd3 , 0xcf , 0xcc , 0xc9 , 0xc5 , 0xc2 , 0xbf , 0xbc , 0xb9 , 0xb6 , 0xb3 , 0xaf , 0xac , 0xa9 , 0xa6 , 0xa2 , 0x9f , 0x9c , 0x99 , 0x96 , 0x92 , 0x8f , 0x8c , 0x89 , 0x86 , 0x83 , 0x80 , 0x7d , 0x7a , 0x77 , 0x73 , 0x70 , 0x6c , 0x69 , 0x65 , 0x62 , 0x5f , 0x5c , 0x59 , 0x56 , 0x53 , 0x50 , 0x4d , 0x4a , 0x46 , 0x43 , 0x40 , 0x3d , 0x3a , 0x36 , 0x33 , 0x30 , 0x2d , 0x29 , 0x26 , 0x23 , 0x20 , 0x1d , 0x1a , 0x17 , 0x13 , 0x10 , 0xd , 0xa , 0x10107 , 0x10104 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x0 , 0x0 , 0x0}

,

/* 17 heart */

{0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x40404 , 0x80808 , 0xc0c0c , 0x101010 , 0x131313 , 0x171717 , 0x1b1b1b , 0x222222 , 0x262626 , 0x2a2a2a , 0x2e2e2e , 0x313131 , 0x353535 , 0x393939 , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3d3d3d , 0x3c3c3c , 0x3b3b3b , 0x3b3b3b , 0x3a3a3a , 0x393939 , 0x383838 , 0x383838 , 0x373737 , 0x363636 , 0x363636 , 0x353535 , 0x343434 , 0x343434 , 0x333333 , 0x333333 , 0x2f2f2f , 0x2c2c2c , 0x292929 , 0x262626 , 0x232323 , 0x202020 , 0x1d1d1d , 0x1a1a1a , 0x161616 , 0x131313 , 0xd0d0d , 0xa0a0a , 0x70707 , 0x40404 , 0x10101 , 0x10104 , 0x10207 , 0x1020a , 0x1030d , 0x10410 , 0x10413 , 0x10516 , 0x1061a , 0x1061d , 0x10720 , 0x10723 , 0x10826 , 0x10929 , 0x1092c , 0x10b33 , 0x10c35 , 0x10d37 , 0x10e39 , 0x2103b , 0x2113d , 0x21240 , 0x31341 , 0x31543 , 0x31645 , 0x41747 , 0x4184a , 0x41a4c , 0x51b4e , 0x51c50 , 0x51d52 , 0x61f55 , 0x62056 , 0x72158 , 0x7225a , 0x9255e , 0x92660 , 0xa2762 , 0xb2964 , 0xb2a65 , 0xc2b67 , 0xc2c69 , 0xd2e6b , 0xe2f6d , 0xe306f , 0xf3171 , 0x103373 , 0x103474 , 0x113576 , 0x113678 , 0x12387a , 0x13397c , 0x133a7e , 0x143b80 , 0x153d82 , 0x163f85 , 0x164087 , 0x174189 , 0x18428b , 0x18438d , 0x19448f , 0x1a4691 , 0x1a4792 , 0x1b4993 , 0x1c4b94 , 0x1d4d96 , 0x1e4f97 , 0x1f5198 , 0x205399 , 0x21559b , 0x22569c , 0x23589d , 0x245a9e , 0x255ca0 , 0x2760a2 , 0x2862a3 , 0x2964a5 , 0x2a65a6 , 0x2c67a7 , 0x2e69a8 , 0x306baa , 0x326dab , 0x346fac , 0x3671ad , 0x3873af , 0x3974b0 , 0x3b76b1 , 0x3d78b2 , 0x3f7ab4 , 0x407cb5 , 0x427eb6 , 0x4480b7 , 0x4682b9 , 0x4783ba , 0x4a87bc , 0x4c89be , 0x4d8bbf , 0x4f8dbf , 0x508fc0 , 0x5291c2 , 0x5492c3 , 0x5594c4 , 0x5796c5 , 0x5898c7 , 0x5a9ac8 , 0x5b9cc9 , 0x5d9eca , 0x5fa0cc , 0x62a3cd , 0x65a6cf , 0x68a9d0 , 0x6bacd2 , 0x6eafd3 , 0x71b2d5 , 0x78b9d8 , 0x7bbcda , 0x7ebfdb , 0x81c1dd , 0x84c4de , 0x87c7e0 , 0x8acae1 , 0x8dcde3 , 0x91d1e5 , 0x98d3e6 , 0x9fd6e7 , 0xa6d9e8 , 0xaddbea , 0xb4deeb , 0xbbe1ec , 0xc1e3ee , 0xc8e6ef , 0xcfe9f0 , 0xd6ebf2 , 0xe0f2fb , 0xe9f7fc , 0xf0fbfd , 0xf6fdfd , 0xfbfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd}

,

/* 18 brain */

{0x10301 , 0x10301 , 0x10502 , 0x20803 , 0x20a03 , 0x30d04 , 0x30f05 , 0x31206 , 0x41406 , 0x41607 , 0x51908 , 0x51b09 , 0x51e09 , 0x6200a , 0x6230b , 0x7250c , 0x7270c , 0x72a0d , 0x82c0e , 0x82f0e , 0x9310f , 0x93410 , 0x93611 , 0xa3811 , 0xa3b12 , 0xb3d13 , 0xb4014 , 0xb4114 , 0xc4415 , 0xc4616 , 0xd4917 , 0xd4915 , 0xd4914 , 0xd4913 , 0xd4912 , 0xd4911 , 0xd4910 , 0xd490f , 0xd490e , 0xd490c , 0xd490b , 0xd490a , 0xd4909 , 0xd4908 , 0xd4907 , 0xd4906 , 0xe4905 , 0x1c4d05 , 0x2b5205 , 0x395705 , 0x475c05 , 0x556105 , 0x646605 , 0x726b05 , 0x817005 , 0x907405 , 0x9e7905 , 0xad7e05 , 0xbb8305 , 0xc98805 , 0xd78d05 , 0xe69205 , 0xf59705 , 0xf5950a , 0xf6930f , 0xf69114 , 0xf78f19 , 0xf78d1e , 0xf88b23 , 0xf88928 , 0xf9882d , 0xfa8632 , 0xfa8437 , 0xfb823c , 0xfb8040 , 0xfc7e45 , 0xfc7c4a , 0xfd7a4f , 0xfe7955 , 0xfe7250 , 0xfe6b4b , 0xfe6446 , 0xfe5e41 , 0xfe573d , 0xfe5038 , 0xfe4933 , 0xfe432e , 0xfe3d29 , 0xfe3624 , 0xfe2f1f , 0xfe291a , 0xfe2215 , 0xfe1b10 , 0xfe140b , 0xfe0e07 , 0xf90d07 , 0xf50c07 , 0xf00b07 , 0xec0b08 , 0xe70a08 , 0xe30908 , 0xde0908 , 0xda0809 , 0xd50709 , 0xd10709 , 0xcc0609 , 0xc8050a , 0xc3050a , 0xbf040a , 0xbb030a , 0xb7030b , 0xb8060f , 0xba0a13 , 0xbb0d17 , 0xbd111b , 0xbe141f , 0xbf1823 , 0xc01b27 , 0xc21f2c , 0xc42230 , 0xc52634 , 0xc72938 , 0xc82d3c , 0xca3040 , 0xcb3443 , 0xcd3747 , 0xcf3b4c , 0xd1384f , 0xd43553 , 0xd73257 , 0xda305b , 0xdd2d5f , 0xe02a63 , 0xe32867 , 0xe6256b , 0xe8226f , 0xeb2073 , 0xee1d77 , 0xf11a7b , 0xf4187f , 0xf71583 , 0xfa1287 , 0xfd108b , 0xfc108c , 0xfb108e , 0xfa1090 , 0xf91092 , 0xf81094 , 0xf71096 , 0xf61098 , 0xf6109a , 0xf5109b , 0xf4109d , 0xf3109f , 0xf210a1 , 0xf110a3 , 0xf010a5 , 0xef10a7 , 0xef10a9 , 0xe112ae , 0xd315b3 , 0xc517b8 , 0xb81abe , 0xaa1dc2 , 0x9c1fc7 , 0x8e22cd , 0x8025d2 , 0x7227d7 , 0x642add , 0x562ce2 , 0x482fe7 , 0x3b32ed , 0x2d34f2 , 0x1f37f7 , 0x113afd , 0x123cfd , 0x133efd , 0x1440fd , 0x1642fd , 0x1744fd , 0x1846fd , 0x1a49fd , 0x1b4bfd , 0x1c4dfd , 0x1e50fd , 0x1f52fd , 0x2054fd , 0x2257fd , 0x2359fd , 0x245bfd , 0x265efd , 0x2460fd , 0x2263fd , 0x2066fd , 0x1e68fd , 0x1c6bfd , 0x1a6efd , 0x1870fd , 0x1673fd , 0x1476fd , 0x1278fd , 0x107bfd , 0xe7efd , 0xc80fd , 0xa83fd , 0x886fd , 0x789fe , 0x68bfd , 0x68dfd , 0x68ffd , 0x691fd , 0x693fd , 0x596fd , 0x598fd , 0x59afd , 0x59cfd , 0x59efd , 0x4a1fd , 0x4a3fd , 0x4a5fd , 0x4a7fd , 0x4a9fd , 0x4acfd , 0xbaffd , 0x13b3fd , 0x1bb6fd , 0x23bafd , 0x2bbefd , 0x32c0fd , 0x3ac4fd , 0x41c8fd , 0x49cbfd , 0x51cffd , 0x58d2fd , 0x60d6fd , 0x68dafd , 0x70ddfd , 0x78e1fd , 0x80e5fe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe}

,

/* 19 001-fire */

{0x0 , 0x70000 , 0xf0000 , 0x160000 , 0x1e0000 , 0x260000 , 0x2d0000 , 0x350000 , 0x3d0000 , 0x410000 , 0x450000 , 0x4a0000 , 0x4e0000 , 0x520000 , 0x570000 , 0x5b0000 , 0x600001 , 0x640004 , 0x680007 , 0x6c000a , 0x71000d , 0x750010 , 0x790013 , 0x7d0016 , 0x820019 , 0x86001c , 0x8a001f , 0x8f0022 , 0x930025 , 0x970028 , 0x9c002b , 0xa0002e , 0xa50031 , 0xa80034 , 0xab0037 , 0xaf003a , 0xb2003d , 0xb50040 , 0xb90043 , 0xbc0046 , 0xc00049 , 0xc3004c , 0xc7004f , 0xca0052 , 0xce0055 , 0xd10058 , 0xd5005b , 0xd8005e , 0xdc0062 , 0xdc0065 , 0xdd0068 , 0xde006b , 0xdf006e , 0xe00071 , 0xe10074 , 0xe20077 , 0xe3007a , 0xe0007d , 0xde0080 , 0xdc0083 , 0xda0086 , 0xd80089 , 0xd6008c , 0xd4008f , 0xd20092 , 0xce0094 , 0xca0096 , 0xc70098 , 0xc3009a , 0xbf009c , 0xbc009e , 0xb800a0 , 0xb500a2 , 0xb100a3 , 0xad00a4 , 0xa900a6 , 0xa600a7 , 0xa200a8 , 0x9e00aa , 0x9a00ab , 0x9700ad , 0x9300ae , 0x8f00af , 0x8c00b1 , 0x8800b2 , 0x8400b3 , 0x8100b5 , 0x7d00b6 , 0x7a00b8 , 0x7600b9 , 0x7200ba , 0x6f00bc , 0x6b00bd , 0x6700be , 0x6400c0 , 0x6000c1 , 0x5d00c3 , 0x5901c4 , 0x5503c6 , 0x5205c7 , 0x4e07c9 , 0x4a08ca , 0x470acc , 0x430ccd , 0x400ecf , 0x3c10d0 , 0x3813d1 , 0x3515d2 , 0x3118d4 , 0x2d1bd5 , 0x2a1dd6 , 0x2620d7 , 0x2323d9 , 0x1f25da , 0x1b28dc , 0x172bdd , 0x142edf , 0x1030e0 , 0xc33e2 , 0x836e3 , 0x539e5 , 0x43be6 , 0x33ee7 , 0x341e9 , 0x244ea , 0x146eb , 0x149ed , 0x4cee , 0x4ff0 , 0x51f1 , 0x54f3 , 0x57f4 , 0x5af6 , 0x5cf7 , 0x5ff9 , 0x62fa , 0x65fc , 0x67fc , 0x69fc , 0x6bfd , 0x6dfd , 0x6ffd , 0x71fe , 0x73fe , 0x75ff , 0x77ff , 0x79ff , 0x7bff , 0x7dff , 0x7fff , 0x81ff , 0x83ff , 0x85ff , 0x86ff , 0x88ff , 0x8aff , 0x8cff , 0x8dff , 0x8fff , 0x91ff , 0x93ff , 0x94ff , 0x96ff , 0x98ff , 0x9aff , 0x9bff , 0x9dff , 0x9fff , 0xa1ff , 0xa2ff , 0xa4ff , 0xa6ff , 0xa8ff , 0xa9ff , 0xabff , 0xadff , 0xafff , 0xb0ff , 0xb2ff , 0xb4ff , 0xb6ff , 0xb8ff , 0xbaff , 0xbcff , 0xbeff , 0xbfff , 0xc1ff , 0xc3ff , 0xc5ff , 0xc7ff , 0xc9ff , 0xcbff , 0xcdff , 0xceff , 0xd0ff , 0xd2ff , 0xd4ff , 0xd5ff , 0xd7ff , 0xd9ff , 0xdbff , 0xdcff , 0xdeff , 0xe0ff , 0xe2ff , 0xe4ff , 0xe6ff , 0xe8ff , 0xeaff , 0x4ebff , 0x8edff , 0xdefff , 0x11f1ff , 0x15f2ff , 0x1af4ff , 0x1ef6ff , 0x23f8ff , 0x2af8ff , 0x32f9ff , 0x3afaff , 0x42fbff , 0x4afcff , 0x52fdff , 0x5afeff , 0x62ffff , 0x69ffff , 0x71ffff , 0x79ffff , 0x81ffff , 0x88ffff , 0x90ffff , 0x98ffff , 0xa0ffff , 0xa7ffff , 0xafffff , 0xb7ffff , 0xbfffff , 0xc7ffff , 0xcfffff , 0xd7ffff , 0xdfffff , 0xe3ffff , 0xe7ffff , 0xebffff , 0xefffff , 0xf3ffff , 0xf7ffff , 0xfbffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff}

,

/* 20 thal_16 */

{0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x800101 , 0x800101 , 0x800101 , 0x800101 , 0x800101 , 0x800101 , 0x800101 , 0x800101 , 0x800101 , 0x800101 , 0x800101 , 0x800101 , 0x800101 , 0x800101 , 0x800101 , 0x800101 , 0xbf0101 , 0xbf0101 , 0xbf0101 , 0xbf0101 , 0xbf0101 , 0xbf0101 , 0xbf0101 , 0xbf0101 , 0xbf0101 , 0xbf0101 , 0xbf0101 , 0xbf0101 , 0xbf0101 , 0xbf0101 , 0xbf0101 , 0xbf0101 , 0xfe0101 , 0xfe0101 , 0xfe0101 , 0xfe0101 , 0xfe0101 , 0xfe0101 , 0xfe0101 , 0xfe0101 , 0xfe0101 , 0xfe0101 , 0xfe0101 , 0xfe0101 , 0xfe0101 , 0xfe0101 , 0xfe0101 , 0xfe0101 , 0x14001 , 0x14001 , 0x14001 , 0x14001 , 0x14001 , 0x14001 , 0x14001 , 0x14001 , 0x14001 , 0x14001 , 0x14001 , 0x14001 , 0x14001 , 0x14001 , 0x14001 , 0x14001 , 0x18001 , 0x18001 , 0x18001 , 0x18001 , 0x18001 , 0x18001 , 0x18001 , 0x18001 , 0x18001 , 0x18001 , 0x18001 , 0x18001 , 0x18001 , 0x18001 , 0x18001 , 0x18001 , 0x1bf01 , 0x1bf01 , 0x1bf01 , 0x1bf01 , 0x1bf01 , 0x1bf01 , 0x1bf01 , 0x1bf01 , 0x1bf01 , 0x1bf01 , 0x1bf01 , 0x1bf01 , 0x1bf01 , 0x1bf01 , 0x1bf01 , 0x1bf01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x10140 , 0x10140 , 0x10140 , 0x10140 , 0x10140 , 0x10140 , 0x10140 , 0x10140 , 0x10140 , 0x10140 , 0x10140 , 0x10140 , 0x10140 , 0x10140 , 0x10140 , 0x10140 , 0x10180 , 0x10180 , 0x10180 , 0x10180 , 0x10180 , 0x10180 , 0x10180 , 0x10180 , 0x10180 , 0x10180 , 0x10180 , 0x10180 , 0x10180 , 0x10180 , 0x10180 , 0x10180 , 0x101bf , 0x101bf , 0x101bf , 0x101bf , 0x101bf , 0x101bf , 0x101bf , 0x101bf , 0x101bf , 0x101bf , 0x101bf , 0x101bf , 0x101bf , 0x101bf , 0x101bf , 0x101bf , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x140c7 , 0x140c7 , 0x140c7 , 0x140c7 , 0x140c7 , 0x140c7 , 0x140c7 , 0x140c7 , 0x140c7 , 0x140c7 , 0x140c7 , 0x140c7 , 0x140c7 , 0x140c7 , 0x140c7 , 0x140c7 , 0x180c7 , 0x180c7 , 0x180c7 , 0x180c7 , 0x180c7 , 0x180c7 , 0x180c7 , 0x180c7 , 0x180c7 , 0x180c7 , 0x180c7 , 0x180c7 , 0x180c7 , 0x180c7 , 0x180c7 , 0x180c7 , 0x1bffe , 0x1bffe , 0x1bffe , 0x1bffe , 0x1bffe , 0x1bffe , 0x1bffe , 0x1bffe , 0x1bffe , 0x1bffe , 0x1bffe , 0x1bffe , 0x1bffe , 0x1bffe , 0x1bffe , 0x1bffe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe}

,

/* 21 HotRed */

{0x0 , 0x0 , 0x0 , 0x0 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x4 , 0x7 , 0xa , 0xd , 0x10 , 0x14 , 0x17 , 0x1b , 0x1e , 0x21 , 0x24 , 0x27 , 0x2a , 0x2d , 0x30 , 0x34 , 0x37 , 0x3a , 0x3d , 0x41 , 0x44 , 0x47 , 0x4a , 0x4e , 0x51 , 0x54 , 0x57 , 0x5a , 0x5d , 0x60 , 0x63 , 0x67 , 0x6a , 0x6e , 0x71 , 0x75 , 0x78 , 0x7b , 0x7e , 0x82 , 0x85 , 0x88 , 0x8b , 0x8e , 0x91 , 0x94 , 0x97 , 0x9b , 0x9e , 0xa1 , 0xa4 , 0xa8 , 0xab , 0xae , 0xb1 , 0xb5 , 0xb8 , 0xbb , 0xbe , 0xc1 , 0xc4 , 0xc8 , 0xcb , 0xcf , 0xd2 , 0xd5 , 0xd8 , 0xdc , 0xdf , 0xe2 , 0xe5 , 0xe9 , 0xec , 0xef , 0xf2 , 0xf5 , 0xf7 , 0x1fa , 0x2fc , 0x3ff , 0x6ff , 0x9ff , 0xcff , 0x10ff , 0x13ff , 0x16ff , 0x19ff , 0x1dff , 0x20ff , 0x23ff , 0x26ff , 0x2aff , 0x2dff , 0x30ff , 0x33ff , 0x37ff , 0x3aff , 0x3dff , 0x40ff , 0x44ff , 0x47ff , 0x4aff , 0x4dff , 0x51ff , 0x54ff , 0x57ff , 0x5aff , 0x5eff , 0x61ff , 0x64ff , 0x67ff , 0x6aff , 0x6dff , 0x70ff , 0x73ff , 0x77ff , 0x7aff , 0x7eff , 0x81ff , 0x85ff , 0x88ff , 0x8bff , 0x8eff , 0x91ff , 0x94ff , 0x97ff , 0x9aff , 0x9eff , 0xa1ff , 0xa4ff , 0xa7ff , 0xabff , 0xaeff , 0xb1ff , 0xb4ff , 0xb8ff , 0xbbff , 0xbeff , 0xc1ff , 0xc5ff , 0xc8ff , 0xcbff , 0xceff , 0xd1ff , 0xd4ff , 0xd8ff , 0xdbff , 0xdfff , 0xe2ff , 0xe5ff , 0xe8ff , 0xecff , 0xefff , 0xf2ff , 0xf5ff , 0xf9ff , 0x1faff , 0x3fcff , 0x4fdff , 0x6ffff , 0x9ffff , 0xcffff , 0xfffff , 0x13ffff , 0x16ffff , 0x19ffff , 0x1cffff , 0x20ffff , 0x23ffff , 0x26ffff , 0x29ffff , 0x2dffff , 0x30ffff , 0x33ffff , 0x36ffff , 0x3affff , 0x3dffff , 0x40ffff , 0x43ffff , 0x47ffff , 0x4affff , 0x4dffff , 0x50ffff , 0x54ffff , 0x57ffff , 0x5affff , 0x5dffff , 0x61ffff , 0x64ffff , 0x67ffff , 0x6affff , 0x6effff , 0x71ffff , 0x74ffff , 0x77ffff , 0x7affff , 0x7dffff , 0x80ffff , 0x83ffff , 0x87ffff , 0x8affff , 0x8effff , 0x91ffff , 0x95ffff , 0x98ffff , 0x9bffff , 0x9effff , 0xa1ffff , 0xa4ffff , 0xa7ffff , 0xaaffff , 0xaeffff , 0xb1ffff , 0xb4ffff , 0xb7ffff , 0xbbffff , 0xbeffff , 0xc1ffff , 0xc4ffff , 0xc8ffff , 0xcbffff , 0xceffff , 0xd1ffff , 0xd5ffff , 0xd8ffff , 0xdbffff , 0xdeffff , 0xe1ffff , 0xe4ffff , 0xe8ffff , 0xebffff , 0xefffff , 0xf2ffff , 0xf5ffff , 0xf8ffff , 0xfcffff , 0xfcffff , 0xfdffff , 0xfeffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff}

,

/* 22 royal */

{0x70101 , 0x70101 , 0xd0101 , 0x120101 , 0x180101 , 0x1e0101 , 0x240101 , 0x2a0101 , 0x2f0101 , 0x350101 , 0x3b0101 , 0x400101 , 0x460101 , 0x4b0101 , 0x510101 , 0x550101 , 0x590101 , 0x5d0101 , 0x620101 , 0x680101 , 0x6d0101 , 0x730101 , 0x790101 , 0x7e0101 , 0x840101 , 0x890101 , 0x8f0101 , 0x950101 , 0x9a0101 , 0xa00101 , 0xa50101 , 0xa90101 , 0xae0101 , 0xb10701 , 0xb60d01 , 0xbc1201 , 0xc01801 , 0xc61e01 , 0xcc2401 , 0xd12a01 , 0xd72f01 , 0xdc3501 , 0xe23b01 , 0xe84001 , 0xed4601 , 0xf34b01 , 0xf85101 , 0xfe5501 , 0xfe5901 , 0xfe5d01 , 0xfe6201 , 0xfe6801 , 0xfe6d01 , 0xfe7301 , 0xfe7901 , 0xfe7e01 , 0xfe8401 , 0xfe8901 , 0xfe8f01 , 0xfe9501 , 0xfe9a01 , 0xfea001 , 0xfea501 , 0xfea901 , 0xfeae01 , 0xfeb101 , 0xfeb601 , 0xfebc01 , 0xfec001 , 0xfec601 , 0xfecc01 , 0xfed101 , 0xfed701 , 0xfedc01 , 0xfee201 , 0xfee801 , 0xfeed01 , 0xfef301 , 0xfef801 , 0xfefe01 , 0xfefe01 , 0xfefe01 , 0xfefe01 , 0xfefe01 , 0xfefe01 , 0xfefe01 , 0xfefe01 , 0xfefe01 , 0xfefe01 , 0xfefe01 , 0xfefe01 , 0xfefe01 , 0xfefe01 , 0xfefe01 , 0xfefe01 , 0xfefe01 , 0xeefe01 , 0xdefe0a , 0xcefe12 , 0xbffe1b , 0xaffe24 , 0x9ffe2d , 0x8ffe35 , 0x7ffe3e , 0x6ffe46 , 0x5ffe4f , 0x4ffe57 , 0x40fe60 , 0x30fe69 , 0x20fe72 , 0x10fe7a , 0x1fe80 , 0x1fe86 , 0x1fc8b , 0x1fa94 , 0x1f89c , 0x1f5a4 , 0x1f3ac , 0x1f1b5 , 0x1efbd , 0x1edc4 , 0x1ebcc , 0x1e9d5 , 0x1e7dd , 0x1e4e5 , 0x1e2ed , 0x1e0f6 , 0x1defe , 0x1defe , 0x1dbfe , 0x1d9fe , 0x1d6fe , 0x1d3fe , 0x1d1fe , 0x1cefe , 0x1cbfe , 0x1c9fe , 0x1c6fe , 0x1c3fe , 0x1c1fe , 0x1bffe , 0x1bcfe , 0x1bafe , 0x1b8fe , 0x1b6fe , 0x1b4fe , 0x1b0fe , 0x1adfe , 0x1a9fe , 0x1a6fe , 0x1a2fe , 0x19ffe , 0x19bfe , 0x198fe , 0x194fe , 0x191fe , 0x18dfe , 0x18afe , 0x186fe , 0x181fe , 0x183fe , 0x17ffe , 0x17afe , 0x176fe , 0x172fe , 0x16efe , 0x169fe , 0x165fe , 0x161fe , 0x15dfe , 0x158fe , 0x154fe , 0x150fe , 0x14cfe , 0x147fe , 0x142fe , 0x144fe , 0x140fe , 0x13bfe , 0x137fe , 0x132fe , 0x12efe , 0x129fe , 0x125fe , 0x120fe , 0x11cfe , 0x117fe , 0x113fe , 0x10efe , 0x10afe , 0x105fe , 0x101fe , 0x101fe , 0xe0efc , 0x1a1afa , 0x2727f8 , 0x3434f5 , 0x4040f3 , 0x4c4cf1 , 0x5959ef , 0x6666ed , 0x7373eb , 0x7f7fe9 , 0x8c8ce7 , 0x9999e4 , 0xa6a6e2 , 0xb2b2e0 , 0xb7b7de , 0xbcbcde , 0xc0c0de , 0xc2c2de , 0xc4c4de , 0xc7c7de , 0xc9c9de , 0xcbcbde , 0xcdcdde , 0xcfcfde , 0xd1d1de , 0xd3d3de , 0xd5d5de , 0xd8d8de , 0xdadade , 0xdcdcde , 0xdedede , 0xdedede , 0xe0e0e0 , 0xe2e2e2 , 0xe4e4e4 , 0xe7e7e7 , 0xe9e9e9 , 0xebebeb , 0xededed , 0xefefef , 0xf1f1f1 , 0xf3f3f3 , 0xf5f5f5 , 0xf8f8f8 , 0xfafafa , 0xfcfcfc , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe}

,

/* 23 grey */

{0xffffff , 0x7b7b7b , 0x787878 , 0x757575 , 0x727272 , 0x6f6f6f , 0x6c6c6c , 0x696969 , 0x666666 , 0x636363 , 0x606060 , 0x5d5d5d , 0x5a5a5a , 0x575757 , 0x545454 , 0x515151 , 0x4e4e4e , 0x4b4b4b , 0x484848 , 0x454545 , 0x434343 , 0x404040 , 0x3d3d3d , 0x3b3b3b , 0x383838 , 0x353535 , 0x333333 , 0x303030 , 0x2e2e2e , 0x2c2c2c , 0x292929 , 0x272727 , 0x252525 , 0x232323 , 0x202020 , 0x1e1e1e , 0x1c1c1c , 0x1a1a1a , 0x181818 , 0x171717 , 0x151515 , 0x131313 , 0x121212 , 0x101010 , 0xe0e0e , 0xd0d0d , 0xc0c0c , 0xa0a0a , 0x90909 , 0x80808 , 0x70707 , 0x60606 , 0x50505 , 0x40404 , 0x30303 , 0x30303 , 0x20202 , 0x10101 , 0x10101 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x10101 , 0x10101 , 0x20202 , 0x30303 , 0x30303 , 0x40404 , 0x50505 , 0x60606 , 0x70707 , 0x80808 , 0x90909 , 0xa0a0a , 0xc0c0c , 0xd0d0d , 0xe0e0e , 0x101010 , 0x121212 , 0x131313 , 0x151515 , 0x171717 , 0x181818 , 0x1a1a1a , 0x1c1c1c , 0x1e1e1e , 0x202020 , 0x232323 , 0x252525 , 0x272727 , 0x292929 , 0x2c2c2c , 0x2e2e2e , 0x303030 , 0x333333 , 0x353535 , 0x383838 , 0x3b3b3b , 0x3d3d3d , 0x404040 , 0x434343 , 0x454545 , 0x484848 , 0x4b4b4b , 0x4e4e4e , 0x515151 , 0x545454 , 0x575757 , 0x5a5a5a , 0x5d5d5d , 0x606060 , 0x636363 , 0x666666 , 0x696969 , 0x6c6c6c , 0x6f6f6f , 0x727272 , 0x757575 , 0x787878 , 0x7b7b7b , 0x7f7f7f , 0x828282 , 0x858585 , 0x888888 , 0x8b8b8b , 0x8e8e8e , 0x919191 , 0x949494 , 0x979797 , 0x9a9a9a , 0x9d9d9d , 0xa0a0a0 , 0xa3a3a3 , 0xa6a6a6 , 0xa9a9a9 , 0xacacac , 0xafafaf , 0xb2b2b2 , 0xb5b5b5 , 0xb8b8b8 , 0xbababa , 0xbdbdbd , 0xc0c0c0 , 0xc2c2c2 , 0xc5c5c5 , 0xc8c8c8 , 0xcacaca , 0xcdcdcd , 0xcfcfcf , 0xd1d1d1 , 0xd4d4d4 , 0xd6d6d6 , 0xd8d8d8 , 0xdadada , 0xdddddd , 0xdfdfdf , 0xe1e1e1 , 0xe3e3e3 , 0xe5e5e5 , 0xe6e6e6 , 0xe8e8e8 , 0xeaeaea , 0xebebeb , 0xededed , 0xefefef , 0xf0f0f0 , 0xf1f1f1 , 0xf3f3f3 , 0xf4f4f4 , 0xf5f5f5 , 0xf6f6f6 , 0xf7f7f7 , 0xf8f8f8 , 0xf9f9f9 , 0xfafafa , 0xfafafa , 0xfbfbfb , 0xfcfcfc , 0xfcfcfc , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfefefe , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfcfcfc , 0xfcfcfc , 0xfbfbfb , 0xfafafa , 0xfafafa , 0xf9f9f9 , 0xf8f8f8 , 0xf7f7f7 , 0xf6f6f6 , 0xf5f5f5 , 0xf4f4f4 , 0xf3f3f3 , 0xf1f1f1 , 0xf0f0f0 , 0xefefef , 0xededed , 0xebebeb , 0xeaeaea , 0xe8e8e8 , 0xe6e6e6 , 0xe5e5e5 , 0xe3e3e3 , 0xe1e1e1 , 0xdfdfdf , 0xdddddd , 0xdadada , 0xd8d8d8 , 0xd6d6d6 , 0xd4d4d4 , 0xd1d1d1 , 0xcfcfcf , 0xcdcdcd , 0xcacaca , 0xc8c8c8 , 0xc5c5c5 , 0xc2c2c2 , 0xc0c0c0 , 0xbdbdbd , 0xbababa , 0xb8b8b8 , 0xb5b5b5 , 0xb2b2b2 , 0xafafaf , 0xacacac , 0xa9a9a9 , 0xa6a6a6 , 0xa3a3a3 , 0xa0a0a0 , 0x9d9d9d , 0x9a9a9a , 0x979797 , 0x949494 , 0x919191 , 0x8e8e8e , 0x8b8b8b , 0x888888 , 0x858585 , 0x0}

,

/* 24 gem-256 */

{0x20201 , 0x20201 , 0x30201 , 0x40301 , 0x50301 , 0x60401 , 0x70401 , 0x80501 , 0x90501 , 0xa0601 , 0xb0601 , 0xc0701 , 0xd0701 , 0xe0801 , 0xf0801 , 0x100901 , 0x110901 , 0x120a01 , 0x130a01 , 0x140b01 , 0x150b01 , 0x160c01 , 0x170c01 , 0x180d01 , 0x190d01 , 0x1a0e01 , 0x1b0e01 , 0x1c0f01 , 0x1d0f01 , 0x1e1001 , 0x1f1001 , 0x201101 , 0x211101 , 0x221201 , 0x231201 , 0x241301 , 0x251301 , 0x261401 , 0x271401 , 0x281501 , 0x291501 , 0x2a1601 , 0x2b1601 , 0x2c1701 , 0x2d1701 , 0x2e1801 , 0x2f1801 , 0x301901 , 0x311901 , 0x321a01 , 0x331a01 , 0x341b01 , 0x351b01 , 0x361c01 , 0x371c01 , 0x381d01 , 0x391d01 , 0x3a1e01 , 0x3b1e01 , 0x3c1f01 , 0x3d1f01 , 0x3e2001 , 0x3f2001 , 0x402101 , 0x402101 , 0x412002 , 0x422004 , 0x431f06 , 0x441f08 , 0x451e0a , 0x461e0c , 0x471d0e , 0x481d10 , 0x491c12 , 0x4a1c14 , 0x4b1b16 , 0x4c1b18 , 0x4d1a1a , 0x4e1a1c , 0x4f191e , 0x501920 , 0x511822 , 0x521824 , 0x531726 , 0x541728 , 0x55162a , 0x56162c , 0x57152e , 0x581530 , 0x591432 , 0x5a1434 , 0x5b1336 , 0x5c1338 , 0x5d123a , 0x5e123c , 0x5f113e , 0x601140 , 0x611041 , 0x621043 , 0x630f45 , 0x640f47 , 0x650e49 , 0x660e4b , 0x670d4d , 0x680d4f , 0x690c51 , 0x6a0c53 , 0x6b0b55 , 0x6c0b57 , 0x6d0a59 , 0x6e0a5b , 0x6f095d , 0x70095f , 0x710861 , 0x720863 , 0x730765 , 0x740767 , 0x750669 , 0x76066b , 0x77056d , 0x78056f , 0x790471 , 0x7a0473 , 0x7b0375 , 0x7c0377 , 0x7d0279 , 0x7e027b , 0x7f017d , 0x7f037f , 0x7d0481 , 0x7b0683 , 0x790785 , 0x770987 , 0x750a89 , 0x730c8b , 0x710d8d , 0x6f0f8f , 0x6d1091 , 0x6b1293 , 0x691395 , 0x671597 , 0x651699 , 0x63189b , 0x61199d , 0x5f1b9f , 0x5d1ca1 , 0x5b1ea3 , 0x591fa5 , 0x5721a7 , 0x5522a9 , 0x5324ab , 0x5125ad , 0x4f27af , 0x4d28b1 , 0x4b2ab3 , 0x492bb5 , 0x472db7 , 0x452eb9 , 0x4330bb , 0x4131bd , 0x4033bf , 0x3e34c0 , 0x3c36c2 , 0x3a37c4 , 0x3839c6 , 0x363ac8 , 0x343cca , 0x323dcc , 0x303fce , 0x2e40d0 , 0x2c41d2 , 0x2a42d4 , 0x2844d6 , 0x2645d8 , 0x2447da , 0x2248dc , 0x204ade , 0x1e4be0 , 0x1c4de2 , 0x1a4ee4 , 0x1850e6 , 0x1651e8 , 0x1453ea , 0x1254ec , 0x1056ee , 0xe57f0 , 0xc59f2 , 0xa5af4 , 0x85cf6 , 0x65df8 , 0x45ffa , 0x15ffc , 0x462fe , 0x864fe , 0xc67fe , 0x1069fe , 0x146cfe , 0x186efe , 0x1c71fe , 0x2073fe , 0x2476fe , 0x2878fe , 0x2c7bfe , 0x307dfe , 0x3480fe , 0x3882fe , 0x3c85fe , 0x4087fe , 0x438afe , 0x478cfe , 0x4b8ffe , 0x4f91fe , 0x5394fe , 0x5796fe , 0x5b99fe , 0x5f9bfe , 0x639efe , 0x67a0fe , 0x6ba3fe , 0x6fa5fe , 0x73a8fe , 0x77aafe , 0x7badfe , 0x7faffe , 0x83b2fe , 0x87b4fe , 0x8bb7fe , 0x8fb9fe , 0x93bcfe , 0x97befe , 0x9bc0fe , 0x9fc2fe , 0xa3c5fe , 0xa7c7fe , 0xabcafe , 0xafccfe , 0xb3cffe , 0xb7d1fe , 0xbbd4fe , 0xbfd6fe , 0xc2d9fe , 0xc6dbfe , 0xcadefe , 0xcee0fe , 0xd2e3fe , 0xd6e5fe , 0xdae8fe , 0xdeeafe , 0xe2edfe , 0xe6effe , 0xeaf2fe , 0xeef4fe , 0xf2f7fe , 0xf6f9fe , 0xfafcfe , 0xfafcfe}

,

/* 25 cequal */

{0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x80701 , 0xf0e01 , 0x161401 , 0x1d1b01 , 0x242201 , 0x2b2801 , 0x322f01 , 0x3a3601 , 0x403c01 , 0x474201 , 0x4e4801 , 0x554f01 , 0x5c5601 , 0x635c01 , 0x6a6301 , 0x726a01 , 0x746a01 , 0x776b01 , 0x7a6b01 , 0x7c6c01 , 0x7f6c01 , 0x826d01 , 0x856d01 , 0x876e01 , 0x8a6f01 , 0x8d6f01 , 0x907001 , 0x927001 , 0x957101 , 0x987101 , 0x9b7201 , 0x9e7301 , 0xa07203 , 0xa37206 , 0xa67209 , 0xa9720c , 0xac710f , 0xae7112 , 0xb17115 , 0xb47118 , 0xb7701b , 0xba701e , 0xbc7021 , 0xbf7024 , 0xc16f27 , 0xc46f2a , 0xc76f2d , 0xce6f30 , 0xd16f34 , 0xd46f38 , 0xd76f3c , 0xdb6f40 , 0xde6f43 , 0xe16f47 , 0xe46f4b , 0xe86f4f , 0xeb6f53 , 0xee6f57 , 0xf16f5b , 0xf56f5f , 0xf86f63 , 0xfb6f67 , 0xfe6f6b , 0xfe7070 , 0xfe6972 , 0xfe6274 , 0xfe5b76 , 0xfe5478 , 0xfe4d7a , 0xfe467c , 0xfe407e , 0xfe3981 , 0xfe3283 , 0xfe2b85 , 0xfe2487 , 0xfe1d89 , 0xfe168b , 0xfe0f8d , 0xfe088f , 0xfe0192 , 0xfb0194 , 0xf90197 , 0xf7019a , 0xf5019d , 0xf301a0 , 0xf101a3 , 0xef01a6 , 0xed01a9 , 0xea01ac , 0xe801af , 0xe601b1 , 0xe401b4 , 0xe201b7 , 0xe001ba , 0xde01bd , 0xdc02af , 0xd707b1 , 0xd20cb4 , 0xcd12b7 , 0xc917ba , 0xc41cbd , 0xbf22bf , 0xbc27c2 , 0xb72dc5 , 0xb232c8 , 0xae37cb , 0xa93dce , 0xa441d1 , 0xa046d4 , 0x9b4cd7 , 0x9651da , 0x9257dd , 0x8f58dd , 0x8c5ade , 0x8a5bdf , 0x875de0 , 0x845fe1 , 0x8260e2 , 0x7f62e3 , 0x7d64e4 , 0x7a65e5 , 0x7767e6 , 0x7568e7 , 0x726ae8 , 0x6f6ce9 , 0x6d6dea , 0x6a6feb , 0x6871ec , 0x6174ed , 0x5b77ee , 0x547aef , 0x4e7ef0 , 0x4781f1 , 0x4184f2 , 0x3b88f3 , 0x358bf5 , 0x2e8ef6 , 0x2892f7 , 0x2195f8 , 0x1b98f9 , 0x149cfa , 0xe9ffb , 0x7a2fc , 0x1a6fe , 0x1aafe , 0x1aefe , 0x1b2fe , 0x1b6fe , 0x1bafe , 0x1bffe , 0x1c2fe , 0x1c6fe , 0x1cafe , 0x1cefe , 0x1d3fe , 0x1d7fe , 0x1dbfe , 0x1dffe , 0x1e3fe , 0x1d4fe , 0xcd6fe , 0x17d8fe , 0x22dafe , 0x2edcfe , 0x39defe , 0x43e0fe , 0x4fe2fe , 0x5ae4fe , 0x65e6fe , 0x71e8fe , 0x7ceafe , 0x87ecfe , 0x93eefe , 0x9ef0fe , 0xa9f2fe , 0x94f5fe , 0x97f5fe , 0x9bf6fe , 0x9ff6fe , 0xa3f7fe , 0xa7f7fe , 0xabf8fe , 0xaff8fe , 0xb3f9fe , 0xb6fafe , 0xbafafe , 0xbefbfe , 0xc1fbfe , 0xc5fcfe , 0xc9fcfe , 0xcdfdfe , 0xd1fefe , 0xd3fefe , 0xd6fefe , 0xd9fefe , 0xdcfefe , 0xdffefe , 0xe1fefe , 0xe4fefe , 0xe7fefe , 0xeafefe , 0xedfefe , 0xeffefe , 0xf2fefe , 0xf5fefe , 0xf8fefe , 0xfbfefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe}

,

/* 26 invert_gray */

{0xffffff , 0xfefefe , 0xfdfdfd , 0xfcfcfc , 0xfbfbfb , 0xfafafa , 0xf9f9f9 , 0xf8f8f8 , 0xf7f7f7 , 0xf6f6f6 , 0xf5f5f5 , 0xf4f4f4 , 0xf3f3f3 , 0xf2f2f2 , 0xf1f1f1 , 0xf0f0f0 , 0xefefef , 0xeeeeee , 0xededed , 0xececec , 0xebebeb , 0xeaeaea , 0xe9e9e9 , 0xe8e8e8 , 0xe7e7e7 , 0xe6e6e6 , 0xe5e5e5 , 0xe4e4e4 , 0xe3e3e3 , 0xe2e2e2 , 0xe1e1e1 , 0xe0e0e0 , 0xdfdfdf , 0xdedede , 0xdddddd , 0xdcdcdc , 0xdbdbdb , 0xdadada , 0xd9d9d9 , 0xd8d8d8 , 0xd7d7d7 , 0xd6d6d6 , 0xd5d5d5 , 0xd4d4d4 , 0xd3d3d3 , 0xd2d2d2 , 0xd1d1d1 , 0xd0d0d0 , 0xcfcfcf , 0xcecece , 0xcdcdcd , 0xcccccc , 0xcbcbcb , 0xcacaca , 0xc9c9c9 , 0xc8c8c8 , 0xc7c7c7 , 0xc6c6c6 , 0xc5c5c5 , 0xc4c4c4 , 0xc3c3c3 , 0xc2c2c2 , 0xc1c1c1 , 0xc0c0c0 , 0xbfbfbf , 0xbebebe , 0xbdbdbd , 0xbcbcbc , 0xbbbbbb , 0xbababa , 0xb9b9b9 , 0xb8b8b8 , 0xb7b7b7 , 0xb6b6b6 , 0xb5b5b5 , 0xb4b4b4 , 0xb3b3b3 , 0xb2b2b2 , 0xb1b1b1 , 0xb0b0b0 , 0xafafaf , 0xaeaeae , 0xadadad , 0xacacac , 0xababab , 0xaaaaaa , 0xa9a9a9 , 0xa8a8a8 , 0xa7a7a7 , 0xa6a6a6 , 0xa5a5a5 , 0xa4a4a4 , 0xa3a3a3 , 0xa2a2a2 , 0xa1a1a1 , 0xa0a0a0 , 0x9f9f9f , 0x9e9e9e , 0x9d9d9d , 0x9c9c9c , 0x9b9b9b , 0x9a9a9a , 0x999999 , 0x989898 , 0x979797 , 0x969696 , 0x959595 , 0x949494 , 0x939393 , 0x929292 , 0x919191 , 0x909090 , 0x8f8f8f , 0x8e8e8e , 0x8d8d8d , 0x8c8c8c , 0x8b8b8b , 0x8a8a8a , 0x898989 , 0x888888 , 0x878787 , 0x868686 , 0x858585 , 0x848484 , 0x838383 , 0x828282 , 0x818181 , 0x808080 , 0x7f7f7f , 0x7e7e7e , 0x7d7d7d , 0x7c7c7c , 0x7b7b7b , 0x7a7a7a , 0x797979 , 0x787878 , 0x777777 , 0x767676 , 0x757575 , 0x747474 , 0x737373 , 0x727272 , 0x717171 , 0x707070 , 0x6f6f6f , 0x6e6e6e , 0x6d6d6d , 0x6c6c6c , 0x6b6b6b , 0x6a6a6a , 0x696969 , 0x686868 , 0x676767 , 0x666666 , 0x656565 , 0x646464 , 0x636363 , 0x626262 , 0x616161 , 0x606060 , 0x5f5f5f , 0x5e5e5e , 0x5d5d5d , 0x5c5c5c , 0x5b5b5b , 0x5a5a5a , 0x595959 , 0x585858 , 0x575757 , 0x565656 , 0x555555 , 0x545454 , 0x535353 , 0x525252 , 0x515151 , 0x505050 , 0x4f4f4f , 0x4e4e4e , 0x4d4d4d , 0x4c4c4c , 0x4b4b4b , 0x4a4a4a , 0x494949 , 0x484848 , 0x474747 , 0x464646 , 0x454545 , 0x444444 , 0x434343 , 0x424242 , 0x414141 , 0x404040 , 0x3f3f3f , 0x3e3e3e , 0x3d3d3d , 0x3c3c3c , 0x3b3b3b , 0x3a3a3a , 0x393939 , 0x383838 , 0x373737 , 0x363636 , 0x353535 , 0x343434 , 0x333333 , 0x323232 , 0x313131 , 0x303030 , 0x2f2f2f , 0x2e2e2e , 0x2d2d2d , 0x2c2c2c , 0x2b2b2b , 0x2a2a2a , 0x292929 , 0x282828 , 0x272727 , 0x262626 , 0x252525 , 0x242424 , 0x232323 , 0x222222 , 0x212121 , 0x202020 , 0x1f1f1f , 0x1e1e1e , 0x1d1d1d , 0x1c1c1c , 0x1b1b1b , 0x1a1a1a , 0x191919 , 0x181818 , 0x171717 , 0x161616 , 0x151515 , 0x141414 , 0x131313 , 0x121212 , 0x111111 , 0x101010 , 0xf0f0f , 0xe0e0e , 0xd0d0d , 0xc0c0c , 0xb0b0b , 0xa0a0a , 0x90909 , 0x80808 , 0x70707 , 0x60606 , 0x50505 , 0x40404 , 0x30303 , 0x20202 , 0x10101 , 0x0}

,

/* 27 32_colors */

{0xdd01a7 , 0xdd01a7 , 0xdd01a7 , 0xdd01a7 , 0xdd01a7 , 0xdd01a7 , 0xdd01a7 , 0xdd01a7 , 0xe40483 , 0xe40483 , 0xe40483 , 0xe40483 , 0xe40483 , 0xe40483 , 0xe40483 , 0xe40483 , 0xe7024a , 0xe7024a , 0xe7024a , 0xe7024a , 0xe7024a , 0xe7024a , 0xe7024a , 0xe7024a , 0xf2022f , 0xf2022f , 0xf2022f , 0xf2022f , 0xf2022f , 0xf2022f , 0xf2022f , 0xf2022f , 0xf71b02 , 0xf71b02 , 0xf71b02 , 0xf71b02 , 0xf71b02 , 0xf71b02 , 0xf71b02 , 0xf71b02 , 0xfb2103 , 0xfb2103 , 0xfb2103 , 0xfb2103 , 0xfb2103 , 0xfb2103 , 0xfb2103 , 0xfb2103 , 0xfe2704 , 0xfe2704 , 0xfe2704 , 0xfe2704 , 0xfe2704 , 0xfe2704 , 0xfe2704 , 0xfe2704 , 0xfe3305 , 0xfe3305 , 0xfe3305 , 0xfe3305 , 0xfe3305 , 0xfe3305 , 0xfe3305 , 0xfe3305 , 0xfe4b06 , 0xfe4b06 , 0xfe4b06 , 0xfe4b06 , 0xfe4b06 , 0xfe4b06 , 0xfe4b06 , 0xfe4b06 , 0xfe6507 , 0xfe6507 , 0xfe6507 , 0xfe6507 , 0xfe6507 , 0xfe6507 , 0xfe6507 , 0xfe6507 , 0xee7f08 , 0xee7f08 , 0xee7f08 , 0xee7f08 , 0xee7f08 , 0xee7f08 , 0xee7f08 , 0xee7f08 , 0xd59709 , 0xd59709 , 0xd59709 , 0xd59709 , 0xd59709 , 0xd59709 , 0xd59709 , 0xd59709 , 0xa8b10a , 0xa8b10a , 0xa8b10a , 0xa8b10a , 0xa8b10a , 0xa8b10a , 0xa8b10a , 0xa8b10a , 0x7bca13 , 0x7bca13 , 0x7bca13 , 0x7bca13 , 0x7bca13 , 0x7bca13 , 0x7bca13 , 0x7bca13 , 0x4ee31e , 0x4ee31e , 0x4ee31e , 0x4ee31e , 0x4ee31e , 0x4ee31e , 0x4ee31e , 0x4ee31e , 0x28fc29 , 0x28fc29 , 0x28fc29 , 0x28fc29 , 0x28fc29 , 0x28fc29 , 0x28fc29 , 0x28fc29 , 0x24fe48 , 0x24fe48 , 0x24fe48 , 0x24fe48 , 0x24fe48 , 0x24fe48 , 0x24fe48 , 0x24fe48 , 0x21fe7e , 0x21fe7e , 0x21fe7e , 0x21fe7e , 0x21fe7e , 0x21fe7e , 0x21fe7e , 0x21fe7e , 0x1efea7 , 0x1efea7 , 0x1efea7 , 0x1efea7 , 0x1efea7 , 0x1efea7 , 0x1efea7 , 0x1efea7 , 0x1efdc2 , 0x1efdc2 , 0x1efdc2 , 0x1efdc2 , 0x1efdc2 , 0x1efdc2 , 0x1efdc2 , 0x1efdc2 , 0x1de0dc , 0x1de0dc , 0x1de0dc , 0x1de0dc , 0x1de0dc , 0x1de0dc , 0x1de0dc , 0x1de0dc , 0x1fcbfd , 0x1fcbfd , 0x1fcbfd , 0x1fcbfd , 0x1fcbfd , 0x1fcbfd , 0x1fcbfd , 0x1fcbfd , 0x19b0fe , 0x19b0fe , 0x19b0fe , 0x19b0fe , 0x19b0fe , 0x19b0fe , 0x19b0fe , 0x19b0fe , 0x1594fe , 0x1594fe , 0x1594fe , 0x1594fe , 0x1594fe , 0x1594fe , 0x1594fe , 0x1594fe , 0x1278fe , 0x1278fe , 0x1278fe , 0x1278fe , 0x1278fe , 0x1278fe , 0x1278fe , 0x1278fe , 0xd5afe , 0xd5afe , 0xd5afe , 0xd5afe , 0xd5afe , 0xd5afe , 0xd5afe , 0xd5afe , 0x93dfe , 0x93dfe , 0x93dfe , 0x93dfe , 0x93dfe , 0x93dfe , 0x93dfe , 0x93dfe , 0x61efe , 0x61efe , 0x61efe , 0x61efe , 0x61efe , 0x61efe , 0x61efe , 0x61efe , 0x202fe , 0x202fe , 0x202fe , 0x202fe , 0x202fe , 0x202fe , 0x202fe , 0x202fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe , 0x101fe}

,

/* 28 rgb-blue */

{0x0 , 0x10000 , 0x20000 , 0x30000 , 0x40000 , 0x50000 , 0x60000 , 0x70000 , 0x80000 , 0x90000 , 0xa0000 , 0xb0000 , 0xc0000 , 0xd0000 , 0xe0000 , 0xf0000 , 0x100000 , 0x110000 , 0x120000 , 0x130000 , 0x140000 , 0x150000 , 0x160000 , 0x170000 , 0x180000 , 0x190000 , 0x1a0000 , 0x1b0000 , 0x1c0000 , 0x1d0000 , 0x1e0000 , 0x1f0000 , 0x200000 , 0x210000 , 0x220000 , 0x230000 , 0x240000 , 0x250000 , 0x260000 , 0x270000 , 0x280000 , 0x290000 , 0x2a0000 , 0x2b0000 , 0x2c0000 , 0x2d0000 , 0x2e0000 , 0x2f0000 , 0x300000 , 0x310000 , 0x320000 , 0x330000 , 0x340000 , 0x350000 , 0x360000 , 0x370000 , 0x380000 , 0x390000 , 0x3a0000 , 0x3b0000 , 0x3c0000 , 0x3d0000 , 0x3e0000 , 0x3f0000 , 0x400000 , 0x410000 , 0x420000 , 0x430000 , 0x440000 , 0x450000 , 0x460000 , 0x470000 , 0x480000 , 0x490000 , 0x4a0000 , 0x4b0000 , 0x4c0000 , 0x4d0000 , 0x4e0000 , 0x4f0000 , 0x500000 , 0x510000 , 0x520000 , 0x530000 , 0x540000 , 0x550000 , 0x560000 , 0x570000 , 0x580000 , 0x590000 , 0x5a0000 , 0x5b0000 , 0x5c0000 , 0x5d0000 , 0x5e0000 , 0x5f0000 , 0x600000 , 0x610000 , 0x620000 , 0x630000 , 0x640000 , 0x650000 , 0x660000 , 0x670000 , 0x680000 , 0x690000 , 0x6a0000 , 0x6b0000 , 0x6c0000 , 0x6d0000 , 0x6e0000 , 0x6f0000 , 0x700000 , 0x710000 , 0x720000 , 0x730000 , 0x740000 , 0x750000 , 0x760000 , 0x770000 , 0x780000 , 0x790000 , 0x7a0000 , 0x7b0000 , 0x7c0000 , 0x7d0000 , 0x7e0000 , 0x7f0000 , 0x800000 , 0x810000 , 0x820000 , 0x830000 , 0x840000 , 0x850000 , 0x860000 , 0x870000 , 0x880000 , 0x890000 , 0x8a0000 , 0x8b0000 , 0x8c0000 , 0x8d0000 , 0x8e0000 , 0x8f0000 , 0x900000 , 0x910000 , 0x920000 , 0x930000 , 0x940000 , 0x950000 , 0x960000 , 0x970000 , 0x980000 , 0x990000 , 0x9a0000 , 0x9b0000 , 0x9c0000 , 0x9d0000 , 0x9e0000 , 0x9f0000 , 0xa00000 , 0xa10000 , 0xa20000 , 0xa30000 , 0xa40000 , 0xa50000 , 0xa60000 , 0xa70000 , 0xa80000 , 0xa90000 , 0xaa0000 , 0xab0000 , 0xac0000 , 0xad0000 , 0xae0000 , 0xaf0000 , 0xb00000 , 0xb10000 , 0xb20000 , 0xb30000 , 0xb40000 , 0xb50000 , 0xb60000 , 0xb70000 , 0xb80000 , 0xb90000 , 0xba0000 , 0xbb0000 , 0xbc0000 , 0xbd0000 , 0xbe0000 , 0xbf0000 , 0xc00000 , 0xc10000 , 0xc20000 , 0xc30000 , 0xc40000 , 0xc50000 , 0xc60000 , 0xc70000 , 0xc80000 , 0xc90000 , 0xca0000 , 0xcb0000 , 0xcc0000 , 0xcd0000 , 0xce0000 , 0xcf0000 , 0xd00000 , 0xd10000 , 0xd20000 , 0xd30000 , 0xd40000 , 0xd50000 , 0xd60000 , 0xd70000 , 0xd80000 , 0xd90000 , 0xda0000 , 0xdb0000 , 0xdc0000 , 0xdd0000 , 0xde0000 , 0xdf0000 , 0xe00000 , 0xe10000 , 0xe20000 , 0xe30000 , 0xe40000 , 0xe50000 , 0xe60000 , 0xe70000 , 0xe80000 , 0xe90000 , 0xea0000 , 0xeb0000 , 0xec0000 , 0xed0000 , 0xee0000 , 0xef0000 , 0xf00000 , 0xf10000 , 0xf20000 , 0xf30000 , 0xf40000 , 0xf50000 , 0xf60000 , 0xf70000 , 0xf80000 , 0xf90000 , 0xfa0000 , 0xfb0000 , 0xfc0000 , 0xfd0000 , 0xfe0000 , 0xff0000}

,

/* 29 nih-image-fire2 */

{0xffffff , 0xfbffff , 0xf7ffff , 0xf3ffff , 0xefffff , 0xecffff , 0xe8ffff , 0xe4ffff , 0xe0ffff , 0xd9ffff , 0xd1ffff , 0xcaffff , 0xc2ffff , 0xbaffff , 0xb3ffff , 0xabffff , 0xa3ffff , 0x9cffff , 0x94ffff , 0x8dffff , 0x85ffff , 0x7effff , 0x76ffff , 0x6fffff , 0x67ffff , 0x60ffff , 0x58feff , 0x50fdff , 0x49fcff , 0x41fbff , 0x39faff , 0x32faff , 0x2af9ff , 0x23f8ff , 0x1ef6ff , 0x1af4ff , 0x16f3ff , 0x12f1ff , 0xdefff , 0x9eeff , 0x5ecff , 0x1eaff , 0xe8ff , 0xe7ff , 0xe5ff , 0xe3ff , 0xe1ff , 0xdfff , 0xddff , 0xdcff , 0xdaff , 0xd8ff , 0xd6ff , 0xd5ff , 0xd3ff , 0xd1ff , 0xd0ff , 0xceff , 0xccff , 0xcaff , 0xc9ff , 0xc7ff , 0xc5ff , 0xc3ff , 0xc1ff , 0xbfff , 0xbeff , 0xbcff , 0xbaff , 0xb8ff , 0xb6ff , 0xb5ff , 0xb3ff , 0xb1ff , 0xafff , 0xadff , 0xacff , 0xaaff , 0xa8ff , 0xa7ff , 0xa5ff , 0xa3ff , 0xa1ff , 0xa0ff , 0x9eff , 0x9cff , 0x9bff , 0x99ff , 0x97ff , 0x96ff , 0x94ff , 0x92ff , 0x90ff , 0x8fff , 0x8dff , 0x8bff , 0x8aff , 0x88ff , 0x86ff , 0x84ff , 0x82ff , 0x81ff , 0x7fff , 0x7dff , 0x7bff , 0x79ff , 0x77ff , 0x75ff , 0x73ff , 0x71fe , 0x6ffe , 0x6dfe , 0x6bfd , 0x69fd , 0x67fc , 0x65fc , 0x63fb , 0x60f9 , 0x5df8 , 0x5bf6 , 0x58f5 , 0x55f3 , 0x53f2 , 0x50f1 , 0x4def , 0x14bee , 0x248ed , 0x245eb , 0x343ea , 0x340e8 , 0x43de7 , 0x53be6 , 0x638e4 , 0xa35e3 , 0xe33e2 , 0x1130e0 , 0x152ddf , 0x192bdd , 0x1c28dc , 0x2025da , 0x2423d9 , 0x2720d8 , 0x2b1dd6 , 0x2e1bd5 , 0x3218d4 , 0x3516d3 , 0x3913d2 , 0x3c11d0 , 0x400ecf , 0x430cce , 0x470bcc , 0x4a09cb , 0x4e07c9 , 0x5106c8 , 0x5504c6 , 0x5802c5 , 0x5c00c3 , 0x6000c2 , 0x6300c1 , 0x6700bf , 0x6a00be , 0x6e00bd , 0x7100bb , 0x7500ba , 0x7800b9 , 0x7c00b7 , 0x7f00b6 , 0x8300b5 , 0x8600b3 , 0x8a00b2 , 0x8d00b1 , 0x9100af , 0x9400ae , 0x9800ad , 0x9c00ab , 0x9f00aa , 0xa300a9 , 0xa700a7 , 0xaa00a6 , 0xae00a5 , 0xb100a3 , 0xb500a2 , 0xb900a0 , 0xbc009e , 0xc0009c , 0xc3009a , 0xc70098 , 0xca0096 , 0xce0094 , 0xd10092 , 0xd40090 , 0xd6008d , 0xd8008a , 0xda0087 , 0xdc0084 , 0xde0081 , 0xe0007e , 0xe2007b , 0xe30078 , 0xe20075 , 0xe10072 , 0xe00070 , 0xdf006d , 0xde006a , 0xdd0067 , 0xdd0064 , 0xdb0061 , 0xd7005e , 0xd4005b , 0xd10058 , 0xcd0055 , 0xca0052 , 0xc6004f , 0xc3004c , 0xc00049 , 0xbc0046 , 0xb90043 , 0xb60040 , 0xb2003d , 0xaf003a , 0xac0037 , 0xa90034 , 0xa50031 , 0xa1002e , 0x9d002b , 0x990029 , 0x940026 , 0x900023 , 0x8c0020 , 0x88001d , 0x83001a , 0x7f0017 , 0x7b0014 , 0x770011 , 0x73000e , 0x6f000b , 0x6b0008 , 0x660006 , 0x620003 , 0x5e0001 , 0x5a0001 , 0x560001 , 0x510001 , 0x4d0000 , 0x490000 , 0x450000 , 0x400000 , 0x3b0000 , 0x340000 , 0x2c0000 , 0x250000 , 0x1e0000 , 0x160000 , 0xf0000 , 0x70000 , 0x0}

,

/* 30 auxctq */

{0x10401 , 0x10401 , 0x10501 , 0x10601 , 0x10701 , 0x10801 , 0x10901 , 0x10a01 , 0x10b01 , 0x10c01 , 0x10d01 , 0x10e01 , 0x10f01 , 0x11001 , 0x11101 , 0x11201 , 0x11301 , 0x11401 , 0x11501 , 0x11601 , 0x11701 , 0x11901 , 0x11a01 , 0x11b01 , 0x11c01 , 0x11d01 , 0x11e01 , 0x11f01 , 0x12001 , 0x12101 , 0x12201 , 0x12301 , 0x12401 , 0x12501 , 0x12601 , 0x12701 , 0x12801 , 0x12901 , 0x12a01 , 0x12b01 , 0x12c01 , 0x12e01 , 0x12f01 , 0x13001 , 0x13101 , 0x13201 , 0x13301 , 0x13401 , 0x13501 , 0x13601 , 0x13701 , 0x13801 , 0x13901 , 0x13a01 , 0x13b01 , 0x13c01 , 0x13d01 , 0x13e01 , 0x13f01 , 0x14001 , 0x14101 , 0x14201 , 0x14301 , 0x14401 , 0x14501 , 0x14601 , 0x14701 , 0x14801 , 0x14901 , 0x14a01 , 0x14b01 , 0x14c01 , 0x14d01 , 0x14e01 , 0x14f01 , 0x15001 , 0x45101 , 0x85201 , 0xb5301 , 0xf5401 , 0x165601 , 0x195701 , 0x1d5801 , 0x205901 , 0x245a01 , 0x275b01 , 0x2b5c01 , 0x2e5d01 , 0x325e01 , 0x355f01 , 0x396001 , 0x3d5b01 , 0x405601 , 0x435101 , 0x464c01 , 0x4a4701 , 0x4d4301 , 0x513f01 , 0x543a01 , 0x583501 , 0x5f2c01 , 0x622701 , 0x662201 , 0x691d01 , 0x6d1801 , 0x711401 , 0x710f01 , 0x720a01 , 0x730501 , 0x740101 , 0x750101 , 0x760101 , 0x770101 , 0x780101 , 0x790101 , 0x7a0101 , 0x7b0101 , 0x7c0101 , 0x7d0101 , 0x7f0101 , 0x80010a , 0x810113 , 0x82011c , 0x830125 , 0x84012e , 0x850137 , 0x860140 , 0x870148 , 0x880151 , 0x89015a , 0x8a0163 , 0x8b016c , 0x8c0175 , 0x8d017e , 0x8e0187 , 0x8f0190 , 0x900191 , 0x910192 , 0x920193 , 0x940195 , 0x950196 , 0x960197 , 0x970198 , 0x980199 , 0x99019a , 0x9a019b , 0x9b019c , 0x9c019d , 0x9d019e , 0x9e019f , 0x9f01a0 , 0xa003a1 , 0x9d05a2 , 0x9b07a3 , 0x980aa4 , 0x960ca5 , 0x930ea6 , 0x9110a7 , 0x8e13a8 , 0x8917aa , 0x8719ab , 0x841cac , 0x821ead , 0x7f20ae , 0x7d22af , 0x7a25b0 , 0x7827b1 , 0x7529b2 , 0x732cb3 , 0x702eb4 , 0x6e30b5 , 0x6b32b6 , 0x6935b7 , 0x6637b8 , 0x6439b9 , 0x613bba , 0x5f3ebb , 0x5c40bc , 0x5743be , 0x5546bf , 0x5248bf , 0x504ac0 , 0x4d4dc1 , 0x4b4fc2 , 0x4851c3 , 0x4653c4 , 0x4356c5 , 0x4158c6 , 0x3f5ac7 , 0x3d5cc8 , 0x3a5fc9 , 0x3861ca , 0x3563cb , 0x3365cc , 0x3068cd , 0x2e6ace , 0x2b6ccf , 0x296fd0 , 0x2473d2 , 0x2175d3 , 0x1f78d4 , 0x1c7ad5 , 0x1a7cd6 , 0x177ed7 , 0x1581d8 , 0x1283d9 , 0x1085da , 0xd87db , 0xb8adc , 0x88cdd , 0x68ede , 0x391df , 0x194e0 , 0x998e1 , 0x129be2 , 0x1a9fe3 , 0x23a2e4 , 0x2ba6e5 , 0x3cade7 , 0x44b0e8 , 0x4cb4e9 , 0x55b8ea , 0x5dbbeb , 0x66bfec , 0x6ec1ed , 0x77c5ee , 0x7fc8ef , 0x88ccf0 , 0x90cff1 , 0x99d3f2 , 0xa1d6f3 , 0xaadaf4 , 0xb2def5 , 0xbbe1f6 , 0xc2e5f7 , 0xcbe8f8 , 0xd3ecf9 , 0xe0f2fb , 0xeaf7fd , 0xf1fdfd , 0xf7fdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd}

,

/* 31 GreenFire */

{0x1f0000 , 0x220000 , 0x260000 , 0x2a0000 , 0x2e0000 , 0x310000 , 0x350000 , 0x390000 , 0x3d0000 , 0x410000 , 0x450000 , 0x4a0000 , 0x4e0000 , 0x520000 , 0x570000 , 0x5b0000 , 0x600100 , 0x640400 , 0x680700 , 0x6c0a00 , 0x710d00 , 0x751000 , 0x791300 , 0x7d1600 , 0x821900 , 0x861c00 , 0x8a1f00 , 0x8f2200 , 0x932500 , 0x972800 , 0x9c2b00 , 0xa02e00 , 0xa53100 , 0xa83400 , 0xab3700 , 0xaf3a00 , 0xb23d00 , 0xb54000 , 0xb94300 , 0xbc4600 , 0xc04900 , 0xc34c00 , 0xc74f00 , 0xca5200 , 0xce5500 , 0xd15800 , 0xd55b00 , 0xd85e00 , 0xdc6200 , 0xdc6500 , 0xdd6800 , 0xde6b00 , 0xdf6e00 , 0xe07100 , 0xe17400 , 0xe27700 , 0xe37a00 , 0xe07d00 , 0xde8000 , 0xdc8300 , 0xda8600 , 0xd88900 , 0xd68c00 , 0xd48f00 , 0xd29200 , 0xce9400 , 0xca9600 , 0xc79800 , 0xc39a00 , 0xbf9c00 , 0xbc9e00 , 0xb8a000 , 0xb5a200 , 0xb1a300 , 0xada400 , 0xa9a600 , 0xa6a700 , 0xa2a800 , 0x9eaa00 , 0x9aab00 , 0x97ad00 , 0x93ae00 , 0x8faf00 , 0x8cb100 , 0x88b200 , 0x84b300 , 0x81b500 , 0x7db600 , 0x7ab800 , 0x76b900 , 0x72ba00 , 0x6fbc00 , 0x6bbd00 , 0x67be00 , 0x64c000 , 0x60c100 , 0x5dc300 , 0x59c401 , 0x55c603 , 0x52c705 , 0x4ec907 , 0x4aca08 , 0x47cc0a , 0x43cd0c , 0x40cf0e , 0x3cd010 , 0x38d113 , 0x35d215 , 0x31d418 , 0x2dd51b , 0x2ad61d , 0x26d720 , 0x23d923 , 0x1fda25 , 0x1bdc28 , 0x17dd2b , 0x14df2e , 0x10e030 , 0xce233 , 0x8e336 , 0x5e539 , 0x4e63b , 0x3e73e , 0x3e941 , 0x2ea44 , 0x1eb46 , 0x1ed49 , 0xee4c , 0xf04f , 0xf151 , 0xf354 , 0xf457 , 0xf65a , 0xf75c , 0xf95f , 0xfa62 , 0xfc65 , 0xfc67 , 0xfc69 , 0xfd6b , 0xfd6d , 0xfd6f , 0xfe71 , 0xfe73 , 0xff75 , 0xff77 , 0xff79 , 0xff7b , 0xff7d , 0xff7f , 0xff81 , 0xff83 , 0xff85 , 0xff86 , 0xff88 , 0xff8a , 0xff8c , 0xff8d , 0xff8f , 0xff91 , 0xff93 , 0xff94 , 0xff96 , 0xff98 , 0xff9a , 0xff9b , 0xff9d , 0xff9f , 0xffa1 , 0xffa2 , 0xffa4 , 0xffa6 , 0xffa8 , 0xffa9 , 0xffab , 0xffad , 0xffaf , 0xffb0 , 0xffb2 , 0xffb4 , 0xffb6 , 0xffb8 , 0xffba , 0xffbc , 0xffbe , 0xffbf , 0xffc1 , 0xffc3 , 0xffc5 , 0xffc7 , 0xffc9 , 0xffcb , 0xffcd , 0xffce , 0xffd0 , 0xffd2 , 0xffd4 , 0xffd5 , 0xffd7 , 0xffd9 , 0xffdb , 0xffdc , 0xffde , 0xffe0 , 0xffe2 , 0xffe4 , 0xffe6 , 0xffe8 , 0xffea , 0x4ffeb , 0x8ffed , 0xdffef , 0x11fff1 , 0x15fff2 , 0x1afff4 , 0x1efff6 , 0x23fff8 , 0x2afff8 , 0x32fff9 , 0x3afffa , 0x42fffb , 0x4afffc , 0x52fffd , 0x5afffe , 0x62ffff , 0x69ffff , 0x71ffff , 0x79ffff , 0x81ffff , 0x88ffff , 0x90ffff , 0x98ffff , 0xa0ffff , 0xa7ffff , 0xafffff , 0xb7ffff , 0xbfffff , 0xc7ffff , 0xcfffff , 0xd7ffff , 0xdfffff , 0xe3ffff , 0xe7ffff , 0xebffff , 0xefffff , 0xf3ffff , 0xf7ffff , 0xfbffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff}

,

/* 32 neon-magenta */

{0x0 , 0x10001 , 0x30003 , 0x50005 , 0x70007 , 0x90009 , 0xb000b , 0xd000d , 0xf000f , 0x110011 , 0x130013 , 0x150015 , 0x170017 , 0x190019 , 0x1b001b , 0x1d001d , 0x1f001f , 0x210021 , 0x230023 , 0x250025 , 0x270027 , 0x290029 , 0x2b002b , 0x2d002d , 0x2f002f , 0x310031 , 0x330033 , 0x350035 , 0x370037 , 0x390039 , 0x3b003b , 0x3d003d , 0x3f003f , 0x410041 , 0x430043 , 0x450045 , 0x470047 , 0x490049 , 0x4b004b , 0x4d004d , 0x4f004f , 0x510051 , 0x530053 , 0x550055 , 0x570057 , 0x590059 , 0x5b005b , 0x5d005d , 0x5f005f , 0x610061 , 0x630063 , 0x650065 , 0x670067 , 0x690069 , 0x6b006b , 0x6d006d , 0x6f006f , 0x710071 , 0x730073 , 0x750075 , 0x770077 , 0x790079 , 0x7b007b , 0x7d007d , 0x7f007f , 0x810081 , 0x830083 , 0x850085 , 0x870087 , 0x890089 , 0x8b008b , 0x8d008d , 0x8f008f , 0x910091 , 0x930093 , 0x950095 , 0x970097 , 0x990099 , 0x9b009b , 0x9d009d , 0x9f009f , 0xa100a1 , 0xa300a3 , 0xa500a5 , 0xa700a7 , 0xa900a9 , 0xab00ab , 0xad00ad , 0xaf00af , 0xb100b1 , 0xb300b3 , 0xb500b5 , 0xb700b7 , 0xb900b9 , 0xbb00bb , 0xbd00bd , 0xbf00bf , 0xc100c1 , 0xc300c3 , 0xc500c5 , 0xc700c7 , 0xc900c9 , 0xcb00cb , 0xcd00cd , 0xcf00cf , 0xd100d1 , 0xd300d3 , 0xd500d5 , 0xd700d7 , 0xd900d9 , 0xdb00db , 0xdd00dd , 0xdf00df , 0xe100e1 , 0xe300e3 , 0xe500e5 , 0xe700e7 , 0xe900e9 , 0xeb00eb , 0xed00ed , 0xef00ef , 0xf100f1 , 0xf300f3 , 0xf500f5 , 0xf700f7 , 0xf900f9 , 0xfb00fb , 0xff00ff , 0xff00ff , 0xff01ff , 0xff03ff , 0xff05ff , 0xff07ff , 0xff09ff , 0xff0bff , 0xff0dff , 0xff0fff , 0xff11ff , 0xff13ff , 0xff15ff , 0xff17ff , 0xff19ff , 0xff1bff , 0xff1dff , 0xff1fff , 0xff21ff , 0xff23ff , 0xff25ff , 0xff27ff , 0xff29ff , 0xff2bff , 0xff2dff , 0xff2fff , 0xff31ff , 0xff33ff , 0xff35ff , 0xff37ff , 0xff39ff , 0xff3bff , 0xff3dff , 0xff3fff , 0xff41ff , 0xff43ff , 0xff45ff , 0xff47ff , 0xff49ff , 0xff4bff , 0xff4dff , 0xff4fff , 0xff51ff , 0xff53ff , 0xff55ff , 0xff57ff , 0xff59ff , 0xff5bff , 0xff5dff , 0xff5fff , 0xff61ff , 0xff63ff , 0xff65ff , 0xff67ff , 0xff69ff , 0xff6bff , 0xff6dff , 0xff6fff , 0xff71ff , 0xff73ff , 0xff75ff , 0xff77ff , 0xff79ff , 0xff7bff , 0xff7dff , 0xff7fff , 0xff81ff , 0xff83ff , 0xff85ff , 0xff87ff , 0xff89ff , 0xff8bff , 0xff8dff , 0xff8fff , 0xff91ff , 0xff93ff , 0xff95ff , 0xff97ff , 0xff99ff , 0xff9bff , 0xff9dff , 0xff9fff , 0xffa1ff , 0xffa3ff , 0xffa5ff , 0xffa7ff , 0xffa9ff , 0xffabff , 0xffadff , 0xffafff , 0xffb1ff , 0xffb3ff , 0xffb5ff , 0xffb7ff , 0xffb9ff , 0xffbbff , 0xffbdff , 0xffbfff , 0xffc1ff , 0xffc3ff , 0xffc5ff , 0xffc7ff , 0xffc9ff , 0xffcbff , 0xffcdff , 0xffcfff , 0xffd1ff , 0xffd3ff , 0xffd5ff , 0xffd7ff , 0xffd9ff , 0xffdbff , 0xffddff , 0xffdfff , 0xffe1ff , 0xffe3ff , 0xffe5ff , 0xffe7ff , 0xffe9ff , 0xffebff , 0xffedff , 0xffefff , 0xfff1ff , 0xfff3ff , 0xfff5ff , 0xfff7ff , 0xfff9ff , 0xfffbff , 0xffffff}

,

/* 33 rgb-red */

{0x0 , 0x1 , 0x2 , 0x3 , 0x4 , 0x5 , 0x6 , 0x7 , 0x8 , 0x9 , 0xa , 0xb , 0xc , 0xd , 0xe , 0xf , 0x10 , 0x11 , 0x12 , 0x13 , 0x14 , 0x15 , 0x16 , 0x17 , 0x18 , 0x19 , 0x1a , 0x1b , 0x1c , 0x1d , 0x1e , 0x1f , 0x20 , 0x21 , 0x22 , 0x23 , 0x24 , 0x25 , 0x26 , 0x27 , 0x28 , 0x29 , 0x2a , 0x2b , 0x2c , 0x2d , 0x2e , 0x2f , 0x30 , 0x31 , 0x32 , 0x33 , 0x34 , 0x35 , 0x36 , 0x37 , 0x38 , 0x39 , 0x3a , 0x3b , 0x3c , 0x3d , 0x3e , 0x3f , 0x40 , 0x41 , 0x42 , 0x43 , 0x44 , 0x45 , 0x46 , 0x47 , 0x48 , 0x49 , 0x4a , 0x4b , 0x4c , 0x4d , 0x4e , 0x4f , 0x50 , 0x51 , 0x52 , 0x53 , 0x54 , 0x55 , 0x56 , 0x57 , 0x58 , 0x59 , 0x5a , 0x5b , 0x5c , 0x5d , 0x5e , 0x5f , 0x60 , 0x61 , 0x62 , 0x63 , 0x64 , 0x65 , 0x66 , 0x67 , 0x68 , 0x69 , 0x6a , 0x6b , 0x6c , 0x6d , 0x6e , 0x6f , 0x70 , 0x71 , 0x72 , 0x73 , 0x74 , 0x75 , 0x76 , 0x77 , 0x78 , 0x79 , 0x7a , 0x7b , 0x7c , 0x7d , 0x7e , 0x7f , 0x80 , 0x81 , 0x82 , 0x83 , 0x84 , 0x85 , 0x86 , 0x87 , 0x88 , 0x89 , 0x8a , 0x8b , 0x8c , 0x8d , 0x8e , 0x8f , 0x90 , 0x91 , 0x92 , 0x93 , 0x94 , 0x95 , 0x96 , 0x97 , 0x98 , 0x99 , 0x9a , 0x9b , 0x9c , 0x9d , 0x9e , 0x9f , 0xa0 , 0xa1 , 0xa2 , 0xa3 , 0xa4 , 0xa5 , 0xa6 , 0xa7 , 0xa8 , 0xa9 , 0xaa , 0xab , 0xac , 0xad , 0xae , 0xaf , 0xb0 , 0xb1 , 0xb2 , 0xb3 , 0xb4 , 0xb5 , 0xb6 , 0xb7 , 0xb8 , 0xb9 , 0xba , 0xbb , 0xbc , 0xbd , 0xbe , 0xbf , 0xc0 , 0xc1 , 0xc2 , 0xc3 , 0xc4 , 0xc5 , 0xc6 , 0xc7 , 0xc8 , 0xc9 , 0xca , 0xcb , 0xcc , 0xcd , 0xce , 0xcf , 0xd0 , 0xd1 , 0xd2 , 0xd3 , 0xd4 , 0xd5 , 0xd6 , 0xd7 , 0xd8 , 0xd9 , 0xda , 0xdb , 0xdc , 0xdd , 0xde , 0xdf , 0xe0 , 0xe1 , 0xe2 , 0xe3 , 0xe4 , 0xe5 , 0xe6 , 0xe7 , 0xe8 , 0xe9 , 0xea , 0xeb , 0xec , 0xed , 0xee , 0xef , 0xf0 , 0xf1 , 0xf2 , 0xf3 , 0xf4 , 0xf5 , 0xf6 , 0xf7 , 0xf8 , 0xf9 , 0xfa , 0xfb , 0xfc , 0xfd , 0xfe , 0xff}

,

/* 34 pastel */

{0x0 , 0xf0f0f , 0x212121 , 0x303030 , 0x404040 , 0x4f4f4f , 0x616161 , 0x707070 , 0x808080 , 0x8f8f8f , 0xa1a1a1 , 0xb0b0b0 , 0xbfbfbf , 0xd1d1d1 , 0xe0e0e0 , 0xf0f0f0 , 0xede3e3 , 0xedd9d9 , 0xedcccc , 0xebc2c2 , 0xebb5b5 , 0xe8abab , 0xe89e9e , 0xe89494 , 0xe68787 , 0xe67d7d , 0xe37070 , 0xe36666 , 0xe35959 , 0xe04f4f , 0xe04545 , 0xdb454f , 0xd9455c , 0xd44569 , 0xd14573 , 0xcc4580 , 0xc7478c , 0xc44796 , 0xc247a3 , 0xbd47ad , 0xba47ba , 0xb54ac7 , 0xb04ad1 , 0xad4ade , 0xa84ae8 , 0xa64af5 , 0x9e4af5 , 0x994af2 , 0x944af2 , 0x8c4af2 , 0x874af0 , 0x824af0 , 0x7a4aed , 0x734aed , 0x6e4aeb , 0x694aeb , 0x614ae8 , 0x5c4ae8 , 0x574ae8 , 0x4f4ae6 , 0x4a4ae6 , 0x4a57e8 , 0x4a61e8 , 0x4a6eeb , 0x4a78eb , 0x4a85ed , 0x4a8fed , 0x4a9cf0 , 0x4aa6f2 , 0x4ab0f2 , 0x4abdf5 , 0x4ac7f5 , 0x4ad4f7 , 0x4adefa , 0x4aebfa , 0x4af5fc , 0x47edf5 , 0x47e3ed , 0x47d9e3 , 0x47cfdb , 0x47c4d4 , 0x47bdcc , 0x47b3c4 , 0x47a8bd , 0x479eb5 , 0x4794ab , 0x478ca3 , 0x47829c , 0x477894 , 0x476e8c , 0x476685 , 0x4a6980 , 0x4a6b7a , 0x4a6e78 , 0x4a7073 , 0x4a736e , 0x4d756b , 0x4d7866 , 0x4d7a61 , 0x4d7d5c , 0x4f8259 , 0x4f8554 , 0x4f874f , 0x4f8a4d , 0x4f8c47 , 0x528f45 , 0x529445 , 0x529647 , 0x529c4a , 0x529e4a , 0x52a34d , 0x52a64d , 0x52ab4f , 0x52b04f , 0x52b352 , 0x52b854 , 0x52ba54 , 0x52bf57 , 0x52c259 , 0x52c759 , 0x54cc5c , 0x5ec959 , 0x69c959 , 0x73c957 , 0x7dc757 , 0x87c754 , 0x91c754 , 0x9cc752 , 0xa6c74f , 0xb0c44f , 0xbac44d , 0xc7c44d , 0xd1c24a , 0xdbc24a , 0xe6c247 , 0xf0c247 , 0xedba45 , 0xedb045 , 0xeda845 , 0xeba145 , 0xeb9945 , 0xe88f45 , 0xe88745 , 0xe88045 , 0xe67845 , 0xe66e45 , 0xe36645 , 0xe35e45 , 0xe35445 , 0xe04d45 , 0xe04545 , 0xde4547 , 0xdb454a , 0xd9454f , 0xd64552 , 0xd44557 , 0xd14559 , 0xcf455c , 0xcc4561 , 0xcc4563 , 0xc74566 , 0xc7456b , 0xc4456e , 0xc24573 , 0xbf4575 , 0xbd4578 , 0xba4582 , 0xba458a , 0xb84591 , 0xb54599 , 0xb545a3 , 0xb347ab , 0xb347b3 , 0xb047ba , 0xb047c4 , 0xad47cc , 0xab4ad4 , 0xab4ade , 0xa84ae6 , 0xa64aed , 0xa64af5 , 0x9e4af2 , 0x9947ed , 0x9145eb , 0x8c42e6 , 0x8540e3 , 0x7d40de , 0x753dd9 , 0x703bd6 , 0x6938d1 , 0x6138cf , 0x5c36c9 , 0x5433c7 , 0x4f30c2 , 0x332ebf , 0x402eba , 0x422ebd , 0x4230bf , 0x4233c2 , 0x4536c7 , 0x4538c9 , 0x4538cc , 0x453bcf , 0x453dd1 , 0x4540d4 , 0x4740d6 , 0x4742d9 , 0x4745de , 0x4747e0 , 0x4a4ae3 , 0x4a4ae6 , 0x4a4fe8 , 0x4a54e8 , 0x4a59eb , 0x475ced , 0x4761ed , 0x4766f0 , 0x476bf2 , 0x476ef5 , 0x4773f5 , 0x4778f5 , 0x477af7 , 0x4580fa , 0x4585fa , 0x4587fc , 0x458cff , 0x4794ff , 0x479cff , 0x47a3ff , 0x47a8ff , 0x47b0ff , 0x47c4fc , 0x47bffc , 0x47c4fc , 0x47ccfc , 0x47d1fc , 0x47d9fc , 0x47e0fc , 0x47e8fc , 0x47edfc , 0x47f5fc , 0x54f5fc , 0x61f7fc , 0x6ef7fc , 0x78f7fc , 0x85f7fc , 0x91fafc , 0x9efafc , 0xabfaff , 0xb5faff , 0xc2faff , 0xcffcff , 0xd9fcff , 0xe6fcff , 0xf2ffff , 0xffffff}

,

/* 35 cool */

{0x20301 , 0x20301 , 0x40501 , 0x60701 , 0x80901 , 0xa0b01 , 0xc0d01 , 0xe0f01 , 0x101101 , 0x121301 , 0x141501 , 0x161701 , 0x181901 , 0x1a1b01 , 0x1c1d01 , 0x1e1f01 , 0x202101 , 0x222301 , 0x242501 , 0x262701 , 0x282901 , 0x2a2b01 , 0x2c2d01 , 0x2e2f01 , 0x303101 , 0x323301 , 0x343501 , 0x363701 , 0x383901 , 0x3a3b01 , 0x3c3d01 , 0x3e3f01 , 0x404001 , 0x414201 , 0x434401 , 0x454601 , 0x474801 , 0x494a01 , 0x4b4c01 , 0x4d4e01 , 0x4f5001 , 0x515201 , 0x535401 , 0x555601 , 0x575801 , 0x595a01 , 0x5b5c01 , 0x5d5e01 , 0x5f6001 , 0x616201 , 0x636401 , 0x656601 , 0x676801 , 0x696a01 , 0x6b6c01 , 0x6d6e01 , 0x6f7001 , 0x717201 , 0x737401 , 0x757601 , 0x777801 , 0x797a01 , 0x7b7c01 , 0x7d7e01 , 0x7f8001 , 0x817e02 , 0x837c04 , 0x857a06 , 0x877808 , 0x89760a , 0x8b740c , 0x8d720e , 0x8f7010 , 0x916e12 , 0x936c14 , 0x956a16 , 0x976818 , 0x99661a , 0x9b641c , 0x9d621e , 0x9f6020 , 0xa15e22 , 0xa35c24 , 0xa55a26 , 0xa75828 , 0xa9562a , 0xab542c , 0xad522e , 0xaf5030 , 0xb14e32 , 0xb34c34 , 0xb54a36 , 0xb74838 , 0xb9463a , 0xbb443c , 0xbd423e , 0xbf4040 , 0xc03f41 , 0xc23d43 , 0xc43b45 , 0xc63947 , 0xc83749 , 0xca354b , 0xcc334d , 0xce314f , 0xd02f51 , 0xd22d53 , 0xd42b55 , 0xd62957 , 0xd82759 , 0xda255b , 0xdc235d , 0xde215f , 0xe01f61 , 0xe21d63 , 0xe41b65 , 0xe61967 , 0xe81769 , 0xea156b , 0xec136d , 0xee116f , 0xf00f71 , 0xf20d73 , 0xf40b75 , 0xf60977 , 0xf80779 , 0xfa057b , 0xfc037d , 0xfe017f , 0xfa0281 , 0xf60483 , 0xf20685 , 0xee0887 , 0xea0a89 , 0xe60c8b , 0xe20e8d , 0xde108f , 0xda1291 , 0xd61493 , 0xd21695 , 0xce1897 , 0xca1a99 , 0xc61c9b , 0xc21e9d , 0xbf209f , 0xbb22a1 , 0xb724a3 , 0xb326a5 , 0xaf28a7 , 0xab2aa9 , 0xa72cab , 0xa32ead , 0x9f30af , 0x9b32b1 , 0x9734b3 , 0x9336b5 , 0x8f38b7 , 0x8b3ab9 , 0x873cbb , 0x833ebd , 0x7f40bf , 0x7b41c0 , 0x7743c2 , 0x7345c4 , 0x6f47c6 , 0x6b49c8 , 0x674bca , 0x634dcc , 0x5f4fce , 0x5b51d0 , 0x5753d2 , 0x5355d4 , 0x4f57d6 , 0x4b59d8 , 0x475bda , 0x435ddc , 0x405fde , 0x3c61e0 , 0x3863e2 , 0x3465e4 , 0x3067e6 , 0x2c69e8 , 0x286bea , 0x246dec , 0x206fee , 0x1c71f0 , 0x1873f2 , 0x1475f4 , 0x1077f6 , 0xc79f8 , 0x87bfa , 0x47dfc , 0x17ffe , 0x481fe , 0x883fe , 0xc85fe , 0x1087fe , 0x1489fe , 0x188bfe , 0x1c8dfe , 0x208ffe , 0x2491fe , 0x2893fe , 0x2c95fe , 0x3097fe , 0x3499fe , 0x389bfe , 0x3c9dfe , 0x409ffe , 0x43a1fe , 0x47a3fe , 0x4ba5fe , 0x4fa7fe , 0x53a9fe , 0x57abfe , 0x5badfe , 0x5faffe , 0x63b1fe , 0x67b3fe , 0x6bb5fe , 0x6fb7fe , 0x73b9fe , 0x77bbfe , 0x7bbdfe , 0x7fbffe , 0x83c0fe , 0x87c2fe , 0x8bc4fe , 0x8fc6fe , 0x93c8fe , 0x97cafe , 0x9bccfe , 0x9fcefe , 0xa3d0fe , 0xa7d2fe , 0xabd4fe , 0xafd6fe , 0xb3d8fe , 0xb7dafe , 0xbbdcfe , 0xbfdefe , 0xc2e0fe , 0xc6e2fe , 0xcae4fe , 0xcee6fe , 0xd2e8fe , 0xd6eafe , 0xdaecfe , 0xdeeefe , 0xe2f0fe , 0xe6f2fe , 0xeaf4fe , 0xeef6fe , 0xf2f8fe , 0xf6fafe , 0xf6fafe}

,

/* 36 topography */

{0x0 , 0x10101 , 0x30000 , 0x60000 , 0x90000 , 0xb0000 , 0xe0000 , 0x110000 , 0x130000 , 0x160000 , 0x190000 , 0x1b0000 , 0x1e0000 , 0x210000 , 0x230000 , 0x260000 , 0x290000 , 0x2b0000 , 0x2e0000 , 0x310000 , 0x330000 , 0x360000 , 0x390000 , 0x3b0000 , 0x3e0000 , 0x410000 , 0x430000 , 0x460000 , 0x490000 , 0x4b0000 , 0x4e0000 , 0x510000 , 0x530000 , 0x560000 , 0x590000 , 0x5b0000 , 0x5e0000 , 0x610000 , 0x630000 , 0x660000 , 0x690000 , 0x6b0000 , 0x6e0000 , 0x710000 , 0x730000 , 0x760000 , 0x790000 , 0x7b0000 , 0x7e0000 , 0x810000 , 0x840000 , 0x860000 , 0x890000 , 0x8c0000 , 0x8e0000 , 0x910000 , 0x940000 , 0x960000 , 0x990000 , 0x9c0000 , 0x9e0000 , 0xa10000 , 0xa40000 , 0xa60000 , 0xa90000 , 0xac0000 , 0xae0000 , 0xb10000 , 0xb40000 , 0xb60000 , 0xb90000 , 0xbc0000 , 0xbe0000 , 0xc10000 , 0xc40000 , 0xc60000 , 0xc90000 , 0xcc0000 , 0xce0000 , 0xd10000 , 0xd40000 , 0xd60000 , 0xd90000 , 0xdc0000 , 0xde0000 , 0xe10000 , 0xe40000 , 0xe60000 , 0xe90000 , 0xec0000 , 0xee0000 , 0xf10000 , 0xf40000 , 0xf60000 , 0xf90000 , 0xfc0000 , 0xff0000 , 0xff0502 , 0xff0b04 , 0xff1106 , 0xff1709 , 0xff1d0b , 0xff230d , 0xff2910 , 0xff2e12 , 0xff3414 , 0xff3a16 , 0xff4019 , 0xff461b , 0xff4c1d , 0xff5220 , 0xff5822 , 0xff5d24 , 0xff6326 , 0xff6929 , 0xff6f2b , 0xff752d , 0xff7b30 , 0xff8132 , 0xff8734 , 0xff8c36 , 0xff9239 , 0xff983b , 0xff9e3d , 0xffa440 , 0xffaa42 , 0xffff , 0xffff , 0x66f5ff , 0x33f7f3 , 0x33fb9e , 0x30fd48 , 0x2dfb47 , 0x2af946 , 0x27f845 , 0x24f644 , 0x21f443 , 0x1ef342 , 0x1bf141 , 0x18ef40 , 0x15ed3f , 0x12ec3e , 0xfea3d , 0xce83c , 0x9e539 , 0x8e338 , 0x8e136 , 0x8df34 , 0x8de32 , 0x7dc30 , 0x7da2f , 0x7d82d , 0x6d62b , 0x6d429 , 0x6d227 , 0x5d026 , 0x5ce24 , 0x5cc22 , 0x5ca20 , 0x4c81e , 0x4c61c , 0x4c41b , 0x4c319 , 0x3c117 , 0x3bf15 , 0x3bd13 , 0x2bb12 , 0x2b910 , 0x2b70e , 0x2b50c , 0x1b30a , 0x1b109 , 0x1af07 , 0xad05 , 0xab03 , 0xa901 , 0xa211 , 0x9b22 , 0x9926 , 0x972a , 0x952f , 0x9333 , 0x9137 , 0x8f3c , 0x8d40 , 0x8c45 , 0x8a49 , 0x884d , 0x8652 , 0x8456 , 0x825a , 0x805f , 0x7e63 , 0x7d68 , 0x7b6c , 0x7970 , 0x7775 , 0x7579 , 0x737d , 0x7182 , 0x6f86 , 0x6e87 , 0x6e89 , 0x6d8a , 0x6d8c , 0x6c8d , 0x6c8f , 0x6b90 , 0x6b92 , 0x6a93 , 0x6a95 , 0x6996 , 0x6998 , 0x6899 , 0x689b , 0x679c , 0x679e , 0x669f , 0x66a1 , 0x4f93ba , 0x629fc1 , 0x76abc9 , 0x89b7d1 , 0x9dc3d8 , 0xb0cfe0 , 0xc4dbe8 , 0xd7e7ef , 0xffffff , 0xfefefe , 0xfefefd , 0xfefefc , 0xfefdfb , 0xfefdfb , 0xfefdfa , 0xfefcf9 , 0xfefcf8 , 0xfefcf8 , 0xfefbf7 , 0xfefbf6 , 0xfefbf5 , 0xfefaf5 , 0xfefaf4 , 0xfefaf3 , 0xfefaf2 , 0xfef9f2 , 0xfef9f1 , 0xfef9f0 , 0xfef8ef , 0xfef8ef , 0xfef8ee , 0xfef7ed , 0xfef7ec , 0xfef7ec , 0xfef6eb , 0xfef6ea , 0xfef6e9}

,

/* 37 iman */

{0x0 , 0x38 , 0x3c , 0x40 , 0x44 , 0x48 , 0x50 , 0x54 , 0x58 , 0x5c , 0x64 , 0x68 , 0x6c , 0x70 , 0x74 , 0x7c , 0x80 , 0x84 , 0x88 , 0x90 , 0x94 , 0x98 , 0x9c , 0xa0 , 0xa8 , 0xac , 0xb0 , 0xb4 , 0xbc , 0xc0 , 0xc4 , 0xc8 , 0xcc , 0xd4 , 0xd8 , 0xdc , 0xe0 , 0xe8 , 0xec , 0xf0 , 0xf4 , 0xfc , 0x4fc , 0xcfc , 0x10fc , 0x18fc , 0x20fc , 0x24fc , 0x2cfc , 0x34fc , 0x38fc , 0x40fc , 0x48fc , 0x4cfc , 0x54fc , 0x5cfc , 0x60fc , 0x68fc , 0x70fc , 0x74fc , 0x7cfc , 0x84fc , 0x88fc , 0x90fc , 0x98fc , 0x9cfc , 0xa4fc , 0xacfc , 0xb0fc , 0xb8fc , 0xc0fc , 0xc4fc , 0xccfc , 0xd4fc , 0xd8fc , 0xe0fc , 0xe8fc , 0xecfc , 0xf4fc , 0xfcfc , 0xfcf8 , 0xfcf0 , 0xf8e8 , 0xf8e0 , 0xf4d8 , 0xf4d0 , 0xf0c8 , 0xf0c0 , 0xf0b8 , 0xecb0 , 0xeca8 , 0xe8a0 , 0xe898 , 0xe490 , 0xe488 , 0x4e080 , 0x4e078 , 0x4e070 , 0x4dc68 , 0x4dc60 , 0x4d858 , 0x4d850 , 0x4d448 , 0x4d440 , 0x4d438 , 0x4d030 , 0x4d028 , 0x4cc20 , 0x4cc18 , 0x4c810 , 0x4c808 , 0x8c400 , 0x18c400 , 0x2cc800 , 0x3cc800 , 0x50cc00 , 0x60cc00 , 0x74d000 , 0x84d000 , 0x98d400 , 0xa8d400 , 0xbcd800 , 0xbcd400 , 0xbcd000 , 0xbccc00 , 0xbcc800 , 0xbcc000 , 0xc0bc00 , 0xc0b800 , 0xc0b000 , 0xc4ac00 , 0xc4a804 , 0xc4a004 , 0xc49c04 , 0xc89404 , 0xc89004 , 0xc88c04 , 0xcc8404 , 0xcc8004 , 0xcc7c08 , 0xd07408 , 0xd07008 , 0xd06c08 , 0xd06408 , 0xd46008 , 0xd45808 , 0xd4540c , 0xd8500c , 0xd8480c , 0xd8440c , 0xdc400c , 0xdc380c , 0xdc340c , 0xe02c10 , 0xe02810 , 0xe42410 , 0xe41c10 , 0xe41810 , 0xe81014 , 0xe81018 , 0xe4101c , 0xe01020 , 0xe01024 , 0xdc1028 , 0xd8102c , 0xd41030 , 0xd41034 , 0xd01038 , 0xcc103c , 0xc81040 , 0xc81040 , 0xc81044 , 0xc81044 , 0xc41448 , 0xc4144c , 0xc4144c , 0xc01450 , 0xc01854 , 0xc01854 , 0xc01858 , 0xbc1858 , 0xbc1c5c , 0xbc1c60 , 0xb81c60 , 0xb81c64 , 0xb82068 , 0xb42068 , 0xb4206c , 0xb42470 , 0xb42470 , 0xb02474 , 0xb02474 , 0xb02878 , 0xac287c , 0xac287c , 0xac2880 , 0xa82c84 , 0xa82c84 , 0xa82c88 , 0xa4308c , 0xa4308c , 0xa4348c , 0xa0388c , 0xa03c8c , 0xa0408c , 0x9c448c , 0x9c448c , 0x984888 , 0x984c88 , 0x985088 , 0x945488 , 0x945888 , 0x945888 , 0x905c88 , 0x906088 , 0x8c6484 , 0x8c6884 , 0x8c6c84 , 0x886c84 , 0x887084 , 0x887484 , 0x847884 , 0x847c84 , 0x808080 , 0x848484 , 0x888888 , 0x8c8c8c , 0x909090 , 0x949494 , 0x989898 , 0x9c9c9c , 0xa0a0a0 , 0xa4a4a4 , 0xa8a8a8 , 0xacacac , 0xb0b0b0 , 0xb4b4b4 , 0xb8b8b8 , 0xbcbcbc , 0xc0c0c0 , 0xc4c4c4 , 0xc8c8c8 , 0xcccccc , 0xd0d0d0 , 0xd4d4d4 , 0xd8d8d8 , 0xdcdcdc , 0xe0e0e0 , 0xe4e4e4 , 0xe8e8e8 , 0xececec , 0xf0f0f0 , 0xf4f4f4 , 0xf8f8f8 , 0xfcfcfc , 0xfcfcfc}

,

/* 38 Green */

{0x11601 , 0x11601 , 0x11701 , 0x11801 , 0x11901 , 0x11a01 , 0x11b01 , 0x11c01 , 0x11d01 , 0x11e01 , 0x11f01 , 0x12001 , 0x12101 , 0x12201 , 0x12301 , 0x12401 , 0x12501 , 0x12601 , 0x12701 , 0x12801 , 0x12901 , 0x12a01 , 0x12b01 , 0x12c01 , 0x12d01 , 0x12e01 , 0x12f01 , 0x13001 , 0x13101 , 0x13201 , 0x13301 , 0x13401 , 0x13501 , 0x13601 , 0x13701 , 0x13801 , 0x13901 , 0x13a01 , 0x13b01 , 0x13c01 , 0x13d01 , 0x13e01 , 0x13f01 , 0x14001 , 0x14001 , 0x14101 , 0x14201 , 0x14301 , 0x14401 , 0x14501 , 0x14701 , 0x14801 , 0x14902 , 0x14a03 , 0x14b04 , 0x14c05 , 0x14d06 , 0x14e07 , 0x14f08 , 0x15009 , 0x1510a , 0x1520b , 0x1530c , 0x1540d , 0x1550e , 0x1560f , 0x15710 , 0x15811 , 0x15912 , 0x15a13 , 0x15b14 , 0x15c15 , 0x15d16 , 0x15e17 , 0x15f18 , 0x16019 , 0x1611a , 0x1621b , 0x1631c , 0x1641d , 0x1651e , 0x1661f , 0x16720 , 0x16821 , 0x16922 , 0x16a23 , 0x16b24 , 0x16c25 , 0x16d26 , 0x16e27 , 0x26f28 , 0x37029 , 0x4712a , 0x5722b , 0x6732c , 0x7742d , 0x8752e , 0x9762f , 0xa7730 , 0xb7831 , 0xc7932 , 0xd7a33 , 0xe7b34 , 0xf7c35 , 0x107d36 , 0x117e37 , 0x127f38 , 0x138039 , 0x14813a , 0x15823b , 0x16833c , 0x17843d , 0x18853e , 0x19863f , 0x1a8740 , 0x1b8840 , 0x1c8941 , 0x1d8a42 , 0x1e8b43 , 0x1f8c44 , 0x208d45 , 0x218e46 , 0x228f47 , 0x239048 , 0x249149 , 0x25924a , 0x26934b , 0x27944c , 0x28954d , 0x29964e , 0x2a974f , 0x2b9850 , 0x2c9951 , 0x2d9a52 , 0x2e9b53 , 0x2f9c54 , 0x309d55 , 0x319e56 , 0x329f57 , 0x33a058 , 0x34a159 , 0x35a25a , 0x36a35b , 0x37a45c , 0x38a55d , 0x39a65e , 0x3aa75f , 0x3ba860 , 0x3ca961 , 0x3daa62 , 0x3eab63 , 0x3fac64 , 0x40ad65 , 0x40ae66 , 0x41af67 , 0x42b068 , 0x43b169 , 0x44b26a , 0x45b36b , 0x46b46c , 0x47b56d , 0x48b66e , 0x49b76f , 0x4ab870 , 0x4bb971 , 0x4cba72 , 0x4dbb73 , 0x4ebc74 , 0x4fbd75 , 0x50be76 , 0x51bf77 , 0x52bf79 , 0x53c07a , 0x54c17b , 0x55c27c , 0x56c37d , 0x57c47e , 0x58c57f , 0x59c680 , 0x5ac781 , 0x5bc882 , 0x5cc983 , 0x5dca84 , 0x5ecb85 , 0x5fcc86 , 0x60cd87 , 0x61ce88 , 0x62cf89 , 0x63d08a , 0x64d18b , 0x65d28c , 0x66d38d , 0x67d48e , 0x68d58f , 0x69d690 , 0x6ad791 , 0x6bd892 , 0x6cd993 , 0x6dda94 , 0x6edb95 , 0x6fdc96 , 0x70dd97 , 0x71de98 , 0x72df99 , 0x73e09a , 0x74e19b , 0x75e29c , 0x76e39d , 0x77e49e , 0x78e59f , 0x79e6a0 , 0x7ae7a1 , 0x7be8a2 , 0x7ce9a3 , 0x7deaa4 , 0x7eeba5 , 0x7feca6 , 0x80eda7 , 0x81eea8 , 0x82efa9 , 0x83f0aa , 0x84f1ab , 0x85f2ac , 0x86f3ad , 0x87f4ae , 0x88f5af , 0x89f6b0 , 0x8af7b1 , 0x8bf8b2 , 0x8cf9b3 , 0x8dfab4 , 0x8efbb5 , 0x8ffcb6 , 0x90fdb7 , 0x91feb8 , 0x92feb9 , 0x93feba , 0x94febb , 0x95febc , 0x96febd , 0x97febe , 0x98febf , 0x99febf , 0x9afec0 , 0x9bfec1 , 0x9cfec2 , 0x9dfec3 , 0x9efec4 , 0x9ffec5 , 0xa0fec6 , 0xa1fec7 , 0xa3fec8 , 0xa4fec9 , 0xa5feca , 0xa6fecb , 0xa6fecb}

,

/* 39 Amber */

{0x9bf3e7 , 0x9bf3e7 , 0x9afef2 , 0x99fef2 , 0x98fef3 , 0x96fef3 , 0x95fef3 , 0x94fef3 , 0x93fef4 , 0x92fef4 , 0x91fef4 , 0x90fef5 , 0x8ffef5 , 0x8efef5 , 0x8dfef6 , 0x8cfef6 , 0x8bfef6 , 0x8afef7 , 0x89fef7 , 0x88fef8 , 0x87fef8 , 0x85fef9 , 0x84fef9 , 0x83fefa , 0x82fefa , 0x81fefb , 0x80fefb , 0x7ffefc , 0x7efefc , 0x7dfefd , 0x7cfefd , 0x7bfefe , 0x7afdfe , 0x79fdfe , 0x78fcfe , 0x77fbfe , 0x76fbfe , 0x74fafe , 0x73f9fe , 0x72f8fe , 0x71f8fe , 0x70f7fe , 0x6ff6fe , 0x6ef5fe , 0x6df5fe , 0x6cf4fe , 0x6bf3fe , 0x6af2fe , 0x69f1fe , 0x68f0fe , 0x67effe , 0x66eefe , 0x65edfe , 0x63edfd , 0x62ecfd , 0x61ebfd , 0x60eafd , 0x5fe9fd , 0x5ee8fd , 0x5de7fd , 0x5ce6fd , 0x5be4fd , 0x5ae3fd , 0x59e2fd , 0x58e1fd , 0x57e0fd , 0x56dffd , 0x55defd , 0x54ddfd , 0x52dbfd , 0x51dafd , 0x50d9fd , 0x4fd8fd , 0x4ed7fd , 0x4dd5fd , 0x4cd4fd , 0x4bd3fd , 0x4ad2fd , 0x49d0fd , 0x48cffd , 0x47cefd , 0x46ccfd , 0x45cbfd , 0x44cafd , 0x43c8fd , 0x41c7fd , 0x40c5fd , 0x3fc4fd , 0x3ec3fd , 0x3dc1fd , 0x3cc0fd , 0x3bbefd , 0x3abdfd , 0x39bbfd , 0x38bafd , 0x37b8fd , 0x36b7fd , 0x35b5fd , 0x34b3fd , 0x33b1fb , 0x32affa , 0x30adf8 , 0x2fabf6 , 0x2ea8f5 , 0x2da6f3 , 0x2ca4f2 , 0x2ba2f0 , 0x2aa0ee , 0x299eed , 0x289ceb , 0x2799e9 , 0x2697e8 , 0x2595e6 , 0x2493e5 , 0x2391e3 , 0x228fe1 , 0x218de0 , 0x1f8bde , 0x1e89dd , 0x1d87db , 0x1c85d9 , 0x1b83d8 , 0x1a81d6 , 0x1980d4 , 0x187ed3 , 0x177cd1 , 0x167ad0 , 0x1578ce , 0x1476cc , 0x1374cb , 0x1273c9 , 0x1171c7 , 0x106fc6 , 0xe6dc4 , 0xd6bc3 , 0xc6ac1 , 0xb68bf , 0xa66be , 0x965bc , 0x863ba , 0x761b9 , 0x660b7 , 0x55eb6 , 0x45cb4 , 0x35bb2 , 0x259b1 , 0x157af , 0x56ad , 0x154ac , 0x153aa , 0x151a9 , 0x150a7 , 0x14ea5 , 0x14da4 , 0x14ba2 , 0x14aa1 , 0x1489f , 0x1479d , 0x1459c , 0x1449a , 0x14398 , 0x14197 , 0x14095 , 0x13f94 , 0x13d92 , 0x13c90 , 0x13b8f , 0x1398d , 0x1388b , 0x1378a , 0x13688 , 0x13487 , 0x13385 , 0x13283 , 0x13182 , 0x13080 , 0x12f7e , 0x12d7d , 0x12c7b , 0x12b7a , 0x12a78 , 0x12976 , 0x12875 , 0x12773 , 0x12671 , 0x12570 , 0x1246e , 0x1236d , 0x226b , 0x2169 , 0x2068 , 0x1f66 , 0x1e64 , 0x1d63 , 0x1c61 , 0x1b60 , 0x1a5e , 0x195c , 0x185b , 0x1859 , 0x1757 , 0x1656 , 0x1554 , 0x1453 , 0x1451 , 0x134f , 0x124e , 0x114c , 0x114a , 0x1049 , 0xf47 , 0xf46 , 0xe44 , 0xd42 , 0xd41 , 0xc3f , 0xc3d , 0xb3c , 0xa3a , 0xa39 , 0x937 , 0x935 , 0x834 , 0x832 , 0x730 , 0x72f , 0x62d , 0x62c , 0x62a , 0x528 , 0x527 , 0x425 , 0x423 , 0x422 , 0x320 , 0x31f , 0x31d , 0x21b , 0x21a , 0x218 , 0x216 , 0x115 , 0x113 , 0x112 , 0x110 , 0x10e , 0xd , 0xb , 0x9 , 0x8 , 0x6 , 0x5 , 0x3 , 0x1 , 0x10101 , 0x10101}

,

/* 40 thal_256 */

{0x90101 , 0x90101 , 0x110101 , 0x190101 , 0x210101 , 0x290101 , 0x310101 , 0x390101 , 0x400101 , 0x480101 , 0x500101 , 0x580101 , 0x600101 , 0x680101 , 0x700101 , 0x780101 , 0x800101 , 0x840101 , 0x880101 , 0x8c0101 , 0x900101 , 0x940101 , 0x980101 , 0x9c0101 , 0xa00101 , 0xa40101 , 0xa80101 , 0xac0101 , 0xb00101 , 0xb40101 , 0xb80101 , 0xbc0101 , 0xbf0101 , 0xc30101 , 0xc70101 , 0xcb0101 , 0xcf0101 , 0xd30101 , 0xd70101 , 0xdb0101 , 0xdf0101 , 0xe30101 , 0xe70101 , 0xeb0101 , 0xef0101 , 0xf30101 , 0xf70101 , 0xfb0101 , 0xfe0101 , 0xef0501 , 0xdf0901 , 0xcf0d01 , 0xbf1101 , 0xb01501 , 0xa01901 , 0x901d01 , 0x802101 , 0x702501 , 0x602901 , 0x502d01 , 0x403101 , 0x313501 , 0x213901 , 0x113d01 , 0x14001 , 0x14401 , 0x14801 , 0x14c01 , 0x15001 , 0x15401 , 0x15801 , 0x15c01 , 0x16001 , 0x16401 , 0x16801 , 0x16c01 , 0x17001 , 0x17401 , 0x17801 , 0x17c01 , 0x18001 , 0x18401 , 0x18801 , 0x18c01 , 0x19001 , 0x19401 , 0x19801 , 0x19c01 , 0x1a001 , 0x1a401 , 0x1a801 , 0x1ac01 , 0x1b001 , 0x1b401 , 0x1b801 , 0x1bc01 , 0x1bf01 , 0x1c301 , 0x1c701 , 0x1cb01 , 0x1cf01 , 0x1d301 , 0x1d701 , 0x1db01 , 0x1df01 , 0x1e301 , 0x1e701 , 0x1eb01 , 0x1ef01 , 0x1f301 , 0x1f701 , 0x1fb01 , 0x1fe01 , 0x1ef05 , 0x1df09 , 0x1cf0d , 0x1bf11 , 0x1b015 , 0x1a019 , 0x1901d , 0x18021 , 0x17025 , 0x16029 , 0x1502d , 0x14031 , 0x13135 , 0x12139 , 0x1113d , 0x10140 , 0x10144 , 0x10148 , 0x1014c , 0x10150 , 0x10154 , 0x10158 , 0x1015c , 0x10160 , 0x10164 , 0x10168 , 0x1016c , 0x10170 , 0x10174 , 0x10178 , 0x1017c , 0x10180 , 0x10184 , 0x10188 , 0x1018c , 0x10190 , 0x10194 , 0x10198 , 0x1019c , 0x101a0 , 0x101a4 , 0x101a8 , 0x101ac , 0x101b0 , 0x101b4 , 0x101b8 , 0x101bc , 0x101bf , 0x101c3 , 0x101c7 , 0x101cb , 0x101cf , 0x101d3 , 0x101d7 , 0x101db , 0x101df , 0x101e3 , 0x101e7 , 0x101eb , 0x101ef , 0x101f3 , 0x101f7 , 0x101fb , 0x101fe , 0x105fa , 0x109f7 , 0x10df3 , 0x111ef , 0x115ec , 0x119e8 , 0x11de5 , 0x121e1 , 0x125dd , 0x129da , 0x12dd6 , 0x131d2 , 0x135d0 , 0x139cd , 0x13dc9 , 0x140c7 , 0x144c7 , 0x148c7 , 0x14cc7 , 0x150c7 , 0x154c7 , 0x158c7 , 0x15cc7 , 0x160c7 , 0x164c7 , 0x168c7 , 0x16cc7 , 0x170c7 , 0x174c7 , 0x178c7 , 0x17cc7 , 0x180c7 , 0x184c9 , 0x188cd , 0x18cd0 , 0x190d2 , 0x194d6 , 0x198da , 0x19cdd , 0x1a0e1 , 0x1a4e5 , 0x1a8e8 , 0x1acec , 0x1b0ef , 0x1b4f3 , 0x1b8f7 , 0x1bcfa , 0x1bffe , 0x11c3fe , 0x21c7fe , 0x31cbfe , 0x40cffe , 0x50d3fe , 0x60d7fe , 0x70dbfe , 0x80dffe , 0x90e3fe , 0xa0e7fe , 0xb0ebfe , 0xbfeffe , 0xcff3fe , 0xdff7fe , 0xeffbfe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe}

,

/* 41 RedHot */

{0x0 , 0x0 , 0x0 , 0x0 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x4 , 0x7 , 0xa , 0xd , 0x10 , 0x14 , 0x17 , 0x1b , 0x1e , 0x21 , 0x24 , 0x27 , 0x2a , 0x2d , 0x30 , 0x34 , 0x37 , 0x3a , 0x3d , 0x41 , 0x44 , 0x47 , 0x4a , 0x4e , 0x51 , 0x54 , 0x57 , 0x5a , 0x5d , 0x60 , 0x63 , 0x67 , 0x6a , 0x6e , 0x71 , 0x75 , 0x78 , 0x7b , 0x7e , 0x82 , 0x85 , 0x88 , 0x8b , 0x8e , 0x91 , 0x94 , 0x97 , 0x9b , 0x9e , 0xa1 , 0xa4 , 0xa8 , 0xab , 0xae , 0xb1 , 0xb5 , 0xb8 , 0xbb , 0xbe , 0xc1 , 0xc4 , 0xc8 , 0xcb , 0xcf , 0xd2 , 0xd5 , 0xd8 , 0xdc , 0xdf , 0xe2 , 0xe5 , 0xe9 , 0xec , 0xef , 0xf2 , 0xf5 , 0xf7 , 0x1fa , 0x2fc , 0x3ff , 0x6ff , 0x9ff , 0xcff , 0x10ff , 0x13ff , 0x16ff , 0x19ff , 0x1dff , 0x20ff , 0x23ff , 0x26ff , 0x2aff , 0x2dff , 0x30ff , 0x33ff , 0x37ff , 0x3aff , 0x3dff , 0x40ff , 0x44ff , 0x47ff , 0x4aff , 0x4dff , 0x51ff , 0x54ff , 0x57ff , 0x5aff , 0x5eff , 0x61ff , 0x64ff , 0x67ff , 0x6aff , 0x6dff , 0x70ff , 0x73ff , 0x77ff , 0x7aff , 0x7eff , 0x81ff , 0x85ff , 0x88ff , 0x8bff , 0x8eff , 0x91ff , 0x94ff , 0x97ff , 0x9aff , 0x9eff , 0xa1ff , 0xa4ff , 0xa7ff , 0xabff , 0xaeff , 0xb1ff , 0xb4ff , 0xb8ff , 0xbbff , 0xbeff , 0xc1ff , 0xc5ff , 0xc8ff , 0xcbff , 0xceff , 0xd1ff , 0xd4ff , 0xd8ff , 0xdbff , 0xdfff , 0xe2ff , 0xe5ff , 0xe8ff , 0xecff , 0xefff , 0xf2ff , 0xf5ff , 0xf9ff , 0x1faff , 0x3fcff , 0x4fdff , 0x6ffff , 0x9ffff , 0xcffff , 0xfffff , 0x13ffff , 0x16ffff , 0x19ffff , 0x1cffff , 0x20ffff , 0x23ffff , 0x26ffff , 0x29ffff , 0x2dffff , 0x30ffff , 0x33ffff , 0x36ffff , 0x3affff , 0x3dffff , 0x40ffff , 0x43ffff , 0x47ffff , 0x4affff , 0x4dffff , 0x50ffff , 0x54ffff , 0x57ffff , 0x5affff , 0x5dffff , 0x61ffff , 0x64ffff , 0x67ffff , 0x6affff , 0x6effff , 0x71ffff , 0x74ffff , 0x77ffff , 0x7affff , 0x7dffff , 0x80ffff , 0x83ffff , 0x87ffff , 0x8affff , 0x8effff , 0x91ffff , 0x95ffff , 0x98ffff , 0x9bffff , 0x9effff , 0xa1ffff , 0xa4ffff , 0xa7ffff , 0xaaffff , 0xaeffff , 0xb1ffff , 0xb4ffff , 0xb7ffff , 0xbbffff , 0xbeffff , 0xc1ffff , 0xc4ffff , 0xc8ffff , 0xcbffff , 0xceffff , 0xd1ffff , 0xd5ffff , 0xd8ffff , 0xdbffff , 0xdeffff , 0xe1ffff , 0xe4ffff , 0xe8ffff , 0xebffff , 0xefffff , 0xf2ffff , 0xf5ffff , 0xf8ffff , 0xfcffff , 0xfcffff , 0xfdffff , 0xfeffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff}

,

/* 42 HiLo */

{0xfe0202 , 0xfe0202 , 0xfe0303 , 0xfe0404 , 0xfe0505 , 0xfe0606 , 0xfe0707 , 0xfe0808 , 0xfe0909 , 0xfe0a0a , 0xfe0b0b , 0xc0c0c , 0xd0d0d , 0xe0e0e , 0xf0f0f , 0x101010 , 0x111111 , 0x121212 , 0x131313 , 0x141414 , 0x151515 , 0x161616 , 0x171717 , 0x181818 , 0x191919 , 0x1a1a1a , 0x1b1b1b , 0x1c1c1c , 0x1d1d1d , 0x1e1e1e , 0x1f1f1f , 0x202020 , 0x212121 , 0x222222 , 0x232323 , 0x242424 , 0x252525 , 0x262626 , 0x272727 , 0x282828 , 0x292929 , 0x2a2a2a , 0x2b2b2b , 0x2c2c2c , 0x2d2d2d , 0x2e2e2e , 0x2f2f2f , 0x303030 , 0x313131 , 0x323232 , 0x333333 , 0x343434 , 0x353535 , 0x363636 , 0x373737 , 0x383838 , 0x393939 , 0x3a3a3a , 0x3b3b3b , 0x3c3c3c , 0x3d3d3d , 0x3e3e3e , 0x3f3f3f , 0x404040 , 0x404040 , 0x414141 , 0x424242 , 0x434343 , 0x444444 , 0x454545 , 0x464646 , 0x474747 , 0x484848 , 0x494949 , 0x4a4a4a , 0x4b4b4b , 0x4c4c4c , 0x4d4d4d , 0x4e4e4e , 0x4f4f4f , 0x505050 , 0x515151 , 0x525252 , 0x535353 , 0x545454 , 0x555555 , 0x565656 , 0x575757 , 0x585858 , 0x595959 , 0x5a5a5a , 0x5b5b5b , 0x5c5c5c , 0x5d5d5d , 0x5e5e5e , 0x5f5f5f , 0x606060 , 0x616161 , 0x626262 , 0x636363 , 0x646464 , 0x656565 , 0x666666 , 0x676767 , 0x686868 , 0x696969 , 0x6a6a6a , 0x6b6b6b , 0x6c6c6c , 0x6d6d6d , 0x6e6e6e , 0x6f6f6f , 0x707070 , 0x717171 , 0x727272 , 0x737373 , 0x747474 , 0x757575 , 0x767676 , 0x777777 , 0x787878 , 0x797979 , 0x7a7a7a , 0x7b7b7b , 0x7c7c7c , 0x7d7d7d , 0x7e7e7e , 0x7f7f7f , 0x808080 , 0x818181 , 0x828282 , 0x838383 , 0x848484 , 0x858585 , 0x868686 , 0x878787 , 0x888888 , 0x898989 , 0x8a8a8a , 0x8b8b8b , 0x8c8c8c , 0x8d8d8d , 0x8e8e8e , 0x8f8f8f , 0x909090 , 0x919191 , 0x929292 , 0x939393 , 0x949494 , 0x959595 , 0x969696 , 0x979797 , 0x989898 , 0x999999 , 0x9a9a9a , 0x9b9b9b , 0x9c9c9c , 0x9d9d9d , 0x9e9e9e , 0x9f9f9f , 0xa0a0a0 , 0xa1a1a1 , 0xa2a2a2 , 0xa3a3a3 , 0xa4a4a4 , 0xa5a5a5 , 0xa6a6a6 , 0xa7a7a7 , 0xa8a8a8 , 0xa9a9a9 , 0xaaaaaa , 0xababab , 0xacacac , 0xadadad , 0xaeaeae , 0xafafaf , 0xb0b0b0 , 0xb1b1b1 , 0xb2b2b2 , 0xb3b3b3 , 0xb4b4b4 , 0xb5b5b5 , 0xb6b6b6 , 0xb7b7b7 , 0xb8b8b8 , 0xb9b9b9 , 0xbababa , 0xbbbbbb , 0xbcbcbc , 0xbdbdbd , 0xbebebe , 0xbfbfbf , 0xbfbfbf , 0xc0c0c0 , 0xc1c1c1 , 0xc2c2c2 , 0xc3c3c3 , 0xc4c4c4 , 0xc5c5c5 , 0xc6c6c6 , 0xc7c7c7 , 0xc8c8c8 , 0xc9c9c9 , 0xcacaca , 0xcbcbcb , 0xcccccc , 0xcdcdcd , 0xcecece , 0xcfcfcf , 0xd0d0d0 , 0xd1d1d1 , 0xd2d2d2 , 0xd3d3d3 , 0xd4d4d4 , 0xd5d5d5 , 0xd6d6d6 , 0xd7d7d7 , 0xd8d8d8 , 0xd9d9d9 , 0xdadada , 0xdbdbdb , 0xdcdcdc , 0xdddddd , 0xdedede , 0xdfdfdf , 0xe0e0e0 , 0xe1e1e1 , 0xe2e2e2 , 0xe3e3e3 , 0xe4e4e4 , 0xe5e5e5 , 0xe6e6e6 , 0xe7e7e7 , 0xe8e8e8 , 0xe9e9e9 , 0xeaeaea , 0xebebeb , 0xececec , 0xededed , 0xeeeeee , 0xefefef , 0xf0f0f0 , 0xf1f1f1 , 0xf2f2f2 , 0xf3f3f3 , 0xf4f4f4 , 0xf5f5f5 , 0xf6f6f6 , 0xf7f7f7 , 0xf8f8f8 , 0x101f9 , 0x101fa , 0x101fb , 0x101fc , 0x101fd , 0x101fd}

,

/* 43 neon-red */

{0x0 , 0x1 , 0x3 , 0x5 , 0x7 , 0x9 , 0xb , 0xd , 0xf , 0x11 , 0x13 , 0x15 , 0x17 , 0x19 , 0x1b , 0x1d , 0x1f , 0x21 , 0x23 , 0x25 , 0x27 , 0x29 , 0x2b , 0x2d , 0x2f , 0x31 , 0x33 , 0x35 , 0x37 , 0x39 , 0x3b , 0x3d , 0x3f , 0x41 , 0x43 , 0x45 , 0x47 , 0x49 , 0x4b , 0x4d , 0x4f , 0x51 , 0x53 , 0x55 , 0x57 , 0x59 , 0x5b , 0x5d , 0x5f , 0x61 , 0x63 , 0x65 , 0x67 , 0x69 , 0x6b , 0x6d , 0x6f , 0x71 , 0x73 , 0x75 , 0x77 , 0x79 , 0x7b , 0x7d , 0x7f , 0x81 , 0x83 , 0x85 , 0x87 , 0x89 , 0x8b , 0x8d , 0x8f , 0x91 , 0x93 , 0x95 , 0x97 , 0x99 , 0x9b , 0x9d , 0x9f , 0xa1 , 0xa3 , 0xa5 , 0xa7 , 0xa9 , 0xab , 0xad , 0xaf , 0xb1 , 0xb3 , 0xb5 , 0xb7 , 0xb9 , 0xbb , 0xbd , 0xbf , 0xc1 , 0xc3 , 0xc5 , 0xc7 , 0xc9 , 0xcb , 0xcd , 0xcf , 0xd1 , 0xd3 , 0xd5 , 0xd7 , 0xd9 , 0xdb , 0xdd , 0xdf , 0xe1 , 0xe3 , 0xe5 , 0xe7 , 0xe9 , 0xeb , 0xed , 0xef , 0xf1 , 0xf3 , 0xf5 , 0xf7 , 0xf9 , 0xfb , 0xff , 0xff , 0x101ff , 0x303ff , 0x505ff , 0x707ff , 0x909ff , 0xb0bff , 0xd0dff , 0xf0fff , 0x1111ff , 0x1313ff , 0x1515ff , 0x1717ff , 0x1919ff , 0x1b1bff , 0x1d1dff , 0x1f1fff , 0x2121ff , 0x2323ff , 0x2525ff , 0x2727ff , 0x2929ff , 0x2b2bff , 0x2d2dff , 0x2f2fff , 0x3131ff , 0x3333ff , 0x3535ff , 0x3737ff , 0x3939ff , 0x3b3bff , 0x3d3dff , 0x3f3fff , 0x4141ff , 0x4343ff , 0x4545ff , 0x4747ff , 0x4949ff , 0x4b4bff , 0x4d4dff , 0x4f4fff , 0x5151ff , 0x5353ff , 0x5555ff , 0x5757ff , 0x5959ff , 0x5b5bff , 0x5d5dff , 0x5f5fff , 0x6161ff , 0x6363ff , 0x6565ff , 0x6767ff , 0x6969ff , 0x6b6bff , 0x6d6dff , 0x6f6fff , 0x7171ff , 0x7373ff , 0x7575ff , 0x7777ff , 0x7979ff , 0x7b7bff , 0x7d7dff , 0x7f7fff , 0x8181ff , 0x8383ff , 0x8585ff , 0x8787ff , 0x8989ff , 0x8b8bff , 0x8d8dff , 0x8f8fff , 0x9191ff , 0x9393ff , 0x9595ff , 0x9797ff , 0x9999ff , 0x9b9bff , 0x9d9dff , 0x9f9fff , 0xa1a1ff , 0xa3a3ff , 0xa5a5ff , 0xa7a7ff , 0xa9a9ff , 0xababff , 0xadadff , 0xafafff , 0xb1b1ff , 0xb3b3ff , 0xb5b5ff , 0xb7b7ff , 0xb9b9ff , 0xbbbbff , 0xbdbdff , 0xbfbfff , 0xc1c1ff , 0xc3c3ff , 0xc5c5ff , 0xc7c7ff , 0xc9c9ff , 0xcbcbff , 0xcdcdff , 0xcfcfff , 0xd1d1ff , 0xd3d3ff , 0xd5d5ff , 0xd7d7ff , 0xd9d9ff , 0xdbdbff , 0xddddff , 0xdfdfff , 0xe1e1ff , 0xe3e3ff , 0xe5e5ff , 0xe7e7ff , 0xe9e9ff , 0xebebff , 0xededff , 0xefefff , 0xf1f1ff , 0xf3f3ff , 0xf5f5ff , 0xf7f7ff , 0xf9f9ff , 0xfbfbff , 0xffffff}

,

/* 44 red-blue */

{0xff , 0x100fe , 0x200fd , 0x300fc , 0x400fb , 0x500fa , 0x600f9 , 0x700f8 , 0x800f7 , 0x900f6 , 0xa00f5 , 0xb00f4 , 0xc00f3 , 0xd00f2 , 0xe00f1 , 0xf00f0 , 0x1000ef , 0x1100ee , 0x1200ed , 0x1300ec , 0x1400eb , 0x1500ea , 0x1600e9 , 0x1700e8 , 0x1800e7 , 0x1900e6 , 0x1a00e5 , 0x1b00e4 , 0x1c00e3 , 0x1d00e2 , 0x1e00e1 , 0x1f00e0 , 0x2000df , 0x2100de , 0x2200dd , 0x2300dc , 0x2400db , 0x2500da , 0x2600d9 , 0x2700d8 , 0x2800d7 , 0x2900d6 , 0x2a00d5 , 0x2b00d4 , 0x2c00d3 , 0x2d00d2 , 0x2e00d1 , 0x2f00d0 , 0x3000cf , 0x3100ce , 0x3200cd , 0x3300cc , 0x3400cb , 0x3500ca , 0x3600c9 , 0x3700c8 , 0x3800c7 , 0x3900c6 , 0x3a00c5 , 0x3b00c4 , 0x3c00c3 , 0x3d00c2 , 0x3e00c1 , 0x3f00c0 , 0x4000bf , 0x4100be , 0x4200bd , 0x4300bc , 0x4400bb , 0x4500ba , 0x4600b9 , 0x4700b8 , 0x4800b7 , 0x4900b6 , 0x4a00b5 , 0x4b00b4 , 0x4c00b3 , 0x4d00b2 , 0x4e00b1 , 0x4f00b0 , 0x5000af , 0x5100ae , 0x5200ad , 0x5300ac , 0x5400ab , 0x5500aa , 0x5600a9 , 0x5700a8 , 0x5800a7 , 0x5900a6 , 0x5a00a5 , 0x5b00a4 , 0x5c00a3 , 0x5d00a2 , 0x5e00a1 , 0x5f00a0 , 0x60009f , 0x61009e , 0x62009d , 0x63009c , 0x64009b , 0x65009a , 0x660099 , 0x670098 , 0x680097 , 0x690096 , 0x6a0095 , 0x6b0094 , 0x6c0093 , 0x6d0092 , 0x6e0091 , 0x6f0090 , 0x70008f , 0x71008e , 0x72008d , 0x73008c , 0x74008b , 0x75008a , 0x760089 , 0x770088 , 0x780087 , 0x790086 , 0x7a0085 , 0x7b0084 , 0x7c0083 , 0x7d0082 , 0x7e0081 , 0x7f0080 , 0x80007f , 0x81007e , 0x82007d , 0x83007c , 0x84007b , 0x85007a , 0x860079 , 0x870078 , 0x880077 , 0x890076 , 0x8a0075 , 0x8b0074 , 0x8c0073 , 0x8d0072 , 0x8e0071 , 0x8f0070 , 0x90006f , 0x91006e , 0x92006d , 0x93006c , 0x94006b , 0x95006a , 0x960069 , 0x970068 , 0x980067 , 0x990066 , 0x9a0065 , 0x9b0064 , 0x9c0063 , 0x9d0062 , 0x9e0061 , 0x9f0060 , 0xa0005f , 0xa1005e , 0xa2005d , 0xa3005c , 0xa4005b , 0xa5005a , 0xa60059 , 0xa70058 , 0xa80057 , 0xa90056 , 0xaa0055 , 0xab0054 , 0xac0053 , 0xad0052 , 0xae0051 , 0xaf0050 , 0xb0004f , 0xb1004e , 0xb2004d , 0xb3004c , 0xb4004b , 0xb5004a , 0xb60049 , 0xb70048 , 0xb80047 , 0xb90046 , 0xba0045 , 0xbb0044 , 0xbc0043 , 0xbd0042 , 0xbe0041 , 0xbf0040 , 0xc0003f , 0xc1003e , 0xc2003d , 0xc3003c , 0xc4003b , 0xc5003a , 0xc60039 , 0xc70038 , 0xc80037 , 0xc90036 , 0xca0035 , 0xcb0034 , 0xcc0033 , 0xcd0032 , 0xce0031 , 0xcf0030 , 0xd0002f , 0xd1002e , 0xd2002d , 0xd3002c , 0xd4002b , 0xd5002a , 0xd60029 , 0xd70028 , 0xd80027 , 0xd90026 , 0xda0025 , 0xdb0024 , 0xdc0023 , 0xdd0022 , 0xde0021 , 0xdf0020 , 0xe0001f , 0xe1001e , 0xe2001d , 0xe3001c , 0xe4001b , 0xe5001a , 0xe60019 , 0xe70018 , 0xe80017 , 0xe90016 , 0xea0015 , 0xeb0014 , 0xec0013 , 0xed0012 , 0xee0011 , 0xef0010 , 0xf0000f , 0xf1000e , 0xf2000d , 0xf3000c , 0xf4000b , 0xf5000a , 0xf60009 , 0xf70008 , 0xf80007 , 0xf90006 , 0xfa0005 , 0xfb0004 , 0xfc0003 , 0xfd0002 , 0xfe0001 , 0xff0000}

,

/* 45 5_ramps */

{0x10000 , 0x50000 , 0xa0000 , 0xf0000 , 0x140000 , 0x190000 , 0x1e0000 , 0x230000 , 0x280000 , 0x2d0000 , 0x320000 , 0x370000 , 0x3c0000 , 0x410000 , 0x460000 , 0x4b0000 , 0x500000 , 0x550000 , 0x5a0000 , 0x5f0000 , 0x640000 , 0x690000 , 0x6e0000 , 0x730000 , 0x780000 , 0x7d0000 , 0x820000 , 0x870000 , 0x8c0000 , 0x910000 , 0x960000 , 0x9b0000 , 0xa00000 , 0xa50000 , 0xaa0000 , 0xaf0000 , 0xb40000 , 0xb90000 , 0xbe0000 , 0xc30000 , 0xc80000 , 0xcd0000 , 0xd20000 , 0xd70000 , 0xdc0000 , 0xe10000 , 0xe60000 , 0xeb0000 , 0xf00000 , 0xf50000 , 0xfa0000 , 0x100 , 0x500 , 0xa00 , 0xf00 , 0x1400 , 0x1900 , 0x1e00 , 0x2300 , 0x2800 , 0x2d00 , 0x3200 , 0x3700 , 0x3c00 , 0x4100 , 0x4600 , 0x4b00 , 0x5000 , 0x5500 , 0x5a00 , 0x5f00 , 0x6400 , 0x6900 , 0x6e00 , 0x7300 , 0x7800 , 0x7d00 , 0x8200 , 0x8700 , 0x8c00 , 0x9100 , 0x9600 , 0x9b00 , 0xa000 , 0xa500 , 0xaa00 , 0xaf00 , 0xb400 , 0xb900 , 0xbe00 , 0xc300 , 0xc800 , 0xcd00 , 0xd200 , 0xd700 , 0xdc00 , 0xe100 , 0xe600 , 0xeb00 , 0xf000 , 0xf500 , 0xfa00 , 0x101 , 0x505 , 0xa0a , 0xf0f , 0x1414 , 0x1919 , 0x1e1e , 0x2323 , 0x2828 , 0x2d2d , 0x3232 , 0x3737 , 0x3c3c , 0x4141 , 0x4646 , 0x4b4b , 0x5050 , 0x5555 , 0x5a5a , 0x5f5f , 0x6464 , 0x6969 , 0x6e6e , 0x7373 , 0x7878 , 0x7d7d , 0x8282 , 0x8787 , 0x8c8c , 0x9191 , 0x9696 , 0x9b9b , 0xa0a0 , 0xa5a5 , 0xaaaa , 0xafaf , 0xb4b4 , 0xb9b9 , 0xbebe , 0xc3c3 , 0xc8c8 , 0xcdcd , 0xd2d2 , 0xd7d7 , 0xdcdc , 0xe1e1 , 0xe6e6 , 0xebeb , 0xf0f0 , 0xf5f5 , 0xfafa , 0x1 , 0x5 , 0xa , 0xf , 0x14 , 0x19 , 0x1e , 0x23 , 0x28 , 0x2d , 0x32 , 0x37 , 0x3c , 0x41 , 0x46 , 0x4b , 0x50 , 0x55 , 0x5a , 0x5f , 0x64 , 0x69 , 0x6e , 0x73 , 0x78 , 0x7d , 0x82 , 0x87 , 0x8c , 0x91 , 0x96 , 0x9b , 0xa0 , 0xa5 , 0xaa , 0xaf , 0xb4 , 0xb9 , 0xbe , 0xc3 , 0xc8 , 0xcd , 0xd2 , 0xd7 , 0xdc , 0xe1 , 0xe6 , 0xeb , 0xf0 , 0xf5 , 0xfa , 0x10101 , 0x50505 , 0xa0a0a , 0xf0f0f , 0x141414 , 0x191919 , 0x1e1e1e , 0x232323 , 0x282828 , 0x2d2d2d , 0x323232 , 0x373737 , 0x3c3c3c , 0x414141 , 0x464646 , 0x4b4b4b , 0x505050 , 0x555555 , 0x5a5a5a , 0x5f5f5f , 0x646464 , 0x696969 , 0x6e6e6e , 0x737373 , 0x787878 , 0x7d7d7d , 0x828282 , 0x878787 , 0x8c8c8c , 0x919191 , 0x969696 , 0x9b9b9b , 0xa0a0a0 , 0xa5a5a5 , 0xaaaaaa , 0xafafaf , 0xb4b4b4 , 0xb9b9b9 , 0xbebebe , 0xc3c3c3 , 0xc8c8c8 , 0xcdcdcd , 0xd2d2d2 , 0xd7d7d7 , 0xdcdcdc , 0xe1e1e1 , 0xe6e6e6 , 0xebebeb , 0xf0f0f0 , 0xf5f5f5 , 0xfafafa , 0xffffff}

,

/* 46 002-spectrum */

{0xff , 0x6ff , 0xcff , 0x12ff , 0x18ff , 0x1eff , 0x24ff , 0x2aff , 0x30ff , 0x36ff , 0x3cff , 0x42ff , 0x48ff , 0x4eff , 0x54ff , 0x5aff , 0x60ff , 0x66ff , 0x6cff , 0x72ff , 0x78ff , 0x7eff , 0x84ff , 0x8aff , 0x90ff , 0x96ff , 0x9cff , 0xa2ff , 0xa8ff , 0xaeff , 0xb4ff , 0xbaff , 0xc0ff , 0xc6ff , 0xccff , 0xd2ff , 0xd8ff , 0xdeff , 0xe4ff , 0xeaff , 0xf0ff , 0xf6ff , 0xfcff , 0xfffc , 0xfff6 , 0xfff0 , 0xffea , 0xffe4 , 0xffde , 0xffd8 , 0xffd2 , 0xffcc , 0xffc6 , 0xffc0 , 0xffba , 0xffb4 , 0xffae , 0xffa8 , 0xffa2 , 0xff9c , 0xff96 , 0xff90 , 0xff8a , 0xff84 , 0xff7e , 0xff78 , 0xff72 , 0xff6c , 0xff66 , 0xff60 , 0xff5a , 0xff54 , 0xff4e , 0xff48 , 0xff42 , 0xff3c , 0xff36 , 0xff30 , 0xff2a , 0xff24 , 0xff1e , 0xff18 , 0xff12 , 0xff0c , 0xff06 , 0xff00 , 0x6ff00 , 0xcff00 , 0x12ff00 , 0x18ff00 , 0x1eff00 , 0x24ff00 , 0x2aff00 , 0x30ff00 , 0x36ff00 , 0x3cff00 , 0x42ff00 , 0x48ff00 , 0x4eff00 , 0x54ff00 , 0x5aff00 , 0x60ff00 , 0x66ff00 , 0x6cff00 , 0x72ff00 , 0x78ff00 , 0x7eff00 , 0x84ff00 , 0x8aff00 , 0x90ff00 , 0x96ff00 , 0x9cff00 , 0xa2ff00 , 0xa8ff00 , 0xaeff00 , 0xb4ff00 , 0xbaff00 , 0xc0ff00 , 0xc6ff00 , 0xccff00 , 0xd2ff00 , 0xd8ff00 , 0xdeff00 , 0xe4ff00 , 0xeaff00 , 0xf0ff00 , 0xf6ff00 , 0xfcff00 , 0xfffc00 , 0xfff600 , 0xfff000 , 0xffea00 , 0xffe400 , 0xffde00 , 0xffd800 , 0xffd200 , 0xffcc00 , 0xffc600 , 0xffc000 , 0xffba00 , 0xffb400 , 0xffae00 , 0xffa800 , 0xffa200 , 0xff9c00 , 0xff9600 , 0xff9000 , 0xff8a00 , 0xff8400 , 0xff7e00 , 0xff7800 , 0xff7200 , 0xff6c00 , 0xff6600 , 0xff6000 , 0xff5a00 , 0xff5400 , 0xff4e00 , 0xff4800 , 0xff4200 , 0xff3c00 , 0xff3600 , 0xff3000 , 0xff2a00 , 0xff2400 , 0xff1e00 , 0xff1800 , 0xff1200 , 0xff0c00 , 0xff0600 , 0xff0000 , 0xff0006 , 0xff000c , 0xff0012 , 0xff0018 , 0xff001e , 0xff0024 , 0xff002a , 0xff0030 , 0xff0036 , 0xff003c , 0xff0042 , 0xff0048 , 0xff004e , 0xff0054 , 0xff005a , 0xff0060 , 0xff0066 , 0xff006c , 0xff0072 , 0xff0078 , 0xff007e , 0xff0084 , 0xff008a , 0xff0090 , 0xff0096 , 0xff009c , 0xff00a2 , 0xff00a8 , 0xff00ae , 0xff00b4 , 0xff00ba , 0xff00c0 , 0xff00c6 , 0xff00cc , 0xff00d2 , 0xff00d8 , 0xff00de , 0xff00e4 , 0xff00ea , 0xff00f0 , 0xff00f6 , 0xff00fc , 0xfc00ff , 0xf600ff , 0xf000ff , 0xea00ff , 0xe400ff , 0xde00ff , 0xd800ff , 0xd200ff , 0xcc00ff , 0xc600ff , 0xc000ff , 0xba00ff , 0xb400ff , 0xae00ff , 0xa800ff , 0xa200ff , 0x9c00ff , 0x9600ff , 0x9000ff , 0x8a00ff , 0x8400ff , 0x7e00ff , 0x7800ff , 0x7200ff , 0x6c00ff , 0x6600ff , 0x6000ff , 0x5a00ff , 0x5400ff , 0x4e00ff , 0x4800ff , 0x4200ff , 0x3c00ff , 0x3600ff , 0x3000ff , 0x2a00ff , 0x2400ff , 0x1e00ff , 0x1800ff , 0x1200ff , 0xc00ff , 0x600ff , 0xff}

,

/* 47 Gold */

{0x9bf3e7 , 0x9bf3e7 , 0x9afef2 , 0x99fef2 , 0x98fef3 , 0x96fef3 , 0x95fef3 , 0x94fef3 , 0x93fef4 , 0x92fef4 , 0x91fef4 , 0x90fef5 , 0x8ffef5 , 0x8efef5 , 0x8dfef6 , 0x8cfef6 , 0x8bfef6 , 0x8afef7 , 0x89fef7 , 0x88fef8 , 0x87fef8 , 0x85fef9 , 0x84fef9 , 0x83fefa , 0x82fefa , 0x81fefb , 0x80fefb , 0x7ffefc , 0x7efefc , 0x7dfefd , 0x7cfefd , 0x7bfefe , 0x7afdfe , 0x79fdfe , 0x78fcfe , 0x77fbfe , 0x76fbfe , 0x74fafe , 0x73f9fe , 0x72f8fe , 0x71f8fe , 0x70f7fe , 0x6ff6fe , 0x6ef5fe , 0x6df5fe , 0x6cf4fe , 0x6bf3fe , 0x6af2fe , 0x69f1fe , 0x68f0fe , 0x67effe , 0x66eefe , 0x65edfe , 0x63edfd , 0x62ecfd , 0x61ebfd , 0x60eafd , 0x5fe9fd , 0x5ee8fd , 0x5de7fd , 0x5ce6fd , 0x5be4fd , 0x5ae3fd , 0x59e2fd , 0x58e1fd , 0x57e0fd , 0x56dffd , 0x55defd , 0x54ddfd , 0x52dbfd , 0x51dafd , 0x50d9fd , 0x4fd8fd , 0x4ed7fd , 0x4dd5fd , 0x4cd4fd , 0x4bd3fd , 0x4ad2fd , 0x49d0fd , 0x48cffd , 0x47cefd , 0x46ccfd , 0x45cbfd , 0x44cafd , 0x43c8fd , 0x41c7fd , 0x40c5fd , 0x3fc4fd , 0x3ec3fd , 0x3dc1fd , 0x3cc0fd , 0x3bbefd , 0x3abdfd , 0x39bbfd , 0x38bafd , 0x37b8fd , 0x36b7fd , 0x35b5fd , 0x34b3fd , 0x33b1fb , 0x32affa , 0x30adf8 , 0x2fabf6 , 0x2ea8f5 , 0x2da6f3 , 0x2ca4f2 , 0x2ba2f0 , 0x2aa0ee , 0x299eed , 0x289ceb , 0x2799e9 , 0x2697e8 , 0x2595e6 , 0x2493e5 , 0x2391e3 , 0x228fe1 , 0x218de0 , 0x1f8bde , 0x1e89dd , 0x1d87db , 0x1c85d9 , 0x1b83d8 , 0x1a81d6 , 0x1980d4 , 0x187ed3 , 0x177cd1 , 0x167ad0 , 0x1578ce , 0x1476cc , 0x1374cb , 0x1273c9 , 0x1171c7 , 0x106fc6 , 0xe6dc4 , 0xd6bc3 , 0xc6ac1 , 0xb68bf , 0xa66be , 0x965bc , 0x863ba , 0x761b9 , 0x660b7 , 0x55eb6 , 0x45cb4 , 0x35bb2 , 0x259b1 , 0x157af , 0x56ad , 0x154ac , 0x153aa , 0x151a9 , 0x150a7 , 0x14ea5 , 0x14da4 , 0x14ba2 , 0x14aa1 , 0x1489f , 0x1479d , 0x1459c , 0x1449a , 0x14398 , 0x14197 , 0x14095 , 0x13f94 , 0x13d92 , 0x13c90 , 0x13b8f , 0x1398d , 0x1388b , 0x1378a , 0x13688 , 0x13487 , 0x13385 , 0x13283 , 0x13182 , 0x13080 , 0x12f7e , 0x12d7d , 0x12c7b , 0x12b7a , 0x12a78 , 0x12976 , 0x12875 , 0x12773 , 0x12671 , 0x12570 , 0x1246e , 0x1236d , 0x226b , 0x2169 , 0x2068 , 0x1f66 , 0x1e64 , 0x1d63 , 0x1c61 , 0x1b60 , 0x1a5e , 0x195c , 0x185b , 0x1859 , 0x1757 , 0x1656 , 0x1554 , 0x1453 , 0x1451 , 0x134f , 0x124e , 0x114c , 0x114a , 0x1049 , 0xf47 , 0xf46 , 0xe44 , 0xd42 , 0xd41 , 0xc3f , 0xc3d , 0xb3c , 0xa3a , 0xa39 , 0x937 , 0x935 , 0x834 , 0x832 , 0x730 , 0x72f , 0x62d , 0x62c , 0x62a , 0x528 , 0x527 , 0x425 , 0x423 , 0x422 , 0x320 , 0x31f , 0x31d , 0x21b , 0x21a , 0x218 , 0x216 , 0x115 , 0x113 , 0x112 , 0x110 , 0x10e , 0xd , 0xb , 0x9 , 0x8 , 0x6 , 0x5 , 0x3 , 0x1 , 0x0 , 0x0}

,

/* 48 neon-blue */

{0x0 , 0x10000 , 0x30000 , 0x50000 , 0x70000 , 0x90000 , 0xb0000 , 0xd0000 , 0xf0000 , 0x110000 , 0x130000 , 0x150000 , 0x170000 , 0x190000 , 0x1b0000 , 0x1d0000 , 0x1f0000 , 0x210000 , 0x230000 , 0x250000 , 0x270000 , 0x290000 , 0x2b0000 , 0x2d0000 , 0x2f0000 , 0x310000 , 0x330000 , 0x350000 , 0x370000 , 0x390000 , 0x3b0000 , 0x3d0000 , 0x3f0000 , 0x410000 , 0x430000 , 0x450000 , 0x470000 , 0x490000 , 0x4b0000 , 0x4d0000 , 0x4f0000 , 0x510000 , 0x530000 , 0x550000 , 0x570000 , 0x590000 , 0x5b0000 , 0x5d0000 , 0x5f0000 , 0x610000 , 0x630000 , 0x650000 , 0x670000 , 0x690000 , 0x6b0000 , 0x6d0000 , 0x6f0000 , 0x710000 , 0x730000 , 0x750000 , 0x770000 , 0x790000 , 0x7b0000 , 0x7d0000 , 0x7f0000 , 0x810000 , 0x830000 , 0x850000 , 0x870000 , 0x890000 , 0x8b0000 , 0x8d0000 , 0x8f0000 , 0x910000 , 0x930000 , 0x950000 , 0x970000 , 0x990000 , 0x9b0000 , 0x9d0000 , 0x9f0000 , 0xa10000 , 0xa30000 , 0xa50000 , 0xa70000 , 0xa90000 , 0xab0000 , 0xad0000 , 0xaf0000 , 0xb10000 , 0xb30000 , 0xb50000 , 0xb70000 , 0xb90000 , 0xbb0000 , 0xbd0000 , 0xbf0000 , 0xc10000 , 0xc30000 , 0xc50000 , 0xc70000 , 0xc90000 , 0xcb0000 , 0xcd0000 , 0xcf0000 , 0xd10000 , 0xd30000 , 0xd50000 , 0xd70000 , 0xd90000 , 0xdb0000 , 0xdd0000 , 0xdf0000 , 0xe10000 , 0xe30000 , 0xe50000 , 0xe70000 , 0xe90000 , 0xeb0000 , 0xed0000 , 0xef0000 , 0xf10000 , 0xf30000 , 0xf50000 , 0xf70000 , 0xf90000 , 0xfb0000 , 0xff0000 , 0xff0000 , 0xff0101 , 0xff0303 , 0xff0505 , 0xff0707 , 0xff0909 , 0xff0b0b , 0xff0d0d , 0xff0f0f , 0xff1111 , 0xff1313 , 0xff1515 , 0xff1717 , 0xff1919 , 0xff1b1b , 0xff1d1d , 0xff1f1f , 0xff2121 , 0xff2323 , 0xff2525 , 0xff2727 , 0xff2929 , 0xff2b2b , 0xff2d2d , 0xff2f2f , 0xff3131 , 0xff3333 , 0xff3535 , 0xff3737 , 0xff3939 , 0xff3b3b , 0xff3d3d , 0xff3f3f , 0xff4141 , 0xff4343 , 0xff4545 , 0xff4747 , 0xff4949 , 0xff4b4b , 0xff4d4d , 0xff4f4f , 0xff5151 , 0xff5353 , 0xff5555 , 0xff5757 , 0xff5959 , 0xff5b5b , 0xff5d5d , 0xff5f5f , 0xff6161 , 0xff6363 , 0xff6565 , 0xff6767 , 0xff6969 , 0xff6b6b , 0xff6d6d , 0xff6f6f , 0xff7171 , 0xff7373 , 0xff7575 , 0xff7777 , 0xff7979 , 0xff7b7b , 0xff7d7d , 0xff7f7f , 0xff8181 , 0xff8383 , 0xff8585 , 0xff8787 , 0xff8989 , 0xff8b8b , 0xff8d8d , 0xff8f8f , 0xff9191 , 0xff9393 , 0xff9595 , 0xff9797 , 0xff9999 , 0xff9b9b , 0xff9d9d , 0xff9f9f , 0xffa1a1 , 0xffa3a3 , 0xffa5a5 , 0xffa7a7 , 0xffa9a9 , 0xffabab , 0xffadad , 0xffafaf , 0xffb1b1 , 0xffb3b3 , 0xffb5b5 , 0xffb7b7 , 0xffb9b9 , 0xffbbbb , 0xffbdbd , 0xffbfbf , 0xffc1c1 , 0xffc3c3 , 0xffc5c5 , 0xffc7c7 , 0xffc9c9 , 0xffcbcb , 0xffcdcd , 0xffcfcf , 0xffd1d1 , 0xffd3d3 , 0xffd5d5 , 0xffd7d7 , 0xffd9d9 , 0xffdbdb , 0xffdddd , 0xffdfdf , 0xffe1e1 , 0xffe3e3 , 0xffe5e5 , 0xffe7e7 , 0xffe9e9 , 0xffebeb , 0xffeded , 0xffefef , 0xfff1f1 , 0xfff3f3 , 0xfff5f5 , 0xfff7f7 , 0xfff9f9 , 0xfffbfb , 0xffffff}

,

/* 49 Siemens */

{0x200102 , 0x200102 , 0x300102 , 0x400103 , 0x4f0103 , 0x5f0104 , 0x5f0104 , 0x5f0104 , 0x5f0104 , 0x5f0104 , 0x5f0104 , 0x5f0104 , 0x5f0104 , 0x5f0104 , 0x5f0104 , 0x5f0104 , 0x630104 , 0x680104 , 0x6c0104 , 0x710104 , 0x760104 , 0x7a0104 , 0x7f0104 , 0x830105 , 0x880105 , 0x8d0105 , 0x910105 , 0x960105 , 0x9a0105 , 0x9f0105 , 0xa40105 , 0xa80106 , 0xad0106 , 0xb20106 , 0xb60106 , 0xbb0106 , 0xbf0106 , 0xc30106 , 0xc80106 , 0xcc0107 , 0xd10107 , 0xd50107 , 0xda0107 , 0xdf0107 , 0xe30107 , 0xe80107 , 0xed0108 , 0xec0108 , 0xe8010d , 0xe30112 , 0xde0117 , 0xd9011d , 0xd40122 , 0xcf0127 , 0xca012c , 0xc50132 , 0xc00137 , 0xbc013c , 0xb70140 , 0xb20146 , 0xad014b , 0xa80150 , 0xa30155 , 0x9e015b , 0x990160 , 0x940165 , 0x8f016a , 0x8a0170 , 0x850175 , 0x80017a , 0x7b017f , 0x770185 , 0x72018a , 0x6d018f , 0x680194 , 0x63019a , 0x5e019f , 0x5901a4 , 0x5401a9 , 0x4f01af , 0x4a01b4 , 0x4501b9 , 0x4001be , 0x3c01c3 , 0x3701c8 , 0x3201cd , 0x2d01d2 , 0x2801d8 , 0x2301dd , 0x1e01e2 , 0x1901e7 , 0x1401ed , 0xf01f2 , 0xa01f7 , 0x601fd , 0xf01f3 , 0xe02f3 , 0xe04f3 , 0xe06f3 , 0xe08f3 , 0xe09f3 , 0xe0bf3 , 0xe0df3 , 0xe0ff3 , 0xe10f3 , 0xe12f3 , 0xd14f3 , 0xd16f3 , 0xd18f4 , 0xd19f4 , 0xd1bf4 , 0xd1df4 , 0xd1ff4 , 0xd20f4 , 0xd22f4 , 0xd24f4 , 0xc26f4 , 0xc28f4 , 0xc29f4 , 0xc2bf4 , 0xc2df5 , 0xc2ff5 , 0xc30f5 , 0xc32f5 , 0xc34f5 , 0xc36f5 , 0xb38f5 , 0xb39f5 , 0xb3bf5 , 0xb3df5 , 0xb3ff5 , 0xb40f5 , 0xb41f6 , 0xb43f6 , 0xb45f6 , 0xb46f6 , 0xa48f6 , 0xa4af6 , 0xa4cf6 , 0xa4ef6 , 0xa4ff6 , 0xa51f6 , 0xa53f6 , 0xa55f6 , 0xa56f6 , 0xa58f7 , 0x95af7 , 0x95cf7 , 0x95ef7 , 0x95ff7 , 0x961f7 , 0x963f7 , 0x965f7 , 0x966f7 , 0x968f7 , 0x96af7 , 0x86cf7 , 0x86ef8 , 0x86ff8 , 0x871f8 , 0x873f8 , 0x875f8 , 0x876f8 , 0x878f8 , 0x87af8 , 0x87cf8 , 0x77ef8 , 0x77ff8 , 0x781f8 , 0x783f9 , 0x785f9 , 0x786f9 , 0x788f9 , 0x78af9 , 0x78cf9 , 0x78df9 , 0x68ff9 , 0x691f9 , 0x693f9 , 0x695f9 , 0x696f9 , 0x698f9 , 0x69afa , 0x69cfa , 0x69dfa , 0x69ffa , 0x5a1fa , 0x5a3fa , 0x5a5fa , 0x5a6fa , 0x5a8fa , 0x5aafa , 0x5acfa , 0x5adfa , 0x5affb , 0x5b1fb , 0x4b3fb , 0x4b5fb , 0x4b6fb , 0x4b8fb , 0x4bafb , 0x4bcfb , 0x4bdfb , 0x4bffb , 0x4c0fb , 0x4c2fb , 0x4c4fc , 0x4c4fc , 0x3c8fb , 0x3ccfb , 0x3d0fb , 0x3d4fa , 0x3d7fa , 0x2dbfa , 0x2dffa , 0x2e3f9 , 0x2e7f9 , 0x2eaf9 , 0x1eef9 , 0x1f2f8 , 0x1f6f8 , 0x1faf8 , 0x1fef8 , 0x1fef8 , 0x9fef8 , 0x11fef8 , 0x19fef8 , 0x21fef8 , 0x2afef8 , 0x32fef9 , 0x3afef9 , 0x41fef9 , 0x4afef9 , 0x52fef9 , 0x5afefa , 0x62fefa , 0x6afefa , 0x73fefa , 0x7bfefa , 0x83fefb , 0x8bfefb , 0x94fefb , 0x9cfefb , 0xa4fefb , 0xacfefc , 0xb4fefc , 0xbdfefc , 0xc4fefc , 0xccfefc , 0xd4fefd , 0xddfefd , 0xe5fefd , 0xedfefd , 0xf5fefd , 0x10101 , 0x10101}

,

/* 50 smart */

{0x20202 , 0x20202 , 0x20202 , 0x30303 , 0x40404 , 0x50505 , 0x50505 , 0x60606 , 0x70707 , 0x80808 , 0x80808 , 0x90909 , 0xa0a0a , 0xb0b0b , 0xb0b0b , 0xc0c0c , 0xd0d0d , 0xe0e0e , 0xe0e0e , 0xf0f0f , 0x101010 , 0x101010 , 0x111111 , 0x121212 , 0x131313 , 0x131313 , 0x141414 , 0x151515 , 0x161616 , 0x161616 , 0x171717 , 0x181818 , 0x191919 , 0x191919 , 0x1a1a1a , 0x1b1b1b , 0x1b1b1b , 0x1c1c1c , 0x1d1d1d , 0x1e1e1e , 0x1e1e1e , 0x1f1f1f , 0x202020 , 0x212121 , 0x212121 , 0x222222 , 0x232323 , 0x242424 , 0x242424 , 0x252525 , 0x262626 , 0x272727 , 0x272727 , 0x282828 , 0x292929 , 0x292929 , 0x2a2a2a , 0x2b2b2b , 0x2c2c2c , 0x2c2c2c , 0x2d2d2d , 0x2e2e2e , 0x2f2f2f , 0x2f2f2f , 0x303030 , 0x313131 , 0x323232 , 0x323232 , 0x333333 , 0x343434 , 0x343434 , 0x353535 , 0x363636 , 0x373737 , 0x373737 , 0x383838 , 0x393939 , 0x3a3a3a , 0x3a3a3a , 0x3b3b3b , 0x3c3c3c , 0x3d3d3d , 0x3d3d3d , 0x3e3e3e , 0x3f3f3f , 0x404040 , 0x404040 , 0x404040 , 0x414141 , 0x414141 , 0x424242 , 0x434343 , 0x444444 , 0x444444 , 0x454545 , 0x464646 , 0x474747 , 0x474747 , 0x484848 , 0x494949 , 0x4a4a4a , 0x4a4a4a , 0x4b4b4b , 0x47474c , 0x42444d , 0x3e404d , 0x3a3e4e , 0x353a4f , 0x303750 , 0x2b3350 , 0x282f51 , 0x262d52 , 0x242b53 , 0x222a53 , 0x202854 , 0x1e2655 , 0x1c2456 , 0x1a2256 , 0x182057 , 0x161e58 , 0x151d59 , 0x141b59 , 0x131a5a , 0x131a5b , 0x12195c , 0x12185c , 0x11185d , 0x11175e , 0x11175f , 0x10165f , 0x101660 , 0xf1661 , 0xe1562 , 0xd1562 , 0xd1463 , 0xd1464 , 0xc1465 , 0xc1365 , 0xb1366 , 0xb1267 , 0xb1266 , 0xa1269 , 0x9116a , 0x8116b , 0x7116d , 0x6106f , 0x61070 , 0x51071 , 0x41072 , 0x40f74 , 0x40f75 , 0x30f77 , 0x3107a , 0x2107d , 0x21080 , 0x21082 , 0x11085 , 0x11088 , 0x1108b , 0x1108e , 0x11191 , 0x11295 , 0x11498 , 0x1159c , 0x1179f , 0x118a3 , 0x11aa6 , 0x11bbd , 0x11dad , 0x11fb1 , 0x11fb3 , 0x11fb6 , 0x120b9 , 0x121bd , 0x122bf , 0x124c2 , 0x125c4 , 0x127c7 , 0x129c9 , 0x12bcb , 0x12fcd , 0x132cf , 0x135d1 , 0x137d2 , 0x139d4 , 0x13bd5 , 0x13ed7 , 0x140d8 , 0x143da , 0x147dc , 0x24add , 0x24ede , 0x253e0 , 0x358de , 0x35ce2 , 0x461e3 , 0x464e4 , 0x467e5 , 0x56be5 , 0x56ee6 , 0x671e7 , 0x679e9 , 0x682ea , 0x788ec , 0x790ee , 0x89cef , 0x8a2f0 , 0x8a7f0 , 0x9adf1 , 0x9b3f2 , 0xab9f2 , 0xabbf3 , 0xbbef5 , 0xbc0f6 , 0xcc3f7 , 0xcc6f8 , 0xdc9f8 , 0xdccf9 , 0xecff9 , 0xed2fa , 0xfd5fa , 0x11d7fb , 0x12d8fc , 0x14d9fc , 0x16dafd , 0x18dbfd , 0x1cdcfe , 0x20defe , 0x24defe , 0x28e0fe , 0x2be1fe , 0x2fe2fe , 0x32e3fe , 0x35e3fe , 0x38e4fe , 0x3be5fe , 0x42e5fe , 0x4ae6fe , 0x52e6fe , 0x59e7fe , 0x65e7fe , 0x71ebfe , 0x7deefe , 0x88f0fe , 0x94f2fe , 0xa4f5fe , 0xb3f5fe , 0xbbf6fe , 0xc6f7fe , 0xd1f8fe , 0xe1f9fe , 0xf1fafe , 0xf5fbfe , 0xf9fcfe , 0xfcfdfe , 0xfcfdfe}

,

/* 51 unionjack */

{0x0 , 0x20000 , 0x50000 , 0x80000 , 0xb0000 , 0xe0000 , 0x110000 , 0x140000 , 0x170000 , 0x1a0000 , 0x1d0000 , 0x200000 , 0x230000 , 0x260000 , 0x290000 , 0x2c0000 , 0x2f0000 , 0x320000 , 0x350000 , 0x380000 , 0x3b0000 , 0x3e0000 , 0x410000 , 0x440000 , 0x470000 , 0x4a0000 , 0x4d0000 , 0x500000 , 0x530000 , 0x560000 , 0x590000 , 0x5c0000 , 0x5f0000 , 0x610000 , 0x640000 , 0x670000 , 0x6a0000 , 0x6d0000 , 0x700000 , 0x730000 , 0x760000 , 0x790000 , 0x7c0000 , 0x7f0000 , 0x820000 , 0x850000 , 0x880000 , 0x8b0000 , 0x8e0000 , 0x910000 , 0x940000 , 0x970000 , 0x9a0000 , 0x9d0000 , 0xa00000 , 0xa30000 , 0xa60000 , 0xa90000 , 0xac0000 , 0xaf0000 , 0xb20000 , 0xb50000 , 0xb80000 , 0xbe0000 , 0xbe0000 , 0xbf0303 , 0xc00707 , 0xc10b0b , 0xc20f0f , 0xc31313 , 0xc41717 , 0xc51b1b , 0xc61f1f , 0xc72323 , 0xc82727 , 0xc92b2b , 0xca2f2f , 0xcb3333 , 0xcc3737 , 0xcd3b3b , 0xce3f3f , 0xcf4343 , 0xd04747 , 0xd14b4b , 0xd24f4f , 0xd35353 , 0xd45757 , 0xd55b5b , 0xd65f5f , 0xd76363 , 0xd86767 , 0xd96b6b , 0xda6f6f , 0xdb7373 , 0xdc7777 , 0xdd7b7b , 0xde7f7f , 0xdf8383 , 0xe08787 , 0xe18b8b , 0xe28f8f , 0xe39393 , 0xe49797 , 0xe59b9b , 0xe69f9f , 0xe7a3a3 , 0xe8a7a7 , 0xe9abab , 0xeaafaf , 0xebb3b3 , 0xecb7b7 , 0xedbbbb , 0xeebfbf , 0xefc3c3 , 0xf0c7c7 , 0xf1cbcb , 0xf2cfcf , 0xf3d3d3 , 0xf4d7d7 , 0xf5dbdb , 0xf6dfdf , 0xf7e3e3 , 0xf8e7e7 , 0xf9ebeb , 0xfaefef , 0xfbf3f3 , 0xfcf7f7 , 0xffffff , 0xffffff , 0xfbfbfd , 0xf7f7fc , 0xf3f3fb , 0xefeffa , 0xebebf9 , 0xe7e7f8 , 0xe3e3f7 , 0xdfdff6 , 0xdbdbf5 , 0xd7d7f4 , 0xd3d3f3 , 0xcfcff2 , 0xcbcbf1 , 0xc7c7f0 , 0xc3c3ef , 0xbfbfee , 0xbbbbed , 0xb7b7ec , 0xb3b3eb , 0xafafea , 0xababe9 , 0xa7a7e8 , 0xa3a3e7 , 0x9f9fe6 , 0x9b9be5 , 0x9797e4 , 0x9393e3 , 0x8f8fe2 , 0x8b8be1 , 0x8787e0 , 0x8383df , 0x7f7fde , 0x7b7bdd , 0x7777dc , 0x7373db , 0x6f6fda , 0x6b6bd9 , 0x6767d8 , 0x6363d7 , 0x5f5fd6 , 0x5b5bd5 , 0x5757d4 , 0x5353d3 , 0x4f4fd2 , 0x4b4bd1 , 0x4747d0 , 0x4343cf , 0x3f3fce , 0x3b3bcd , 0x3737cc , 0x3333cb , 0x2f2fca , 0x2b2bc9 , 0x2727c8 , 0x2323c7 , 0x1f1fc6 , 0x1b1bc5 , 0x1717c4 , 0x1313c3 , 0xf0fc2 , 0xb0bc1 , 0x707c0 , 0xbe , 0xbe , 0x303bf , 0x707c0 , 0xb0bc1 , 0xf0fc2 , 0x1313c3 , 0x1717c4 , 0x1b1bc5 , 0x1f1fc6 , 0x2323c7 , 0x2727c8 , 0x2b2bc9 , 0x2f2fca , 0x3333cb , 0x3737cc , 0x3b3bcd , 0x3f3fce , 0x4343cf , 0x4747d0 , 0x4b4bd1 , 0x4f4fd2 , 0x5353d3 , 0x5757d4 , 0x5b5bd5 , 0x5f5fd6 , 0x6363d7 , 0x6767d8 , 0x6b6bd9 , 0x6f6fda , 0x7373db , 0x7777dc , 0x7b7bdd , 0x7f7fde , 0x8383df , 0x8787e0 , 0x8b8be1 , 0x8f8fe2 , 0x9393e3 , 0x9797e4 , 0x9b9be5 , 0x9f9fe6 , 0xa3a3e7 , 0xa7a7e8 , 0xababe9 , 0xafafea , 0xb3b3eb , 0xb7b7ec , 0xbbbbed , 0xbfbfee , 0xc3c3ef , 0xc7c7f0 , 0xcbcbf1 , 0xcfcff2 , 0xd3d3f3 , 0xd7d7f4 , 0xdbdbf5 , 0xdfdff6 , 0xe3e3f7 , 0xe7e7f8 , 0xebebf9 , 0xefeffa , 0xf3f3fb , 0xf7f7fc , 0xffffff}

,

/* 52 20_colors */

{0x940131 , 0x940131 , 0x940131 , 0x940131 , 0x940131 , 0x940131 , 0x940131 , 0x940131 , 0x940131 , 0x940131 , 0x940131 , 0x940131 , 0x940131 , 0x9e0101 , 0x9e0101 , 0x9e0101 , 0x9e0101 , 0x9e0101 , 0x9e0101 , 0x9e0101 , 0x9e0101 , 0x9e0101 , 0x9e0101 , 0x9e0101 , 0x9e0101 , 0x9e0101 , 0xcf0101 , 0xcf0101 , 0xcf0101 , 0xcf0101 , 0xcf0101 , 0xcf0101 , 0xcf0101 , 0xcf0101 , 0xcf0101 , 0xcf0101 , 0xcf0101 , 0xcf0101 , 0xcf0101 , 0xf00101 , 0xf00101 , 0xf00101 , 0xf00101 , 0xf00101 , 0xf00101 , 0xf00101 , 0xf00101 , 0xf00101 , 0xf00101 , 0xf00101 , 0xf00101 , 0xf00101 , 0xf06201 , 0xf06201 , 0xf06201 , 0xf06201 , 0xf06201 , 0xf06201 , 0xf06201 , 0xf06201 , 0xf06201 , 0xf06201 , 0xf06201 , 0xf06201 , 0xc58401 , 0xc58401 , 0xc58401 , 0xc58401 , 0xc58401 , 0xc58401 , 0xc58401 , 0xc58401 , 0xc58401 , 0xc58401 , 0xc58401 , 0xc58401 , 0xc58401 , 0x9e8501 , 0x9e8501 , 0x9e8501 , 0x9e8501 , 0x9e8501 , 0x9e8501 , 0x9e8501 , 0x9e8501 , 0x9e8501 , 0x9e8501 , 0x9e8501 , 0x9e8501 , 0x9e8501 , 0x809e01 , 0x809e01 , 0x809e01 , 0x809e01 , 0x809e01 , 0x809e01 , 0x809e01 , 0x809e01 , 0x809e01 , 0x809e01 , 0x809e01 , 0x809e01 , 0x809e01 , 0x19401 , 0x19401 , 0x19401 , 0x19401 , 0x19401 , 0x19401 , 0x19401 , 0x19401 , 0x19401 , 0x19401 , 0x19401 , 0x19401 , 0x1c501 , 0x1c501 , 0x1c501 , 0x1c501 , 0x1c501 , 0x1c501 , 0x1c501 , 0x1c501 , 0x1c501 , 0x1c501 , 0x1c501 , 0x1c501 , 0x1c501 , 0x1e301 , 0x1e301 , 0x1e301 , 0x1e301 , 0x1e301 , 0x1e301 , 0x1e301 , 0x1e301 , 0x1e301 , 0x1e301 , 0x1e301 , 0x1e301 , 0x1e301 , 0x1ed80 , 0x1ed80 , 0x1ed80 , 0x1ed80 , 0x1ed80 , 0x1ed80 , 0x1ed80 , 0x1ed80 , 0x1ed80 , 0x1ed80 , 0x1ed80 , 0x1ed80 , 0x1ed80 , 0x1f0c5 , 0x1f0c5 , 0x1f0c5 , 0x1f0c5 , 0x1f0c5 , 0x1f0c5 , 0x1f0c5 , 0x1f0c5 , 0x1f0c5 , 0x1f0c5 , 0x1f0c5 , 0x1f0c5 , 0x1f0f0 , 0x1f0f0 , 0x1f0f0 , 0x1f0f0 , 0x1f0f0 , 0x1f0f0 , 0x1f0f0 , 0x1f0f0 , 0x1f0f0 , 0x1f0f0 , 0x1f0f0 , 0x1f0f0 , 0x1f0f0 , 0x1cff0 , 0x1cff0 , 0x1cff0 , 0x1cff0 , 0x1cff0 , 0x1cff0 , 0x1cff0 , 0x1cff0 , 0x1cff0 , 0x1cff0 , 0x1cff0 , 0x1cff0 , 0x1cff0 , 0x1b2f0 , 0x1b2f0 , 0x1b2f0 , 0x1b2f0 , 0x1b2f0 , 0x1b2f0 , 0x1b2f0 , 0x1b2f0 , 0x1b2f0 , 0x1b2f0 , 0x1b2f0 , 0x1b2f0 , 0x1b2f0 , 0x18ef0 , 0x18ef0 , 0x18ef0 , 0x18ef0 , 0x18ef0 , 0x18ef0 , 0x18ef0 , 0x18ef0 , 0x18ef0 , 0x18ef0 , 0x18ef0 , 0x18ef0 , 0x162f0 , 0x162f0 , 0x162f0 , 0x162f0 , 0x162f0 , 0x162f0 , 0x162f0 , 0x162f0 , 0x162f0 , 0x162f0 , 0x162f0 , 0x162f0 , 0x162f0 , 0x101f0 , 0x101f0 , 0x101f0 , 0x101f0 , 0x101f0 , 0x101f0 , 0x101f0 , 0x101f0 , 0x101f0 , 0x101f0 , 0x101f0 , 0x101f0 , 0x101f0 , 0xc5c5c5 , 0xc5c5c5 , 0xc5c5c5 , 0xc5c5c5 , 0xc5c5c5 , 0xc5c5c5 , 0xc5c5c5 , 0xc5c5c5 , 0xc5c5c5 , 0xc5c5c5 , 0xc5c5c5 , 0xc5c5c5 , 0xc5c5c5}

,

/* 53 haze */

{0xffffff , 0x0 , 0x7f7f7f , 0x3f003f , 0x430043 , 0x460046 , 0x490049 , 0x4d004d , 0x500050 , 0x530053 , 0x570057 , 0x5a005a , 0x5d005d , 0x610061 , 0x640064 , 0x680068 , 0x6b006b , 0x6e006e , 0x720072 , 0x750075 , 0x780078 , 0x7c007c , 0x7f007f , 0x7f0082 , 0x7f0085 , 0x7f0089 , 0x7f008c , 0x7f008f , 0x7f0092 , 0x7f0095 , 0x7f0099 , 0x7f009c , 0x7f009f , 0x7f00a2 , 0x7f00a5 , 0x7f00a8 , 0x7f00ac , 0x7f00af , 0x7f00b2 , 0x7f00b5 , 0x7f00b8 , 0x7f00bc , 0x7f00bf , 0x7f00c2 , 0x7f00c5 , 0x7f00c8 , 0x7f00cc , 0x7f00cf , 0x7f00d2 , 0x7f00d5 , 0x7f00d8 , 0x7f00db , 0x7f00df , 0x7f00e2 , 0x7f00e5 , 0x7f00e8 , 0x7f00eb , 0x7f00ef , 0x7f00f2 , 0x7f00f5 , 0x7f00f8 , 0x7f00fb , 0x7f00ff , 0x8203ff , 0x8506ff , 0x8909ff , 0x8c0cff , 0x8f0fff , 0x9213ff , 0x9516ff , 0x9919ff , 0x9c1cff , 0x9f1fff , 0xa223ff , 0xa526ff , 0xa829ff , 0xac2cff , 0xaf2fff , 0xb233ff , 0xb536ff , 0xb839ff , 0xbc3cff , 0xbf3fff , 0xc242ff , 0xc546ff , 0xc849ff , 0xcc4cff , 0xcf4fff , 0xd252ff , 0xd556ff , 0xd859ff , 0xdb5cff , 0xdf5fff , 0xe262ff , 0xe566ff , 0xe869ff , 0xeb6cff , 0xef6fff , 0xf272ff , 0xf575ff , 0xf879ff , 0xfb7cff , 0xff7fff , 0xff80ff , 0xff81ff , 0xff81ff , 0xff82ff , 0xff83ff , 0xff84ff , 0xff85ff , 0xff86ff , 0xff86ff , 0xff87ff , 0xff88ff , 0xff89ff , 0xff8aff , 0xff8bff , 0xff8bff , 0xff8cff , 0xff8dff , 0xff8eff , 0xff8fff , 0xff90ff , 0xff90ff , 0xff91ff , 0xff92ff , 0xff93ff , 0xff94ff , 0xff95ff , 0xff95ff , 0xff96ff , 0xff97ff , 0xff98ff , 0xff99ff , 0xff99ff , 0xff9aff , 0xff9bff , 0xff9cff , 0xff9dff , 0xff9eff , 0xff9eff , 0xff9fff , 0xffa0ff , 0xffa1ff , 0xffa2ff , 0xffa3ff , 0xffa3ff , 0xffa4ff , 0xffa5ff , 0xffa6ff , 0xffa7ff , 0xffa8ff , 0xffa8ff , 0xffa9ff , 0xffaaff , 0xffabff , 0xffacff , 0xffadff , 0xffadff , 0xffaeff , 0xffafff , 0xffb0ff , 0xffb1ff , 0xffb2ff , 0xffb2ff , 0xffb3ff , 0xffb4ff , 0xffb5ff , 0xffb6ff , 0xffb6ff , 0xffb7ff , 0xffb8ff , 0xffb9ff , 0xffbaff , 0xffbbff , 0xffbbff , 0xffbcff , 0xffbdff , 0xffbeff , 0xffbfff , 0xffc0ff , 0xffc0ff , 0xffc1ff , 0xffc2ff , 0xffc3ff , 0xffc4ff , 0xffc5ff , 0xffc5ff , 0xffc6ff , 0xffc7ff , 0xffc8ff , 0xffc9ff , 0xffcaff , 0xffcaff , 0xffcbff , 0xffccff , 0xffcdff , 0xffceff , 0xffceff , 0xffcfff , 0xffd0ff , 0xffd1ff , 0xffd2ff , 0xffd3ff , 0xffd3ff , 0xffd4ff , 0xffd5ff , 0xffd6ff , 0xffd7ff , 0xffd8ff , 0xffd8ff , 0xffd9ff , 0xffdaff , 0xffdbff , 0xffdcff , 0xffddff , 0xffddff , 0xffdeff , 0xffdfff , 0xffe0ff , 0xffe1ff , 0xffe2ff , 0xffe2ff , 0xffe3ff , 0xffe4ff , 0xffe5ff , 0xffe6ff , 0xffe6ff , 0xffe7ff , 0xffe8ff , 0xffe9ff , 0xffeaff , 0xffebff , 0xffebff , 0xffecff , 0xffedff , 0xffeeff , 0xffefff , 0xfff0ff , 0xfff0ff , 0xfff1ff , 0xfff2ff , 0xfff3ff , 0xfff4ff , 0xfff5ff , 0xfff5ff , 0xfff6ff , 0xfff7ff , 0xfff8ff , 0xfff9ff , 0xfffaff , 0xfffaff , 0xfffbff , 0xfffcff , 0xfffdff , 0x0}

,

/* 54 log_up */

{0x202020 , 0x202020 , 0x323232 , 0x404040 , 0x4a4a4a , 0x525252 , 0x595959 , 0x606060 , 0x656565 , 0x6a6a6a , 0x6e6e6e , 0x727272 , 0x767676 , 0x797979 , 0x7d7d7d , 0x808080 , 0x828282 , 0x858585 , 0x878787 , 0x8a8a8a , 0x8c8c8c , 0x8e8e8e , 0x909090 , 0x929292 , 0x949494 , 0x969696 , 0x989898 , 0x999999 , 0x9b9b9b , 0x9d9d9d , 0x9e9e9e , 0x9f9f9f , 0xa1a1a1 , 0xa2a2a2 , 0xa4a4a4 , 0xa5a5a5 , 0xa6a6a6 , 0xa7a7a7 , 0xa9a9a9 , 0xaaaaaa , 0xababab , 0xacacac , 0xadadad , 0xaeaeae , 0xafafaf , 0xb0b0b0 , 0xb1b1b1 , 0xb2b2b2 , 0xb3b3b3 , 0xb4b4b4 , 0xb5b5b5 , 0xb6b6b6 , 0xb7b7b7 , 0xb8b8b8 , 0xb9b9b9 , 0xb9b9b9 , 0xbababa , 0xbbbbbb , 0xbcbcbc , 0xbdbdbd , 0xbdbdbd , 0xbebebe , 0xbfbfbf , 0xbfbfbf , 0xc0c0c0 , 0xc1c1c1 , 0xc2c2c2 , 0xc2c2c2 , 0xc3c3c3 , 0xc4c4c4 , 0xc4c4c4 , 0xc5c5c5 , 0xc6c6c6 , 0xc6c6c6 , 0xc7c7c7 , 0xc7c7c7 , 0xc8c8c8 , 0xc9c9c9 , 0xc9c9c9 , 0xcacaca , 0xcacaca , 0xcbcbcb , 0xcccccc , 0xcccccc , 0xcdcdcd , 0xcdcdcd , 0xcecece , 0xcecece , 0xcfcfcf , 0xcfcfcf , 0xd0d0d0 , 0xd0d0d0 , 0xd1d1d1 , 0xd1d1d1 , 0xd2d2d2 , 0xd2d2d2 , 0xd3d3d3 , 0xd3d3d3 , 0xd4d4d4 , 0xd4d4d4 , 0xd5d5d5 , 0xd5d5d5 , 0xd5d5d5 , 0xd6d6d6 , 0xd6d6d6 , 0xd7d7d7 , 0xd7d7d7 , 0xd8d8d8 , 0xd8d8d8 , 0xd9d9d9 , 0xd9d9d9 , 0xd9d9d9 , 0xdadada , 0xdadada , 0xdbdbdb , 0xdbdbdb , 0xdbdbdb , 0xdcdcdc , 0xdcdcdc , 0xdddddd , 0xdddddd , 0xdddddd , 0xdedede , 0xdedede , 0xdedede , 0xdfdfdf , 0xdfdfdf , 0xdfdfdf , 0xe0e0e0 , 0xe0e0e0 , 0xe1e1e1 , 0xe1e1e1 , 0xe1e1e1 , 0xe2e2e2 , 0xe2e2e2 , 0xe2e2e2 , 0xe3e3e3 , 0xe3e3e3 , 0xe3e3e3 , 0xe4e4e4 , 0xe4e4e4 , 0xe4e4e4 , 0xe5e5e5 , 0xe5e5e5 , 0xe5e5e5 , 0xe6e6e6 , 0xe6e6e6 , 0xe6e6e6 , 0xe7e7e7 , 0xe7e7e7 , 0xe7e7e7 , 0xe7e7e7 , 0xe8e8e8 , 0xe8e8e8 , 0xe8e8e8 , 0xe9e9e9 , 0xe9e9e9 , 0xe9e9e9 , 0xeaeaea , 0xeaeaea , 0xeaeaea , 0xeaeaea , 0xebebeb , 0xebebeb , 0xebebeb , 0xebebeb , 0xececec , 0xececec , 0xececec , 0xededed , 0xededed , 0xededed , 0xededed , 0xeeeeee , 0xeeeeee , 0xeeeeee , 0xeeeeee , 0xefefef , 0xefefef , 0xefefef , 0xefefef , 0xf0f0f0 , 0xf0f0f0 , 0xf0f0f0 , 0xf1f1f1 , 0xf1f1f1 , 0xf1f1f1 , 0xf1f1f1 , 0xf1f1f1 , 0xf2f2f2 , 0xf2f2f2 , 0xf3f3f3 , 0xf2f2f2 , 0xf3f3f3 , 0xf3f3f3 , 0xf3f3f3 , 0xf3f3f3 , 0xf4f4f4 , 0xf4f4f4 , 0xf4f4f4 , 0xf4f4f4 , 0xf5f5f5 , 0xf5f5f5 , 0xf5f5f5 , 0xf5f5f5 , 0xf5f5f5 , 0xf6f6f6 , 0xf6f6f6 , 0xf6f6f6 , 0xf6f6f6 , 0xf7f7f7 , 0xf7f7f7 , 0xf7f7f7 , 0xf7f7f7 , 0xf7f7f7 , 0xf8f8f8 , 0xf8f8f8 , 0xf8f8f8 , 0xf8f8f8 , 0xf9f9f9 , 0xf9f9f9 , 0xf9f9f9 , 0xf9f9f9 , 0xf8f8f8 , 0xfafafa , 0xfafafa , 0xfafafa , 0xfafafa , 0xfafafa , 0xfbfbfb , 0xfbfbfb , 0xfbfbfb , 0xfbfbfb , 0xfbfbfb , 0xfcfcfc , 0xfcfcfc , 0xfcfcfc , 0xfcfcfc , 0xfcfcfc , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff}

,

/* 55 hue */

{0xff , 0x7f8 , 0xdf2 , 0x13ec , 0x19e6 , 0x1ee1 , 0x23dc , 0x27d8 , 0x2cd3 , 0x30cf , 0x33cc , 0x37c8 , 0x3bc4 , 0x3ec1 , 0x41be , 0x44bb , 0x47b8 , 0x4ab5 , 0x4db2 , 0x4fb0 , 0x52ad , 0x54ab , 0x57a8 , 0x59a6 , 0x5ba4 , 0x5da2 , 0x609f , 0x629d , 0x649b , 0x6699 , 0x6897 , 0x6a95 , 0x6c93 , 0x6e91 , 0x708f , 0x718e , 0x738c , 0x758a , 0x7788 , 0x7986 , 0x7b84 , 0x7c83 , 0x7e81 , 0x807f , 0x827d , 0x847b , 0x8679 , 0x8778 , 0x8976 , 0x8b74 , 0x8d72 , 0x8f70 , 0x916e , 0x936c , 0x956a , 0x9669 , 0x9867 , 0x9b64 , 0x9d62 , 0x9f60 , 0xa15e , 0xa35c , 0xa55a , 0xa857 , 0xaa55 , 0xac53 , 0xaf50 , 0xb24d , 0xb44b , 0xb748 , 0xba45 , 0xbd42 , 0xc03f , 0xc33c , 0xc738 , 0xca35 , 0xce31 , 0xd22d , 0xd629 , 0xdb24 , 0xdf20 , 0xe41b , 0xea15 , 0xf00f , 0xf609 , 0xfd02 , 0x5fa00 , 0xbf400 , 0x11ee00 , 0x17e800 , 0x1ce300 , 0x21de00 , 0x26d900 , 0x2ad500 , 0x2ed100 , 0x32cd00 , 0x36c900 , 0x39c600 , 0x3dc200 , 0x40bf00 , 0x43bc00 , 0x46b900 , 0x49b600 , 0x4cb300 , 0x4eb100 , 0x51ae00 , 0x53ac00 , 0x56a900 , 0x58a700 , 0x5aa500 , 0x5da200 , 0x5fa000 , 0x619e00 , 0x639c00 , 0x659a00 , 0x679800 , 0x699600 , 0x6b9400 , 0x6d9200 , 0x6f9000 , 0x718e00 , 0x738c00 , 0x758a00 , 0x768900 , 0x788700 , 0x7a8500 , 0x7c8300 , 0x7e8100 , 0x7f8000 , 0x817e00 , 0x837c00 , 0x857a00 , 0x877800 , 0x897600 , 0x8a7500 , 0x8c7300 , 0x8e7100 , 0x906f00 , 0x926d00 , 0x946b00 , 0x966900 , 0x986700 , 0x9a6500 , 0x9c6300 , 0x9e6100 , 0xa05f00 , 0xa25d00 , 0xa55a00 , 0xa75800 , 0xa95600 , 0xac5300 , 0xae5100 , 0xb14e00 , 0xb34c00 , 0xb64900 , 0xb94600 , 0xbc4300 , 0xbf4000 , 0xc23d00 , 0xc63900 , 0xc93600 , 0xcd3200 , 0xd12e00 , 0xd52a00 , 0xd92600 , 0xde2100 , 0xe31c00 , 0xe81700 , 0xee1100 , 0xf40b00 , 0xfa0500 , 0xfd0002 , 0xf60009 , 0xf0000f , 0xea0015 , 0xe4001b , 0xdf0020 , 0xdb0024 , 0xd60029 , 0xd2002d , 0xce0031 , 0xca0035 , 0xc70038 , 0xc3003c , 0xc0003f , 0xbd0042 , 0xba0045 , 0xb70048 , 0xb4004b , 0xb2004d , 0xaf0050 , 0xac0053 , 0xaa0055 , 0xa80057 , 0xa5005a , 0xa3005c , 0xa1005e , 0x9f0060 , 0x9d0062 , 0x9b0064 , 0x980067 , 0x960069 , 0x95006a , 0x93006c , 0x91006e , 0x8f0070 , 0x8d0072 , 0x8b0074 , 0x890076 , 0x870078 , 0x860079 , 0x84007b , 0x82007d , 0x80007f , 0x7e0081 , 0x7c0083 , 0x7b0084 , 0x790086 , 0x770088 , 0x75008a , 0x73008c , 0x71008e , 0x70008f , 0x6e0091 , 0x6c0093 , 0x6a0095 , 0x680097 , 0x660099 , 0x64009b , 0x62009d , 0x60009f , 0x5d00a2 , 0x5b00a4 , 0x5900a6 , 0x5700a8 , 0x5400ab , 0x5200ad , 0x4f00b0 , 0x4d00b2 , 0x4a00b5 , 0x4700b8 , 0x4400bb , 0x4100be , 0x3e00c1 , 0x3b00c4 , 0x3700c8 , 0x3300cc , 0x3000cf , 0x2c00d3 , 0x2700d8 , 0x2300dc , 0x1e00e1 , 0x1900e6 , 0x1300ec , 0xd00f2 , 0x700f8}

,

/* 56 log_down */

{0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x20202 , 0x20202 , 0x20202 , 0x20202 , 0x20202 , 0x20202 , 0x30303 , 0x30303 , 0x30303 , 0x30303 , 0x30303 , 0x40404 , 0x40404 , 0x40404 , 0x40404 , 0x40404 , 0x50505 , 0x50505 , 0x50505 , 0x50505 , 0x50505 , 0x60606 , 0x60606 , 0x60606 , 0x60606 , 0x70707 , 0x70707 , 0x70707 , 0x70707 , 0x70707 , 0x80808 , 0x80808 , 0x80808 , 0x80808 , 0x80808 , 0x90909 , 0x90909 , 0x90909 , 0x90909 , 0xa0a0a , 0xa0a0a , 0xa0a0a , 0xa0a0a , 0xa0a0a , 0xb0b0b , 0xb0b0b , 0xb0b0b , 0xb0b0b , 0xc0c0c , 0xc0c0c , 0xc0c0c , 0xc0c0c , 0xd0d0d , 0xd0d0d , 0xd0d0d , 0xd0d0d , 0xe0e0e , 0xe0e0e , 0xe0e0e , 0xe0e0e , 0xe0e0e , 0xf0f0f , 0xf0f0f , 0xf0f0f , 0x101010 , 0x101010 , 0x101010 , 0x101010 , 0x111111 , 0x111111 , 0x111111 , 0x111111 , 0x121212 , 0x121212 , 0x121212 , 0x121212 , 0x131313 , 0x131313 , 0x131313 , 0x141414 , 0x141414 , 0x141414 , 0x141414 , 0x151515 , 0x151515 , 0x151515 , 0x151515 , 0x161616 , 0x161616 , 0x161616 , 0x171717 , 0x171717 , 0x171717 , 0x181818 , 0x181818 , 0x181818 , 0x181818 , 0x191919 , 0x191919 , 0x191919 , 0x1a1a1a , 0x1a1a1a , 0x1a1a1a , 0x1b1b1b , 0x1b1b1b , 0x1b1b1b , 0x1c1c1c , 0x1c1c1c , 0x1c1c1c , 0x1d1d1d , 0x1d1d1d , 0x1d1d1d , 0x1e1e1e , 0x1e1e1e , 0x1e1e1e , 0x1f1f1f , 0x1f1f1f , 0x202020 , 0x202020 , 0x202020 , 0x212121 , 0x212121 , 0x212121 , 0x222222 , 0x222222 , 0x222222 , 0x232323 , 0x232323 , 0x242424 , 0x242424 , 0x242424 , 0x252525 , 0x252525 , 0x262626 , 0x262626 , 0x262626 , 0x272727 , 0x272727 , 0x282828 , 0x282828 , 0x292929 , 0x292929 , 0x2a2a2a , 0x2a2a2a , 0x2a2a2a , 0x2b2b2b , 0x2b2b2b , 0x2c2c2c , 0x2c2c2c , 0x2d2d2d , 0x2d2d2d , 0x2e2e2e , 0x2e2e2e , 0x2f2f2f , 0x2f2f2f , 0x303030 , 0x303030 , 0x313131 , 0x313131 , 0x323232 , 0x323232 , 0x333333 , 0x343434 , 0x343434 , 0x353535 , 0x353535 , 0x363636 , 0x363636 , 0x373737 , 0x383838 , 0x383838 , 0x393939 , 0x393939 , 0x3a3a3a , 0x3b3b3b , 0x3b3b3b , 0x3c3c3c , 0x3d3d3d , 0x3d3d3d , 0x3e3e3e , 0x3f3f3f , 0x404040 , 0x404040 , 0x414141 , 0x424242 , 0x424242 , 0x434343 , 0x444444 , 0x454545 , 0x464646 , 0x464646 , 0x474747 , 0x484848 , 0x494949 , 0x4a4a4a , 0x4b4b4b , 0x4c4c4c , 0x4d4d4d , 0x4e4e4e , 0x4f4f4f , 0x505050 , 0x515151 , 0x525252 , 0x535353 , 0x545454 , 0x555555 , 0x565656 , 0x585858 , 0x595959 , 0x5a5a5a , 0x5b5b5b , 0x5d5d5d , 0x5e5e5e , 0x606060 , 0x616161 , 0x626262 , 0x646464 , 0x666666 , 0x676767 , 0x696969 , 0x6b6b6b , 0x6d6d6d , 0x6f6f6f , 0x717171 , 0x737373 , 0x757575 , 0x787878 , 0x7a7a7a , 0x7d7d7d , 0x808080 , 0x828282 , 0x868686 , 0x898989 , 0x8d8d8d , 0x919191 , 0x959595 , 0x9a9a9a , 0xa0a0a0 , 0xa6a6a6 , 0xadadad , 0xb5b5b5 , 0xc0c0c0 , 0xcdcdcd , 0xe0e0e0 , 0xe0e0e0}

,

/* 57 ether */

{0xffffff , 0xf60002 , 0xed0103 , 0xe40104 , 0xda0105 , 0xd10106 , 0xc80107 , 0xbf0108 , 0xb7010a , 0xae000b , 0xa6000c , 0x9e000a , 0x96000e , 0x8e010f , 0x870210 , 0x800211 , 0x790312 , 0x730414 , 0x6d0515 , 0x670616 , 0x620717 , 0x5d0718 , 0x580819 , 0x54091a , 0x500a1b , 0x4d0b1d , 0x4a0c1e , 0x470a1f , 0x450e20 , 0x440e21 , 0x420f22 , 0x421023 , 0x411024 , 0x421125 , 0x421127 , 0x441128 , 0x451229 , 0x47122a , 0x4a122b , 0x4d112c , 0x50112d , 0x54112e , 0x58102f , 0x5d0f31 , 0x620f32 , 0x670e33 , 0x6d0a34 , 0x730b35 , 0x790a36 , 0x800837 , 0x870638 , 0x8e053a , 0x96023b , 0x9e003c , 0xa6013d , 0xae043e , 0xb7063f , 0xbf0940 , 0xc80c41 , 0xd10f42 , 0xda1244 , 0xe41645 , 0xed1946 , 0xf61d47 , 0x2148 , 0x92449 , 0x12284a , 0x1b2c4b , 0x25304d , 0x2e344e , 0x37394f , 0x403d50 , 0x484151 , 0x514552 , 0x594a53 , 0x614e54 , 0x695255 , 0x715657 , 0x785b58 , 0x7f5f59 , 0x86635a , 0x8c675b , 0x926b5c , 0x986f5d , 0x9d725e , 0xa2765f , 0xa77a61 , 0xab7d62 , 0xaf8063 , 0xb28364 , 0xb58665 , 0xb88966 , 0xba8b67 , 0xbb8e68 , 0xbd906a , 0xbd926b , 0xbe936c , 0xbd956d , 0xbd966e , 0xbb976f , 0xba9770 , 0xb89871 , 0xb59872 , 0xb29874 , 0xaf9775 , 0xab9776 , 0xa79677 , 0xa29478 , 0x9d9379 , 0x98917a , 0x928f7b , 0x8c8d7c , 0x868a7e , 0x7f877f , 0x788480 , 0x718181 , 0x697d82 , 0x617983 , 0x597584 , 0x517185 , 0x486c87 , 0x406788 , 0x376289 , 0x2e5d8a , 0x25588b , 0x1b528c , 0x124d8d , 0x9478e , 0x418f , 0xf63b91 , 0xed3592 , 0xe42f93 , 0xda2994 , 0xd12295 , 0xc81c96 , 0xbf1697 , 0xb71098 , 0xae0a9a , 0xa6049b , 0x9e029c , 0x96079d , 0x8e0a9e , 0x87139f , 0x8019a0 , 0x791ea1 , 0x7323a2 , 0x6d28a4 , 0x672da5 , 0x6232a6 , 0x5d36a7 , 0x583aa8 , 0x543ea9 , 0x5041aa , 0x4d45ab , 0x4a47ac , 0x474aae , 0x454caf , 0x444eb0 , 0x424fb1 , 0x4251b2 , 0x4151b3 , 0x4252b4 , 0x4252b5 , 0x4451b7 , 0x4550b8 , 0x474fb9 , 0x4a4dba , 0x4d4bbb , 0x5048bc , 0x5445bd , 0x5842be , 0x5d3ebf , 0x623ac1 , 0x6735c2 , 0x6d30c3 , 0x732bc4 , 0x7925c5 , 0x801fc6 , 0x8718c7 , 0x8e11c8 , 0x960ac9 , 0x9e02cb , 0xa605cc , 0xae0acd , 0xb716ce , 0xbf1ecf , 0xc827d0 , 0xd131d1 , 0xda3ad2 , 0xe444d4 , 0xed4ed5 , 0xf657d6 , 0x62d7 , 0x96cd8 , 0x1276d9 , 0x1b80da , 0x258bdb , 0x2e95dc , 0x379fde , 0x40aadf , 0x48b4e0 , 0x51bee1 , 0x59c8e2 , 0x61d2e3 , 0x69dbe4 , 0x71e5e5 , 0x78eee7 , 0x7ff7e8 , 0x86fee9 , 0x8cf6ea , 0x92eeeb , 0x98e6ec , 0x9ddeed , 0xa2d7ee , 0xa7d1ef , 0xabcaf1 , 0xafc5f2 , 0xb2bff3 , 0xb5baf4 , 0xb8b6f5 , 0xbab2f6 , 0xbbaff7 , 0xbdacf8 , 0xbda9f9 , 0xbea8fb , 0xbda6fc , 0xbda6fd , 0xbba6fe , 0xbaa6ff , 0xb8a7fe , 0xb5a9fd , 0xb2abfc , 0xafadfa , 0xabb1f9 , 0xa7b5f8 , 0xa2b9f7 , 0x9dbef6 , 0x98c3f5 , 0x92c9f4 , 0x8cd0f3 , 0x86d7f2 , 0x7fdef0 , 0x78e6ef , 0x71efee , 0x69f8ed , 0x61fdec , 0x59f3eb , 0x51e9ea , 0x48dfe9 , 0x40d4e7 , 0x37c9e6 , 0x2ebee5 , 0x25b2e4 , 0x1ba6e3 , 0x129ae2 , 0x0}

,

/* 58 sepia */

{0x0 , 0x41c2b , 0x52339 , 0x62341 , 0x72442 , 0x82543 , 0x92644 , 0xa2745 , 0xb2846 , 0xc2947 , 0xd2a48 , 0xe2b49 , 0xf2c4a , 0x102d4b , 0x112e4c , 0x122f4d , 0x13304e , 0x14314f , 0x153250 , 0x163351 , 0x173452 , 0x183553 , 0x183553 , 0x193654 , 0x1a3755 , 0x1b3856 , 0x1c3957 , 0x1d3a58 , 0x1e3b59 , 0x1f3c5a , 0x203d5b , 0x213e5c , 0x223f5d , 0x23405e , 0x24415f , 0x254260 , 0x264361 , 0x274462 , 0x284563 , 0x294664 , 0x2a4765 , 0x2b4866 , 0x2c4967 , 0x2d4a68 , 0x2e4b69 , 0x2f4c6a , 0x304d6b , 0x304d6b , 0x314e6c , 0x324f6d , 0x33506e , 0x34516f , 0x355270 , 0x365371 , 0x375472 , 0x385573 , 0x395674 , 0x3a5775 , 0x3b5876 , 0x3c5977 , 0x3d5a78 , 0x3e5b79 , 0x3f5c7a , 0x405d7b , 0x415e7c , 0x425f7d , 0x425f7d , 0x43607e , 0x44617f , 0x456280 , 0x466381 , 0x476482 , 0x486583 , 0x496684 , 0x4a6785 , 0x4b6886 , 0x4c6987 , 0x4d6a88 , 0x4e6b89 , 0x4f6c8a , 0x506d8b , 0x516e8c , 0x526f8d , 0x53708e , 0x54718f , 0x557290 , 0x567391 , 0x577492 , 0x587593 , 0x597694 , 0x5a7795 , 0x5b7896 , 0x5c7997 , 0x5d7a98 , 0x5e7b99 , 0x5f7c9a , 0x607d9b , 0x617e9c , 0x627f9d , 0x63809e , 0x64819f , 0x6582a0 , 0x6683a1 , 0x6784a2 , 0x6885a3 , 0x6986a4 , 0x6a87a5 , 0x6b88a6 , 0x6c89a7 , 0x6d8aa8 , 0x6e8ba9 , 0x6f8caa , 0x708dab , 0x718eac , 0x728fad , 0x7390ae , 0x7491af , 0x7592b0 , 0x7693b1 , 0x7794b2 , 0x7895b3 , 0x7996b4 , 0x7a97b5 , 0x7b98b6 , 0x7c99b7 , 0x7d9ab8 , 0x7e9bb9 , 0x7f9cba , 0x809dbb , 0x809dbb , 0x819ebc , 0x829fbd , 0x83a0be , 0x84a1bf , 0x85a2c0 , 0x86a3c1 , 0x87a4c2 , 0x88a5c3 , 0x89a6c4 , 0x8aa7c5 , 0x8ba8c6 , 0x8ca9c7 , 0x8daac8 , 0x8eabc9 , 0x8facca , 0x90adcb , 0x91aecc , 0x92afcd , 0x93b0ce , 0x94b1cf , 0x95b2d0 , 0x96b3d1 , 0x97b4d2 , 0x98b5d3 , 0x99b6d4 , 0x9ab7d5 , 0x9bb8d6 , 0x9cb9d7 , 0x9dbad8 , 0x9ebbd9 , 0x9fbcda , 0xa0bddb , 0xa1bedc , 0xa2bfdd , 0xa3c0de , 0xa4c1df , 0xa5c2e0 , 0xa6c3e1 , 0xa7c4e2 , 0xa8c5e3 , 0xa9c6e4 , 0xaac7e5 , 0xabc8e6 , 0xacc9e7 , 0xadcae8 , 0xaecbe9 , 0xafccea , 0xb0cdeb , 0xb1ceec , 0xb2cfed , 0xb3d0ee , 0xb4d1ef , 0xb5d2f0 , 0xb6d3f1 , 0xb7d4f2 , 0xb8d5f3 , 0xb9d6f4 , 0xbad7f5 , 0xbbd8f6 , 0xbcd9f7 , 0xbddaf8 , 0xbedbf9 , 0xbfdcfa , 0xc0ddfb , 0xc1defc , 0xc2dffd , 0xc3e0fe , 0xc4e1ff , 0xc4e1ff , 0xc4e1ff , 0xc4e1ff , 0xc4e1ff , 0xc5e1ff , 0xc5e1ff , 0xc5e1ff , 0xc5e1ff , 0xc6e2ff , 0xc6e2ff , 0xc6e2ff , 0xc6e2ff , 0xc6e2ff , 0xc7e2ff , 0xc7e2ff , 0xc7e2ff , 0xc7e2ff , 0xc7e2ff , 0xc8e3ff , 0xc8e3ff , 0xc8e3ff , 0xc8e3ff , 0xc8e3ff , 0xc9e3ff , 0xc9e3ff , 0xc9e3ff , 0xc9e3ff , 0xc9e3ff , 0xcae4ff , 0xcae4ff , 0xcae4ff , 0xcae4ff , 0xcae4ff , 0xcae4ff , 0xcbe4ff , 0xcbe4ff , 0xcbe4ff , 0xcbe4ff , 0xcbe4ff , 0xcce5ff , 0xcce5ff , 0xcce5ff , 0xcce5ff , 0xcce5ff , 0xcce5ff , 0xcde5ff , 0xcde5ff , 0xcde5ff , 0xcde5ff , 0xcde5ff , 0xcde6ff , 0xcde6ff , 0xcde7ff , 0xd1e7ff , 0xd6e9ff , 0xd9e9ff , 0xe9f2ff , 0xffffff}

,

/* 59 neon-green */

{0x0 , 0x100 , 0x300 , 0x500 , 0x700 , 0x900 , 0xb00 , 0xd00 , 0xf00 , 0x1100 , 0x1300 , 0x1500 , 0x1700 , 0x1900 , 0x1b00 , 0x1d00 , 0x1f00 , 0x2100 , 0x2300 , 0x2500 , 0x2700 , 0x2900 , 0x2b00 , 0x2d00 , 0x2f00 , 0x3100 , 0x3300 , 0x3500 , 0x3700 , 0x3900 , 0x3b00 , 0x3d00 , 0x3f00 , 0x4100 , 0x4300 , 0x4500 , 0x4700 , 0x4900 , 0x4b00 , 0x4d00 , 0x4f00 , 0x5100 , 0x5300 , 0x5500 , 0x5700 , 0x5900 , 0x5b00 , 0x5d00 , 0x5f00 , 0x6100 , 0x6300 , 0x6500 , 0x6700 , 0x6900 , 0x6b00 , 0x6d00 , 0x6f00 , 0x7100 , 0x7300 , 0x7500 , 0x7700 , 0x7900 , 0x7b00 , 0x7d00 , 0x7f00 , 0x8100 , 0x8300 , 0x8500 , 0x8700 , 0x8900 , 0x8b00 , 0x8d00 , 0x8f00 , 0x9100 , 0x9300 , 0x9500 , 0x9700 , 0x9900 , 0x9b00 , 0x9d00 , 0x9f00 , 0xa100 , 0xa300 , 0xa500 , 0xa700 , 0xa900 , 0xab00 , 0xad00 , 0xaf00 , 0xb100 , 0xb300 , 0xb500 , 0xb700 , 0xb900 , 0xbb00 , 0xbd00 , 0xbf00 , 0xc100 , 0xc300 , 0xc500 , 0xc700 , 0xc900 , 0xcb00 , 0xcd00 , 0xcf00 , 0xd100 , 0xd300 , 0xd500 , 0xd700 , 0xd900 , 0xdb00 , 0xdd00 , 0xdf00 , 0xe100 , 0xe300 , 0xe500 , 0xe700 , 0xe900 , 0xeb00 , 0xed00 , 0xef00 , 0xf100 , 0xf300 , 0xf500 , 0xf700 , 0xf900 , 0xfb00 , 0xff00 , 0xff00 , 0x1ff01 , 0x3ff03 , 0x5ff05 , 0x7ff07 , 0x9ff09 , 0xbff0b , 0xdff0d , 0xfff0f , 0x11ff11 , 0x13ff13 , 0x15ff15 , 0x17ff17 , 0x19ff19 , 0x1bff1b , 0x1dff1d , 0x1fff1f , 0x21ff21 , 0x23ff23 , 0x25ff25 , 0x27ff27 , 0x29ff29 , 0x2bff2b , 0x2dff2d , 0x2fff2f , 0x31ff31 , 0x33ff33 , 0x35ff35 , 0x37ff37 , 0x39ff39 , 0x3bff3b , 0x3dff3d , 0x3fff3f , 0x41ff41 , 0x43ff43 , 0x45ff45 , 0x47ff47 , 0x49ff49 , 0x4bff4b , 0x4dff4d , 0x4fff4f , 0x51ff51 , 0x53ff53 , 0x55ff55 , 0x57ff57 , 0x59ff59 , 0x5bff5b , 0x5dff5d , 0x5fff5f , 0x61ff61 , 0x63ff63 , 0x65ff65 , 0x67ff67 , 0x69ff69 , 0x6bff6b , 0x6dff6d , 0x6fff6f , 0x71ff71 , 0x73ff73 , 0x75ff75 , 0x77ff77 , 0x79ff79 , 0x7bff7b , 0x7dff7d , 0x7fff7f , 0x81ff81 , 0x83ff83 , 0x85ff85 , 0x87ff87 , 0x89ff89 , 0x8bff8b , 0x8dff8d , 0x8fff8f , 0x91ff91 , 0x93ff93 , 0x95ff95 , 0x97ff97 , 0x99ff99 , 0x9bff9b , 0x9dff9d , 0x9fff9f , 0xa1ffa1 , 0xa3ffa3 , 0xa5ffa5 , 0xa7ffa7 , 0xa9ffa9 , 0xabffab , 0xadffad , 0xafffaf , 0xb1ffb1 , 0xb3ffb3 , 0xb5ffb5 , 0xb7ffb7 , 0xb9ffb9 , 0xbbffbb , 0xbdffbd , 0xbfffbf , 0xc1ffc1 , 0xc3ffc3 , 0xc5ffc5 , 0xc7ffc7 , 0xc9ffc9 , 0xcbffcb , 0xcdffcd , 0xcfffcf , 0xd1ffd1 , 0xd3ffd3 , 0xd5ffd5 , 0xd7ffd7 , 0xd9ffd9 , 0xdbffdb , 0xddffdd , 0xdfffdf , 0xe1ffe1 , 0xe3ffe3 , 0xe5ffe5 , 0xe7ffe7 , 0xe9ffe9 , 0xebffeb , 0xedffed , 0xefffef , 0xf1fff1 , 0xf3fff3 , 0xf5fff5 , 0xf7fff7 , 0xf9fff9 , 0xfbfffb , 0xffffff}

,

/* 60 blue_orange */

{0x30302 , 0x30302 , 0x50503 , 0x70604 , 0x80805 , 0xa0a06 , 0xc0c07 , 0xe0e08 , 0x101009 , 0x12110a , 0x14130c , 0x16150d , 0x17170e , 0x19190f , 0x1b1a10 , 0x1d1c11 , 0x1f1e12 , 0x212013 , 0x232214 , 0x252415 , 0x262516 , 0x282717 , 0x2a2918 , 0x2c2b19 , 0x2e2d1a , 0x302f1b , 0x32301c , 0x34321d , 0x35341e , 0x37361f , 0x393821 , 0x3b3922 , 0x3d3b23 , 0x3f3d24 , 0x403f25 , 0x424026 , 0x434227 , 0x454328 , 0x474529 , 0x49472a , 0x4a462b , 0x4c442c , 0x4d432c , 0x4e422d , 0x4f412e , 0x51402f , 0x523f2f , 0x533e30 , 0x543d31 , 0x563b32 , 0x573a32 , 0x583933 , 0x593834 , 0x5b3635 , 0x5c3535 , 0x5d3436 , 0x5f3237 , 0x603138 , 0x613038 , 0x622f39 , 0x642d3a , 0x652c3b , 0x662b3b , 0x672a3c , 0x69283d , 0x6a273e , 0x6b263e , 0x6c243f , 0x6e2340 , 0x6f2240 , 0x702140 , 0x721f41 , 0x731e42 , 0x741d43 , 0x751c43 , 0x771a44 , 0x781945 , 0x791846 , 0x7a1746 , 0x7c1547 , 0x7d1448 , 0x7e1349 , 0x801149 , 0x81104a , 0x820f4b , 0x830e4c , 0x850c4c , 0x860b4d , 0x870a4e , 0x88094f , 0x8a074f , 0x8b0650 , 0x8c0551 , 0x8d0452 , 0x8f0252 , 0x900153 , 0x900355 , 0x8f0457 , 0x8f0659 , 0x8e085b , 0x8e0a5d , 0x8d0b5f , 0x8d0d61 , 0x8c0f63 , 0x8c1166 , 0x8c1268 , 0x8b146a , 0x8b166c , 0x8a176e , 0x8a1970 , 0x891b72 , 0x891d74 , 0x881e76 , 0x882078 , 0x88227a , 0x87247c , 0x87257e , 0x862780 , 0x862982 , 0x852b84 , 0x852c86 , 0x852e89 , 0x84308b , 0x84318d , 0x83338f , 0x833591 , 0x823793 , 0x823895 , 0x813a97 , 0x813c99 , 0x813e9b , 0x803f9d , 0x80409f , 0x7f42a1 , 0x7f44a3 , 0x7e45a5 , 0x7e47a7 , 0x7d49a9 , 0x7d4aab , 0x7d4cae , 0x7c4eb0 , 0x7c50b2 , 0x7b51b4 , 0x7b53b6 , 0x7a55b8 , 0x7a57ba , 0x7958bc , 0x795abe , 0x775abf , 0x755bbf , 0x735bc0 , 0x715cc1 , 0x6f5cc2 , 0x6d5cc4 , 0x6b5dc5 , 0x695dc6 , 0x675ec7 , 0x655ec8 , 0x635ec9 , 0x615fca , 0x5f5fcb , 0x5d60cc , 0x5b60cd , 0x5960ce , 0x5761cf , 0x5561d1 , 0x5362d2 , 0x5162d3 , 0x4f62d4 , 0x4d63d5 , 0x4b63d6 , 0x4964d7 , 0x4764d8 , 0x4564d9 , 0x4365da , 0x4165db , 0x4066dc , 0x3e66de , 0x3b66df , 0x3967e0 , 0x3767e1 , 0x3568e2 , 0x3368e3 , 0x3168e4 , 0x2f69e5 , 0x2d69e6 , 0x2b6ae7 , 0x296ae8 , 0x276ae9 , 0x256beb , 0x236bec , 0x216ced , 0x1f6cee , 0x1d6cef , 0x1b6df0 , 0x196df1 , 0x176ef2 , 0x156ef3 , 0x136ef4 , 0x116ff5 , 0xf6ff6 , 0xd70f8 , 0xb70f9 , 0x970fa , 0x771fb , 0x571fc , 0x372fd , 0x172fe , 0x675fe , 0xc78fe , 0x117bfe , 0x167efe , 0x1c81fe , 0x2184fe , 0x2687fe , 0x2c8afe , 0x318cfe , 0x368ffe , 0x3b92fe , 0x4095fe , 0x4598fe , 0x4a9bfe , 0x509efe , 0x55a1fe , 0x5aa4fe , 0x60a7fe , 0x65aafe , 0x6aadfe , 0x70b0fe , 0x75b3fe , 0x7ab6fe , 0x80b9fe , 0x85bbfe , 0x8abefe , 0x8fc0fe , 0x95c3fe , 0x9ac6fe , 0x9fc9fe , 0xa5ccfe , 0xaacffe , 0xafd2fe , 0xb5d5fe , 0xbad8fe , 0xbfdbfe , 0xc4defe , 0xc9e1fe , 0xcee4fe , 0xd4e7fe , 0xd9e9fe , 0xdeecfe , 0xe3effe , 0xe9f2fe , 0xeef5fe , 0xf3f8fe , 0xf9fbfe , 0xf9fbfe}

,

/* 61 16_colors */

{0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0xab0101 , 0xab0101 , 0xab0101 , 0xab0101 , 0xab0101 , 0xab0101 , 0xab0101 , 0xab0101 , 0xab0101 , 0xab0101 , 0xab0101 , 0xab0101 , 0xab0101 , 0xab0101 , 0xab0101 , 0xab0101 , 0xe00101 , 0xe00101 , 0xe00101 , 0xe00101 , 0xe00101 , 0xe00101 , 0xe00101 , 0xe00101 , 0xe00101 , 0xe00101 , 0xe00101 , 0xe00101 , 0xe00101 , 0xe00101 , 0xe00101 , 0xe00101 , 0xff6e00 , 0xff6e00 , 0xff6e00 , 0xff6e00 , 0xff6e00 , 0xff6e00 , 0xff6e00 , 0xff6e00 , 0xff6e00 , 0xff6e00 , 0xff6e00 , 0xff6e00 , 0xff6e00 , 0xff6e00 , 0xff6e00 , 0xff6e00 , 0xfeab01 , 0xfeab01 , 0xfeab01 , 0xfeab01 , 0xfeab01 , 0xfeab01 , 0xfeab01 , 0xfeab01 , 0xfeab01 , 0xfeab01 , 0xfeab01 , 0xfeab01 , 0xfeab01 , 0xfeab01 , 0xfeab01 , 0xfeab01 , 0xfee001 , 0xfee001 , 0xfee001 , 0xfee001 , 0xfee001 , 0xfee001 , 0xfee001 , 0xfee001 , 0xfee001 , 0xfee001 , 0xfee001 , 0xfee001 , 0xfee001 , 0xfee001 , 0xfee001 , 0xfee001 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0x1fe01 , 0xffbe , 0xffbe , 0xffbe , 0xffbe , 0xffbe , 0xffbe , 0xffbe , 0xffbe , 0xffbe , 0xffbe , 0xffbe , 0xffbe , 0xffbe , 0xffbe , 0xffbe , 0xffbe , 0xffff , 0xffff , 0xffff , 0xffff , 0xffff , 0xffff , 0xffff , 0xffff , 0xffff , 0xffff , 0xffff , 0xffff , 0xffff , 0xffff , 0xffff , 0xffff , 0xe0ff , 0xe0ff , 0xe0ff , 0xe0ff , 0xe0ff , 0xe0ff , 0xe0ff , 0xe0ff , 0xe0ff , 0xe0ff , 0xe0ff , 0xe0ff , 0xe0ff , 0xe0ff , 0xe0ff , 0xe0ff , 0x8dff , 0x8dff , 0x8dff , 0x8dff , 0x8dff , 0x8dff , 0x8dff , 0x8dff , 0x8dff , 0x8dff , 0x8dff , 0x8dff , 0x8dff , 0x8dff , 0x8dff , 0x8dff , 0x5efa , 0x5efa , 0x5efa , 0x5efa , 0x5efa , 0x5efa , 0x5efa , 0x5efa , 0x5efa , 0x5efa , 0x5efa , 0x5efa , 0x5efa , 0x5efa , 0x5efa , 0x5efa , 0xf5 , 0xf5 , 0xf5 , 0xf5 , 0xf5 , 0xf5 , 0xf5 , 0xf5 , 0xf5 , 0xf5 , 0xf5 , 0xf5 , 0xf5 , 0xf5 , 0xf5 , 0xf5 , 0x8700f5 , 0x8d00f5 , 0x9300f5 , 0x9900f5 , 0xa000f5 , 0xa600f5 , 0xac00f5 , 0xb200f5 , 0xb900f5 , 0xbf00f5 , 0xc500f5 , 0xcb00f5 , 0xd200f5 , 0xd800f5 , 0xde00f5 , 0xe400f5 , 0xdeb4de , 0xdeb4de , 0xdeb4de , 0xdeb4de , 0xdeb4de , 0xdeb4de , 0xdeb4de , 0xdeb4de , 0xdeb4de , 0xdeb4de , 0xdeb4de , 0xdeb4de , 0xdeb4de , 0xdeb4de , 0xdeb4de , 0xdeb4de , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff , 0xffffff}

,

/* 62 thallium */

{0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10101 , 0x10201 , 0x10201 , 0x10301 , 0x10401 , 0x10401 , 0x10501 , 0x10601 , 0x10601 , 0x10701 , 0x10701 , 0x10801 , 0x10901 , 0x10901 , 0x10a01 , 0x10b01 , 0x10c01 , 0x10d01 , 0x10d01 , 0x10e01 , 0x10e01 , 0x10f01 , 0x11001 , 0x11001 , 0x11101 , 0x11201 , 0x11301 , 0x11301 , 0x11301 , 0x11301 , 0x11301 , 0x11401 , 0x11401 , 0x11401 , 0x11401 , 0x11501 , 0x11501 , 0x11501 , 0x11601 , 0x11601 , 0x11601 , 0x11601 , 0x11701 , 0x11701 , 0x11701 , 0x11701 , 0x11801 , 0x11801 , 0x11801 , 0x11801 , 0x11901 , 0x11901 , 0x11901 , 0x11901 , 0x11a01 , 0x11a01 , 0x11b01 , 0x11b01 , 0x11c01 , 0x11c01 , 0x11d01 , 0x11d01 , 0x11e01 , 0x11e01 , 0x11f01 , 0x11f01 , 0x12001 , 0x12001 , 0x12101 , 0x12101 , 0x12201 , 0x12201 , 0x12301 , 0x12301 , 0x12401 , 0x12501 , 0x12501 , 0x12601 , 0x12601 , 0x12701 , 0x12801 , 0x12801 , 0x12901 , 0x22a02 , 0x22a02 , 0x32b03 , 0x32c03 , 0x42d04 , 0x42d04 , 0x52e05 , 0x52f05 , 0x63006 , 0x63006 , 0x63106 , 0x73307 , 0x83308 , 0x83408 , 0x93509 , 0x93609 , 0xa360a , 0xa370a , 0xb380b , 0xb380b , 0xb390b , 0xc3a0c , 0xc3b0c , 0xd3b0d , 0xd3c0d , 0xe3d0e , 0xe3d0e , 0xe3e0e , 0xf3f0f , 0xf400f , 0x104010 , 0x114211 , 0x114311 , 0x124412 , 0x124512 , 0x134613 , 0x134613 , 0x144714 , 0x144814 , 0x154915 , 0x154a15 , 0x164b16 , 0x164c16 , 0x174d17 , 0x174e17 , 0x184f18 , 0x185018 , 0x195119 , 0x195119 , 0x1a521a , 0x1a531a , 0x1c561c , 0x1d571d , 0x1e581e , 0x1f5a1f , 0x215b21 , 0x225d22 , 0x235e23 , 0x246024 , 0x266126 , 0x276327 , 0x286428 , 0x296629 , 0x2b672b , 0x2c692c , 0x2d6a2d , 0x2e6c2e , 0x306d30 , 0x316f31 , 0x327032 , 0x357335 , 0x367536 , 0x377637 , 0x397739 , 0x3a793a , 0x3b7a3b , 0x3c7c3c , 0x3e7d3e , 0x3f7f3f , 0x408040 , 0x418241 , 0x438443 , 0x468646 , 0x488848 , 0x4b8b4b , 0x4d8d4d , 0x508f50 , 0x529152 , 0x559355 , 0x579657 , 0x5c9a5c , 0x5f9c5f , 0x619f61 , 0x64a164 , 0x66a366 , 0x69a569 , 0x6ba76b , 0x6eaa6e , 0x70ac70 , 0x73ae73 , 0x75b075 , 0x78b278 , 0x7ab57a , 0x7db77d , 0x7fb97f , 0x82bb82 , 0x86bf86 , 0x8ac08a , 0x8ec28e , 0x93c493 , 0x9cc99c , 0xa0cba0 , 0xa4cea4 , 0xa9d0a9 , 0xadd2ad , 0xb2d5b2 , 0xb6d7b6 , 0xbad9ba , 0xbfdcbf , 0xc2dec2 , 0xc7e0c7 , 0xcbe3cb , 0xcfe5cf , 0xd4e7d4 , 0xd8ead8 , 0xddecdd , 0xe1eee1 , 0xe5f1e5 , 0xeaf3ea , 0xf4f2fb , 0xfbf7fc , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd , 0xfdfdfd}

,

/* 63 Hot_Iron */

{0x40101 , 0x40101 , 0x80101 , 0xc0101 , 0x100101 , 0x140101 , 0x180101 , 0x1c0101 , 0x200101 , 0x240101 , 0x280101 , 0x2c0101 , 0x300101 , 0x340101 , 0x380101 , 0x3c0101 , 0x400101 , 0x430101 , 0x470101 , 0x4b0101 , 0x4f0101 , 0x530101 , 0x570101 , 0x5b0101 , 0x5f0101 , 0x630101 , 0x670101 , 0x6b0101 , 0x6f0101 , 0x730101 , 0x770101 , 0x7b0101 , 0x7f0101 , 0x830101 , 0x870101 , 0x8b0101 , 0x8f0101 , 0x930101 , 0x970101 , 0x9b0101 , 0x9f0101 , 0xa30101 , 0xa70101 , 0xab0101 , 0xaf0101 , 0xb30101 , 0xb70101 , 0xbb0101 , 0xbf0101 , 0xc20101 , 0xc60101 , 0xca0101 , 0xce0101 , 0xd20101 , 0xd60101 , 0xda0101 , 0xde0101 , 0xe20101 , 0xe60101 , 0xea0101 , 0xee0101 , 0xf20101 , 0xf60101 , 0xfa0101 , 0xfe0101 , 0xfe0101 , 0xfa0104 , 0xf60108 , 0xf2010c , 0xee0110 , 0xea0114 , 0xe60118 , 0xe2011c , 0xde0120 , 0xda0124 , 0xd60128 , 0xd2012c , 0xce0130 , 0xca0134 , 0xc60138 , 0xc2013c , 0xbf0140 , 0xbb0143 , 0xb70147 , 0xb3014b , 0xaf014f , 0xab0153 , 0xa70157 , 0xa3015b , 0x9f015f , 0x9b0163 , 0x970167 , 0x93016b , 0x8f016f , 0x8b0173 , 0x870177 , 0x83017b , 0x7f017f , 0x7b0183 , 0x770187 , 0x73018b , 0x6f018f , 0x6b0193 , 0x670197 , 0x63019b , 0x5f019f , 0x5b01a3 , 0x5701a7 , 0x5301ab , 0x4f01af , 0x4b01b3 , 0x4701b7 , 0x4301bb , 0x4001bf , 0x3c01c2 , 0x3801c6 , 0x3401ca , 0x3001ce , 0x2c01d2 , 0x2801d6 , 0x2401da , 0x2001de , 0x1c01e2 , 0x1801e6 , 0x1401ea , 0x1001ee , 0xc01f2 , 0x801f6 , 0x401fa , 0x101fe , 0x104fe , 0x108fe , 0x10cfe , 0x110fe , 0x114fe , 0x118fe , 0x11cfe , 0x120fe , 0x124fe , 0x128fe , 0x12cfe , 0x130fe , 0x134fe , 0x138fe , 0x13cfe , 0x140fe , 0x143fe , 0x147fe , 0x14bfe , 0x14ffe , 0x153fe , 0x157fe , 0x15bfe , 0x15ffe , 0x163fe , 0x167fe , 0x16bfe , 0x16ffe , 0x173fe , 0x177fe , 0x17bfe , 0x17ffe , 0x183fe , 0x187fe , 0x18bfe , 0x18ffe , 0x193fe , 0x197fe , 0x19bfe , 0x19ffe , 0x1a3fe , 0x1a7fe , 0x1abfe , 0x1affe , 0x1b3fe , 0x1b7fe , 0x1bbfe , 0x1bffe , 0x1c2fe , 0x1c6fe , 0x1cafe , 0x1cefe , 0x1d2fe , 0x1d6fe , 0x1dafe , 0x1defe , 0x1e2fe , 0x1e6fe , 0x1eafe , 0x1eefe , 0x1f2fe , 0x1f6fe , 0x1fafe , 0x1fefe , 0x5fefe , 0x9fefe , 0xdfefe , 0x11fefe , 0x15fefe , 0x19fefe , 0x1dfefe , 0x21fefe , 0x25fefe , 0x29fefe , 0x2dfefe , 0x31fefe , 0x35fefe , 0x39fefe , 0x3dfefe , 0x40fefe , 0x44fefe , 0x48fefe , 0x4cfefe , 0x50fefe , 0x55fefe , 0x59fefe , 0x5dfefe , 0x61fefe , 0x65fefe , 0x69fefe , 0x6dfefe , 0x71fefe , 0x75fefe , 0x79fefe , 0x7dfefe , 0x81fefe , 0x85fefe , 0x89fefe , 0x8dfefe , 0x91fefe , 0x95fefe , 0x99fefe , 0x9dfefe , 0xa1fefe , 0xa5fefe , 0xaafefe , 0xaefefe , 0xb2fefe , 0xb6fefe , 0xbafefe , 0xbefefe , 0xc1fefe , 0xc5fefe , 0xc9fefe , 0xcdfefe , 0xd1fefe , 0xd5fefe , 0xd9fefe , 0xddfefe , 0xe1fefe , 0xe5fefe , 0xe9fefe , 0xedfefe , 0xf1fefe , 0xf5fefe , 0xf5fefe}

,

/* 64 blue_orange_icb */

{0x10201 , 0x10201 , 0x10301 , 0x10401 , 0x10501 , 0x10601 , 0x10701 , 0x10801 , 0x10901 , 0x10a01 , 0x10b01 , 0x10c01 , 0x10d01 , 0x10e01 , 0x10f01 , 0x11001 , 0x11101 , 0x11201 , 0x11301 , 0x11401 , 0x11501 , 0x11601 , 0x11701 , 0x11801 , 0x11901 , 0x11a01 , 0x11b01 , 0x11c01 , 0x11d01 , 0x11e01 , 0x11f01 , 0x12001 , 0x12101 , 0x12201 , 0x12301 , 0x12401 , 0x12501 , 0x12601 , 0x12701 , 0x12801 , 0x12901 , 0x12a01 , 0x12b01 , 0x12c01 , 0x12d01 , 0x12e01 , 0x12f01 , 0x13001 , 0x13101 , 0x13201 , 0x13301 , 0x13401 , 0x13501 , 0x13601 , 0x13701 , 0x13801 , 0x13901 , 0x13a01 , 0x13b01 , 0x13c01 , 0x13d01 , 0x13e01 , 0x13f01 , 0x14001 , 0x24002 , 0x43f04 , 0x63e06 , 0x83d08 , 0xa3c0a , 0xc3b0c , 0xe3a0e , 0x103910 , 0x123812 , 0x143714 , 0x163616 , 0x183518 , 0x1a341a , 0x1c331c , 0x1e321e , 0x203120 , 0x223022 , 0x242f24 , 0x262e26 , 0x282d28 , 0x2a2c2a , 0x2c2b2c , 0x2e2a2e , 0x302930 , 0x322832 , 0x342734 , 0x362636 , 0x382538 , 0x3a243a , 0x3c233c , 0x3e223e , 0x402140 , 0x412041 , 0x431f43 , 0x451e45 , 0x471d47 , 0x491c49 , 0x4b1b4b , 0x4d1a4d , 0x4f194f , 0x511851 , 0x531753 , 0x551655 , 0x571557 , 0x591459 , 0x5b135b , 0x5d125d , 0x5f115f , 0x611061 , 0x630f63 , 0x650e65 , 0x670d67 , 0x690c69 , 0x6b0b6b , 0x6d0a6d , 0x6f096f , 0x710871 , 0x730773 , 0x750675 , 0x770577 , 0x790479 , 0x7b037b , 0x7d027d , 0x7f017f , 0x7d0181 , 0x7b0383 , 0x790585 , 0x770787 , 0x750989 , 0x730b8b , 0x710d8d , 0x6f0f8f , 0x6d1191 , 0x6b1393 , 0x691595 , 0x671797 , 0x651999 , 0x631b9b , 0x611d9d , 0x5f1f9f , 0x5d21a1 , 0x5b23a3 , 0x5925a5 , 0x5727a7 , 0x5529a9 , 0x532bab , 0x512dad , 0x4f2faf , 0x4d31b1 , 0x4b33b3 , 0x4935b5 , 0x4737b7 , 0x4539b9 , 0x433bbb , 0x413dbd , 0x403fbf , 0x3e40c0 , 0x3c42c2 , 0x3a44c4 , 0x3846c6 , 0x3648c8 , 0x344aca , 0x324ccc , 0x304ece , 0x2e50d0 , 0x2c52d2 , 0x2a54d4 , 0x2856d6 , 0x2658d8 , 0x245ada , 0x225cdc , 0x205ede , 0x1e60e0 , 0x1c62e2 , 0x1a64e4 , 0x1866e6 , 0x1668e8 , 0x146aea , 0x126cec , 0x106eee , 0xe70f0 , 0xc72f2 , 0xa74f4 , 0x876f6 , 0x678f8 , 0x47afa , 0x27cfc , 0x17efe , 0x480fe , 0x882fe , 0xc84fe , 0x1086fe , 0x1488fe , 0x188afe , 0x1c8cfe , 0x208efe , 0x2490fe , 0x2892fe , 0x2c94fe , 0x3096fe , 0x3498fe , 0x389afe , 0x3c9cfe , 0x409efe , 0x43a0fe , 0x47a2fe , 0x4ba4fe , 0x4fa6fe , 0x53a8fe , 0x57aafe , 0x5bacfe , 0x5faefe , 0x63b0fe , 0x67b2fe , 0x6bb4fe , 0x6fb6fe , 0x73b8fe , 0x77bafe , 0x7bbcfe , 0x7fbefe , 0x83bffe , 0x87c1fe , 0x8bc3fe , 0x8fc5fe , 0x93c7fe , 0x97c9fe , 0x9bcbfe , 0x9fcdfe , 0xa3cffe , 0xa7d1fe , 0xabd3fe , 0xafd5fe , 0xb3d7fe , 0xb7d9fe , 0xbbdbfe , 0xbfddfe , 0xc2dffe , 0xc6e1fe , 0xcae3fe , 0xcee5fe , 0xd2e7fe , 0xd6e9fe , 0xdaebfe , 0xdeedfe , 0xe2effe , 0xe6f1fe , 0xeaf3fe , 0xeef5fe , 0xf2f7fe , 0xf6f9fe , 0xfafbfe , 0xfafbfe}

,

/* 65 HotGreen */

{0x10302 , 0x10302 , 0x10501 , 0x10701 , 0x10901 , 0x10b01 , 0x10d01 , 0x10f01 , 0x11101 , 0x11301 , 0x11501 , 0x11701 , 0x11901 , 0x11b01 , 0x11d01 , 0x11f01 , 0x12101 , 0x12301 , 0x12501 , 0x12701 , 0x12901 , 0x12b01 , 0x12d01 , 0x12f01 , 0x13101 , 0x13301 , 0x13501 , 0x13701 , 0x13901 , 0x13b01 , 0x13d01 , 0x13f01 , 0x14001 , 0x14201 , 0x14401 , 0x14601 , 0x14801 , 0x14a01 , 0x14c01 , 0x14e01 , 0x15001 , 0x15201 , 0x15401 , 0x15601 , 0x15801 , 0x15a01 , 0x15c01 , 0x15e01 , 0x16001 , 0x16201 , 0x16401 , 0x16601 , 0x16801 , 0x16a01 , 0x16c01 , 0x16e01 , 0x17001 , 0x17201 , 0x17401 , 0x17601 , 0x17801 , 0x17a01 , 0x17c01 , 0x17e01 , 0x18001 , 0x18201 , 0x18401 , 0x18601 , 0x18801 , 0x18a01 , 0x18c01 , 0x18e01 , 0x19001 , 0x19201 , 0x19401 , 0x19601 , 0x19801 , 0x19a01 , 0x19c01 , 0x19e01 , 0x1a001 , 0x1a201 , 0x1a401 , 0x1a601 , 0x1a801 , 0x1aa01 , 0x1ac01 , 0x1ae01 , 0x1b001 , 0x1b201 , 0x1b401 , 0x1b601 , 0x1b801 , 0x1ba01 , 0x1bc01 , 0x1be01 , 0x1bf01 , 0x1c102 , 0x1c304 , 0x1c507 , 0x1c70a , 0x1c90c , 0x1cb0f , 0x1cd12 , 0x1cf14 , 0x1d117 , 0x1d31a , 0x1d51d , 0x1d71f , 0x1d922 , 0x1db25 , 0x1dd27 , 0x1df2a , 0x1e12d , 0x1e330 , 0x1e532 , 0x1e735 , 0x1e938 , 0x1eb3a , 0x1ed3d , 0x1ef40 , 0x1f141 , 0x1f344 , 0x1f547 , 0x1f74a , 0x1f94c , 0x1fb4f , 0x1fd52 , 0x1fd54 , 0x1fd57 , 0x1fd5a , 0x1fd5d , 0x1fd5f , 0x1fd62 , 0x1fd65 , 0x1fd67 , 0x1fd6a , 0x1fd6d , 0x1fd6f , 0x1fd72 , 0x1fd75 , 0x1fd78 , 0x1fd7a , 0x1fd7d , 0x1fd80 , 0x1fd82 , 0x1fd85 , 0x1fd88 , 0x1fd8b , 0x1fd8d , 0x1fd90 , 0x1fd93 , 0x1fd95 , 0x1fd98 , 0x1fd9b , 0x1fd9d , 0x1fda0 , 0x1fda3 , 0x1fda6 , 0x1fda8 , 0x1fdab , 0x1fdae , 0x1fdb0 , 0x1fdb3 , 0x1fdb6 , 0x1fdb9 , 0x1fdbb , 0x1fdbe , 0x1fdc0 , 0x1fdc2 , 0x1fdc5 , 0x1fdc8 , 0x1fdca , 0x1fdcd , 0x1fdd0 , 0x1fdd3 , 0x1fdd5 , 0x1fdd8 , 0x1fddb , 0x1fddd , 0x1fde0 , 0x1fde3 , 0x1fde6 , 0x1fde8 , 0x1fdeb , 0x1fdee , 0x1fdf0 , 0x1fdf3 , 0x1fdf6 , 0x1fdf9 , 0x1fdfd , 0x1fdfd , 0x4fdfd , 0x8fdfd , 0xcfdfd , 0x10fdfd , 0x14fdfd , 0x18fdfd , 0x1cfdfd , 0x20fdfd , 0x24fdfd , 0x28fdfd , 0x2cfdfd , 0x30fdfd , 0x34fdfd , 0x38fdfd , 0x3cfdfd , 0x40fdfd , 0x43fdfd , 0x47fdfd , 0x4bfdfd , 0x4ffdfd , 0x53fdfd , 0x57fdfd , 0x5bfdfd , 0x5ffdfd , 0x63fdfd , 0x67fdfd , 0x6bfdfd , 0x6ffdfd , 0x73fdfd , 0x77fdfd , 0x7bfdfd , 0x7efdfd , 0x82fdfd , 0x86fdfd , 0x8afdfd , 0x8efdfd , 0x92fdfd , 0x96fdfd , 0x9afdfd , 0x9efdfd , 0xa2fdfd , 0xa6fdfd , 0xaafdfd , 0xaefdfd , 0xb2fdfd , 0xb6fdfd , 0xbafdfd , 0xbefdfd , 0xc1fdfd , 0xc5fdfd , 0xc9fdfd , 0xcdfdfd , 0xd1fdfd , 0xd5fdfd , 0xd9fdfd , 0xddfdfd , 0xe1fdfd , 0xe5fdfd , 0xe9fdfd , 0xedfdfd , 0xf1fdfe , 0xf5fdfd , 0xf9fdfe , 0xf9fdfe}

,

/* 66 gem-16 */

{0x110901 , 0x110901 , 0x110901 , 0x110901 , 0x110901 , 0x110901 , 0x110901 , 0x110901 , 0x110901 , 0x110901 , 0x110901 , 0x110901 , 0x110901 , 0x110901 , 0x110901 , 0x110901 , 0x110901 , 0x211101 , 0x211101 , 0x211101 , 0x211101 , 0x211101 , 0x211101 , 0x211101 , 0x211101 , 0x211101 , 0x211101 , 0x211101 , 0x211101 , 0x211101 , 0x211101 , 0x211101 , 0x211101 , 0x311901 , 0x311901 , 0x311901 , 0x311901 , 0x311901 , 0x311901 , 0x311901 , 0x311901 , 0x311901 , 0x311901 , 0x311901 , 0x311901 , 0x311901 , 0x311901 , 0x311901 , 0x311901 , 0x402101 , 0x402101 , 0x402101 , 0x402101 , 0x402101 , 0x402101 , 0x402101 , 0x402101 , 0x402101 , 0x402101 , 0x402101 , 0x402101 , 0x402101 , 0x402101 , 0x402101 , 0x402101 , 0x501920 , 0x501920 , 0x501920 , 0x501920 , 0x501920 , 0x501920 , 0x501920 , 0x501920 , 0x501920 , 0x501920 , 0x501920 , 0x501920 , 0x501920 , 0x501920 , 0x501920 , 0x501920 , 0x601140 , 0x601140 , 0x601140 , 0x601140 , 0x601140 , 0x601140 , 0x601140 , 0x601140 , 0x601140 , 0x601140 , 0x601140 , 0x601140 , 0x601140 , 0x601140 , 0x601140 , 0x601140 , 0x70095f , 0x70095f , 0x70095f , 0x70095f , 0x70095f , 0x70095f , 0x70095f , 0x70095f , 0x70095f , 0x70095f , 0x70095f , 0x70095f , 0x70095f , 0x70095f , 0x70095f , 0x70095f , 0x7f017f , 0x7f017f , 0x7f017f , 0x7f017f , 0x7f017f , 0x7f017f , 0x7f017f , 0x7f017f , 0x7f017f , 0x7f017f , 0x7f017f , 0x7f017f , 0x7f017f , 0x7f017f , 0x7f017f , 0x7f017f , 0x5f1a9f , 0x5f1a9f , 0x5f1a9f , 0x5f1a9f , 0x5f1a9f , 0x5f1a9f , 0x5f1a9f , 0x5f1a9f , 0x5f1a9f , 0x5f1a9f , 0x5f1a9f , 0x5f1a9f , 0x5f1a9f , 0x5f1a9f , 0x5f1a9f , 0x5f1a9f , 0x4033bf , 0x4033bf , 0x4033bf , 0x4033bf , 0x4033bf , 0x4033bf , 0x4033bf , 0x4033bf , 0x4033bf , 0x4033bf , 0x4033bf , 0x4033bf , 0x4033bf , 0x4033bf , 0x4033bf , 0x4033bf , 0x204ade , 0x204ade , 0x204ade , 0x204ade , 0x204ade , 0x204ade , 0x204ade , 0x204ade , 0x204ade , 0x204ade , 0x204ade , 0x204ade , 0x204ade , 0x204ade , 0x204ade , 0x204ade , 0x462fe , 0x462fe , 0x462fe , 0x462fe , 0x462fe , 0x462fe , 0x462fe , 0x462fe , 0x462fe , 0x462fe , 0x462fe , 0x462fe , 0x462fe , 0x462fe , 0x462fe , 0x462fe , 0x438afe , 0x438afe , 0x438afe , 0x438afe , 0x438afe , 0x438afe , 0x438afe , 0x438afe , 0x438afe , 0x438afe , 0x438afe , 0x438afe , 0x438afe , 0x438afe , 0x438afe , 0x438afe , 0x83b2fe , 0x83b2fe , 0x83b2fe , 0x83b2fe , 0x83b2fe , 0x83b2fe , 0x83b2fe , 0x83b2fe , 0x83b2fe , 0x83b2fe , 0x83b2fe , 0x83b2fe , 0x83b2fe , 0x83b2fe , 0x83b2fe , 0x83b2fe , 0xc2d9fe , 0xc2d9fe , 0xc2d9fe , 0xc2d9fe , 0xc2d9fe , 0xc2d9fe , 0xc2d9fe , 0xc2d9fe , 0xc2d9fe , 0xc2d9fe , 0xc2d9fe , 0xc2d9fe , 0xc2d9fe , 0xc2d9fe , 0xc2d9fe , 0xc2d9fe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe , 0xfefefe}

} ;
